c-----------------------------------------------------------------------
c     file surface_ints.f 
c     module that includes integrand routines associated with surface
c     integrals.  all subroutines included here must use the same
c     interface block.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     code organization.
c-----------------------------------------------------------------------
c     1.  no_surf_int.
c     2.  e_tangential.
c     3.  mag_tension.
c-----------------------------------------------------------------------
c     declarations.
c-----------------------------------------------------------------------
      MODULE surface_ints

      USE local
      USE rblock_type_mod
      USE tblock_type_mod
      USE lagr_quad_mod
      USE tri_linear
      USE math_tran
      USE physdat
      USE input
      USE global
      USE fft_mod
      USE generic_evals
      USE dump
      USE pardata
      IMPLICIT NONE

      CONTAINS
c-----------------------------------------------------------------------
c     subprogram 1. no_surf_int.
c     dummy routine with the correct interface.
c-----------------------------------------------------------------------
      SUBROUTINE no_surf_int(int,rb,tb,x,y,bigr,norm,
     $                       ijcell,alpha,dxdr,dydr,dxdz,dydz)

      COMPLEX(r8), DIMENSION(:,:,:), INTENT(OUT) :: int 
      TYPE(rblock_type), INTENT(INOUT) :: rb
      TYPE(tblock_type), INTENT(INOUT) :: tb
      REAL(r8), INTENT(IN) :: x,y,bigr,dxdr,dydr,dxdz,dydz
      REAL(r8), DIMENSION(2), INTENT(IN) :: norm
      REAL(r8), DIMENSION(:), INTENT(IN) :: alpha
      INTEGER(i4), DIMENSION(2) :: ijcell

      int=0

      RETURN
      END SUBROUTINE no_surf_int
c-----------------------------------------------------------------------
c     subprogram 2. e_tangential.
c     compute the integrand for tangential electric field.
c-----------------------------------------------------------------------
      SUBROUTINE e_tangential(int,rb,tb,x,y,bigr,norm,
     $                        ijcell,alpha,dxdr,dydr,dxdz,dydz)

      !USE input, ONLY: periodicity

      COMPLEX(r8), DIMENSION(:,:,:), INTENT(OUT) :: int 
      TYPE(rblock_type), INTENT(INOUT) :: rb
      TYPE(tblock_type), INTENT(INOUT) :: tb
      REAL(r8), INTENT(IN) :: x,y,bigr,dxdr,dydr,dxdz,dydz
      REAL(r8), DIMENSION(2), INTENT(IN) :: norm
      REAL(r8), DIMENSION(:), INTENT(IN) :: alpha
      INTEGER(i4), DIMENSION(2) :: ijcell

      REAL(r8) :: e_tor,r,z, freq, amp, zeta, zmid, kzeff, damp_dz

      COMPLEX(r8), DIMENSION(44) ::  phi_coef, Ez_coef
      COMPLEX(r8), DIMENSION(1,SIZE(int,3)) :: tele,ter,tez
      INTEGER(i4) :: imode, mm, nz
      COMPLEX(r8), DIMENSION(22) :: tibbar_coef
      REAL(r8), DIMENSION(500, 89) :: phiDCB2, EzDCB2
      REAL(r8), DIMENSION(499) :: ddz  
 
      COMPLEX(r8)::bthi,btni,bthx,btnx,eth,etn
      COMPLEX(r8),EXTERNAL::intp

      REAL(r8),PARAMETER::ez0=100.

      INTEGER(i4)::iz,zloc

      REAL(r8)::dz,xi
      REAL(r8),DIMENSION(rb%nth)::bphi,bder

c-----------------------------------------------------------------------
c     find toroidal applied E from loop voltage.
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        e_tor=volt/(twopi*bigr)
      ELSE
        e_tor=volt/per_length
      ENDIF
c---------------------------------------------------------------------
c    do a lagrange quad eval to obtain the Z coord.  
c-----------------------------------------------------------------------
      CALL lagr_quad_eval(rb%rz,x,y,0_i4)
      r = rb%rz%f(1)
      z = rb%rz%f(2)
c    amplitude of the enveloping function to smoothly make Etang drop to
c    0 at z = ymin, ymax
      zmid = 0.5*(ymax + ymin)
      zeta = ABS(z-zmid)/(ymax-zmid)
      amp = (1. - TANH( (zeta-frac)/epsz) )
      amp = amp/( 1. + TANH(frac/epsz) )
c    analytically compute the derivative of amp since d/dz (amp) contributes
c    to Ez
      damp_dz = -SIGN(1., z - zmid)/(epsz*(ymax-zmid) )
     $        *  COSH( (zeta-frac)/epsz )**(-2.)
      damp_dz = damp_dz/( 1. + TANH(frac/epsz) )
c      IF ( abs(damp_dz) >1e-1) WRITE(*,*) 'df_dz = ', damp_dz
c     $      , ' at z = ', z/per_length
      IF ( (periodicity.eq. 'y-dir').OR. (frac > 1.e3)) THEN
          amp = 1.0
          damp_dz = 0.
      ENDIF    
c    collect all the factors into a single one including time ramp
c    Note 0.5 is a fudge factor
      amp = 0.5*phi0*amp*TANH( t/t_ramp) 
c      amp = 0.5*phi0*TANH( t/t_ramp) 
      damp_dz = 0.5*phi0*damp_dz*TANH( t/t_ramp)
      !IF (frac > 1.e3) damp_dz = 0. ! zero it for the periodic geometry
c-----------------------------------------------------------------------
      int=0._r8
c-----------------------------------------------------------------------
c      IF (ABS(phi0)<1e-10) THEN
c-----------------------------------------------------------------------
c     apply to n=0 only.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
        IF (keff(imode)/=0) CYCLE mode_loop
        int(1,:,imode)=-alpha*dt*e_tor*norm(2)
        int(2,:,imode)= alpha*dt*e_tor*norm(1)
        int(3,:,imode)=-alpha*dt*e_vert*norm(1)
      ENDDO mode_loop
c DCB
      IF (phi0.ge.1e-10.AND.z > ymin .AND. z < ymax) THEN
c-----------------------------------------------------------------------
c   TIBBAR primary electrode implementation
c     apply to n=1 (m=1 physicaly) only if phi0 is ON
c       int2 -> -Ephi zhat = im/rw*Phi = i (keff/rw)*Phi
c       int3 ->  Ez phi_hat  = -ikPhi = -i(n/R)*Phi 
c                            = -i*2pi*klock*Phi/length
c-----------------------------------------------------------------------
c       tibbar_coef =(0.,1.0)*(/ (0., 0.) ,(0., 0.70823881), (0., 0.),
c     $      (0., 0.56906566), (0., 0.), (0., 0.34238405), (0., 0.),
c     $      (0., 0.10766845), (0., 0.), (0., 0.06302926), (0., 0.),
c     $      (0., 0.13384927), (0., 0.), (0., 0.11412448), (0., 0.),
c     $      (0., 0.04563614), (0., 0.), (0., 0.02256472), (0., 0.),
c     $      (0., 0.05746561), (0., 0.), (0., 0.05253371) /)

c CA, March 2017 -- new coefficients to implement the localized Tibbar
c   primary electrodes. Note the odd modes are purely real while the
c   even ones purely imaginary. This waveform for the phi results in a
c   localized normal current that carries current from one electrode
c   to the other
        tibbar_coef = (/ (-1.102460902804933e-08, 0.),
     $      (-4.15059180e-01, 0.), (0., 5.99480453e-03), 
     $      (-4.99397435e-02, 0.), (0., -3.02987786e-03),
     $      (-1.70376223e-02, 0.), (0., -1.85373457e-03),
     $      (-4.23288636e-03, 0.), (0., -4.53331312e-04),
     $      (-2.17594734e-03, 0.), (0., -5.11425023e-05),
     $      (-4.24859404e-03, 0.), (0., -8.18493020e-05),
     $      (-3.37563008e-03, 0.), (0., -7.53709125e-05),
     $      (-1.26916773e-03, 0.), (0., -2.97097467e-05),
     $      (-5.92858283e-04, 0.), (0., -3.70496725e-06),
     $      (-1.43987979e-03, 0.), (0., -7.43501125e-06),
     $      (-1.25700795e-03, 0.) /)

c ---------------------------------------------------------------------
c  CA Sept 2017 -- read in DCB's python=processed potential and Ez from 
c    a txt file. The files is nz x ntheta where nz = 500 amd ntheta = 89,
c    corresponding to real and imaginary parts of 44 Fourier amplitudes 
c    for each z postion. There is an extra column because the very
c    first column contains the z-positions from DCB's phi solver.
        phiDCB2 = RESHAPE( phiDCB, (/500,89/) )
        EzDCB2 = RESHAPE( EzDCB, (/500,89/) )

        ddz = phiDCB(2:500) - phiDCB(1:499)
c    Find the index of phiDCB closet to z
        zloc =  MINLOC( ABS(phiDCB2(:,1) - z), 1 )
c    fix the index so that it lies between i + 1 and i'th DCB point
        IF ( phiDCB2(zloc, 1) > z) zloc = zloc - 1
        dz = z - phiDCB2(zloc, 1); ! linear weight
c    construct the Fourier amplitudes up to 44 modes for phi
        phi_coef = (1. - dz/ddz(zloc))*phiDCB2( zloc, 2:45) 
     $           + dz/ddz(zloc)*phiDCB2( zloc + 1, 2:45) 
c    Add the imaginary part with '-' sign because of NIMROD's left-
c    handiness
        phi_coef = phi_coef - (0.,1.)* 
     $           ((1. - dz/ddz(zloc))*phiDCB2( zloc, 46:) 
     $           + dz/ddz(zloc)*phiDCB2( zloc + 1, 46:) )
c    construct the Fourier amplitudes up to 44 modes for Ez
        Ez_coef = (1. - dz/ddz(zloc))*EzDCB2( zloc, 2:45) 
     $           + dz/ddz(zloc)*EzDCB2( zloc + 1, 2:45) 
c    Add the imaginary part with '-' sign because of NIMROD's left-
c    handiness
        Ez_coef = Ez_coef - (0.,1.)* 
     $           ((1. - dz/ddz(zloc))*EzDCB2( zloc, 46:) 
     $           + dz/ddz(zloc)*EzDCB2( zloc + 1, 46:) )
c ---------------------------------------------------------------------

        freq = twopi*60 ! AC frequency
c    Note a bunch of factors have been absorbed into amp including phi0        
        DO imode=1,nmodes
          IF (local_electrodes) THEN ! use tibbar_coef to construct the E_tang
            IF ( MOD( NINT( keff(imode) ), 2)== 0 ) CYCLE !only do the odd modes 
            kzeff = twopi*keff(imode)/ABS( ymax- ymin ) ! set n = -m
            mm = mode_lo + imode - 1  ! mode number variable
c     Program the coefficients of the DCB potential
c       el_width = electrode width in radians
            phi_coef = 1./( twopi*mm*(pi - el_width/2.) )
     $             * ( 2.*el_width*SIN(el_width*mm/2.) + 
     $                (8./mm)*COS(el_width*mm/2.) )
            int(2,:,imode) = alpha*dt*tibbar_coef(mm)*
     $         amp*(0.0,1.0)*(keff(imode)/bigr)*
     $         EXP((0.0,1.0)*kzeff*(z - ymin) ) !*SIN(freq*t)
            int(3,:,imode)= -amp*alpha*dt*tibbar_coef(mm)*
     $        (0.0,1.0)*kzeff*EXP( (0.0,1.0)*kzeff*(z - ymin) )
c     alternate formulation using DCB's potential
c            int(3,:,imode)= -amp*alpha*dt*Ez_coef(mm)
          ELSE
            IF (keff(imode)/= mlock) CYCLE 
c    Use per_length as the full periodic length to keep klock as an integer
c    E.g., if you want 1/6 of twist over 10 cm, leave klock = -1 but set
c    per_length = 0.6. 
            kzeff = -twopi*klock/per_length  ! the axial wave number
            int(2,:,imode) = alpha*dt*
     $         amp*(0.0,1.0)*(keff(imode)/bigr)*
     $         EXP((0.0,1.0)*kzeff*(z - ymin) ) !*SIN(freq*t)
            int(3,:,imode) = -alpha*dt*( amp*(0.0,1.0)*kzeff + damp_dz )
     $                     * EXP( (0.0,1.0)*kzeff*(z - ymin) )
c     $        *SIN(freq*t)
          ENDIF ! END local_electrodes
        ENDDO
      ELSEIF(phi0<0) THEN
c use surface potential read from file
c ends -- nothing to do
         IF(z>ymin.AND.z<ymax) THEN
            dz=rb%z_phi(2)-rb%z_phi(1); xi=(z-rb%z_phi(1))/dz+1
            iz=xi; iz=MIN(iz,rb%nt-1); xi=xi-iz
c            write(*,*) 'z, iz, xi = ',z, iz, xi
            write(*,*) 'shape bphi =', shape(rb%b_phi)
            bphi= rb%b_phi(:,iz  )*(1+xi*xi*(-3+2*xi))
     $        +dz*rb%d_phi(:,iz  )*(xi*(1+xi*(-2+xi)))
     $           +rb%b_phi(:,iz+1)*(xi*xi*(3-2*xi))
     $        +dz*rb%d_phi(:,iz+1)*(xi*xi*(-1+xi))
            bder= rb%b_phi(:,iz  )*      xi*(-6+6*xi)/dz
     $           +rb%d_phi(:,iz  )*    (1+xi*(-4+3*xi))
     $           +rb%b_phi(:,iz+1)*    xi*(6-6*xi)/dz
     $           +rb%d_phi(:,iz+1)*    xi*(-2+3*xi)
cc            WRITE(*,*) node,r,z,iz,xi,bder(3),bder(4)
            DO imode=1,nmodes
c               IF ( keff(imode) == 1 ) THEN !only do m = 1 
                 mm = NINT(keff(imode)) ! mode number variable
c                 int(1,:,imode)=int(1,:,imode)
c     $                        -alpha*dt*norm(2)*keff(imode)/bigr
c     $                        *CMPLX(-bphi(2*mm+2),bphi(2*mm+1))
c                 int(2,:,imode)=int(2,:,imode)
                 int(2,:,imode) = alpha*dt*norm(1)*keff(imode)/bigr
     $                        *CMPLX(-bphi(2*mm+2),bphi(2*mm+1))
c                 int(3,:,imode)=int(3,:,imode)
                 int(3,:, imode) = 
     $               -alpha*dt*CMPLX(bder(2*mm+1),bder(2*mm+2))
c              ENDIF
           ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     for two-fluid computations, the electric field in the B-advance
c     is actually E+kboltz*grad(Te)/e, which eliminates the purely
c     electrostatic term.  This term, therefore, needs to be added to
c     the surface integrand.
c-----------------------------------------------------------------------
      IF (ohms/='mhd'.AND.beta>0) THEN
        CALL generic_eval(rb%tele,tb%tele,dxdr,dydr,dxdz,dydz,x,y,
     $                    tb%tgeom,ijcell,tele,ter,tez,1_i4)
        ter=dt*kboltz*(1._r8-meomi)/(elementary_q*(1._r8+meomi))*ter
        tez=dt*kboltz*(1._r8-meomi)/(elementary_q*(1._r8+meomi))*tez
        tele=dt*kboltz*(1._r8-meomi)/(elementary_q*(1._r8+meomi))*tele

        estat_loop: DO imode=1,nmodes
          int(1,:,imode)=int(1,:,imode)-alpha*
     $                   norm(2)*(0,1)*keff(imode)*tele(1,imode)/bigr
          int(2,:,imode)=int(2,:,imode)+alpha*
     $                   norm(1)*(0,1)*keff(imode)*tele(1,imode)/bigr
          int(3,:,imode)=int(3,:,imode)+alpha*
     $                   (ter(1,imode)*norm(2)-tez(1,imode)*norm(1))
        ENDDO estat_loop
      ENDIF

c DCB
      IF(b_segs>0) THEN
cc         IF(node==0) WRITE(*,*) MAXVAL(ABS(phrx))
c do resistive boundary for n=1
         DO imode=1,nmodes
            IF(keff(imode)/=1) CYCLE
c internal field
            CALL lagr_quad_eval(rb%be,x,y,0_i4)
            bthi=rb%be%f(3,imode)
            btni=(-norm(2)*rb%be%f(1,imode)+norm(1)*rb%be%f(2,imode))
c external field            
            bthx=-(0_r8,1_r8)*intp(rb,r,z,phrx,dprx,btnx)/bigr
c wall current induced electric field
            eth=r_wall/mu0*( btnx-btni); etn=r_wall/mu0*(-bthx+bthi)
            int(1,:,imode)=int(1,:,imode)
     $           -alpha*dt*eth*norm(2)
            int(2,:,imode)=int(2,:,imode)
     $           +alpha*dt*eth*norm(1)
            int(3,:,imode)=int(3,:,imode)-alpha*dt*etn
         ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE e_tangential
c-----------------------------------------------------------------------
c     subprogram 3. mag_tension.
c     compute the integrand for Maxwell stress tension at the 
c     surface, -dS.[BB-0.5*B^2*I].alpha .
c-----------------------------------------------------------------------
      SUBROUTINE mag_tension(int,rb,tb,x,y,bigr,norm,
     $                       ijcell,alpha,dxdr,dydr,dxdz,dydz)

      COMPLEX(r8), DIMENSION(:,:,:), INTENT(OUT) :: int 
      TYPE(rblock_type), INTENT(INOUT) :: rb
      TYPE(tblock_type), INTENT(INOUT) :: tb
      REAL(r8), INTENT(IN) :: x,y,bigr,dxdr,dydr,dxdz,dydz
      REAL(r8), DIMENSION(2), INTENT(IN) :: norm
      REAL(r8), DIMENSION(:), INTENT(IN) :: alpha
      INTEGER(i4), DIMENSION(2) :: ijcell

      REAL(r8), DIMENSION(3) :: be_eq,dumr
      COMPLEX(r8), DIMENSION(3,SIZE(int,3)) :: be
      COMPLEX(r8), DIMENSION(1,1) :: dumc

      REAL(r8) :: dtmu0,nbeq
      COMPLEX(r8) :: dot,nb
      INTEGER(i4) :: im
c-----------------------------------------------------------------------
c     point-wise evaluations.
c-----------------------------------------------------------------------
      CALL generic_eval(rb%be_eq,tb%be_eq,dxdr,dydr,dxdz,dydz,x,y,
     $                  tb%tgeom,ijcell,be_eq,dumr,dumr,0_i4)
      CALL generic_eval(rb%be,tb%be,dxdr,dydr,dxdz,dydz,x,y,
     $                  tb%tgeom,ijcell,be,dumc,dumc,0_i4)
      IF (geom=='tor') be_eq(3)=be_eq(3)/bigr
      dtmu0=dt/mu0
c-----------------------------------------------------------------------
c     linear terms only so far.
c-----------------------------------------------------------------------
      int=0
      nbeq=norm(1)*be_eq(1)+norm(2)*be_eq(2)
      DO im=1,nmodes
        dot=be_eq(1)*be(1,im)+be_eq(2)*be(2,im)+be_eq(3)*be(3,im)
        nb=norm(1)*be(1,im)+norm(2)*be(2,im)
        int(1,:,im)=dtmu0*(dot*norm(1)-nbeq*be(1,im)-nb*be_eq(1))*alpha
        int(2,:,im)=dtmu0*(dot*norm(2)-nbeq*be(2,im)-nb*be_eq(2))*alpha
        int(3,:,im)=dtmu0*(           -nbeq*be(3,im)-nb*be_eq(3))*alpha
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE mag_tension
c-----------------------------------------------------------------------
c     close module.
      END MODULE surface_ints
