module dagmg_mod

  !----------------------------------------------------------------------------!
  !                     BRIEF DESCRIPTION OF THE MODULE
  !----------------------------------------------------------------------------!
  ! The main driver subroutine dagmg below provides a SEQUENTIAL implementation
  ! of the method presented in [1], where the used algorithms are described
  ! in detail.
  !
  ! See the accompanying userguide (Ref.[2]) for more details on how to use the
  ! software and the README file for copyright notice and installation
  ! instructions.
  !
  ! REFERENCES:
  ! ==========
  !
  ! [1] http://homepages.ulb.ac.be/~ynotay/AGMG/index.html
  !
  ! [2] Yvan Notay, User's Guide to AGMG, March 2009.
  !
  ! [3] Yvan Notay, An aggregation-based algebraic multigrid method,
  !     Report GANMN 08-02, Revised March 2009.
  !
  !----------------------------------------------------------------------------!

  use numeric_kinds
  implicit none
  
!  private
!  public :: dagmg

  save

  !----------------------------------------------------------------------------!
  ! Parameters for subroutine dagmg that may be tuned by expert users
  !
  ! INTEGER
  !
  !  maxlev is the maximal number of levels
  !         (should be large enough - much larger than needed is armless)
  !  real_len is the length of 1 REAL(kind(0.0d0)) in byte
  !        (used only to display information on memory usage)

  integer, parameter :: maxlev = 40, real_len = 8

  !  nsmooth is the number of smoothing steps (pre + post);
  !          each cycle includes (nsmooth+1)/2 pre-smoothing step(s)
  !          and nsmooth/2 post-smoothing step(s)
  !  nstep is the maximum number of coarsening steps
  !        nstep<0 means that coarsening is stopped only according to
  !        the matrix size, see parameter maxcoarsesize
  !  nlvcyc  is the number of coarse levels (from bottom) on which V-cycle
  !          formulation is enforced (Rmk: K-cycle always allowed at first
  !          coarse level)

  integer , parameter :: nsmooth = 2, nstep = -1, nlvcyc = 0

  !  maxcoarsesize: when the size of the coarse grid matrix is less than or
  !                 equal to maxcoarsesize,  it is factorized exactly and
  !                 coarsening is stopped.

  integer , parameter :: maxcoarsesize = 400

  ! REAL
  !
  !  resi is the threshold t for the relative residual error in inner FCG & GCR
  !       iterations, see Algorithm 3.2 in [1]
  !  checkdd is the threshold for strongly diagonally dominant rows,
  !        see the first line of "Initialization" in Algorithm 2.1 in [1]
  !  trswc is the Strong/Weak coupling threshold beta in Algorithms 2.1 & 2.2
  !        in [1]
  !  trspos is a threshold: if a row has a positive offidiagonal entry larger
  !         than trspos times the diagonal entry, the corresponding node is kept
  !         out of the aggregation
  !  scalcg is the scaling factor applied to the coarse grid matrices
  
  real (kind=rk), parameter :: resi = 0.25_rk
  real (kind=rk), parameter :: checkdd = 5.0_rk, &
                               trswc = 0.25_rk, &
                               trspos = 0.45_rk, &
                               scalcg = 1.0_rk

  !----------------------------------------------------------------------------!
  ! Internal variables (do not modify).
  
  type :: InnerData
     real (kind=rk), dimension(:), pointer :: a
     integer, dimension(:), pointer :: ja
     integer, dimension(:), pointer :: ia
     real (kind=rk), dimension(:), pointer :: p
     integer, dimension(:), pointer :: idiag
     integer, dimension(:), pointer :: ind
     integer, dimension(:), pointer :: inloc
     integer, dimension(:), pointer :: ilstout
     integer, dimension(:), pointer :: lstout
     integer, dimension(:), pointer :: ilstin
     integer, dimension(:), pointer :: lstin
  end type InnerData
  
  type (InnerData), dimension(maxlev) :: dt
  integer, dimension(maxlev) :: nn, innermax
  integer, dimension(3,maxlev) :: kstat = 0
  integer :: nlev, nwrkcum, iout, nrst
  real (kind=rk) :: memi = 0.0_rk, &
                    memax = 0.0_rk, &
                    memr = 0.0_rk
  real (kind=rk) :: mritr, rlenilen
  logical :: spd, wfo, wff, allzero
  integer, parameter :: IRANK = -9999

  INTEGER,PARAMETER::big_int=99999999

  !----------------------------------------------------------------------------!
  !                                 PROCEDURES
  !----------------------------------------------------------------------------!

contains

  !============================================================================!

  recursive subroutine dagmg_setup( l, &
                                    n, &
                                    a, &
                                    ja, &
                                    ia )
   
    ! Arguments.
    integer :: l, n, ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk) :: a(*)

    ! Local variables.
    integer :: nnext, ierr, nzan, nmax, i, j, k, llsto, lw, lwn, lw0, kw, nwprev
    integer, pointer, dimension(:) :: iw, iw0, ja2, lcg, iwp
    real (kind=rk), allocatable, dimension(:) :: a2
    integer, dimension(2), save :: nlc, nlcp, nlc1
    integer, save :: icum
    real (kind=rk), dimension(2), save :: ngl, nglp, ngltot, nlctot, ngl1
    logical, save :: slowcoarse = .false.
    real (kind=rk) :: ops, ff, xsi, eta, checkddl
    real (kind=rk), dimension(2) :: dum
    character (len=13) :: prtint
    real (kind=rk), dimension(n) :: fff
    integer , parameter :: ione = 1
    integer :: allocation_status
    
    nn(l) = n
    nlc(1) = n
    nlc(2) = ia(n+1) - ia(1)
    
    ngl = nlc

    if ( l > 1 ) then

       nlctot = nlctot + real(nlc,rk)
       ngltot = ngltot + real(ngl,rk)

       if ( wfo ) then

          if ( n > big_int ) then
             write(prtint(1:12),'(1pe12.5)') dble(n)
          else
             write(prtint(1:12),'(i12)') n
          end if

          write(iout,920) prtint(1:12), real(nlcp(1),rk) / real(n,rk)

          if ( nlc(2) > big_int ) then

             write(prtint(1:12),'(1pe12.5)') real(nlc(2),rk)

          else

             write(prtint(1:12),'(i12)') nlc(2)

          end if

          write(iout,921) prtint(1:12),real(nlc(2),rk)/real(n,rk), &
               real(nlcp(2),rk)/real(nlc(2),rk)

       end if

       xsi = ( 6.0_rk * 1.0_rk ) / ( 10.0_rk * 1.0_rk )
       eta = 0.5_rk
       ff = ( ngl1(2) / ngl(2) ) * ( xsi**(l-1) ) / real(icum,rk)

       if ( ff < 2.0_rk - eta ) then
          innermax(l) = 1
       else
          innermax(l) = 2
       end if

       icum = icum * innermax(l)

    else

       nlctot = real(nlc,rk)
       ngltot = real(ngl,rk)
       ngl1 = ngl
       nlc1 = nlc
       icum = 1
       innermax(1) = 1
       slowcoarse = .false.
       allzero = .false.
       nglp(1) = 0.0_rk

       if ( wfo ) then

          if ( n > big_int ) then
             write(prtint(1:12),'(1pe12.5)') real(n,rk)
          else
             write(prtint(1:12),'(i12)') n
          end if

          write(iout,'()')
          write(iout,918) prtint(1:12)

          if ( nlc(2) > big_int ) then
             write(prtint(1:12),'(1pe12.5)') real(nlc(2),rk)
          else
             write(prtint(1:12),'(i12)') nlc(2)
          end if

          write(iout,919) prtint(1:12),real(nlc(2),rk)/real(n,rk)
          write(iout,'()')

       end if

    end if

    if ( l == nstep + 1 .or. l == maxlev .or. &
         ngl(1) <= maxcoarsesize .or. &
         ( slowcoarse .and. ( 3.0_rk * nglp(1) < 5.0_rk * ngl(1) ) ) .or. &
         nglp(1) == ngl(1) ) then

       nlev = l
       innermax(l) = 0

    end if

    slowcoarse = .false.

    if ( l > 1 .and. 3.0_rk * nglp(1) < 5.0_rk * ngl(1) ) slowcoarse = .true.

    nlcp = nlc
    nglp = ngl

    if ( n == 0 ) then

       nnext = 0
       allzero = .true.

       ! n = 0 and nlev /= l possible with the parallel version.

       If ( nlev /= l ) then

          if ( wfo ) then
             write(iout,914) IRANK,l+1
          end if

          call dagmg_setup( l + 1, &  !
                            nnext, &  !
                            a, &      !
                            ja, &     !
                            ia )      !

          return

       end if

       if ( allzero ) goto 500 ! skip factorization on last level

       goto 300

    end if

    lw = 0
    lw0 = 0

    if ( l == nlev ) goto 200

    lw = 3 * n + 1
    lw0 = n

    allocate( dt(l) % p(n), &
              dt(l) % ind(n), &
              dt(l) % idiag(n+1), &
              lcg(4*n), &
              a2(nlc(2)), &
              ja2(nlc(2)), &
              iw(lw), &
              iw0(lw0), &
              stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_setup."
       stop
    end if

    memi = memi + 6 * n + 1 + nlc(2) + lw + lw0
    memr = memr + n + nlc(2)
    memax = max( memax, memr + memi * rlenilen )
    iwp => iw
    kw = n + 1

    if ( l > 1 ) then

       checkddl = -1.0_rk

    else

       if ( wfo ) then
          write (iout,901) IRANK
       end if

       if ( wff ) then
          write (iout,902) checkdd, trswc
          write (iout,903) trspos
       end if

       checkddl = checkdd

    end if

    call dagmg_aggl4( n, &                !
                      nnext, &            !
                      a, &                !
                      ja, &               !
                      ia, &               !
                      lcg, &              !
                      dt(l) % ind, &      !
                      dt(l) % idiag, &    !
                      iw0, &              !
                      iwp, &              !
                      iw(kw:kw+n-1), &    !
                      dt(l) % p, &        !
                      trswc, &            !
                      a2, &               !
                      ja2, &              !
                      iw(n+kw:n+kw+n), &  !
                      checkddl, &         !
                      trspos )            !

200 continue

    if ( lw > 0 ) deallocate(iw)

    memi = memi - lw
    lw = 0

    if ( l == nlev ) goto 300

    call dagmg_setsgs( n, &              !
                       a, &              !
                       ja, &             !
                       ia, &             !
                       dt(l) % p, &      !
                       dt(l) % idiag, &  !
                       a2, &             !
                       ja2, &            !
                       dt(l) % inloc )   !

    allocate( dt(l+1) % ia(nnext+1), &
              stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_setup."
       stop
    end if

    memi = memi + nnext + 1
    memax = max( memax, memr + memi * rlenilen )
    j = nnext

    if ( j > lw0 ) then

       if ( lw0 > 0 ) deallocate(iw0)
       
       allocate( iw0(j), &
                 stat = allocation_status )

       if ( allocation_status > 0 ) then
          print *, "Allocation error in dagmg_mod.f90/dagmg_setup."
          stop
       end if
       
       memi = memi - lw0 + j
       memax = max( memax, memr + memi * rlenilen )
       lw0 = j

    end if

    call dagmg_setacg( nnext, &         !
                       lcg, &           !
                       a, &             !
                       ja, &            !
                       ia, &            !
                       a2, &            !
                       ja2, &           !
                       dt(l+1) % ia, &  !
                       nzan, &          !
                       dt(l) % ind, &   !
                       iw0 )            !

    deallocate( iw0, &
                lcg )

    allocate( dt(l+1) % a(nzan), &
              dt(l+1) % ja(nzan), &
              stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_setup."
       stop
    end if

    memi = memi - lw0 - 4 * n + nzan
    memr = memr + nzan
    memax = max( memax, memr + memi * rlenilen )
    dt(l+1) % a(1:nzan) = a2(1:nzan)
    dt(l+1) % ja(1:nzan) = ja2(1:nzan)

    deallocate( a2, &
                ja2 )

    memi = memi - nlc(2)
    memr = memr - nlc(2)

    if ( scalcg < 0.0_rk ) then
       ff = -scalcg * real(nnext,rk) / real(n,rk)
    else
       ff = scalcg
    end if

    if ( ff /= 1.0_rk ) then

       if ( wfo ) then
          write(iout,'()')
          write(iout,915) IRANK,l+1,ff
       end if

       call dscal( nzan, &         !
                   ff, &           !
                   dt(l+1) % a, &  !
                   IONE )          !

    else

       if ( wfo ) then

          write(iout,'()')
          write(iout,914) IRANK,l+1

       end if

    end if
    
250 continue

    call dagmg_setup( l + 1, &         !
                      nnext, &         !
                      dt(l+1) % a, &   !
                      dt(l+1) % ja, &  !
                      dt(l+1) % ia )   !
    
    goto 500
    
300 continue
    
    call dagmg_LAPACK( n, &    !
                       a, &    !
                       ja, &   !
                       ia, &   !
                       fff, &  !
                       1, &    !
                       ops )   !

    if ( wfo ) then

       write(iout,911) IRANK,ops/real(11*nn(1)+2*nlc1(2),rk)
       write(iout,'()')

    end if
    
500 continue

    if ( l == 1 ) then

       if ( wfo ) then

          write(iout,'()')
          write(iout,954) nlctot(1)/real(nlc1(1),rk)
          write(iout,955) nlctot(2)/real(nlc1(2),rk)

       end if

       do i = 1, nlvcyc
          k = nlev - i
          if ( k > 2 ) innermax(k) = 1
       end do

       nwrkcum = 0
       nwprev = 0

       do i = 2, nlev - 1

          if ( innermax(i) > 1 )  then

             nwrkcum = nwrkcum + max( nwprev, 6 * nn(i) )
             nwprev  = nn(i)

          else

             nwrkcum = nwrkcum + max( nwprev, 4 * nn(i) )
             nwprev  = 0

          end if

       end do

       nwrkcum = nwrkcum + max( nwprev, 2 * nn(nlev) )

    end if
    
    return

901 format(i3,'*SETUP: Coarsening by double pairwise aggregation')
902 format('****       Strong diag. dom. trs:',f5.1, &
         ' ; Strong/Weak coupling trs:',f5.2)
903 format('****',18x,'Threshold for rows with large pos. offdiag.:',f5.2)
911 format(i3,'*','        Exact factorization:',f12.3, &
         ' equiv. CG iterations')
914 format(i3,'*','                      Level:',I12)
915 format(i3,'*','                      Level:',I12, &
         ' ;  applied scaling factor:',f6.3)
918 format('****','        Number of unknowns:', A12)
919 format('****','                 Nonzeros :', A12, &
         ' (per row:',f7.2,')')
920 format('****','        Number of variables:',A12, &
         '          (reduction ratio:',f5.2,')')
921 format('****','                   Nonzeros:',A12, &
         ' (per row:',f4.1, &
         '; red. ratio:',f5.2,')')
954 format('****',' relative complexity (grid):',f9.2)
955 format('****',' relative complexity (nnzs):',f9.2)

  end subroutine dagmg_setup

  !============================================================================!

  subroutine dagmg_relmem

    ! Local variables.
    integer :: l

    do l = 1, nlev

       if ( l > 1 ) then

          if ( nn(l-1) > 0 ) then

             deallocate( dt(l) % a, &
                         dt(l) % ja, &
                         dt(l) % ia )

          end if

       end if

       if ( l < nlev .and. nn(l) > 0 ) then

          deallocate( dt(l) % p, &
                      dt(l) % ind, &
                      dt(l) % idiag )

       end if

    end do

    memi = 0
    memr = 0
    memax = 0

    return

  end subroutine dagmg_relmem

  !============================================================================!

  subroutine dagmg_LAPACK( N, &
                           a, &
                           ja, &
                           ia, &
                           f, &
                           ijb, &
                           flop )

    ! Arguments.
    integer :: N
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer :: ia(n+1)
    real (kind=rk) :: f(N)
    integer :: ijb
    real (kind=rk) :: flop

    ! Local variables.
    real (kind=rk), dimension(:,:), allocatable, save :: ac
    integer, dimension(:), allocatable, save :: ipiv
    integer, save :: iflop
    integer :: i, kk, inloc
    integer :: ierr
    integer, parameter :: IONE = 1
    integer :: allocation_status
    
    ierr = 0

    if ( ijb == -2 ) then
       
       deallocate ( ac, &
                    ipiv )
       
    else if ( ijb == 1 ) then
       
       allocate ( ac(n,n), ipiv(n), &
                  stat = allocation_status )

       if ( allocation_status > 0 ) then
          print *, "Allocation error in dagmg_mod.f90/dagmg_LAPACK."
          stop
       end if
       
       memi = memi + n
       memr = memr + n * n
       memax = max( memax, memr + memi * rlenilen )
       ac = 0.0_rk

       do i = 1, n
          do kk = ia(i), ia(i+1) - 1
             ac(i,ja(kk)) = a(kk)
          end do
       end do

       call dgetrf( N, &     !
                    N, &     !
                    ac, &    !
                    N, &     !
                    ipiv, &  !
                    ierr )   !

       if ( ierr /= 0 ) then
          write(iout, *) ' FATAL ERROR in GETRF: ierror=',ierr
          stop
       end if

       iflop = ( 2 * n * n - n )
       flop = ( 2.0_rk * 1.0_rk ) / ( 3.0_rk * 1.0_rk ) * ( real(n,rk)**3 )
       
    else if ( ijb == 2 ) then
       
       call dgetrs( 'N', &   !
                    N, &     !
                    IONE, &  !
                    ac, &    !
                    N, &     !
                    ipiv, &  !
                    f, &     !
                    N, &     !
                    ierr )   !

       if ( ierr /= 0 ) then
          write(iout, *) ' FATAL ERROR in GETRS: ierror=',ierr
          stop
       end if

       flop = flop + real(iflop,rk)
       
    end if
    
    return

  end subroutine dagmg_LAPACK

  !============================================================================!

  recursive subroutine dagmg_CGcorr( N, &
                                     X, &
                                     R, &
                                     l, &
                                     flop, &
                                     w )
    
    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: X
    real (kind=rk), dimension(N) :: R
    integer :: l
    real (kind=rk) :: flop
    real (kind=rk) :: w(*)

    ! Local variables.
    integer :: idum
    real (kind=rk) :: dum
    integer :: nnext, l1, RN, XN

    l1 = l + 1
    nnext = nn(l1)

    if ( nnext > 0 ) then

       XN = 1
       RN = XN + nnext

       call dagmg_restaggl( N, &            !
                            nnext, &        !
                            R, &            !
                            w(RN), &        !
                            dt(l) % ind, &  !
                            flop )          !

       if ( l1 == nlev ) then

          w(XN:XN+nnext-1) = w(RN:RN+nnext-1)

          call dagmg_LAPACK( nnext, &         !
                             dt(l1) % a, &    !
                             dt(l1) % ja, &   !
                             dt(l1) % ia,  &  !
                             w(XN), &         !
                             2, &             !
                             flop )           !

       else if ( innermax(l1) <= 1 ) then

          call dagmg_prec( nnext, &        !
                           w(XN), &        !
                           w(RN), &        !
                           dt(l1) % a, &   !
                           dt(l1) % ja, &  !
                           dt(l1) % ia, &  !
                           l1, &           !
                           flop, &         !
                           w(RN+nnext) )   !

          kstat(2,l+1) = max(kstat(2,l+1),1)
          kstat(3,l+1) = kstat(3,l+1) + 1
          kstat(1,l+1) = kstat(1,l+1) + 1

       else if ( spd ) then

          call dagmg_FlexCG_inner( nnext, &  !
                                   w(XN), &  !
                                   w(RN), &  !
                                   l + 1, &  !
                                   flop )    !

       else

          call dagmg_GCR_inner( nnext, &  !
                                w(XN), &  !
                                w(RN), &  !
                                l + 1, &  !
                                flop )    !

       end if

       call dagmg_prolaggl( N, &           !
                            nnext, &       !
                            X, &           !
                            w(XN), &       !
                            dt(l) % ind )  !

    else

       X(1:N) = 0.0_rk

    end if

    return

  end subroutine dagmg_CGcorr

  !============================================================================!
 
  recursive subroutine dagmg_prec( N, &
                                   X, &
                                   B, &
                                   a, &
                                   ja, &
                                   ia, &
                                   l, &
                                   flop, &
                                   R )

    ! Arguments.
    integer :: n
    real (kind=rk), dimension(N) :: X
    real (kind=rk), dimension(N) :: B
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(N+1) :: ia
    integer :: l
    real (kind=rk) :: flop
    real (kind=rk) :: R(N,*)

    ! Local variables.
    integer :: is

    if ( nsmooth == 2 ) then

       call dagmg_sgsolve1( N, &                !
                            B, &                !
                            R(1,2), &           !
                            R, &                !
                            X, &                !
                            a, &                !
                            ja, &               !
                            ia, &               !
                            dt(l) % p, &        !
                            dt(l) % idiag, &    !
                            flop, &             !
                            dt(l) % inloc, &    !
                            dt(l) % lstout, &   !
                            dt(l) % ilstout, &  !
                            dt(l) % lstin, &    !
                            dt(l) % ilstin )    !

       call dagmg_CGcorr( N, &      !
                          X, &      !
                          R, &      !
                          l, &      !
                          flop, &   !
                          R(1,3) )  !

       call dagmg_sgsolve2( N, &                !
                            X, &                !
                            R(1,2), &           !
                            R, &                !
                            a, &                !
                            ja, &               !
                            ia, &               !
                            dt(l) % p, &        !
                            dt(l) % idiag, &    !
                            flop, &             !
                            dt(l) % inloc, &    !
                            dt(l) % lstout, &   !
                            dt(l) % ilstout, &  !
                            dt(l) % lstin, &    !
                            dt(l) % ilstin )    !

    else

       X(1:N) = B(1:N)

       call dagmg_sgsolve( N, &              !
                           X, &              !
                           a, &              !
                           ja, &             !
                           ia, &             !
                           dt(l) % p, &      !
                           dt(l) % idiag, &  !
                           flop, &           !
                           dt(l) % inloc )   !

       do is = 2, ( nsmooth + 1 ) / 2

          call dagmg_rescalc( N, &                !
                              X, &                !
                              R(1,2), &           !
                              B, &                !
                              a, &                !
                              ja, &               !
                              ia, &               !
                              flop, &             !
                              dt(l) % inloc, &    !
                              dt(l) % lstout, &   !
                              dt(l) % ilstout, &  !
                              dt(l) % lstin, &    !
                              dt(l) % ilstin )    !

          call dagmg_sgsolve( N, &              !
                              R(1,2), &         !
                              a, &              !
                              ja, &             !
                              ia, &             !
                              dt(l) % p, &      !
                              dt(l) % idiag, &  !
                              flop, &           !
                              dt(l) % inloc )   !

          X(1:N) = X(1:N) + R(1:N,2)
          flop = flop + real(N,rk)

       end do

       call dagmg_rescalc( N, &                !
                           X, &                !
                           R, &                !
                           B, &                !
                           a, &                !
                           ja, &               !
                           ia, &               !
                           flop, &             !
                           dt(l) % inloc, &    !
                           dt(l) % lstout, &   !
                           dt(l) % ilstout, &  !
                           dt(l) % lstin, &    !
                           dt(l) % ilstin )    !

       call dagmg_CGcorr( N, &       !
                          R(1,2), &  !
                          R, &       !
                          l, &       !
                          flop, &    !
                          R(1,3) )   !

       X(1:N) = X(1:N) + R(1:N,2)
       flop = flop + real(N,rk)

       do is = 1, nsmooth / 2

          call dagmg_rescalc( N, &                !
                              X, &                !
                              R(1,2), &           !
                              B, &                !
                              a, &                !
                              ja, &               !
                              ia, &               !
                              flop, &             !
                              dt(l) % inloc, &    !
                              dt(l) % lstout, &   !
                              dt(l) % ilstout, &  !
                              dt(l) % lstin, &    !
                              dt(l) % ilstin )    !

          call dagmg_sgsolve( N, &              !
                              R(1,2), &         !
                              a, &              !
                              ja, &             !
                              ia, &             !
                              dt(l) % p, &      !
                              dt(l) % idiag, &  !
                              flop, &           !
                              dt(l) % inloc )   !

          X(1:N) = X(1:N) + R(1:N,2)
          flop = flop + real(N,rk)

       end do

    end if
    
    return

  end subroutine dagmg_prec

  !============================================================================!
 
  subroutine dagmg_applyprec( N, &
                              f, &
                              X, &
                              a, &
                              ja, &
                              ia, &
                              flop )
 
    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: f
    real (kind=rk), dimension(N) :: X
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer :: ia(N+1)
    real (kind=rk) :: flop

    ! Local variables.
    real (kind=rk), allocatable :: S(:)
    integer :: allocation_status
    
    mritr = nwrkcum + 2 * N

    allocate ( S(nwrkcum+2*N), &
               stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_applyprec."
       stop
    end if

    call dagmg_prec( n, &     !
                     X, &     !
                     f, &     !
                     a, &     !
                     ja, &    !
                     ia, &    !
                     1, &     !
                     flop, &  !
                     S )      !

    kstat(3,1) = kstat(3,1) + 1

    deallocate (S)
    
    return

  end subroutine dagmg_applyprec

  !============================================================================!

  subroutine dagmg_FlexCG( N, &
                           f, &
                           X, &
                           ITER, &
                           RESID, &
                           a, &
                           ja, &
                           ia, &
                           init, &
                           flop )

    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: f
    real (kind=rk), dimension(N) :: X
    integer :: ITER
    real (kind=rk) :: RESID
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer :: ia(N+1)
    integer :: init
    real (kind=rk) :: flop

    ! Local variables.
    integer :: MAXIT, ierr
    real (kind=rk) :: TOL, BNORM, dum0
    real (kind=rk) :: ALPHA, BET0, RHO, dum(6)
    real (kind=rk), allocatable :: S(:)
    real (kind=rk), external :: ddot
    real (kind=rk), external :: dnrm2
    integer , parameter :: IONE = 1
    integer :: allocation_status
    
    mritr = nwrkcum + 5 * N

    allocate ( S(n+1:nwrkcum+6*N), &
               stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_FlexCG."
       stop
    end if
    
    flop = 0.0_rk
    kstat = 0

    if ( wfo ) then
       write(iout,940) IRANK
    end if

    if ( wff ) then
       write(iout,946) nsmooth
    end if
    
    TOL = RESID
    MAXIT = ITER
    RESID = 1.0_rk
    ITER = 0
    dum(6) = dnrm2( N, f, IONE )**2
    flop = flop + real(2*N,rk)
    
    ! compute initial residual
    
    if ( init == 1 ) then

       call dagmg_rescalc( N, &                !
                           x, &                !
                           f, &                !
                           f, &                !
                           a, &                !
                           ja, &               !
                           ia, &               !
                           flop, &             !
                           dt(1) % inloc, &    !
                           dt(1) % lstout, &   !
                           dt(1) % ilstout, &  !
                           dt(1) % lstin, &    !
                           dt(1) % ilstin )    !

       dum(5) = dnrm2(N, f, IONE)**2
       dum(2:3) = dum(5:6)
       BNORM = sqrt(dum(3))
       RESID = sqrt(dum(2))
       RESID = RESID / BNORM

       if ( wff .and. ( MAXIT <= 0 .or. RESID <= TOL ) ) then
          write(iout,900) 0, resid*bnorm, resid
       end if
    end if
    
    do while ( ITER < MAXIT .and. RESID > TOL )

       ITER = ITER + 1
       
       ! Preconditioner Solve.
       
       call dagmg_prec( n, &         !
                        S(1+3*N), &  !
                        f, &         !
                        a, &         !
                        ja, &        !
                        ia, &        !
                        1, &         !
                        flop, &      !
                        S(1+4*N) )   !
       
       ! Compute direction vector.
       
       if ( ITER > 1 ) then

          dum(1) = - ddot( N, S(1+N), IONE, S(1+3*N), IONE )
          BET0 = dum(1)
          BET0 = BET0 / RHO

          call daxpy( N, &         !
                      BET0, &      !
                      S(1+2*N), &  !
                      IONE, &      !
                      S(1+3*N), &  !
                      IONE )       !

          flop = flop + real(4*N,rk)

       endif

       call dcopy( N, &         !
                   S(1+3*N), &  !
                   IONE, &      !
                   S(1+2*N), &  !
                   IONE )       !

       call dagmg_matv( N, &                !
                        S(1+2*N), &         !
                        S(1+N), &           !
                        a, &                !
                        ja, &               !
                        ia, &               !
                        flop, &             !
                        dt(1) % inloc, &    !
                        dt(1) % lstout, &   !
                        dt(1) % ilstout, &  !
                        dt(1) % lstin, &    !
                        dt(1) % ilstin )    !

       dum(4) =  ddot( N, S(1+2*N), IONE, S(1+N), IONE )
       dum(5) =  ddot( N, S(1+2*N), IONE, f, IONE )
       flop = flop + real(4*N,rk)

       if ( ITER == 1 ) then

          dum(1:3) = dum(4:6)

          if ( init == 0 ) then
             BNORM = sqrt(dum(3))
          end if

          if ( wff ) then
             write(iout,900) 0, resid*bnorm, resid
          end if

       else

          dum(1:2) = dum(4:5)

       end if

       RHO = dum(1)
       ALPHA = dum(2) / RHO
       
       if ( ITER == 1 .and. init == 0 ) then

          call dcopy( N, &         !
                      S(1+2*N), &  !
                      IONE, &      !
                      X, &         !
                      IONE )       !

          call dscal( N, &      !
                      ALPHA, &  !
                      X, &      !
                      IONE )    !

          flop = flop + real(N,rk)

       else

          call daxpy( N, &         !
                      ALPHA, &     !
                      S(1+2*N), &  !
                      IONE, &      !
                      X, &         !
                      IONE )       !

          flop = flop + real(2*N,rk)

       end if

       call daxpy( N, &       !
                   -ALPHA, &  !
                   S(1+N), &  !
                   IONE, &    !
                   f, &       !
                   IONE )     !

       dum0 = dnrm2(N,f,IONE)**2
       RESID = dum0
       RESID = sqrt(RESID)
       RESID = RESID / BNORM

       if ( wff ) then
          write(iout,900) iter, resid*bnorm, resid
       end if

       flop = flop + real(4*N,rk)

    end do
    
    if ( resid > tol ) then
       if ( IRANK <= 0 ) then

          write(iout,'()')
          write(iout,950) iter
          write(iout,951)
          write(iout,'()')

       end if

       iter = -iter

    else

       if ( wff ) then
          write(iout,952) iter
          write(iout,'()')
       end if

    end if
    
    kstat(3,1) = abs(iter)
    
    deallocate(s)

    return

900 format('****  Iter=',i5,'        Resid=',e9.2,                 &
         '        Relat. res.=',e9.2)
940 format(i3,                                                     &
         '*SOLUTION: flexible conjugate gradient iterations (FCG(1))')
946 format(  '****          (with',i2,' SGS smoothing step per cycle)')
950 format('**** !!!   WARNING   !!!',I5,' ITERATIONS WERE')
951 format('**** INSUFFICIENT TO ACHIEVE THE DESIRED ACCURACY')
952 format('****  - Convergence reached in',I5,' iterations -')
    
  end subroutine dagmg_FlexCG

  !============================================================================!

  recursive subroutine dagmg_FlexCG_inner( N, &
                                           X, &
                                           R, &
                                           l, &
                                           flop )
 
    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: X
    real (kind=rk) :: R(N,*)
    integer :: l
    real (kind=rk) :: flop

    ! Local variables.
    integer :: ITER, ierr
    real (kind=rk) :: RESID, BNORM, dum0
    real (kind=rk) :: alpha1, alpha2, bet0, rho1, rho2, gamm0, dum(6)
    real (kind=rk), external :: ddot
    real (kind=rk), external :: dnrm2
    integer, parameter :: IONE = 1
    
    ! AT MOST 2 ITERATIONS
    
    dum(6) = dnrm2(N, R, IONE)**2
    ITER = 1
    
    ! Preconditioner Solve.
    
    call dagmg_prec( N, &           !
                     X, &           !
                     R, &           !
                     dt(l) % a, &   !
                     dt(l) % ja, &  !
                     dt(l) % ia, &  !
                     l, &           !
                     flop, &        !
                     R(1,2) )       !
    
    call dagmg_matv( N, &                !
                     X, &                !
                     R(1,2), &           !
                     dt(l) % a, &        !
                     dt(l) % ja, &       !
                     dt(l) % ia, &       !
                     flop, &             !
                     dt(l) % inloc, &    !
                     dt(l) % lstout, &   !
                     dt(l) % ilstout, &  !
                     dt(l) % lstin, &    !
                     dt(l) % ilstin )    !
    
    dum(4) = ddot( N, X, IONE, R(1,2), IONE )
    dum(5) = ddot( N, X, IONE, R, IONE )
    dum(1:3) = dum(4:6)
    BNORM = sqrt(dum(3))
    rho1 = dum(1)
    alpha1 = dum(2)
    
    bet0 = alpha1 / rho1

    call daxpy( N, &       !
                -bet0, &   !
                R(1,2), &  !
                IONE, &    !
                R, &       !
                IONE )     !

    dum0 = dnrm2(N,R,IONE)**2
    RESID = dum0
    RESID = sqrt(RESID) / BNORM

    if ( RESID <= resi ) then

       call dscal( N, &     !
                   bet0, &  !
                   X, &     !
                   IONE )   !

       kstat(2,l) = max(kstat(2,l),iter)
       kstat(3,l) = kstat(3,l) + iter
       flop = flop + real(11*N,rk)

       return

    end if
    
    ITER = 2
    
    ! Preconditioner Solve.
    
    call dagmg_prec( N, &           !
                     R(1,3), &      !
                     R, &           !
                     dt(l) % a, &   !
                     dt(l) % ja, &  !
                     dt(l) % ia, &  !
                     l, &           !
                     flop, &        !
                     R(1,4) )       !
    
    call dagmg_matv( N,R(1,3), &         !
                     R(1,4), &           !
                     dt(l) % a, &        !
                     dt(l) % ja, &       !
                     dt(l) % ia, &       !
                     flop, &             !
                     dt(l) % inloc, &    !
                     dt(l) % lstout, &   !
                     dt(l) % ilstout, &  !
                     dt(l) % lstin, &    !
                     dt(l) % ilstin )    !
    
    dum(4) = ddot( N, R(1,2), IONE, R(1,3), IONE )
    dum(5) = ddot( N, R(1,3), IONE, R, IONE )
    dum(6) = ddot( N, R(1,3), IONE, R(1,4), IONE ) 
    dum(1:3) = dum(4:6)
    gamm0 = dum(1)
    alpha2 = dum(2)
    rho2 = dum(3)
    
    rho2 = rho2 - gamm0 * gamm0 / rho1
    
    bet0 = ( alpha1 - alpha2 * gamm0 / rho2 ) / rho1

    call dscal( N, &     !
                bet0, &  !
                X, &     !
                IONE )   !

    bet0 = alpha2 / rho2

    call daxpy( N, &       !
                bet0, &    !
                R(1,3), &  !
                IONE, &    !
                X, &       !
                IONE )     !

    RESID = RESID * RESID         ! crude estimation for statistic only
    flop = flop + real(19*N,rk)
    
    kstat(2,l) = max(kstat(2,l),iter)
    kstat(3,l) = kstat(3,l) + iter

    if ( RESID > resi )  kstat(1,l) = kstat(1,l) + 1
    
    return
    
  end subroutine dagmg_FlexCG_inner

  !============================================================================!

  subroutine dagmg_GCR( N, &
                        f, &
                        X, &
                        ITER, &
                        RESID, &
                        a, &
                        ja, &
                        ia, &
                        init, &
                        flop )
  
    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: f
    real (kind=rk), dimension(N) :: X
    integer :: ITER
    real (kind=rk) :: RESID
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(N+1) :: ia
    integer :: init
    real (kind=rk) :: flop

    ! Local variables.
    integer :: MAXIT, i, itm, kc, ku, kfree, iv, iR, ifree, irst, ierr
    real (kind=rk) :: ALPHA, BET0
    real (kind=rk) :: RESID2, BNORM2,  RHO, TRS
    real (kind=rk) ::  TOL, TOL2BNORM2, TOLT, dum0
    real (kind=rk) :: dum(6)
    real (kind=rk), allocatable :: S(:)
    real (kind=rk), external :: ddot
    real (kind=rk), external :: dnrm2
    integer, parameter :: IONE = 1
    integer :: itm1, m, info
    integer :: allocation_status
    
    mritr = nwrkcum + N * ( 2 * nrst + 2 ) + ( ( nrst + 1 ) * nrst ) / 2 + nrst

    allocate( S(n+1:nwrkcum+N*(2*nrst+3)+((nrst+1)*nrst)/2+nrst), &
              stat = allocation_status )

    if ( allocation_status > 0 ) then
       print *, "Allocation error in dagmg_mod.f90/dagmg_GCR."
       stop
    end if
    
    flop = 0.0_rk
    kstat = 0

    if ( wfo ) then
       write(iout,938) IRANK, nrst
    end if

    if ( wff ) then
       write(iout,946) nsmooth
    end if
    
    m = min(nrst,ITER,N)
    kc = 1
    ku = kc + m
    kfree = ku + m
    iR = 0
    iv = iR + ( m * ( m + 1 ) ) / 2
    ifree = iv + m
    TRS = epsilon(1.0_rk)
    TRS = sqrt( sqrt(TRS) )
    TOL = RESID
    TOL2BNORM2 = TOL
    MAXIT = ITER
    RESID2 = 1.0_rk
    ITER = 0
    dum(6) = dnrm2(N, f, IONE)**2
    flop = flop + real(2*N,rk)
    
    ! compute initial residual.
    
    if ( init == 1 ) then

       call dagmg_rescalc( N, &                !
                           x, &                !
                           f, &                !
                           f, &                !
                           a, &                !
                           ja, &               !
                           ia, &               !
                           flop, &             !
                           dt(1) % inloc, &    !
                           dt(1) % lstout, &   !
                           dt(1) % ilstout, &  !
                           dt(1) % lstin, &    !
                           dt(1) % ilstin )    !

       dum(5) = dnrm2(N, f, IONE)**2
       dum(2:3) = dum(5:6)
       BNORM2 = dum(3)
       RESID2 = dum(2)
       TOL2BNORM2 = TOL * TOL * BNORM2

       if ( wff .and. ( MAXIT <= 0 .or. RESID2 <= TOL2BNORM2 ) ) then
          write(iout,900) 0, 0, sqrt(resid2), sqrt(resid2/bnorm2)
       end if

    end if
    
    itm  = -1
    irst = 0

    do while ( ITER < MAXIT .and. RESID2 > TOL2BNORM2 )

       itm  = itm  + 1
       ITER = ITER + 1
       
       ! Restarting

       if ( itm == m ) then

          call dtptrs( 'U', &              !
                       'N', &              !
                       'U', &              !
                       m, &                !
                       IONE, &             !
                       S(1+iR+kfree*N), &  !
                       S(1+iv+kfree*N), &  !
                       m, &                !
                       info )              !

          if ( irst == 0 .and. init == 0 ) then

             call dgemv( 'N', &              !
                         N, &                !
                         m, &                !
                         1.0_rk, &           !
                         S(1+ku*N), &        !
                         N, &                !
                         S(1+iv+kfree*N), &  !
                         IONE, &             !
                         0.0_rk, &           !
                         X, &                !
                         IONE )              !

             flop = flop + real(2*m*N+m*(m+1),rk)

          else

             call dgemv( 'N', &             !
                         N, &               !
                         m, &               !
                         1.0_rk, &          !
                         S(1+ku*N), &       !
                         N, &               !
                         S(1+iv+kfree*N), & !
                         IONE, &            !
                         0.0_rk, &          !
                         S(1+kc*N), &       !
                         IONE )             !

             do i = 1, N
                X(i) = X(i) + S(1+i-1+kc*N)
             end do

             flop = flop + real((2*m+1)*N+m*(m+1))

          end if

          itm = 0
          irst = irst + 1

       end if
       
       ! Preconditioner Solve & MATVEC.
       
       call dagmg_prec( N, &                  !
                        S(1+(ku+itm)*N), &    !
                        f, &                  !
                        a, &                  !
                        ja, &                 !
                        ia, &                 !
                        1, &                  !
                        flop, &               !
                        S(1+ifree+kfree*N) )  !

       call dagmg_matv( N, &                !
                        S(1+(ku+itm)*N), &  !
                        S(1+(kc+itm)*N), &  !
                        a, &                !
                        ja, &               !
                        ia, &               !
                        flop, &             !
                        dt(1) % inloc, &    !
                        dt(1) % lstout, &   !
                        dt(1) % ilstout, &  !
                        dt(1) % lstin, &    !
                        dt(1) % ilstin )    !
       
       ! Gram-Schmidt.
       
       if ( itm > 0 ) then

          do i = 0, itm - 1

             dum(1) = ddot( N, S(1+(kc+i)*N), IONE, S(1+(kc+itm)*N), IONE )
             bet0 = dum(1)
             bet0 = bet0 / S(1+iR+i+(i*(i+1))/2+kfree*N)
             S(1+iR+i+(itm*(itm+1))/2+kfree*N) = bet0

             call daxpy( N, &                !
                         -bet0, &            !
                         S(1+(kc+i)*N), &    !
                         IONE, &             !
                         S(1+(kc+itm)*N), &  !
                         IONE )              !

             flop = flop + real(4*N,rk)

          end do

       end if
       
       ! no normalisation: record norm instead.
       
       dum(4) = dnrm2( N, S(1+(kc+itm)*N), IONE )**2
       dum(5) = ddot( N, S(1+(kc+itm)*N), IONE, f, IONE )

       if ( ITER == 1 ) then

          dum(1:3) = dum(4:6)

          if ( init == 0 ) then

             BNORM2 = dum(3)
             RESID2 = BNORM2
             TOL2BNORM2 = TOL * TOL * BNORM2

          end if

          if ( wff ) then
             write(iout,900) 0, 0,sqrt(resid2),sqrt(resid2/bnorm2)
          end if

          TOLT = max( TOL2BNORM2, TRS * RESID2 )

       else

          dum(1:2) = dum(4:5)

       end if

       rho = dum(1)
       alpha = dum(2)
       S(1+iR+itm+(itm*(itm+1))/2+kfree*N) = rho
       bet0 = alpha / rho
       S(1+iv+itm+kfree*N) = bet0
       
       call daxpy( N, &                !
                   -bet0, &            !
                   S(1+(kc+itm)*N), &  !
                   IONE, &             !
                   f, &                !
                   IONE )              !

       flop = flop + real(6*N,rk)
       
       RESID2 = RESID2 - alpha * alpha / rho

       if ( RESID2 <= TOLT ) then

          dum0 = dnrm2( N, f, IONE )**2
          RESID2 = dum0
          flop = flop + real(2*N,rk)
          TOLT = max( TOL2BNORM2, TRS * RESID2 )

       end if

       if ( wff ) then
          write(iout,900) iter,irst,sqrt(resid2),sqrt(resid2/bnorm2)
       end if
       
    end do
    
    if ( itm >= 0 ) then

       itm1=itm+1

       call dtptrs( 'U', &
                    'N', &
                    'U', &
                    itm1, &
                    IONE, &
                    S(1+iR+kfree*N), &
                    S(1+iv+kfree*N), &
                    m, &
                    info )

       if ( irst == 0 .and. init == 0 ) then

          call dgemv( 'N', &              !
                      N, &                !
                      itm1, &             !
                      1.0_rk, &           !
                      S(1+ku*N), &        !
                      N, &                !
                      S(1+iv+kfree*N), &  !
                      IONE, &             !
                      0.0_rk, &           !
                      X, &                !
                      IONE )              !

          flop = flop + real(2*(itm+1)*N+(itm+1)*(itm+2),rk)

       else

          call dgemv( 'N', &              !
                      N, &                !
                      itm1, &             !
                      1.0_rk, &           !
                      S(1+ku*N), &        !
                      N, &                !
                      S(1+iv+kfree*N), &  !
                      IONE, &             !
                      0.0_rk, &           !
                      S(1+kc*N), &        !
                      IONE )              !

          do i = 1, N
             X(i) = X(i) + S(1+i-1+kc*N)
          end do

          flop = flop + real((2*(itm+1)+1)*N+(itm+1)*(itm+2),rk)

       end if
    end if
    
    RESID = sqrt( RESID2 / BNORM2 )

    if ( resid > tol ) then

       if ( IRANK <= 0 ) then

          write(iout,'()')
          write(iout,950) iter
          write(iout,951)
          write(iout,'()')

       end if

       iter = -iter

    else

       if ( wff ) then
          write(iout,952) iter
          write(iout,'()')
       end if

    end if
    
    kstat(3,1) = abs(iter)
    
    deallocate (S)
    
    return

900 format('****  Iter=',i5,' (',i2,' rest.)        Resid=',e9.2,    &
         '        Relat. res.=',e9.2)
938 format(i3,'*SOLUTION: GCR iterations (GCR(',i2,'))')
946 format(  '****          (with',i2,' SGS smoothing step per cycle)')
950 format('**** !!!   WARNING   !!!',I5,' ITERATIONS WERE')
951 format('**** INSUFFICIENT TO ACHIEVE THE DESIRED ACCURACY')
952 format('****  - Convergence reached in',I5,' iterations -')
    
  end subroutine dagmg_GCR

  !============================================================================!
 
  recursive subroutine dagmg_GCR_inner( N, &
                                        X, &
                                        R, &
                                        l, &
                                        flop )
    
    ! Arguments.
    integer :: N
    real (kind=rk), dimension(N) :: X
    real (kind=rk) :: R(N,*)
    integer :: l
    real (kind=rk) :: flop

    ! Local variables.
    integer :: ITER, ierr
    real (kind=rk) :: RESID, BNORM, rho1, rho2
    real (kind=rk) :: alpha1, alpha2, bet0, gamm0, dum(6)
    real (kind=rk), external :: ddot
    real (kind=rk), external :: dnrm2
    integer, parameter :: IONE = 1
    
    ! AT MOST 2 ITERATIONS
    
    dum(6) = dnrm2(N, R, IONE)**2
    ITER = 1
    
    ! Preconditioner Solve & MATVEC.
    
    call dagmg_prec( N, &           !
                     X, &           !
                     R, &           !
                     dt(l) % a, &   !
                     dt(l) % ja, &  !
                     dt(l) % ia, &  !
                     l, &           !
                     flop, &        !
                     R(1,2) )       !

    call dagmg_matv( N, &                !
                     X, &                !
                     R(1,2), &           !
                     dt(l) % a, &        !
                     dt(l) % ja, &       !
                     dt(l) % ia, &       !
                     flop, &             ! 
                     dt(l) % inloc, &    !
                     dt(l) % lstout, &   !
                     dt(l) % ilstout, &  !
                     dt(l) % lstin, &    !
                     dt(l) % ilstin )    !
    
    dum(4) = dnrm2( N, R(1,2), IONE )**2
    dum(5) = ddot( N, R(1,2), IONE, R, IONE )
    dum(1:3) = dum(4:6)
    BNORM = dum(3)
    rho1 = dum(1)
    alpha1 = dum(2)
    
    bet0 = alpha1 / rho1
    RESID = BNORM - alpha1 * bet0
    
    if ( RESID <= resi * resi * BNORM ) then

       call dscal( N, &     !
                   bet0, &  !
                   X, &     !
                   IONE )   !

       flop = flop + real(7*N,rk)
       kstat(2,l) = max(kstat(2,l),iter)
       kstat(3,l) = kstat(3,l) + iter
       RESID = sqrt( max( 0.0_rk, RESID / BNORM ) )

       return

    end if
    
    call daxpy( N, &       !
                -bet0, &   !
                R(1,2), &  !
                IONE, &    !
                R, &       !
                IONE )     !
    
    ITER = 2
    
    ! Preconditioner Solve & MATVEC.
    
    call dagmg_prec( N, &           !
                     R(1,3), &      !
                     R, &           !
                     dt(l) % a, &   !
                     dt(l) % ja, &  !
                     dt(l) % ia, &  !
                     l, &           !
                     flop, &        !
                     R(1,4) )       !

    call dagmg_matv( N, &                !
                     R(1,3), &           !
                     R(1,4), &           !
                     dt(l) % a, &        !
                     dt(l) % ja, &       !
                     dt(l) % ia, &       !
                     flop, &             !
                     dt(l) % inloc, &    !
                     dt(l) % lstout, &   !
                     dt(l) % ilstout, &  !
                     dt(l) % lstin, &    !
                     dt(l) % ilstin )    !
    
    dum(4) = ddot(N,R(1,2),IONE,R(1,4),IONE)
    dum(5) = ddot(N,R(1,4),IONE,R,IONE)
    dum(6) = dnrm2(N,R(1,4),IONE)**2
    dum(1:3) = dum(4:6)
    gamm0 = dum(1)
    alpha2 = dum(2)
    rho2 = dum(3)
    
    rho2 = rho2 - gamm0 * gamm0 / rho1
    bet0 = ( alpha1 - alpha2 * gamm0 / rho2 ) / rho1

    call dscal( N, &     !
                bet0, &  !
                X, &     !
                IONE )   !

    bet0 = alpha2 / rho2

    call daxpy( N, &       !
                bet0, &    !
                R(1,3), &  !
                IONE, &    !
                X, &       !
                IONE )     !

    RESID = RESID - alpha2 * bet0
    RESID = sqrt( max( 0.0_rk, RESID / BNORM ) )
    
    flop = flop + real(17*N,rk)
    
    kstat(2,l) = max( kstat(2,l), iter )
    kstat(3,l) = kstat(3,l) + iter

    if ( RESID > resi )  kstat(1,l) = kstat(1,l) + 1
    
    return
    
  end subroutine dagmg_GCR_inner

  !============================================================================!

  subroutine dagmg_aggl2 ( n, &
                           nc, &
                           a, &
                           ja, &
                           ia, &
                           lcg, &
                           ind, &
                           lpair, &
                           deg, &
                           next, &
                           prev, &
                           first, &
                           odmax, &
                           trs, &
                           checkdd, &
                           trspos, &
                           maxdg )
    
    ! Arguments.
    integer :: n
    integer :: nc
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    integer, dimension(4*n) :: lcg
    integer, dimension(n) :: ind
    integer, dimension(n) :: lpair
    integer, dimension(n) :: deg
    integer, dimension(n) :: next
    integer, dimension(n) :: prev
    integer, dimension(0:n-1) :: first
    real (kind=rk), dimension(n) :: odmax
    real (kind=rk) :: trs
    real (kind=rk) :: checkdd
    real (kind=rk) :: trspos
    integer :: maxdg

    ! Local variables.
    real (kind=rk) :: vald
    real (kind=rk) :: val, tent, odm, valp
    integer :: mindg, i, j, jj, k, ipair, dg, isel, nmark, nm1, nm2, kd

    nmark = 0
    nm1 = - n - 1
    nm2 = - n - 2
    maxdg = 0
    nc = 0

    if ( checkdd .gt. 0.0_rk ) then

       do i = 1, n

          j    = ia(i)
          jj   = ia(i + 1) - 1
          dg   = jj - j
          val  = 0.0_rk
          odm  = 0.0_rk
          valp = 0.0_rk

          do k = j, jj

             if ( ja(k) .eq. i ) then
                vald = a(k)
             else
                odm  = max( odm, abs(a(k)) )
                valp = max( valp, a(k) )
                val  = val + abs(a(k))
             endif

          enddo

          if ( dg .eq. 0 .or. abs(vald) .gt. checkdd * val ) then

             ind (i) = nm1
             nmark = nmark + 1
             odmax (i) = 0.0_rk
             deg (i) = nm1

          else if ( valp .gt. trspos * vald ) then
 
             ind(i) = nm2
             deg (i) = 0
             odmax (i) = - trs * odm

          else

             ind (i) = 0
             deg (i) = 0
             odmax (i) = - trs * odm

          endif

       enddo

    else

       do i = 1, n

          j = ia (i)
          jj = ia (i + 1) - 1
          dg = jj - j
          odm = 0.0_rk
          valp= 0.0_rk

          do k = j, jj

             if ( ja(k) .eq. i ) then
                vald = a(k)
             else
                odm  = max( odm, abs(a(k)) )
                valp = max( valp, a(k) )
             endif

          enddo

          if ( dg .eq. 0 ) then

             ind (i) = nm1
             nmark = nmark + 1
             odmax (i) = 0.0_rk
             deg (i) = nm1

          else if ( valp .gt. trspos * vald ) then

             ind(i) = nm2
             deg (i) = 0

          else

             ind (i) = 0
             deg (i) = 0
             odmax (i) = - trs * odm

          endif

       enddo
    endif

    do i = 1, n
       do k = ia (i), ia (i + 1) - 1
          if ( ( ind(i) .eq. 0 .or. ind(i) .eq. nm2 ) .and. &
               a(k) .lt. odmax(i) .and. &
               ja(k) .ne. i  ) then

             dg = deg (ja (k) ) + 1
             deg (ja (k) ) = dg
             maxdg = max( maxdg, dg )

          endif
       enddo
    enddo

    IF(n>0) THEN
    do i = 0, maxdg
       first (i) = 0
    enddo
    ENDIF

    do i = n, 1, - 1

       if ( ind(i) .eq. 0 .or. ind(i) .eq. nm2 ) then

          dg = deg(i)

          if ( first(dg) .gt. 0 ) prev ( first(dg) ) = i

          next (i) = first (dg)
          prev (i) = 0
          first (dg) = i

       endif

    enddo

    do while ( nmark .lt. n )

       mindg = - 1
       ipair = 0

       do while ( ipair .eq. 0 )

          mindg = mindg + 1

          ! if ( mindg .gt. maxdg ) stop 'isel'

          if ( first(mindg) .gt. 0 ) then

             ipair = first(mindg)
             first(mindg) = next (ipair)

             if ( next(ipair) .gt. 0 ) prev (next (ipair) ) = 0

          endif

       enddo

       nc = nc + 1
       ind (ipair) = nc
       isel = 0

       if ( ind(ipair) .eq. nm2 ) goto 20

       val = 0.0_rk

       do i = ia (ipair), ia(ipair+1) - 1

          j = ja (i)

          if ( ind(j) .eq. 0 .and. a(i) .lt. odmax(ipair) ) then

             tent = a(i)

             if ( tent .lt. 1.0001_rk * val ) then
                isel = j
                val = tent
             endif

          endif

       enddo

20     continue

       if ( isel .eq. 0 ) then

          lcg (nc) = ipair
          lpair (nc) = 0
          nmark = nmark + 1

       else

          ind (ipair) = - nc
          ind (isel) = nc
          lcg (nc) = isel
          nmark = nmark + 2
          lpair(nc) = ipair
          dg = deg(isel)

          if ( prev(isel) .gt. 0 ) then
             next( prev(isel) ) = next(isel)
          else
             first(dg) = next(isel)
          endif

          if ( next(isel) .gt. 0 ) prev(next (isel) ) = prev(isel)

          do i = ia(isel), ia (isel + 1) - 1

             j = ja(i)

             if ( ( ind(j) .eq. 0 .or. ind(j) .eq. nm2 ) .and. &
                  a(i) .lt. odmax(isel) ) then

                dg = deg (j)
                dg = dg - 1
                deg (j) = dg

                ! IF (dg .LT. maxdeg) THEN

                if ( prev(j) .gt. 0 ) then
                   next( prev(j) ) = next(j)
                else
                   first(dg+1) = next(j)
                endif

                if ( next (j) .gt. 0 ) prev(next(j)) = prev (j)
                if ( first(dg) .gt. 0 ) prev(first(dg)) = j

                next(j) = first(dg)
                prev(j) = 0
                first(dg) = j

                ! ENDIF

             end if
          enddo
       endif

       do i = ia (ipair), ia(ipair + 1) - 1

          j = ja(i)

          IF ( (ind(j) .EQ. 0 .OR. ind(j) .EQ. nm2 ) .AND. &
               a(i) .LT. odmax(ipair) ) THEN

             dg = deg (j)
             dg = dg - 1
             deg (j) = dg

             ! if ( dg .lt. maxdeg ) then

             if ( prev(j) .gt. 0 ) then
                next(prev(j) ) = next(j)
             else
                first(dg+1) = next(j)
             endif

             if ( next(j) .gt. 0 ) prev(next(j)) = prev(j)
             if ( first(dg) .gt. 0 ) prev(first(dg)) = j

             next(j) = first(dg)
             prev(j) = 0
             first (dg) = j

             ! ENDIF

          endif
       enddo
    enddo

    return

  end subroutine dagmg_aggl2

  !============================================================================!

  subroutine dagmg_aggl4 ( n, &
                           nc, &
                           a, &
                           ja, &
                           ia, &
                           lcg, &
                           ind, &
                           lcg1, &
                           lpair1, &
                           lcg2, &
                           lpair2, &
                           odmax, &
                           trs, &
                           a2, &
                           ja2, &
                           ia2, &
                           checkdd, &
                           trspos )
    
    ! Arguments.
    integer :: n
    integer :: nc
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    integer, dimension(4*n) :: lcg
    integer, dimension(n) :: ind
    integer, dimension(n) :: lcg1
    integer, dimension(n) :: lpair1
    integer, dimension(n) :: lcg2
    integer, dimension(n) :: lpair2
    real (kind=rk), dimension(n) :: odmax
    real (kind=rk) :: trs
    real (kind=rk) :: a2(*)
    integer :: ja2(*)
    integer, dimension(n+1) :: ia2
    real (kind=rk) :: checkdd
    real (kind=rk) :: trspos

    ! Local variables.
    real (kind=rk) :: vald, wextmx, wextmn
    real (kind=rk) :: wextmax, wextmin
    integer :: nc1, i, jj, jc, jcol, kb, jpos, nz, maxdg
    integer, pointer, dimension(:) :: ifirst
    integer :: allocation_status

    call dagmg_aggl2( n, &        !
                      nc1, &      !
                      a, &        !
                      ja, &       !
                      ia, &       !
                      lcg1, &     !
                      ind, &      !
                      lpair1, &   !
                      lcg2,  &    !
                      lpair2, &   !
                      ia2, &      !
                      lcg, &      !
                      odmax, &    !
                      trs, &      !
                      checkdd, &  !
                      trspos, &   !
                      maxdg )     !

    nz = 0
    ia2 (1) = 1

    do i = 1, nc1
       lcg(i) = 0
    enddo

    do i = 1, nc1

       jc = nc1 + 1
       jj = lcg1(i)

       do kb = ia(jj), ia(jj+1) - 1

          jcol = ja(kb)
          jcol = abs( ind(jcol) )

          ! if ( jcol .ne. i .and. jcol .le. nc1 ) then

          if ( jcol .le. nc1 ) then

             jpos = lcg(jcol)

             if ( jpos .eq. 0 ) then

                nz = nz + 1
                ja2(nz) = jcol
                lcg(jcol) = nz
                a2(nz) = a(kb)

             else

                a2(jpos) = a2(jpos) + a(kb)

             endif
          endif
       enddo

       jj = lpair1 (i)

       if ( jj .gt. 0 ) then

          do kb = ia(jj), ia(jj+1) - 1

             jcol = ja (kb)
             jcol = abs( ind(jcol) )

             ! if ( jcol .ne. i .and. jcol .le. nc1 ) then

             if ( jcol .le. nc1 ) then

                jpos = lcg(jcol)

                if ( jpos .eq. 0 ) then

                   nz = nz + 1
                   ja2 (nz) = jcol
                   lcg (jcol) = nz
                   a2 (nz) = a (kb)

                else

                   a2(jpos) = a2(jpos) + a(kb)

                endif

             endif

          enddo

       endif

       do kb = ia2(i), nz
          lcg (ja2(kb) ) = 0
       enddo

       ia2 (i + 1) = nz + 1

    enddo

    if ( maxdg + 1 .le. ia(n+1) - nz - 1 ) then

       call dagmg_aggl2( nc1, &         !
                         nc, &          !
                         a2, &          !
                         ja2, &         !
                         ia2, &         !
                         lcg2, &        !
                         lcg, &         !
                         lpair2, &      !
                         lcg(n+1), &    !
                         lcg(2*n+1), &  !
                         lcg(3*n+1), &  !
                         ja2(nz+1), &   !
                         odmax, &       !
                         trs, &         !
                         -1.0_rk, &     !
                         trspos, &      !
                         maxdg )        !

    else

       allocate( ifirst(maxdg+1), &
                 stat = allocation_status )

       if ( allocation_status > 0 ) then
          print *, "Allocation error in dagmg_mod.f90/dagmg_aggl4."
          stop
       end if

       call dagmg_aggl2 ( nc1, &         !
                          nc, &          !
                          a2, &          !
                          ja2, &         !
                          ia2, &         !
                          lcg2, &        !
                          lcg, &         !
                          lpair2, &      !
                          lcg(n+1), &    !
                          lcg(2*n+1), &  !
                          lcg(3*n+1), &  !
                          ifirst, &      !
                          odmax, &       !
                          trs, &         !
                          -1.0_rk, &     !
                          trspos, &      !
                          maxdg )        !

       deallocate(ifirst)

    end if

    do i = 1, n

       jc = ind (i)
       jcol = ABS (jc)

       if ( jcol .le. nc1 ) then
          if ( jc .gt. 0 ) then
             ind (i) = lcg(jcol)
          else
             ind (i) = - abs( lcg(jcol) )
          endif
       endif

    enddo

    call dagmg_setlcg ( nc, &      !
                        lcg, &     !
                        lcg1, &    !
                        lpair1, &  !
                        lcg2, &    !
                        lpair2 )   !

    return

  end subroutine dagmg_aggl4

  !============================================================================!
  
  subroutine dagmg_setlcg ( nc, &
                            lcg, &
                            lcg1, &
                            lpair1, &
                            lcg2, &
                            lpair2 )
  
    ! Arguments.
    integer :: nc
    integer, dimension(4,nc) :: lcg
    integer :: lcg1(*)
    integer :: lpair1(*)
    integer :: lcg2(*)
    integer :: lpair2 (*)

    ! Local variables.
    integer :: i, j1, j2


    do i = 1, nc

       j1 = lcg2 (i)
       j2 = lpair2 (i)
       lcg (1, i) = lcg1 (j1)
       lcg (2, i) = lpair1 (j1)

       if ( j2 .gt. 0 ) then

          lcg (3,i) = lcg1 (j2)
          lcg (4,i) = lpair1 (j2)

       else

          lcg (3,i) = 0
          lcg (4,i) = 0

       endif

    enddo

    return

  end subroutine dagmg_setlcg

  !============================================================================!
 
  subroutine dagmg_setsgs ( n, &
                            a, &
                            ja, &
                            ia, &
                            p, &
                            idiag, &
                            w, &
                            iw, &
                            inloc )

    ! Arguments.
    integer :: n
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk), dimension(n) :: p
    integer, dimension(n) :: idiag
    real (kind=rk) :: w(*)
    integer :: iw(*)
    integer :: inloc(n)

    ! Local variables.
    real (kind=rk) :: t
    integer :: i, j, k, ipos, nzu

    do i = 1, n

       ! find diag, save upper part, concatenate lower part.

       ipos = ia (i)
       nzu = 0

       do k = ia (i), ia (i + 1) - 1

          j = ja (k)

          if (j.gt.i) then

             nzu = nzu + 1
             w (nzu) = a (k)
             iw (nzu) = j

          elseif (j.lt.i) then

             a (ipos) = a (k)
             ja (ipos) = j
             ipos = ipos + 1

          else

             p (i) = a (k)

          endif

       enddo
       
       ! copy back diagonal entry.

       idiag (i) = ipos
       a (ipos) = p (i)
       ja (ipos) = i
       
       ! copy back upper part.

       do k = 1, nzu
          ipos = ipos + 1
          a (ipos) = w (k)
          ja (ipos) = iw (k)
       enddo
       
       ! save inverse.

       t = p(i)
       p (i) = 1.0_rk / t

    enddo

    return

  end subroutine dagmg_setsgs

  !============================================================================!
 
  subroutine dagmg_matv ( n, &
                          x, &
                          y, &
                          a, &
                          ja, &
                          ia, &
                          flop, &
                          inloc, &
                          lstout, &
                          ilstout, &
                          lstin, &
                          ilstin )
    
    ! Arguments.
    integer :: n
    real (kind=rk), dimension(n) :: x
    real (kind=rk), dimension(n) :: y
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk) :: flop
    integer, dimension(n) :: inloc
    integer :: lstout(*)
    integer :: ilstout(*)
    integer :: lstin(0:*)
    integer :: ilstin(*)
    
    ! Local variables.
    integer :: i, kk, k1, k2, ier
    real (kind=rk) :: t

    do i = 1, n

       k1 = ia (i)
       t = a (k1) * x (ja (k1) )
       k2 = ia (i+1)

       do kk = k1 + 1, k2 - 1
          t = t + a (kk) * x (ja (kk) )
       enddo

       y (i) = t

    enddo

    flop = flop + real( 2 * ( ia(n + 1) - 1 ) - n, rk )

    return

  end subroutine dagmg_matv

  !============================================================================!

  subroutine dagmg_sgsolve1( n, &
                             b, &
                             t1, &
                             r, &
                             v, &
                             a, &
                             ja, &
                             ia, &
                             p, &
                             idiag, &
                             flop, &
                             inloc, &
                             lstout, &
                             ilstout, &
                             lstin, &
                             ilstin )
  
    ! Arguments.
    integer :: n
    real (kind=rk), dimension(n) :: b
    real (kind=rk), dimension(n) :: t1
    real (kind=rk), dimension(n) :: r
    real (kind=rk), dimension(n) :: v
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk), dimension(n) :: p
    integer, dimension(n) :: idiag
    real (kind=rk) :: flop
    integer, dimension(n) :: inloc
    integer :: lstout(*)
    integer :: ilstout(*)
    integer :: lstin(0:*)
    integer :: ilstin(*)

    ! Local variables.
    integer :: kk, j, k2, i, ier
    real (kind=rk) :: t
    
    t1(1) = p(1) * b(1)
    v(1) = b(1)

    do kk = 2, n

       t = b(kk)

       do j = ia(kk), idiag(kk) - 1
          t = t - a(j) * t1(ja(j))
       enddo

       v(kk) = t
       t1(kk) = p(kk) * t

    enddo
    
    do kk = n - 1, 1, - 1

       t = 0.0_rk
       k2 = ia(kk+1)

       do j = idiag(kk) + 1, k2 - 1
          t = t - a(j) * t1(ja(j))
       enddo

       t1(kk) = t1(kk) + p(kk) * t

    enddo

    r(1) = b(1) - v(1)

    do kk = 2, n

       t = b(kk) - v(kk)

       do j = ia(kk), idiag(kk) - 1
          t = t - a(j) * t1(ja(j))
       enddo

       r(kk) = t

    enddo
 
    flop = flop + real( 3 * (ia (n + 1) - 1), rk )

    return

  end subroutine dagmg_sgsolve1

  !============================================================================!

  subroutine dagmg_sgsolve2( n, &
                             x, &
                             t1, &
                             t3, &
                             a, &
                             ja, &
                             ia, &
                             p, &
                             idiag, &
                             flop, &
                             inloc, &
                             lstout, &
                             ilstout, &
                             lstin, &
                             ilstin )
 
    ! Arguments.
    integer :: n
    real (kind=rk), dimension(n) :: x
    real (kind=rk), dimension(n) :: t1
    real (kind=rk), dimension(n) :: t3
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk), dimension(n) :: p
    integer, dimension(n) :: idiag
    real (kind=rk) :: flop
    integer, dimension(n) :: inloc
    integer :: lstout(*)
    integer :: ilstout(*)
    integer :: lstin(0:*)
    integer :: ilstin(*)
    
    ! Local variables.
    integer :: kk, j, i, k2, ier
    real (kind=rk) :: t

    ! input : x=t2 ; converted to t1+t2
    ! output: x=t1+t1+t3
    
    x (n) = x (n) + t1 (n)
    t3 (n) = 0.0_rk

    do kk = n - 1, 1, - 1

       x(kk) = x(kk) + t1(kk)
       t = 0.0_rk

       do j = idiag(kk) + 1, ia(kk + 1) - 1
          t = t + a(j) * x(ja(j))
       enddo

       t3 (kk) = t

    enddo
    
    t3 (1) = p (1) * t3 (1)

    do kk = 2, n

       t = t3(kk)

       do j = ia(kk), idiag(kk) - 1
          t = t - a(j) * t3(ja(j))
       enddo

       t3(kk) = p(kk) * t

    enddo
    
    do kk = 1, n
       t3 (kk) = x(kk) + t3(kk)
    enddo
    
    do kk = n - 1, 1, - 1

       t = 0.0_rk
       k2 = ia(kk+1)

       do j = idiag(kk) + 1, k2 - 1
          t = t - a(j) * t3(ja(j))
       enddo

       t3(kk) = t3(kk) + p(kk) * t

    enddo
    
    do kk = 1, n
       x (kk) = x (kk) + t1 (kk) - t3 (kk)
    enddo

    flop = flop + real( 3 * ( ia(n+1) - 1 + n), rk )

    return

  end subroutine dagmg_sgsolve2

  !============================================================================!

  subroutine dagmg_sgsolve ( n, &
                             x, &
                             a, &
                             ja, &
                             ia, &
                             p, &
                             idiag, &
                             flop, &
                             inloc )
   
    ! Arguments.
    integer :: n
    real(kind=rk), dimension(n) :: x
    real(kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real(kind=rk), dimension(n) :: p
    integer, dimension(n) :: idiag
    real(kind=rk) :: flop
    integer, dimension(n) :: inloc

    ! Local variables.
    integer ::  kk, j, k2
    real(kind=rk) :: t


    x(1) = p(1) * x(1)

    do kk = 2, n

       t = x(kk)

       do j = ia(kk), idiag(kk) - 1

          t = t - a(j) * x(ja(j))

       enddo

       x(kk) = p(kk) * t

    enddo

    do kk = n - 1, 1, - 1

       t = 0.0_rk
       k2 = ia(kk+1)

       do j = idiag(kk) + 1, k2 - 1
          t = t - a(j) * x(ja(j))
       enddo

       x(kk) = x(kk) + p(kk) * t

    enddo

    flop = flop + real( 2 * ( ia(n + 1) - 1 ), rk )

    return

  end subroutine dagmg_sgsolve

  !============================================================================!

  subroutine dagmg_rescalc ( n, &
                             x, &
                             y, &
                             b, &
                             a, &
                             ja, &
                             ia, &
                             flop, &
                             inloc, &
                             lstout, &
                             ilstout, &
                             lstin, &
                             ilstin )
  
    ! Arguments.
    integer :: n
    real (kind=rk), dimension(n) :: x
    real (kind=rk), dimension(n) :: y
    real (kind=rk), dimension(n) :: b
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk) :: flop
    integer, dimension(n) :: inloc
    integer :: lstout(*)
    integer :: ilstout(*)
    integer :: lstin(0:*)
    integer :: ilstin(*)

    ! Local variables.
    integer :: i, kk, k1, k2, ier
    real (kind=rk) :: t

    do i = 1, n

       k1 = ia(i)
       t = b(i) - a(k1) * x(ja(k1))
       k2 = ia(i+1)

       do kk = k1 + 1, k2 - 1
          t = t - a(kk) * x(ja(kk))
       enddo

       y (i) = t

    enddo

    flop = flop + real( 2 * ( ia(n + 1) - 1 ), rk )

    return

  end subroutine dagmg_rescalc

  !============================================================================!

  subroutine dagmg_prolaggl ( n, &
                              nc, &
                              V, &
                              B, &
                              ind )

    ! Arguments.
    integer :: n
    integer :: nc
    real (kind=rk), dimension(n) :: V
    real (kind=rk), dimension(nc) :: B
    integer, dimension(n) :: ind

    ! Local variables.
    integer :: k, i

    do i = 1, n

       k = abs( ind(i) )

       if ( k .le. nc ) then
          v (i) = b(k)
       else
          v (i) = 0.0_rk
       endif

    enddo

    return

  end subroutine dagmg_prolaggl

  !============================================================================!

  subroutine dagmg_restaggl( n, &
                             nc, &
                             V, &
                             B, &
                             ind, &
                             flop )
    ! Arguments.
    integer :: n
    integer :: nc
    real (kind=rk), dimension(n) :: V
    real (kind=rk), dimension(nc) :: B
    integer, dimension(n) :: ind
    real (kind=rk) :: flop

    ! Local variables.
    integer :: k, i

    B(1:nc) = 0.0_rk

    do i = 1, n
       k = abs (ind (i) )
       if ( k .le. nc ) B(k) = B(k) + v(i)
    enddo

    flop = flop + real(n,rk)

    return

  end subroutine dagmg_restaggl

  !============================================================================!

  subroutine dagmg_setacg ( nc, &
                            lcg, &
                            a, &
                            ja, &
                            ia, &
                            ac, &
                            jac, &
                            iac, &
                            nzac, &
                            ind, &
                            iw )
   
    ! Arguments.
    integer :: nc
    integer, dimension(4,nc) :: lcg
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer :: ia(*)
    real (kind=rk) :: ac(*)
    integer :: jac(*)
    integer, dimension(nc+1) :: iac
    integer :: nzac
    integer :: ind(*)
    integer :: iw(*)

    ! Local variables.
    integer :: i, kk, jj, jc, kb, jcol, jpos

    do i = 1, nc
       iw (i) = 0
    enddo

    nzac = 0
    iac (1) = 1

    loop_i : do i = 1, nc

       loop_kk : do kk = 1, 4

          jj = lcg(kk,i)

          if ( jj .ne. 0 ) then

             do kb = ia(jj), ia(jj+1) - 1

                jc = ja(kb)
                jcol = abs(ind(jc))

                if ( jcol .le. nc ) then

                   jpos = iw(jcol)

                   if ( jpos .eq. 0 ) then

                      nzac = nzac + 1
                      jac(nzac) = jcol
                      iw(jcol) = nzac
                      ac(nzac) = a(kb)

                   else

                      ac(jpos) = ac(jpos) + a(kb)

                   endif

                endif

             enddo

          end if

       end do loop_kk

       do kb = iac (i), nzac
          iw (jac (kb) ) = 0
       enddo

       iac (i + 1) = nzac + 1

    end do loop_i

    return

  end subroutine dagmg_setacg

  !============================================================================!

  subroutine agmg( n, &
                   a, &
                   ja, &
                   ia, &
                   f, &
                   ijob, &
                   iprint, &
                   nrest, &
                   iter, &
                   tol )

    !--------------------------------------------------------------------------!
    ! Routine for bacward compatibility.
    !--------------------------------------------------------------------------!
 
    ! Arguments.
    integer :: n
    real (kind=rk) :: a(*)
    integer :: ja(*)
    integer, dimension(n+1) :: ia
    real (kind=rk), dimension(n) :: f
    integer :: ijob
    integer :: iprint
    integer :: nrest
    integer :: iter
    real (kind=rk) :: tol
    
    ! Local variables.
    real (kind=rk), dimension(n) :: x

    call dagmg( n, &       !
                a, &       !
                ja, &      !
                ia, &      !
                f, &       !
                x, &       !
                ijob, &    !
                iprint, &  !
                nrest, &   !
                iter, &    !
                tol )      !

    call dcopy( n, &  !
                x, &  !
                1, &  ! 
                f, &  !
                1 )   !

    return

  end subroutine agmg

  !============================================================================!

  subroutine dagmg_mestime ( id, &
                             cputm, &
                             eltm )

    !--------------------------------------------------------------------------!
    ! Timings.
    !--------------------------------------------------------------------------!

    ! Arguments.
    integer, intent(in) :: id
    real (kind=rk), intent(out), optional :: cputm, eltm
    
    ! Local variables.
    integer, save :: cpt_init(10) = -1, cpt_fin, cpt_max, freq,cpt
    real, save :: t1(10), t2
   


    if ( id > 0 ) then

       call cpu_time(t2)

       call system_clock( cpt_fin, &  ! out
                          freq, &     ! out
                          cpt_max )   ! out
 
       cpt = cpt_fin - cpt_init(id)
       if ( cpt_fin < cpt_init(id) ) cpt = cpt + cpt_max
       eltm = real( cpt, rk ) / freq
       cputm = real( t2 - t1(id), rk )
       
    else
       
       call system_clock( cpt_init(-id), &  ! out
                          freq, &           ! out
                          cpt_max )         ! out
  
       call cpu_time( t1(-id) )
       
    end if

    return

  end subroutine dagmg_mestime

  !============================================================================!

end module dagmg_mod

  !============================================================================!

  subroutine dagmg( n, &
                    a, &
                    ja, &
                    ia, &
                    f, &
                    x, &
                    ijob, &
                    iprint, &
                    nrest, &
                    iter, &
                    tol )
     
    use dagmg_mod

    !--------------------------------------------------------------------------!
    !
    !  Arguments
    !  =========
    !
    !  N       (input) INTEGER.
    !          The dimension of the matrix.
    !
    !  A       (input/output) REAL (kind=rk). Numerical values of the matrix.
    !  IA      (input/output) INTEGER. Pointers for every row.
    !  JA      (input/output) INTEGER. Column indices.
    !
    !              dagmg ASSUMES THAT ALL DIAGONAL ENTRIES ARE POSITIVE
    !
    !          Detailed description of the matrix format
    !
    !              On input, IA(I), I=1,...,N, refers to the physical start
    !              of row I. That is, the entries of row I are located
    !              in A(K), where K=IA(I),...,IA(I+1)-1. JA(K) carries the
    !              associated column indices. IA(N+1) must be defined in such
    !              a way that the above rule also works for I=N (that is,
    !              the last valid entry in arrays A,JA should correspond to
    !              index K=IA(N+1)-1). According to what is written
    !              above, dagmg assumes that some of these JA(K) (for
    !              IA(I)<= K < IA(I+1) ) is equal to I with corresponding
    !              A(K) carrying the value of the diagonal element,
    !              which must be positive.
    !
    !              A,IA,JA are "output" parameters because on exit the
    !              entries of each row may occur in a different order (The
    !              matrix is mathematically the same, but stored in
    !              different way).
    !
    !  F       (input/output) REAL (kind=rk).
    !          On input, the right hand side vector f.
    !          Overwritten on output.
    !          Significant only if IJOB==0, IJOB==10, IJOB==2 or IJOB==12
    !
    !  X       (input/output) REAL (kind=rk).
    !          On input and if IJOB==10 or IJOB==12, initial guess
    !             (for other values of IJOB, the default is used: the zero vector).
    !          On output, the computed solution.
    !
    ! IJOB     (input) INTEGER. Tells dagmg what has to be done.
    !          0: preprocessing + solve + release memory, no initial guess
    !         10: preprocessing + solve + release memory, initial guess in x(1:n)
    !          1: preprocessing only
    !          2: solve only, no initial guess
    !         12: solve only, initial guess in x(1:n)
    !          3: one application of the preconditioner
    !         -1: release memory
    !
    ! IPRINT   (input) INTEGER.
    !              Indicates the unit number where information is to be printed
    !              (N.B.: 5 is converted to 6). If nonpositive, only error
    !              messages are printed on standard output.
    !
    ! NREST    (input) INTEGER.
    !             Restart parameter for GCR (an implementation of GMRES)
    !             Nonpositive values are converted to NREST=10 (default)
    !         !!  If NREST==1, Flexible CG is used instead of GCR
    !             Should be used if and only if the matrix is
    !             symmetric and positive definite.
    !         !!  Significant only if IJOB == 0 or IJOB >= 2.
    !             If IJOB == 3, determines only the type of inner iterations
    !             (FCG if NREST == 1, GCR otherwise).
    !
    !  ITER    (input/output) INTEGER.
    !          On input, the maximum number of iterations. Should be positive.
    !          On output, actual number of iterations.
    !            If this number of iteration was insufficient to meet convergence
    !            criterion, ITER will be returned negative and equal to the
    !            opposite of the number of iterations performed.
    !          Significant only if IJOB==0, IJOB==10, IJOB==2 or IJOB==12
    !
    !  TOL     (input) REAL (kind=rk).
    !          The tolerance on residual norm. Iterations are stopped whenever
    !               || A*x-f || <= TOL* || f ||
    !          Should be positive and less than 1.0
    !          Significant only if IJOB==0, IJOB==10, IJOB==2 or IJOB==12
    !
    ! Remark !!!! Except insufficient number of iterations to achieve
    !             convergence (characterized by a negative value returned
    !             in ITER), all other detected errors are fatal and lead
    !             to a STOP statement.
    !--------------------------------------------------------------------------!

    ! Arguments.
    integer, intent(in) :: n
    real (kind=rk), intent(inout) :: a(*)
    integer, intent(inout) :: ja(*)
    integer, dimension(n+1), intent(inout) :: ia
    real (kind=rk), dimension(n), intent(inout) :: f, x
    integer, intent(in) :: ijob
    integer, intent(in) :: iprint
    integer, intent(in) :: nrest
    integer, intent(inout) :: iter
    real (kind=rk), intent(in) :: tol
    
    ! Local variables.
    integer, save :: nza
    real (kind=rk), save :: cputm = 0.0_rk, eltm = 0.0_rk, flop = 0.0_rk, memh
    logical, save :: preprocessed = .false., &
                     solve = .false.
    integer :: i, init
    real (kind=rk) :: cputmp, eltmp, resid
  
    wfo = .true.
    iout = iprint

    if ( iprint <= 0 ) then
       iout = 6
       wfo = .false.
    else if ( iprint == 5 ) then
       iout = 6
    end if
    
    wff = wfo .and. ( IRANK <= 0 )
    
    if ( mod( ijob, 10 ) >= 2 .and. .not. preprocessed ) then
       write (iout,1001) IRANK,ijob
       stop
    end if
    
    if ( ijob < 0 .and. solve ) goto 500
    if ( ijob < 0 ) goto 450
    if ( mod(ijob,10) >= 2 ) goto 300

    if ( preprocessed ) then

       call dagmg_relmem

       if ( .not. allzero ) then

          call dagmg_LAPACK( nn(nlev), &  !
                             a, &         !
                             ja, &        !
                             ia, &        !
                             f, &         !
                             -2, &        !
                             flop )       !

       end if

       preprocessed = .false.
       solve = .false.
       eltm = 0.0_rk
       cputm = 0.0_rk
       flop = 0.0_rk
       kstat = 0

    end if

    call dagmg_mestime( id = -1 )    ! in
    
    ! Initial settings.
    
    nza = ia(n+1) - ia(1)

    if ( huge(n) > 1.0e10_rk ) then

       rlenilen = real(8,rk) / real(real_len,rk)

    else

       rlenilen = real(4,rk) / real(real_len,rk)

    end if
    
    ! PREPROCESSING
    
    nlev = 0

    if (wfo) then
       write(iout,900) IRANK
    end if

    call dagmg_setup( 1, &   !
                      n, &   !
                      a, &   !
                      ja, &  !
                      ia )   !

    preprocessed = .true.
    memh = memr + memi * rlenilen
    
    call dagmg_mestime( 1, &       ! in
                        cputm, &   ! inout
                        eltmp )    ! inout

    if ( wfo ) then

       write(iout,960) IRANK,memax/nza,real_len,memax*real_len/(2**20)

       if ( mod(ijob,10) == 1 ) then
          write(iout,961) IRANK,memh/nza,real_len,memh*real_len/(2**20)
       end if

       ! CPU_TIME: next line may be uncommented if implemented
       !          (see subroutine mestime)
       !              write(iout,996) IRANK, cputmp

       write(iout,997) IRANK,eltmp
       write(iout,'()')

    end if
    
    if ( mod(ijob,10) == 1 ) return
    
    ! SOLUTION PROCESS.
    
300 continue

    call dagmg_mestime( id = -2 ) ! in

    resid = tol
    nrst = nrest
    init = max( ijob / 10, 0 )

    IF ( nrst <= 0 ) nrst = 10 ! Default restart paramater for GMRESR
    spd = .false.              ! nrst=1 => system assumed SPD, use FCG instead of GCR
    if ( nrst == 1 ) spd = .true.
    
    ! DIRECT SOLVE IF ONLY ONE LEVEL.

    if ( nlev == 1 ) then
       
       x(1:n)=f(1:n)

       call dagmg_LAPACK( n, &    !
                          a, &    !
                          ja, &   !
                          ia, &   !
                          x, &    !
                          2, &    !
                          flop )  !

       mritr = 0
       
       ! END OF DIRECT SOLVE.
       
       ! SINGLE APPLICATION OF MG PRECONDITIONER.

    else if ( mod(ijob,10) >= 3 ) then

       if ( wfo ) then
          write(iout,901) IRANK
       end if

       call dagmg_applyprec( n, &    !
                             f, &    !
                             x, &    !
                             a, &    !
                             ja, &   !
                             ia, &   !
                             flop )  !

       call dagmg_mestime( 2, &       ! in
                           cputmp, &  ! inout
                           eltmp )    ! inout

       cputm = cputm + cputmp
       eltm = eltm + eltmp

       return
       
       ! END OF MG PREC
       
       ! GCR SOLUTION

    else if ( nrst > 1 ) then
       
       call dagmg_GCR( n, &      !
                       f, &      !
                       x, &      !
                       iter, &   !
                       resid, &  !
                       a, &      !
                       ja, &     !
                       ia, &     !
                       init, &   !
                       flop )    !
       
       ! END OF GCR SOLVE
       
       ! FLEXIBLE CG SOLUTION

    else
       
       call dagmg_FlexCG( n, &      !
                          f, &      !
                          x, &      !
                          iter, &   !
                          resid, &  !
                          a, &      !
                          ja, &     !
                          ia, &     !
                          init, &   !
                          flop )    !
       
       ! END OF FCG SOLVE
       
    end if
    
    solve = .true.

    call dagmg_mestime( 2, &      ! in
                        cputm, &  ! inout
                        eltm )    ! inout

450 continue

    if ( wfo ) then

       if ( wff .and. iter .ne. 0 ) then

          do i = 2, nlev - 1

             write(iout,955) i,kstat(3,i-1),kstat(3,i), &
                  real(kstat(3,i),rk) / real(kstat(3,i-1),rk),kstat(2,i)

          end do

       end if

       write(iout,'()')
       write(iout,954) IRANK,flop/real(11*n+2*nza,rk)
       write(iout,962) IRANK,(memh+mritr)/nza,real_len, &
            (memh+mritr)*real_len/(2**20)

       !CPU_TIME: next line may be uncommented if implemented
       !          (see subroutine mestime)
       !            write(iout,998) IRANK,cputm

       write(iout,999) IRANK,eltm
       write(iout,'()')

    end if
    
    if ( mod(ijob,10) > 0 ) return
    
    ! RELEASE MEMORY
    
500 continue

    call dagmg_relmem

    if ( .not. allzero ) then

       call dagmg_LAPACK( n, &    !
                          a, &    !
                          ja, &   !
                          ia, &   !
                          f, &    !
                          -2, &   !
                          flop )  !

    end if

    preprocessed = .false.
    solve        = .false.
    eltm         = 0.0_rk
    cputm        = 0.0_rk
    flop         = 0.0_rk
    kstat        = 0

    if ( wfo ) then
       write (iout,902) IRANK
    end if
    
    continue
    
    return

900 format(i3,'*ENTERING AGMG **********************************',&
         '***************************')
901 format(i3,'*ONE APPLICATION OF AGMG PRECONDITIONER')
902 format(i3,'*LEAVING AGMG * (MEMORY RELEASED) ***************',&
         '***************************')
954 format(i3,'*','  Equiv. number of CG iter.:',f9.2)
955 format('****     level',i2,'   #call=',i6,'   #cycle=',i6,    &
         '   mean=',f7.2,'    max=',i3)
960 format(  i3,'*','         memory used (peak):',f9.2,        &
         ' real(',i1,') words per nnz (',f8.2,' Mb)')
961 format(  i3,'*','     memory still allocated:',f9.2,        &
         ' real(',i1,') words per nnz (',f8.2,' Mb)')
962 format(  i3,'*','   memory used for solution:',f9.2,        &
         ' real(',i1,') words per nnz (',f8.2,' Mb)')
996 format(i3,'*','           Setup time (CPU):   ',1PE10.2,     &
         ' seconds')
997 format(i3,'*','       Setup time (Elapsed):   ',1PE10.2,     &
         ' seconds')
998 format(i3,'*','        Solution time (CPU):   ',1PE10.2,     &
         ' seconds')
999 format(i3,'*','    Solution time (Elapsed):   ',1PE10.2,     &
         ' seconds')
1001 format(i3,'*',' FATAL ERROR: preprocessing not done - ijob=',i3, &
          'is not allowed')

  end subroutine dagmg
