c#######################################################################
c
c     NIMbnd code suite
c     Copyright by Coronado Consulting
c     2017
c
c     Programmed by D. C. Barnes
c     Coronado Consulting
c
c#######################################################################
c
c     Compute boundary integrals of Green's functions (bint.f)
c
c#######################################################################
c
c     version 1.0 -- 09/03/16
c     adapted from Cygnus suite
c
c#######################################################################
c
      MODULE bint_m
c
      USE serv_m
      USE seg_m
c
      IMPLICIT NONE
      INCLUDE 'params.f'
      INCLUDE 'priv.f'
c
      CHARACTER(LEN=2)::geom='rz',prob='es'
c of_s log integral exact, but only gives 2nd order accuracy
c of_d 4th order accuracy, avoiding diagonal
      REAL(rn),PARAMETER::of_s=0.22313016,of_d=0.40824829
     .     ,of_1d=0.4330127019
c
      CONTAINS
c
c#######################################################################
c     get symmetric matrix for given Green's function
c-----------------------------------------------------------------------
      SUBROUTINE sym_mat(gg,mm)
c arguments
      REAL(rn),EXTERNAL::gg
      REAL(rn),DIMENSION(:,:),INTENT(OUT)::mm
c local variables
      INTEGER::sz,l,lp
      REAL(rn)::r0l,z0l,r1l,z1l,dsl,rsl
      REAL(rn)::r0p,z0p,r1p,z1p,dsp,rsp
      REAL(rn)::ellp
      REAL(rn),PARAMETER::eps=1.e-3
c begin
      sz=SIZE(mm,1)
      IF(SIZE(mm,2)/=sz) 
     .     CALL bomb('sym_mat called with non-square matrix')
c do integrals for upper triangle
      DO l=1,sz
         r0l=seg_r0(l ); z0l=seg_z0(l ); r1l=seg_r1(l ); z1l=seg_z1(l )
         dsl=SQRT((r1l-r0l)**2+(z1l-z0l)**2); rsl=(r0l+r1l)/2
c$OMP PARALLEL DO PRIVATE(r0p,z0p,r1p,z1p,dsp,rsp)
         DO lp=l,sz
            r0p=seg_r0(lp); z0p=seg_z0(lp)
            r1p=seg_r1(lp); z1p=seg_z1(lp)
            dsp=SQRT((r1p-r0p)**2+(z1p-z0p)**2); rsp=(r0p+r1p)/2
            ellp=seg_segi(r0l,z0l,r1l,z1l,r0p,z0p,r1p,z1p,gg,eps)
            IF(prob=='ms') THEN
               mm(l,lp)=ellp        /dsl/dsp
            ELSE
               mm(l,lp)=ellp/rsl/rsp/dsl/dsp
            ENDIF
         ENDDO
c$OMP END PARALLEL DO
      ENDDO
c extend to lower triangle by symmetry
      DO l=2,sz
         DO lp=1,l-1
            mm(l,lp)=mm(lp,l)
         ENDDO
      ENDDO
c terminate
      RETURN
      END SUBROUTINE sym_mat
c
c#######################################################################
c     get asymmetric matrix for given Green's function
c-----------------------------------------------------------------------
      SUBROUTINE asm_mat(gg,mm)
c arguments
      REAL(rn),EXTERNAL::gg
      REAL(rn),DIMENSION(:,:),INTENT(OUT)::mm
c local variables
      INTEGER::sz1,sz2,l,lp
      REAL(rn)::r0l,z0l,r1l,z1l,dsl,rsl
      REAL(rn)::r0p,z0p,r1p,z1p,dsp,rsp
      REAL(rn)::ellp
      REAL(rn),PARAMETER::eps=1.e-3
c begin
      sz1=SIZE(mm,1); sz2=SIZE(mm,2)
c do integrals
      DO l=1,sz1
         r0l=seg_r0(l ); z0l=seg_z0(l ); r1l=seg_r1(l ); z1l=seg_z1(l )
         dsl=SQRT((r1l-r0l)**2+(z1l-z0l)**2); rsl=(r0l+r1l)/2
c$OMP PARALLEL DO PRIVATE(r0p,z0p,r1p,z1p,dsp,rsp,ellp)
         DO lp=1,sz2
            r0p=seg_r0(lp); z0p=seg_z0(lp)
            r1p=seg_r1(lp); z1p=seg_z1(lp)
            dsp=SQRT((r1p-r0p)**2+(z1p-z0p)**2); rsp=(r0p+r1p)/2
            ellp=seg_segi(r0l,z0l,r1l,z1l,r0p,z0p,r1p,z1p,gg,eps)
            IF(prob=='ms') THEN
               mm(l,lp)=ellp        /dsl/dsp
            ELSE
               mm(l,lp)=ellp/rsl/rsp/dsl/dsp
            ENDIF
         ENDDO
c$OMP END PARALLEL DO
      ENDDO
c terminate
      RETURN
      END SUBROUTINE asm_mat
c
c#######################################################################
c     contribution to boundary integral from one pair of segments
c-----------------------------------------------------------------------
      FUNCTION seg_segi(r00,z00,r10,z10,r01,z01,r11,z11,gh,eps)
c type function
      REAL(rn)::seg_segi
c arguments
      REAL(rn),INTENT(IN)::r00,z00,r10,z10,r01,z01,r11,z11,eps
      REAL(rn),EXTERNAL::gh
c local variables
      INTEGER::levmx,lev
      REAL(rn)::dmn,ex,xdiv,ln
      INTEGER,PARAMETER::ndiv=2
c begin -- get error control quantities
c     prepare for recursive refinement
      dmn=1/SQRT(eps); ex=1; xdiv=ndiv; ln=1/log(xdiv)
      DO levmx=1,40
         ex=ex/xdiv**2
         if(ex*levmx*ln<eps)  exit
      ENDDO
      lev=1
      seg_segi=refine_seg_segi(r00,z00,r10,z10,r01,z01,r11,z11,gh
     .                        ,ndiv,dmn,levmx,lev)
c terminate
      RETURN
      END FUNCTION seg_segi
c
c#######################################################################
c     refinement of term of boundary integral from one pair of segments
c-----------------------------------------------------------------------
      RECURSIVE FUNCTION refine_seg_segi(r00,z00,r10,z10,r01,z01,r11,z11
     .                                  ,gh,ndiv,dmn,levmx,lev) 
     .          RESULT (sum)
c type function
      REAL(rn)::sum
c arguments
      INTEGER,INTENT(IN)::ndiv,levmx,lev
      REAL(rn),INTENT(IN)::r00,z00,r10,z10,r01,z01,r11,z11,dmn
      REAL(rn),EXTERNAL::gh
c local variables
      INTEGER::n
      REAL(rn)::dlr0,dlz0,dlr1,dlz1
      REAL(rn)::ds02,ds0,rs0,zs0,ds12,ds1,rs1,zs1,dsi2,of
      REAL(rn),DIMENSION(7)::aa
      INTEGER::i,j,idiv,jdiv
      REAL(rn)::rdiv,f0,f1,r00p,z00p,r10p,z10p,r01p,z01p,r11p,z11p,sumj
      REAL(rn),PARAMETER::epsam=1.e-10
      LOGICAL::same
c begin
      dlr0=r10-r00; dlz0=z10-z00; ds02=dlr0**2+dlz0**2; ds0=sqrt(ds02)
      rs0=(r00+r10)/2; zs0=(z00+z10)/2
      dlr1=r11-r01; dlz1=z11-z01; ds12=dlr1**2+dlz1**2; ds1=sqrt(ds12)
      rs1=(r01+r11)/2; zs1=(z01+z11)/2
      dsi2=MAX(ds02,ds12)
c check if sufficiently refined
      same=((r00-r01)**2+(r01-r11)**2+(z00-z01)**2+(z10-z11)**2)
     .     /dsi2<epsam
      IF((lev>levmx).OR.((rs0-rs1)**2+(zs0-zs1)**2>dmn*dsi2))
     .   THEN
         of=of_d; IF(same) of=of_s
         n=7
         aa(5)=ds1; aa(6)= dlz0/ds0; aa(7)=-dlr0/ds0
         sum=0
         DO i=-1,1,2
            aa(1)=rs0+i*of*dlr0; aa(2)=zs0+i*of*dlz0
            aa(3)=rs1          ; aa(4)=zs1
            IF(geom=='rz'.AND.prob/='ms') THEN
               sum=sum+rs0*rs1*ds0*ds1*gh(aa,n)/4
            ELSE
               sum=sum+ds0*ds1*gh(aa,n)/4
            ENDIF
         ENDDO
c
         DO i=-1,1,2
            aa(1)=rs0          ; aa(2)=zs0
            aa(3)=rs1+i*of*dlr1; aa(4)=zs1+i*of*dlz1
            IF(geom=='rz'.AND.prob/='ms') THEN
               sum=sum+rs0*rs1*ds0*ds1*gh(aa,n)/4
            ELSE
               sum=sum+ds0*ds1*gh(aa,n)/4
            ENDIF
         ENDDO
      ELSE
         rdiv=1./ndiv
         sum=0
         DO idiv=1,ndiv
            f0=rdiv*(idiv-1); f1=f0+rdiv
            r00p=r00+f0*(r10-r00); z00p=z00+f0*(z10-z00)
            r10p=r00+f1*(r10-r00); z10p=z00+f1*(z10-z00)
            DO jdiv=1,ndiv
               f0=rdiv*(jdiv-1); f1=f0+rdiv
               r01p=r01+f0*(r11-r01); z01p=z01+f0*(z11-z01)
               r11p=r01+f1*(r11-r01); z11p=z01+f1*(z11-z01)
               sum=sum
     .            +refine_seg_segi(r00p,z00p,r10p,z10p
     .                            ,r01p,z01p,r11p,z11p,gh
     .                            ,ndiv,dmn,levmx,lev+1)
            ENDDO
         ENDDO
      ENDIF
c terminate
      RETURN
      END FUNCTION refine_seg_segi
c
c#######################################################################
c     contribution to boundary integral from one segment
c-----------------------------------------------------------------------
      FUNCTION segi(r,z,r0,z0,r1,z1,ro,xnr,xnz,gh,n,eps)
c type function
      REAL(rn)::segi
c arguments
      INTEGER,INTENT(IN)::n
      REAL(rn),INTENT(IN)::r,z,r0,z0,r1,z1,ro,xnr,xnz,eps
      REAL(rn),EXTERNAL::gh
c local variables
      INTEGER::levmx,lev
      REAL(rn)::dmn,ex,xdiv,ln
      INTEGER,PARAMETER::ndiv=2
c begin -- get error control quantities
c     prepare for recursive refinement
      dmn=1/eps; ex=1; xdiv=ndiv; ln=1/LOG(xdiv)
      DO levmx=1,40
         ex=ex/xdiv**2
         if(ex*levmx*ln<eps)  EXIT
      ENDDO
      lev=1
      segi=refine_segi(r,z,r0,z0,r1,z1,ro,xnr,xnz,gh,7
     .                ,ndiv,dmn,levmx,lev)
c terminate
      RETURN
      END FUNCTION segi
c
c#######################################################################
c     refinement of term of boundary integral from one segment
c-----------------------------------------------------------------------
      RECURSIVE FUNCTION refine_segi(r,z,r0,z0,r1,z1,ro,xnr,xnz,gh,n
     .                              ,ndiv,dmn,levmx,lev) RESULT (sum)
c type function
      REAL(rn)::sum
c arguments
      INTEGER,INTENT(IN)::n,ndiv,levmx,lev
      REAL(rn),INTENT(IN)::r,z,r0,z0,r1,z1,ro,xnr,xnz,dmn
      REAL(rn),EXTERNAL::gh
c local variables
      REAL(rn)::drr,dzz,dss2,dss,rss,zss
      REAL(rn),DIMENSION(n)::aa
      INTEGER::i,idiv
      REAL(rn)::rdiv,f0,f1,r0p,z0p,r1p,z1p
c begin
      drr=r1-r0; dzz=z1-z0; dss2=drr**2+dzz**2; dss=sqrt(dss2)
      rss=(r0+r1)/2; zss=(z0+z1)/2
c check if sufficiently refined
      IF((lev>levmx).OR.((r-rss)**2+(z-zss)**2>dmn*dss2))  then
         aa(1)=r; aa(2)=z
         aa(5)=dss; aa(6)=xnr; aa(7)=xnz
         sum=0
         DO i=-1,1,2
            aa(3)=rss+i*of_1d*drr; aa(4)=zss+i*of_1d*dzz
            IF(geom=='rz'.AND.prob/='ms') THEN
               sum=sum+ro*rss*dss*gh(aa,n)/2
            ELSE
               sum=sum+ro*dss*gh(aa,n)/2
            ENDIF
         ENDDO
      ELSE
         rdiv=1./ndiv
         sum=0
         DO idiv=1,ndiv
            f0=rdiv*(idiv-1); f1=f0+rdiv
            r0p=r0+f0*(r1-r0); z0p=z0+f0*(z1-z0)
            r1p=r0+f1*(r1-r0); z1p=z0+f1*(z1-z0)
            sum=sum+refine_segi(r,z,r0p,z0p,r1p,z1p,ro,xnr,xnz
     .                         ,gh,n,ndiv,dmn,levmx,lev+1)
         ENDDO
      ENDIF
c terminate
      RETURN
      END FUNCTION refine_segi
c
      END MODULE bint_m
