# File make.inc for AIX operating systems and 32-bit addressing:

# This file contains machine-specific definitions:  

# 0) Home directory locations--done first to avoid repitition.
# 1) The first group of definitions determines the compilers and
#    compiler options.
# 2) The machine-dependent 'local' file to use in nimlib.
# 3) The next group determines whether mpi is used in order to run
#    the physics kernel in parallel.
# 4) The third group determines which mathematical libraries are linked
#    and, hence, which linear solver options may be used.



# PART 0: HOME LOCATIONS FOR LIBRARIES, ETC.

# See the respective library definitions in parts 3 and 4.

BLAS_HOME =
MPI_HOME = /usr/lpp/ppe.poe
SLU_HOME = $(SUPERLU_DIST_SRC)/..



# PART 1: GENERAL COMPILER DEFINITIONS

# General fortran 90 compiler and loader definitions:

FMACH = -qarch=pwr3 -qtune=pwr3 -qcache=auto -qmaxmem=-1
FFLAGS = $(FMACH) -O3 -qstrict
FO2FLAGS = $(FMACH) -O2
FFLAGS_DBG = -qcheck -qsigtrap -qflttrap=overflow:zerodivide:invalid:enable -g
FFLAGS_NODBL = $(FFLAGS)
FFLAGS_NOOPT = $(FMACH) -qnoopt
F90 = xlf90 -q32 -qfixed -qrealsize=8  -qxlf90=autodealloc
MPIF90 = mpxlf90 -q32 -qfixed -qrealsize=8  -qxlf90=autodealloc

# Fortran loader:
LDR_FLAGS = -bmaxstack:0x40000000 -bmaxdata:0x80000000
LDR = $(F90)
MPILDR = $(MPIF90)

# General C compiler definitions:
CFLAGS =
CFLAGS_DBG = -qcheck -qflttrap=overflow:zerodivide:invalid:enable -g
CC = xlc -q32

#--------------------------------------------------------------------
# Notes on above compiler flags.
#       -q32 needed to make nimrod compatible with mpi libs
#       -qrealsize-8  nimrod always does auto doubling
#       -qfixed  nimrod is fixed source
#       -qxlf90=autodealloc  we tend to get memory leaks unless
#        arrays are deallocated.  Default behavior on most compilers.
#
#   -qmaxmem=-1 => no limit on memory compiler uses for optimizations
#
# Debugging/checking flags
#  Note that xlf90 flag -qextchk doesn't seem to work to well
#  -qcheck bounds checking (also -C)
#  Note that unlike some compilers, the -g can be used with all
#   optimization flags
#--------------------------------------------------------------------

#Archive utilities.
# AR creates the archive and RANLIB creates the index
AR = ar -X32 -r
RANLIB = ranlib



# PART 2: MACHINE-DEPENDENT LOCAL FILE IN NIMLIB

LOCAL_FILE = local_ibm.f



# PART 3: MPI DEFINITIONS

# MPI library and include paths need to be set when creating a
#  executable for parallel computation.  First, comment-out the
#  mpi_serial.f link, and uncomment the mpi_parallel.f link.  Then,
#  set the mpi library and include paths and library path names.
#  These names may not need to be defined when the compiler is
#  an "mpi*" version of the F90 compiler.  The default definitions
#  create a serial version of NIMROD.

#MPILINK = mpi_serial.f
MPILINK = mpi_parallel.f
MPI_LIB = $(MPI_HOME)/lib/libmpi.a
MPI_INCL = -I$(MPI_HOME)/include



# PART 4: MATH LIBRARY DEFINITIONS

# Links to an appropriate math library are needed to run the NIMROD
#   physics kernel using direct solves for the linear systems, i.e.
#   setting the namelist input item solver to any of the following:

#   1) 'lapack'    --  banded serial LAPACK solves
#   2) 'seq_slu'   --  sparse serial solves via Sequential SuperLU
#   3) 'slu_dist'  --  sparse parallel solves via SuperLU_DIST
 
#   [SuperLU_DIST is our present recommendation for large parallel
#   simulations.  The SuperLU packages were written by Xiaoye Li (LBL),
#   James Demmel (UC-Berkeley), and John Gilbert (UC-SB).  To find
#   documentation, see the ACTS Collection web page at NERSC,
#   http://acts.nersc.gov/superlu/]

# To use one of the solver libraries, one needs to:

#   A. Specify a BLAS library.
#   B. Set definitions for compiling the C bridge routines.
#   C. Set library names for the math libraries.

# Dummy external links are also available for compiling the kernel
#   without any external libraries (but don't expect to use the
#   solver options listed above).  They are also used for NIMPLOT.
#   Leave all of the following variables blank to use the dummy
#   externals in nimrod.

# A. BLAS library:
#   Set the full BLAS library path name in BLAS_LIB.

BLAS_LIB =  -lblas

# B. Specify definitions for the C bridge routines:
#   Sequential SuperLU and SuperLU_DIST are written in C, and
#   there are 'bridge' routines written in C for calling these
#   libraries.  NIMROD only calls these bridge libraries (or their
#   dummy stubs).  To compile a set of actual bridge
#   routines, uncomment the respective SLU_LINK variable below, set the 
#   compiler options for the bridge routines, and set the source paths
#   for the libraries so that the C compiler can
#   find library-specific include files.  There may be conflicts
#   in the include path files if one tries to link both SuperLU
#   libraries.

SLU_SRC = $(SUPERLU_DIST_SRC)
#SLU_LINK = sslu_link
SLU_LINK = slud_link
CFLAGS_SLU = -O3 -DNoChange
CFLAGS_SLU_DBG = $(CFLAGS_DBG) -DNoChange
CINCL_SLU = -I$(SLU_SRC) $(MPI_INCL)


# C. Set the full path names for the math libraries:
#   Leaving the paths as "$(NIMHOME)/externals/lib*_dummy.a" loads the
#   physics kernel with the nonfunctioning dummy external for the respective
#   library.  Include the SLU_LINK library with one of the functioning
#   SuperLU libraries.  You will be able to the actual LAPACK library and/or
#   one of the actual SuperLU libraries.

#   The new version of SuperLU_DIST (2.1) requires ParMETIS to be
#   loaded, and the METIS_LIB variable helps with this.  If you use an older
#   version or the sequential SuperLU, METIS_LIB can be blank.

LAP_LIB = $(NIMHOME)/externals/liblapack_dummy.a
SSLU_LIB = $(NIMHOME)/externals/libsslu_dummy.a
#SLUD_LIB = $(NIMHOME)/externals/libslud_dummy.a
SLUD_LIB = $(NIMHOME)/externals/lib$(SLU_LINK).a \
	   $(SUPERLU_DIST)
METIS_LIB = $(METIS) -lmetis -lparmetis


# THAT'S ALL (FOLKS).
