# File make.inc for Linux operating systems:

# This file contains machine-specific definitions:  

# 0) Home directory locations--done first to avoid repetition.
# 1) The first group of definitions determines the compilers and
#    compiler options.
# 2) The machine-dependent 'local' file to use in nimlib.
# 3) The next group determines whether mpi is used in order to run
#    the physics kernel in parallel.
# 4) The third group determines which mathematical libraries are linked
#    and, hence, which linear solver options may be used.



# PART 0: HOME LOCATIONS FOR LIBRARIES, ETC.

# See the respective library definitions in parts 3 and 4.

# The native BLAS for the architecture may not need an explicit path.
# If the native blas is not available, the CBLAS library from SLU
# can be used.

BLAS_HOME = /opt/zeep/atlas-3.6.0_gnu/lib/Linux_PII

# Set the MPI home directory if an explicit path is needed.

MPI_HOME = /opt/zeep/mpich-1.2.6_lahey

# The NIMEQ program now uses the sequential SLU library, so its path
# is needed if NIMEQ is compiled or if NIMROD uses that library.  If
# NIMROD is built for parallel computation, the path for SuperLU_DIST
# is needed.

SSLU_HOME = /home/inertial/sovinec/lib/SuperLU
SLUD_HOME = /home/inertial/sovinec/lib/SuperLU_DIST_2.0



# PART 1: GENERAL COMPILER DEFINITIONS

# General fortran 90 compiler and loader definitions:

#Fujitsu/Lahey compiler
FFLAGS = -O --dbl
FO2FLAGS = $(FFLAGS)
FFLAGS_DBG = --chk -g --dbl
FFLAGS_NODBL = -O
FFLAGS_NOOPT = -O0
F90 = lf95 --ap
MPIF90 = $(MPI_HOME)/bin/mpif90 --ap

#Intel compiler
#QUIET = -cm -w
#PREC = -autodouble
#PENT4 = -tpp7 -xW
#FFLAGS = -O2 -132 $(PENT4) $(PREC) $(QUIET)
#FO2FLAGS = $(FFLAGS)
#FFLAGS_DBG = -g -132 $(PREC) $(QUIET)
#FFLAGS_NDBL = -O2 -132 $(PENT4) $(PREC) $(QUIET)
#FFLAGS_NOOPT = -O0 -132 $(PENT4) $(PREC) $(QUIET)
#F90 = ifc -mp
#MPIF90 = $(F90)

#Absoft compiler
#FFLAGS = -B100 -O
#FO2FLAGS = $(FFLAGS)
#FFLAGS_DBG = -g
#FFLAGS_NODBL = $(FFLAGS)
#FFLAGS_NOOPT = -O0
#F90 = f90
#MPIF90 = $(F90)

#Portland Group compiler
#FFLAGS = -fast -byteswapio -r8
#FFLAGS = $(FFLAGS)
#FFLAGS_DBG = -g -C -r8 -trapuv
#FFLAGS_NODBL = -fast -byteswapio
#FFLAGS_NOOPT = -byteswapio
#F90 = pgf90 
#MPIF90 = $(F90)

#gfortran
#FFLAGS = -O3 -fdefault-real-8
#FO2FLAGS = -O2 -fdefault-real-8
#FFLAGS_DBG = -g -fdefault-real-8 -fbounds-check
#FFLAGS_NODBL = -O2
#FFLAGS_NOOPT = -O0 -fdefault-real-8
#F90 = gfortran -fconvert='big-endian'
#MPIF90 = $(F90)

# Fortran loader:
LDR_FLAGS = -Wl,-R/opt/zeep/lf/lib
LDR = $(F90)
MPILDR = $(LDR)

# General C compiler definitions:
CFLAGS =
CFLAGS_DBG = -g
CC = $(MPI_HOME)/bin/mpicc

#Archive utilities.
# AR creates the archive and RANLIB creates the index
AR = ar -r
RANLIB = ranlib



# PART 2: MACHINE-DEPENDENT LOCAL FILE IN NIMLIB

LOCAL_FILE = local_linux.f



# PART 3: MPI DEFINITIONS

# MPI library and include paths need to be set when creating an
#  executable for parallel computation.  First, comment-out the
#  mpi_serial.f link, and uncomment the mpi_parallel.f link.  Then,
#  set the mpi library and include paths and library path names.
#  These names may not need to be defined when the compiler is
#  an "mpi*" version of the F90 compiler.  The default definitions
#  create a serial version of NIMROD.

#MPILINK = mpi_serial.f
MPILINK = mpi_parallel.f
MPI_LIB = $(MPI_HOME)/lib/libmpich.a \
          $(MPI_HOME)/lib/libpmpich.a
MPI_INCL = -I$(MPI_HOME)/include



# PART 4: MATH LIBRARY DEFINITIONS

# Links to an appropriate math library are needed to run the NIMROD
#   physics kernel using direct solves for the linear systems, i.e.
#   setting the namelist input item solver to any of the following:

#   1) 'lapack'    --  banded serial LAPACK solves
#   2) 'seq_slu'   --  sparse serial solves via Sequential SuperLU
#   3) 'slu_dist'  --  sparse parallel solves via SuperLU_DIST
 
#   [SuperLU_DIST is our present recommendation for large parallel
#   simulations.  The SuperLU packages were written by Xiaoye Li (LBL),
#   James Demmel (UC-Berkeley), and John Gilbert (UC-SB).  To find
#   documentation, see the ACTS Collection web page at NERSC,
#   http://acts.nersc.gov/superlu/]

# The NIMEQ program now uses the sequential SLU library, so the SSLU_*
#   variables need to be defined if NIMEQ is used or if NIMROD is built
#   for sequential computations.  If NIMROD is built for parallel
#   computation, the SLUD_* variables need to be set for the SuperLU_DIST
#   library.

# To use one of the solver libraries, one needs to:

#   A. Specify a BLAS library.
#   B. Set definitions for compiling the C bridge routines.
#   C. Set library names for the math libraries.

# Dummy external links are also available for compiling the kernel
#   without any external libraries (but don't expect to use the
#   solver options listed above).  They are also used for NIMPLOT.
#   Leave all of the following variables blank to use the dummy
#   externals in nimrod.

# A. BLAS library:
#   Set the full BLAS library path name in BLAS_LIB.

BLAS_LIB = $(BLAS_HOME)/libf77blas.a $(BLAS_HOME)/libatlas.a

# B. Specify definitions for the C bridge routines:
#   Sequential SuperLU and SuperLU_DIST are written in C, and
#   there are 'bridge' routines written in C for calling these
#   libraries.  NIMROD only calls these bridge libraries (or their
#   dummy stubs).  To compile the actual bridge routines for
#   NIMROD, uncomment the respective SLU_LINK variable below, set the 
#   compiler options for the bridge routines, and set the source paths
#   for the libraries so that the C compiler can
#   find library-specific include files.

SSLU_SRC = $(SSLU_HOME)/SRC
SLUD_SRC = $(SLUD_HOME)/SRC
#SLU_LINK = sslu_link
SLU_LINK = slud_link
CFLAGS_SLU = -O3 -DAdd_ -DSuperLU_2
CFLAGS_SLU_DBG = $(CFLAGS_DBG) -DAdd_ -DSuperLU_2
CINCL_SLU = -I$(SLUD_SRC) $(MPI_INCL)


# C. Set the full path names for the math libraries:
#   Leaving the paths as "$(NIMHOME)/externals/lib*_dummy.a" loads the physics
#   kernel with the nonfunctioning dummy external for the respective
#   library.  Include the SLU_LINK library with one of the functioning
#   SuperLU libraries.  You will be able run NIMROD with the actual LAPACK
#   library and/or one of the actual SuperLU libraries.

#   Note: use SLU_NIM_LIB to list the sequential SLU or SuperLU_DIST **and**
#   the dummy library for the one that is not used with NIMROD.  Use
#   SLU_NEQ_LIB to always list the sequential SLU and the dummy for
#   the distributed-memory calls for NIMEQ.

#   Recent versions of SuperLU_DIST (>=2.1) requires ParMETIS to be
#   loaded, and the METIS_LIB variable helps with this.  If you use an older
#   version or the sequential SuperLU, METIS_LIB can be blank.

LAP_LIB = $(NIMHOME)/externals/liblapack_dummy.a

SSLU_LIB = $(NIMHOME)/externals/libsslu_link.a \
	   $(SSLU_HOME)/superlu_LINUX.a

SLUD_LIB =  $(NIMHOME)/externals/lib$(SLU_LINK).a \
            $(SLU_HOME)/superlu_LINUX.a

METIS_LIB = 

SLU_NIM_LIB = $(SSLU_LIB) $(SLUD_LIB)
SLU_NEQ_LIB = $(SSLU_LIB) $(NIMHOME)/externals/libslud_dummy.a

# THAT'S ALL (FOLKS).
