c-----------------------------------------------------------------------
c     file iter_3d_cg.f
c     module that contains routines for a matrix-free 3D Krylov-space
c     solvers.  originally, this only had CG but it now allows GMRES
c     for non-Hermitian systems.  preconditioning is performed with
c     the 2D routines available in iter_cg_f90 and iter_cg_comp.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     code organization.
c-----------------------------------------------------------------------
c     module iter_3d_cg_mod
c     1.  iter_3d_ky_solve.
c     2.  iter_3d_cg_init.
c     3.  iter_3d_cg_dealloc.
c     4.  iter_3d_cg_dot.
c     5.  iter_3d_cg_err.
c     6.  iter_3d_cg_m2v.
c     7.  iter_3d_cg_pre.
c     8.  iter_3d_ortho.
c     9.  iter_3d_grot.
c     10. iter_3d_min.
c     11. iter_3d_hcheck.
c     12. iter_3d_offload.
c     13. iter_3d_offunld.
c     14. iter_3d_offrready.
c-----------------------------------------------------------------------
c     module containing the 3D cg routines.
c-----------------------------------------------------------------------
      MODULE iter_3d_cg_mod
      USE vector_type_mod
      USE local
      USE mpi_nim
      USE pardata
      IMPLICIT NONE

      LOGICAL, PRIVATE :: symm,novec
      LOGICAL, PRIVATE, PARAMETER :: usefom=.false.,fixm=.false.,
     $         hcheck=.false.,leftprec=.false.
      INTEGER(i4), PRIVATE, PARAMETER :: npoln=0_i4

      TYPE :: gm3_basis_type
        TYPE(cvector_type), POINTER, DIMENSION(:) :: base
      END TYPE gm3_basis_type

      TYPE :: iter_3d_off_type
        COMPLEX(r8), DIMENSION(:), POINTER :: offsend,offrecv 
        INTEGER(i4), DIMENSION(:), POINTER :: sendpr,sendreq
        INTEGER(i4) :: datapass,recvpr,recvreq,size,nsend
      END TYPE iter_3d_off_type

      INTEGER(i4), PRIVATE :: n_gm
      REAL(r8), DIMENSION(:,:), ALLOCATABLE, PRIVATE :: hess,givn
      REAL(r8), DIMENSION(:), ALLOCATABLE, PRIVATE :: errvec,minvec

      TYPE(gm3_basis_type), DIMENSION(:), POINTER, PRIVATE :: ep_d,pinv
      TYPE(cvector_type), DIMENSION(:), POINTER, PRIVATE :: cvec,ovec
      TYPE(iter_3d_off_type), DIMENSION(:), POINTER, PRIVATE :: 
     $                        offhi,offlo
      INTEGER(i4), PRIVATE :: noff,nftotal,nsym_pass
      INTEGER(i4), PARAMETER, PRIVATE :: noffmx=20
      INTEGER(i4), DIMENSION(mpi_status_size,noffmx), PRIVATE :: mpistat
      REAL(r8), PRIVATE :: nsym_rfac
      CHARACTER(12), PRIVATE :: nsym_rtype
      LOGICAL, PRIVATE :: it3deal

      CONTAINS
c-----------------------------------------------------------------------
c     subprogram 1. iter_3d_ky_solve
c     perform conjugate gradient iterations to solve a 3D matrix
c     equation using a matrix-free computation of the matrix-vector
c     product.  preconditioning is accomplished with nimrod's routines
c     that are set-up for 2D matrices (diagonal in n).  elimination of
c     cell-interior data is used for the preconditioning step only.
c
c     the following are notes regarding usage:
c
c     1.  the direction vector (dir) must be the storage used in the
c     finite element computation of the matrix-vector product.
c
c     2.  the integrand routine for the finite element computation of
c     the matrix-vector product is the input "integrand".
c
c     3.  any non-homogeneous Dirichlet essential conditions should be
c     satisfied by the initial guess, so that direction vectors
c     satisfy the corresponding homogeneous conditions (supplied by
c     bc_routine with bc_flag).  
c     *** this option has been removed.  the boundary conditions are
c         also applied to the initial guess for consistency. *** 
c
c     4.  the mat_str and fac_str are matrix and preconditioner data
c     structures, respectively, for the 2D matrices used to precondition
c     the 3D system.
c
c     5.  the optional input old_vec is used to get a norm of the rhs.
c     this is used so that the solved equation can be of the form,
c     A.(x_new-x_old)=b, where the rhs for the x_new equation is
c     b+A.x_old.  the tolerance for converence is then based on the
c     norm of the residual relative to the norm of rhs (not b).
c
c     6.  the optional input mat2d is used to speed the matrix-vector
c     computations.  it is a structure of 2d matrices that can hold the
c     diagonal-in-n part of the operator, so that these terms do
c     not have to be recomputed by finite element computations in
c     integrand.
c-----------------------------------------------------------------------
c     change log:

c     1. A restart capability has been added that checks if the error
c     increases after every nrst iterations.  If so, it goes back to the
c     old solution, throwing out the direction vector used at that step.
c     The code also compares the recursive and actual residuals.  If the
c     discrepancy is larger than resid_tol, the direction vector is
c     thrown out restarting the computation from that iterate as a
c     guess.
c       5/8/02, C. Sovinec

c     2. Calls to regular_ave and the new regular_zero_phi have been
c     added to handle n=1 regularity conditions on vectors.
c	6/11/02, C. Sovinec

c     3. A check of the actual residual has been added after the
c     iteration loop to see if the problem has converged when its
c     hits maxits.
c       6/8/03, C. Sovinec
 
c     4. This version "_ky_" uses either CG or QGMRES, depending
c     on symmetry of preconditioning matrix.
c       10/31/05 D. C. Barnes
  
c     5. The nonsymmetric solve has been modified to use the
c     GMRES algorithm from "Iterative Methods for Sparse Linear
c     Systems," 2nd ed., Y. Saad, SIAM (2003).
c     The old restart capability never worked well and has been removed
c     for better readability.
c       6/7/07, C. Sovinec
c
c     6. The iter_3d_cg_init routine now performs a seaming operation
c     to ensure that the provided guess has absolutely no discrepancies
c     among different images along block borders.  This prevents an
c     unusual error that otherwise appears in some computations.
c     11/26/07, C. Sovinec
 
c     7. Difficulties converging on a computation with time-step much
c     larger than tau_A prompted a check of the entire solver.  The
c     root problem appears to be that when the preconditioner matrix (P)
c     itself is ill-conditioned, P^-1 dotted into a linear combination
c     of the basis vectors is not equivalent to the same linear 
c     combination of P^-1 dotted into each basis vector due to finite-
c     precision math.  This is remedied by saving P^-1 dotted into
c     each basis in the new pinv data structure, as motivated by the
c     Barnes code that did not use the Hessenberg matrix but did not
c     exhibit the same problem.
 
c     In the process of sorting this out, other numerical options were
c     implemented as part of the testing.  The logical parameter usefom
c     may be coded to use Arnoldi's full orthogonalization method (FOM)
c     instead of the GMRES minimization when choosing the linear
c     combination of vectors for the solution.  Another, fixm, may be
c     set to skip the tolerance checks and run a full maxit steps
c     unless the orthogonalization runs out of new directions.  This
c     option also skips the Givens rotations and solves either the
c     orthogonalization equation (FOM) or the minimization problem
c     (GMRES) after all vectors are determined.  Finally, one may
c     choose left preconditioning by setting leftprec, but this
c     does not seem to offer the effectiveness of right preconditioning
c     when P is ill conditioned.
c     1/3/08, C. Sovinec

c     8. A polynomial approximation as been added to the preconditioning
c     routine.  As implemented, it amounts to Jacobi-like iteration
c     with the preconditioning matrix as the generalized diagonal block.
c     To apply the preconditioner more than once per GMRES iteration,
c     increase the npoln parameter.
c     1/17/08, C. Sovinec

c     9. Optional matrix structures can be supplied for off diagonal
c     in Fourier component contributions that are used in SOR and SSOR
c     preconditioning steps.
c     6/15/08, C. Sovinec

c     10. The GMRES algorithm in iter_3d_cg now includes coefficients
c     from auxiliary discontinuous fields.  The corresponding rhs
c     coefficients are passed in the arrtmp part of the vector structure
c     for the rhs.  There are also two new integers, nqdsc and nbdsc,
c     that are passed into iter_3d_ky_solve to specify, respectively,
c     the size of the quantity dimension and the number of bases for the
c     discontinuous fields.
c     8/19/13, C. Sovinec

c     11. Changes for discontinuous fields are being merged into the
c     main trunk.  At this time, the unused mat2d option is being
c     removed.
c     9/4/13, C. Sovinec

c     12. A higher-level layer of dot-product management routines is
c     now used for the get_rhs, regularity, and boundary-condition
c     calls.  This has been ported from the vblock branch, where it
c     has been in place since 7/2/14.  Also at this stage,
c     the iter_3d_ky solves no longer use input parameters through
c     the finite_element module.  With this and the external direction-
c     vector computation, the iter_3d_cg module does not depend on
c     the f_e module at all.
c     8/7/15, C. Sovinec
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_ky_solve(dot_routine,mat_str,fac_str,
     $                            sln,rhs,dir,nqty,poly_deg,nqdsc,nbdsc,
     $                            etol,maxits,precon,dealiase,nm_total,
     $                            nsp_band,nsp_rpass,nsp_rfac,nsp_rtype,
     $                            err,its,seed,
     $                            old_vec,minus_str,plus_str)
      USE matrix_mod
      USE seam_storage_mod
      USE edge
      USE iter_cg
      USE time

      TYPE(complex_matrix_type), DIMENSION(:), INTENT(IN) :: mat_str
      TYPE(complex_matrix_type), DIMENSION(:,:), INTENT(IN), OPTIONAL ::
     $                           minus_str,plus_str
      TYPE(complex_factor_type), DIMENSION(:), INTENT(IN) :: fac_str
      TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: rhs,sln,dir
      TYPE(cvector_type), DIMENSION(:), INTENT(IN), OPTIONAL :: old_vec
      INTEGER(i4), INTENT(IN) :: nqty,poly_deg,maxits,nsp_band,
     $             nsp_rpass,nm_total
      INTEGER(i4), INTENT(IN) :: nqdsc,nbdsc  ! nqty and # of bases for
                                              ! discontinuous fields
      REAL(r8), INTENT(IN) :: etol,nsp_rfac
      CHARACTER(*), INTENT(IN) :: precon,nsp_rtype
      LOGICAL, INTENT(IN) :: dealiase
      INTEGER(i4), INTENT(OUT) :: its
      REAL(r8), INTENT(OUT) :: err
      CHARACTER(8), INTENT(OUT) :: seed

      TYPE(cvector_type), DIMENSION(:), POINTER :: ctmp,res
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2zee,c2rhs,c2tmp,
     $                       c2pck

      INTEGER(i4) :: ibl,it3d,ierror,nfour,ntotb,nrb

      REAL(r8), PARAMETER :: resid_tol=1._r8,resid_margin=0.9_r8
      REAL(r8) :: ztr=0.,ztro,alp,er0,erg,rhsnorm
      REAL(r8) :: timestart_cg,timeend_cg
      LOGICAL, SAVE :: first_call=.true.

      LOGICAL, PARAMETER :: residcheck=.true.
      LOGICAL, PARAMETER :: everycheck=.false.
c-----------------------------------------------------------------------
c     interface block for the external subroutine that calls get_rhs
c     for the matrix-free dot-product computation.
c-----------------------------------------------------------------------
      INTERFACE
        SUBROUTINE dot_routine(oper,prod,bc_oper)
        USE vector_type_mod
        USE local
        TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: oper,prod
        LOGICAL, INTENT(IN) :: bc_oper
        END SUBROUTINE dot_routine
      END INTERFACE
c-----------------------------------------------------------------------
c     find out if preconditioner is symmetric (hermitian).
c-----------------------------------------------------------------------
      CALL timer(timestart_cg)
      symm=mat_str(SIZE(mat_str))%hermitian
      IF (first_call.AND..NOT.symm) THEN
        ALLOCATE(hess(maxits+1,maxits))
        ALLOCATE(errvec(maxits+1))
        ALLOCATE(minvec(maxits))
        ALLOCATE(givn(2,maxits+1))
        ALLOCATE(ep_d(maxits+1))
        IF (.NOT.leftprec) ALLOCATE(pinv(SIZE(ep_d)))
        first_call=.false.
      ENDIF
c-----------------------------------------------------------------------
c     copy input data which is accessible through the finite_element
c     module, then call the initialization routine.
c-----------------------------------------------------------------------
      nftotal=nm_total
      nsym_pass=nsp_rpass
      nsym_rfac=nsp_rfac
      nsym_rtype=nsp_rtype
      IF (nsym_rtype=='Gauss-Seidel') nsym_rtype='Gauss Seidel'
      IF (PRESENT(minus_str)) THEN
        IF (.NOT.symm) THEN
          noff=nsp_band
        ELSE
          noff=0
        ENDIF
      ELSE
        noff=0
      ENDIF
      it3deal=dealiase
      CALL iter_3d_cg_init(mat_str,fac_str,ctmp,c2zee,c2rhs,c2tmp,c2pck,
     $                     nfour,nrb,ntotb,poly_deg,nqty,nbdsc,nqdsc,
     $                     sln,res)
      DO ibl=1,ntotb
        res(ibl)=rhs(ibl)
      ENDDO
c-----------------------------------------------------------------------
c     find the norm for rhs.
c-----------------------------------------------------------------------
      IF (PRESENT(old_vec)) THEN
        DO ibl=1,ntotb
          dir(ibl)=old_vec(ibl)
        ENDDO
        CALL dot_routine(dir,ctmp,.false.)
        DO ibl=1,ntotb
          CALL vector_add(ctmp(ibl),res(ibl))
        ENDDO
        CALL iter_3d_cg_err(rhsnorm,nrb,ntotb,poly_deg,ctmp)
      ELSE
        CALL iter_3d_cg_err(rhsnorm,nrb,ntotb,poly_deg,res)
      ENDIF
      rhsnorm=MAX(rhsnorm,SQRT(TINY(rhsnorm)))
      CALL iter_3d_cg_err(er0,nrb,ntotb,poly_deg,res)
c-----------------------------------------------------------------------
c     check if the guess provides a better starting point, i.e. minimize
c     error in the direction of the guess, a la DCB
c-----------------------------------------------------------------------
      CALL iter_3d_cg_dot(erg,sln,sln,seam,ntotb)
      IF (erg>0._r8) THEN
        DO ibl=1,ntotb
          dir(ibl)=sln(ibl)
        ENDDO
        CALL dot_routine(dir,ctmp,.true.)

        CALL iter_3d_cg_dot(alp,ctmp,res,seam,ntotb)
        CALL iter_3d_cg_dot(erg,ctmp,ctmp,seam,ntotb)
        alp=alp/erg
        DO ibl=1,ntotb
          CALL vector_mult(sln(ibl),alp)
          CALL vector_add(ctmp(ibl),res(ibl),v1fac=-alp)
        ENDDO
      ELSE
        DO ibl=1,ntotb
          ctmp(ibl)=res(ibl)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     decide whether to use this improved guess.
c-----------------------------------------------------------------------
      CALL iter_3d_cg_err(erg,nrb,ntotb,poly_deg,ctmp)
      DO ibl=1,ntotb
        IF (erg<er0) THEN
          res(ibl)=ctmp(ibl)
c         sln(ibl)=dir(ibl)
          seed="guess"
          err=erg
        ELSE
          sln(ibl)=0._r8
          seed="0 vec"
          err=er0
        ENDIF
      ENDDO
      er0=rhsnorm
c-----------------------------------------------------------------------
c     catch trivial cases.
c-----------------------------------------------------------------------
      IF (err==0._r8) THEN
        CALL iter_3d_cg_dealloc(ctmp,c2zee,c2rhs,c2tmp,c2pck,res,nqdsc)
        its=0
        CALL timer(timeend_cg)
        time_iter = time_iter + timeend_cg-timestart_cg
        RETURN
      ENDIF
c-----------------------------------------------------------------------
c     GMRES-specific initialization.  left preconditioning is applied
c     here before saving the initial residual and error.
c-----------------------------------------------------------------------
      IF (residcheck.AND.node==0) WRITE(nim_wr,'(2(a,es15.8))')
     $  '   Initial resid=',err,' rhsnorm=',rhsnorm
      IF (.NOT.symm) THEN
        IF (leftprec) THEN
          CALL iter_3d_cg_pre(mat_str,fac_str,res,ctmp,c2rhs,c2zee,
     $                        c2tmp,c2pck,poly_deg,nqty,nbdsc,nqdsc,nrb,
     $                        ntotb,nfour,precon,dot_routine,
     $                        dir,minus_str=minus_str,plus_str=plus_str)
          DO ibl=1,ntotb
            res(ibl)=ctmp(ibl)
          ENDDO
          CALL iter_3d_cg_err(err,nrb,ntotb,poly_deg,res)
          err=MAX(err,SQRT(TINY(err)))
          IF (residcheck.AND.node==0) WRITE(nim_wr,'(2(a,es15.8))')
     $      '   Left resid=',err
          er0=err
        ENDIF

        errvec(1)=err
        DO ibl=1,ntotb
          CALL vector_mult(res(ibl),1._r8/err)
          ep_d(1)%base(ibl)=res(ibl)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     start iterations.
c-----------------------------------------------------------------------
      threed_loop: DO it3d=1,maxits
c-----------------------------------------------------------------------
c       precondition the 3D equation by solving separate 2D equations.
c       the result is placed in ctmp.  for GMRES, this is a right-
c       preconditioning step.
c-----------------------------------------------------------------------
        IF (symm.OR..NOT.leftprec) THEN
          CALL iter_3d_cg_pre(mat_str,fac_str,res,ctmp,c2rhs,c2zee,
     $                        c2tmp,c2pck,poly_deg,nqty,nbdsc,nqdsc,nrb,
     $                        ntotb,nfour,precon,dot_routine,
     $                        dir,minus_str=minus_str,plus_str=plus_str)
          IF (.NOT.symm) THEN
            DO ibl=1,ntotb
              pinv(it3d)%base(ibl)=ctmp(ibl)
            ENDDO
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       CG-specific computations  find z^t.r
c-----------------------------------------------------------------------
        IF (symm) CALL iter_3d_cg_dot(ztr,ctmp,res,seam,ntotb)
c-----------------------------------------------------------------------
c       find the new direction vector.
c-----------------------------------------------------------------------
        IF (it3d==1.OR..NOT.symm) THEN
          IF (leftprec.AND..NOT.symm) THEN
            DO ibl=1,ntotb
              dir(ibl)=res(ibl)
            ENDDO
          ELSE
            DO ibl=1,ntotb
              dir(ibl)=ctmp(ibl)
            ENDDO
          ENDIF
        ELSE
          DO ibl=1,ntotb
            CALL vector_add(dir(ibl),ctmp(ibl),ztr/ztro,1._r8)
          ENDDO
        ENDIF
        ztro=ztr
c-----------------------------------------------------------------------
c       find the dot product of the 3D matrix and the new direction
c       vector.
c-----------------------------------------------------------------------
        CALL dot_routine(dir,ctmp,.false.)
c-----------------------------------------------------------------------
c       CG-specific: find the new solution iterate and residual.
c-----------------------------------------------------------------------
        IF (symm) THEN
          CALL iter_3d_cg_dot(alp,dir,ctmp,seam,ntotb)
          alp=ztr/alp
          DO ibl=1,ntotb
            CALL vector_add(sln(ibl),dir(ibl),1._r8,alp)
            CALL vector_add(res(ibl),ctmp(ibl),1._r8,-alp)
          ENDDO
c-----------------------------------------------------------------------
c       GMRES: orthogonalize and add another column to the
c       Hessenberg matrix.  then, find the Givens rotation for the last
c       Hessenberg entry and determine the minimizing vector of
c       coefficients.
c
c       the left preconditioning step is applied here, after the matrix-
c       vector product.  here, ep_d is only used for temporary storage.
c-----------------------------------------------------------------------
        ELSE
          IF (leftprec) THEN
            DO ibl=1,ntotb
              ep_d(it3d+1)%base(ibl)=ctmp(ibl)
            ENDDO
            CALL iter_3d_cg_pre(mat_str,fac_str,ep_d(it3d+1)%base,ctmp,
     $                          c2rhs,c2zee,c2tmp,c2pck,poly_deg,nqty,
     $                          nbdsc,nqdsc,nrb,ntotb,nfour,precon,
     $                          dot_routine,dir,minus_str=minus_str,
     $                          plus_str=plus_str)
          ENDIF
          CALL iter_3d_ortho(it3d,ntotb,ctmp)
          IF (.NOT.fixm) CALL iter_3d_grot(it3d)
        ENDIF
c-----------------------------------------------------------------------
c       check error level.
c-----------------------------------------------------------------------
        IF (symm) THEN
          CALL iter_3d_cg_err(err,nrb,ntotb,poly_deg,res)
          n_gm=it3d
          IF (everycheck.AND.node==0) WRITE(nim_wr,'(a,i4,a,es15.8)')
     $       '   CG at step',it3d,' is ',err
        ELSE
          err=ABS(errvec(n_gm+1))
          IF (everycheck.AND.node==0) WRITE(nim_wr,'(a,i4,a,es15.8)')
     $       '   GMR at step',it3d,' is ',err
        ENDIF
        IF ((symm.OR..NOT.fixm.OR.novec).AND.
     $      err/er0<=resid_margin*etol) THEN
          EXIT threed_loop
        ELSE IF (.NOT.symm) THEN
          DO ibl=1,ntotb
            res(ibl)=ctmp(ibl)
          ENDDO
        ENDIF
      ENDDO threed_loop 
c-----------------------------------------------------------------------
c     recompute the Hessenberg matrix for testing.
c-----------------------------------------------------------------------
      IF (hcheck.AND..NOT.symm.AND.node==0) 
     $  CALL iter_3d_hcheck(dot_routine,mat_str,fac_str,ctmp,dir,
     $                      nqty,poly_deg,nqdsc,nbdsc,nrb,ntotb,nfour,
     $                      precon,c2zee,c2rhs,c2tmp,c2pck,
     $                      minus_str=minus_str,plus_str=plus_str)
c-----------------------------------------------------------------------
c     for GMRES, determine the minimizing vector of coefficients, and
c     construct the solution from either the basis vectors for left
c     preconditioning or from P^-1 dotted into the basis vectors for
c     right preconditioning, where P^-1 is the preconditioning
c     operation.
c-----------------------------------------------------------------------
      IF (n_gm>=1.AND..NOT.symm) THEN
        CALL iter_3d_min
        IF (leftprec) THEN
          DO it3d=1,n_gm
            DO ibl=1,ntotb
              CALL vector_add(sln(ibl),ep_d(it3d)%base(ibl),
     $                        v2fac=minvec(it3d))
            ENDDO
          ENDDO
        ELSE
          DO it3d=1,n_gm
            DO ibl=1,ntotb
              CALL vector_add(sln(ibl),pinv(it3d)%base(ibl),
     $                        v2fac=minvec(it3d))
            ENDDO
          ENDDO
        ENDIF
      ENDIF
      its=n_gm
c-----------------------------------------------------------------------
c     residual check.
c-----------------------------------------------------------------------
      IF (residcheck) THEN
        IF (node==0)
     $    WRITE(nim_wr,'(a,es15.8)') '   Final iterate  3d resid=',err
        DO ibl=1,ntotb
          dir(ibl)=sln(ibl)
        ENDDO
        CALL dot_routine(dir,ctmp,.false.)
        DO ibl=1,ntotb
          CALL vector_add(ctmp(ibl),rhs(ibl),v1fac=-1._r8)
        ENDDO
        CALL iter_3d_cg_err(err,nrb,ntotb,poly_deg,ctmp)
        IF (node==0)
     $    WRITE(nim_wr,'(a,es15.8)') '   Final separate 3d resid=',err
      ENDIF
      err=err/rhsnorm
c-----------------------------------------------------------------------
c     deallocate space.
c-----------------------------------------------------------------------
      CALL iter_3d_cg_dealloc(ctmp,c2zee,c2rhs,c2tmp,c2pck,res,nqdsc)
      CALL timer(timeend_cg)
      time_iter = time_iter + timeend_cg-timestart_cg
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_ky_solve
c-----------------------------------------------------------------------
c     subprogram 2. iter_3d_cg_init. 
c     allocate space for the linear algebra.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_init(mat_str,fac_str,ctmp,c2zee,c2rhs,c2tmp,
     $                           c2pck,nfour,nrb,ntotb,poly_deg,nqty,
     $                           nbdsc,nqdsc,sln,res)
      USE matrix_type_mod
      USE seam_storage_mod
      USE edge

      TYPE(complex_matrix_type), DIMENSION(:), INTENT(IN) :: mat_str
      TYPE(complex_factor_type), DIMENSION(:), INTENT(IN) :: fac_str
      TYPE(cvector_type), DIMENSION(:), POINTER :: ctmp,res
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2zee,c2rhs,c2tmp,
     $                       c2pck
      TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: sln
      INTEGER(i4), INTENT(OUT) :: nfour,nrb,ntotb
      INTEGER(i4), INTENT(IN) :: poly_deg,nqty,nbdsc,nqdsc

      INTEGER(i4) :: ibl,mxb,myb,igm,iq,ix,iy,ibase,im,jm,iv,nbf,
     $               nsnd,isnd,ioff,nsize,ncomb
      INTEGER(i4), DIMENSION(0:SIZE(layer2proc)-1) :: sndchk
c-----------------------------------------------------------------------
c     find array dimensions.
c-----------------------------------------------------------------------
      nfour=SIZE(mat_str)
      ntotb=SIZE(sln)
      nrb=SIZE(mat_str(1)%rbl_mat)
      ncomb=nqty*(poly_deg-1)**2+nqdsc*nbdsc
c-----------------------------------------------------------------------
c     ensure that the guess has exactly the same value in different
c     images of the same node along block borders.
c-----------------------------------------------------------------------
      DO ibl=1,ntotb
        DO iv=1,seam(ibl)%nvert
          ix=seam(ibl)%vertex(iv)%intxy(1)
          iy=seam(ibl)%vertex(iv)%intxy(2)
          iq=1
          DO im=1,nfour
            seam(ibl)%vertex(iv)%seam_cin(iq:nqty+iq-1)=
     $        sln(ibl)%arr(:,ix,iy,im)*seam(ibl)%vertex(iv)%ave_factor
            iq=iq+nqty
          ENDDO
          IF (poly_deg>1) THEN
            ix=seam(ibl)%segment(iv)%intxys(1)
            iy=seam(ibl)%segment(iv)%intxys(2)
            iq=1
            IF (seam(ibl)%segment(iv)%h_side) THEN
              DO im=1,nfour
                DO ibase=1,poly_deg-1
                  seam(ibl)%segment(iv)%seam_cin(iq:iq+nqty-1)=
     $              sln(ibl)%arrh(:,ibase,ix,iy,im)*
     $                seam(ibl)%segment(iv)%ave_factor
                  iq=iq+nqty
                ENDDO
              ENDDO
            ELSE
              DO im=1,nfour
                DO ibase=1,poly_deg-1
                  seam(ibl)%segment(iv)%seam_cin(iq:iq+nqty-1)=
     $              sln(ibl)%arrv(:,ibase,ix,iy,im)*
     $                seam(ibl)%segment(iv)%ave_factor
                  iq=iq+nqty
                ENDDO
              ENDDO
            ENDIF
          ENDIF
        ENDDO
      ENDDO
      CALL edge_network(nqty,nfour,poly_deg-1_i4,.false.)
      DO ibl=1,ntotb
        CALL edge_unload_carr(sln(ibl),nqty,1_i4,nfour,
     $                        poly_deg-1_i4,seam(ibl))
      ENDDO
c-----------------------------------------------------------------------
c     allocate vector structures.
c-PRE at this time, we assume that Fourier off-diagonal matrices do
c     not contribute to coefficients of discontinous fields.
c-----------------------------------------------------------------------
      ALLOCATE(c2zee(ntotb))
      ALLOCATE(c2rhs(ntotb))
      ALLOCATE(c2tmp(ntotb))
      ALLOCATE(c2pck(ntotb))
      ALLOCATE(ctmp(ntotb))
      ALLOCATE(res(ntotb))
      NULLIFY(offlo,offhi,cvec,ovec)
      IF (noff>0) THEN
        ALLOCATE(cvec(ntotb))
        ALLOCATE(ovec(ntotb))
        IF (mode_lo>1) ALLOCATE(offlo(noff))
        IF (mode_hi<nftotal) ALLOCATE(offhi(noff))
      ENDIF
      DO ibl=1,nrb
        mxb=SIZE(sln(ibl)%arr,2)-1
        myb=SIZE(sln(ibl)%arr,3)-1
        IF (nqdsc>0) THEN
          CALL vector_type_alloc(c2zee(ibl),poly_deg,mxb,myb,nqty,
     $                           nbdsc,nqdsc)
          CALL vector_type_alloc(c2rhs(ibl),poly_deg,mxb,myb,nqty,
     $                           nbdsc,nqdsc)
          CALL vector_type_alloc(c2tmp(ibl),poly_deg,mxb,myb,nqty,
     $                           nbdsc,nqdsc)
          CALL vector_type_alloc(ctmp(ibl),poly_deg,mxb,myb,nqty,nfour,
     $                           nbdsc,nqdsc)
          CALL vector_type_alloc(res(ibl),poly_deg,mxb,myb,nqty,nfour,
     $                           nbdsc,nqdsc)
          ALLOCATE(c2pck(ibl)%arri(ncomb,1,mxb,myb))
          ALLOCATE(c2pck(ibl)%arrtmp(ncomb,1,mxb,myb))
          NULLIFY(c2pck(ibl)%arr,c2pck(ibl)%arrh,c2pck(ibl)%arrv)
        ELSE
          CALL vector_type_alloc(c2zee(ibl),poly_deg,mxb,myb,nqty)
          CALL vector_type_alloc(c2rhs(ibl),poly_deg,mxb,myb,nqty)
          CALL vector_type_alloc(c2tmp(ibl),poly_deg,mxb,myb,nqty)
          CALL vector_type_alloc(ctmp(ibl),poly_deg,mxb,myb,nqty,nfour)
          CALL vector_type_alloc(res(ibl),poly_deg,mxb,myb,nqty,nfour)
        ENDIF
        IF (noff>0) THEN
          CALL vector_type_alloc(cvec(ibl),poly_deg,mxb,myb,nqty,nfour)
          CALL vector_type_alloc(ovec(ibl),poly_deg,mxb,myb,nqty,nfour)
        ENDIF
      ENDDO
      DO ibl=nrb+1,ntotb
        mxb=SIZE(sln(ibl)%arr,2)-1
        CALL vector_type_alloc(c2zee(ibl),1_i4,mxb,0_i4,nqty)
        CALL vector_type_alloc(c2rhs(ibl),1_i4,mxb,0_i4,nqty)
        CALL vector_type_alloc(c2tmp(ibl),1_i4,mxb,0_i4,nqty)
        CALL vector_type_alloc(ctmp(ibl),1_i4,mxb,0_i4,nqty,nfour)
        CALL vector_type_alloc(res(ibl),1_i4,mxb,0_i4,nqty,nfour)
        IF (noff>0) THEN
          CALL vector_type_alloc(cvec(ibl),1_i4,mxb,0_i4,nqty,nfour)
          CALL vector_type_alloc(ovec(ibl),1_i4,mxb,0_i4,nqty,nfour)
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     communication buffers for off-diagonal in Fourier preconditioning.
c     also determine where to find the data coming to this processor.
c-----------------------------------------------------------------------
      DO im=1,noff
        nbf=0
        DO ibl=1,ntotb
          nbf=nbf+SIZE(c2zee(ibl)%arr)
          IF (ASSOCIATED(c2zee(ibl)%arrh)) nbf=nbf+SIZE(c2zee(ibl)%arrh)
          IF (ASSOCIATED(c2zee(ibl)%arrv)) nbf=nbf+SIZE(c2zee(ibl)%arrv)
          IF (ASSOCIATED(c2zee(ibl)%arri)) nbf=nbf+SIZE(c2zee(ibl)%arri)
        ENDDO
        IF (ASSOCIATED(offlo)) THEN
          ALLOCATE(offlo(im)%offsend(nbf))
          ALLOCATE(offlo(im)%offrecv(nbf))
          NULLIFY(offlo(im)%sendpr)
          NULLIFY(offlo(im)%sendreq)
          offlo(im)%size=nbf
          jm=mode_lo-im
          IF (jm>=1) offlo(im)%recvpr=layer2proc(mode2layer(jm))
        ENDIF
        IF (ASSOCIATED(offhi)) THEN
          ALLOCATE(offhi(im)%offsend(nbf))
          ALLOCATE(offhi(im)%offrecv(nbf))
          NULLIFY(offhi(im)%sendpr)
          NULLIFY(offhi(im)%sendreq)
          offhi(im)%size=nbf
          jm=mode_hi+im
          IF (jm<=nftotal) offhi(im)%recvpr=layer2proc(mode2layer(jm))
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     determine the list of processors that need data from this one
c     for the Fourier off-diagonal preconditioning.  the ioff index
c     is ordered to select memory in the order that it becomes
c     available.  (there will be extra space for layers that send
c     and the band goes past the total index range.)
c-----------------------------------------------------------------------
      IF (ASSOCIATED(offlo)) THEN
        DO im=mode_lo,MIN(mode_lo+noff-1,mode_hi)
          sndchk=0
          ioff=mode_lo+noff-im
          DO jm=im-noff,im-1
            IF (jm<mode_lo.AND.jm>=1) sndchk(mode2layer(jm))=1
          ENDDO
          nsnd=SUM(sndchk)
          IF (nsnd>noffmx) CALL nim_stop
     $     ('Iter_3d_cg_init: increase noffmx for this pre_band value.')
          offlo(ioff)%nsend=nsnd
          IF (nsnd>0) THEN
            ALLOCATE(offlo(ioff)%sendpr(nsnd))
            ALLOCATE(offlo(ioff)%sendreq(nsnd))
            isnd=1
            sndchk=0
            DO jm=im-noff,im-1
              IF (jm<mode_lo.AND.jm>=1.AND.
     $            sndchk(mode2layer(jm))==0) THEN
                offlo(ioff)%sendpr(isnd)=layer2proc(mode2layer(jm))
                isnd=isnd+1
                sndchk(mode2layer(jm))=1
              ENDIF
            ENDDO
          ENDIF
        ENDDO
      ENDIF
      IF (ASSOCIATED(offhi)) THEN
        DO im=MAX(mode_lo,mode_hi-noff+1),mode_hi
          sndchk=0
          ioff=im+noff-mode_hi
          DO jm=im+1,im+noff
            IF (jm>mode_hi.AND.jm<=nftotal) sndchk(mode2layer(jm))=1
          ENDDO
          nsnd=SUM(sndchk)
          IF (nsnd>noffmx) CALL nim_stop
     $     ('Iter_3d_cg_init: increase noffmx for this pre_band value.')
          offhi(ioff)%nsend=nsnd
          IF (nsnd>0) THEN
            ALLOCATE(offhi(ioff)%sendpr(nsnd))
            ALLOCATE(offhi(ioff)%sendreq(nsnd))
            isnd=1
            sndchk=0
            DO jm=im+1,im+noff
              IF (jm>mode_hi.AND.jm<=nftotal.AND.
     $            sndchk(mode2layer(jm))==0) THEN
                offhi(ioff)%sendpr(isnd)=layer2proc(mode2layer(jm))
                isnd=isnd+1
                sndchk(mode2layer(jm))=1
              ENDIF
            ENDDO
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     GMRES data structures.
c-----------------------------------------------------------------------
      IF (.NOT.symm) THEN
        DO igm=1,SIZE(ep_d)
          ALLOCATE(ep_d(igm)%base(ntotb))
          IF (.NOT.leftprec) ALLOCATE(pinv(igm)%base(ntotb))
          DO ibl=1,nrb
            mxb=SIZE(sln(ibl)%arr,2)-1
            myb=SIZE(sln(ibl)%arr,3)-1
            IF (nqdsc>0) THEN
              CALL vector_type_alloc(ep_d(igm)%base(ibl),
     $                               poly_deg,mxb,myb,nqty,nfour,
     $                               nbdsc,nqdsc)
              IF (.NOT.leftprec)
     $          CALL vector_type_alloc(pinv(igm)%base(ibl),
     $                                 poly_deg,mxb,myb,nqty,nfour,
     $                                 nbdsc,nqdsc)
            ELSE
              CALL vector_type_alloc(ep_d(igm)%base(ibl),
     $                               poly_deg,mxb,myb,nqty,nfour)
              IF (.NOT.leftprec)
     $          CALL vector_type_alloc(pinv(igm)%base(ibl),
     $                                 poly_deg,mxb,myb,nqty,nfour)
            ENDIF
          ENDDO
          DO ibl=nrb+1,ntotb
            mxb=SIZE(sln(ibl)%arr,2)-1
            CALL vector_type_alloc(ep_d(igm)%base(ibl),
     $                             1_i4,mxb,0_i4,nqty,nfour)
            IF (.NOT.leftprec)
     $        CALL vector_type_alloc(pinv(igm)%base(ibl),
     $                               1_i4,mxb,0_i4,nqty,nfour)
          ENDDO
        ENDDO
        hess=0._r8
        errvec=0._r8
        minvec=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     initialize the novec flag.
c-----------------------------------------------------------------------
      novec=.false.
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_init
c-----------------------------------------------------------------------
c     subprogram 3. iter_3d_cg_dealloc. 
c     deallocate temporary space.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_dealloc(ctmp,c2zee,c2rhs,c2tmp,c2pck,res,
     $                              nqdsc)

      TYPE(cvector_type), DIMENSION(:), POINTER :: ctmp,res
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2zee,c2rhs,c2tmp,
     $                       c2pck
      INTEGER(i4), INTENT(IN) :: nqdsc

      INTEGER(i4) :: ibl,igm,im
c-----------------------------------------------------------------------
c     deallocate vector structures.
c-----------------------------------------------------------------------
      DO ibl=1,SIZE(c2zee)
        CALL vector_type_dealloc(c2zee(ibl))
        CALL vector_type_dealloc(c2rhs(ibl))
        CALL vector_type_dealloc(c2tmp(ibl))
        CALL vector_type_dealloc(ctmp(ibl))
        CALL vector_type_dealloc(res(ibl))
        IF (ASSOCIATED(cvec)) CALL vector_type_dealloc(cvec(ibl))
        IF (ASSOCIATED(ovec)) CALL vector_type_dealloc(ovec(ibl))
        IF (nqdsc>0) THEN
          DEALLOCATE(c2pck(ibl)%arri)
          DEALLOCATE(c2pck(ibl)%arrtmp)
        ENDIF
      ENDDO
      IF (.NOT.symm) THEN
        DO igm=1,SIZE(ep_d)
          DO ibl=1,SIZE(c2zee)
            CALL vector_type_dealloc(ep_d(igm)%base(ibl))
            IF (.NOT.leftprec)
     $        CALL vector_type_dealloc(pinv(igm)%base(ibl))
          ENDDO
          DEALLOCATE(ep_d(igm)%base)
          IF (.NOT.leftprec) DEALLOCATE(pinv(igm)%base)
        ENDDO
      ENDIF
      DEALLOCATE(c2zee,c2rhs,c2tmp,c2pck,ctmp,res)
      IF (ASSOCIATED(cvec)) DEALLOCATE(cvec)
      IF (ASSOCIATED(ovec)) DEALLOCATE(ovec)
      IF (ASSOCIATED(offlo)) THEN
        DO im=1,noff
          DEALLOCATE(offlo(im)%offsend)
          DEALLOCATE(offlo(im)%offrecv)
          IF (ASSOCIATED(offlo(im)%sendpr)) THEN
            DEALLOCATE(offlo(im)%sendpr)
            DEALLOCATE(offlo(im)%sendreq)
          ENDIF
        ENDDO
        DEALLOCATE(offlo)
      ENDIF
      IF (ASSOCIATED(offhi)) THEN
        DO im=1,noff
          DEALLOCATE(offhi(im)%offsend)
          DEALLOCATE(offhi(im)%offrecv)
          IF (ASSOCIATED(offhi(im)%sendpr)) THEN
            DEALLOCATE(offhi(im)%sendpr)
            DEALLOCATE(offhi(im)%sendreq)
          ENDIF
        ENDDO
        DEALLOCATE(offhi)
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_dealloc
c-----------------------------------------------------------------------
c     subprogram 4. iter_3d_cg_dot. 
c     compute the inner product of two 3d vectors.
c
c     include the contribution of the conjugate terms.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_dot(dprod,v1,v2,seam,nbl)
      USE edge_type_mod

      REAL(r8), INTENT(OUT) :: dprod
      TYPE(cvector_type), DIMENSION(:) :: v1,v2
      TYPE(edge_type), DIMENSION(:), POINTER :: seam
      INTEGER(i4), INTENT(IN) :: nbl

      INTEGER(i4) :: ix,iy,iv,ni,ibl,ierror
      REAL(r8) :: tmp
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: p1,p2
      COMPLEX(r8), DIMENSION(:,:,:,:,:), POINTER :: ph1,ph2
c-----------------------------------------------------------------------
c     loop over blocks, and collect the on-processor contributions
c     from all Fourier components.
c-----------------------------------------------------------------------
      dprod=0._r8
      DO ibl=1,nbl
c-----------------------------------------------------------------------
c       find the dot product at each grid vertex.
c-----------------------------------------------------------------------
        p1=>v1(ibl)%arr
        p2=>v2(ibl)%arr
        IF (ilayer==0) THEN
          dprod=dprod+SUM(CONJG(p1(:,:,:,1))*p2(:,:,:,1))+
     $          2._r8*SUM(CONJG(p1(:,:,:,2:))*p2(:,:,:,2:))
        ELSE
          dprod=dprod+2._r8*SUM(CONJG(p1)*p2)
        ENDIF
c-----------------------------------------------------------------------
c       boundary points must be divided by the number of internal 
c       representations to get the correct sum over all blocks.
c       remove what has been contributed above, also.
c-----------------------------------------------------------------------
        IF (ilayer==0) THEN
          DO iv=1,seam(ibl)%nvert
            ix=seam(ibl)%vertex(iv)%intxy(1)
            iy=seam(ibl)%vertex(iv)%intxy(2)
              dprod=dprod+(seam(ibl)%vertex(iv)%ave_factor-1)*
     $               (      SUM(CONJG(p1(:,ix,iy,1))*p2(:,ix,iy,1))+
     $                2._r8*SUM(CONJG(p1(:,ix,iy,2:))*p2(:,ix,iy,2:)) )
          ENDDO
        ELSE
          DO iv=1,seam(ibl)%nvert
            ix=seam(ibl)%vertex(iv)%intxy(1)
            iy=seam(ibl)%vertex(iv)%intxy(2)
              dprod=dprod+2._r8*(seam(ibl)%vertex(iv)%ave_factor-1)*
     $                          SUM(CONJG(p1(:,ix,iy,:))*p2(:,ix,iy,:))
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       higher order contributions.
c-----------------------------------------------------------------------
        IF (ASSOCIATED(v1(ibl)%arrh)) THEN
          IF (ilayer==0) THEN
            ph1=>v1(ibl)%arrh
            ph2=>v2(ibl)%arrh
            dprod=dprod+SUM(CONJG(ph1(:,:,:,:,1))*ph2(:,:,:,:,1))+
     $            2._r8*SUM(CONJG(ph1(:,:,:,:,2:))*ph2(:,:,:,:,2:))
            ph1=>v1(ibl)%arrv
            ph2=>v2(ibl)%arrv
            dprod=dprod+SUM(CONJG(ph1(:,:,:,:,1))*ph2(:,:,:,:,1))+
     $            2._r8*SUM(CONJG(ph1(:,:,:,:,2:))*ph2(:,:,:,:,2:))
            ph1=>v1(ibl)%arri
            ph2=>v2(ibl)%arri
            dprod=dprod+SUM(CONJG(ph1(:,:,:,:,1))*ph2(:,:,:,:,1))+
     $            2._r8*SUM(CONJG(ph1(:,:,:,:,2:))*ph2(:,:,:,:,2:))
          ELSE
            ph1=>v1(ibl)%arrh
            ph2=>v2(ibl)%arrh
            dprod=dprod+2._r8*SUM(CONJG(ph1)*ph2)
            ph1=>v1(ibl)%arrv
            ph2=>v2(ibl)%arrv
            dprod=dprod+2._r8*SUM(CONJG(ph1)*ph2)
            ph1=>v1(ibl)%arri
            ph2=>v2(ibl)%arri
            dprod=dprod+2._r8*SUM(CONJG(ph1)*ph2)
          ENDIF
c-----------------------------------------------------------------------
c         coefficients for discontinuous bases.
c-----------------------------------------------------------------------
          ph1=>v1(ibl)%arrtmp
          ph2=>v2(ibl)%arrtmp
          IF (ASSOCIATED(ph1)) THEN
            IF (ilayer==0) THEN
              dprod=dprod+SUM(CONJG(ph1(:,:,:,:,1))*ph2(:,:,:,:,1))+
     $              2._r8*SUM(CONJG(ph1(:,:,:,:,2:))*ph2(:,:,:,:,2:))
            ELSE
              dprod=dprod+2._r8*SUM(CONJG(ph1)*ph2)
            ENDIF
          ENDIF
c-----------------------------------------------------------------------
c         boundary segment-centered data must be divided by 2.
c-----------------------------------------------------------------------
          IF (ilayer==0) THEN
            DO iv=1,seam(ibl)%nvert
              ix=seam(ibl)%segment(iv)%intxys(1)
              iy=seam(ibl)%segment(iv)%intxys(2)
              IF (seam(ibl)%segment(iv)%h_side) THEN
                dprod=dprod+(seam(ibl)%segment(iv)%ave_factor-1)*
     $              (      SUM(CONJG(v1(ibl)%arrh(:,:,ix,iy,1))*
     $                               v2(ibl)%arrh(:,:,ix,iy,1))+
     $               2._r8*SUM(CONJG(v1(ibl)%arrh(:,:,ix,iy,2:))*
     $                               v2(ibl)%arrh(:,:,ix,iy,2:)) )
              ELSE
                dprod=dprod+(seam(ibl)%segment(iv)%ave_factor-1)*
     $              (      SUM(CONJG(v1(ibl)%arrv(:,:,ix,iy,1))*
     $                               v2(ibl)%arrv(:,:,ix,iy,1))+
     $               2._r8*SUM(CONJG(v1(ibl)%arrv(:,:,ix,iy,2:))*
     $                               v2(ibl)%arrv(:,:,ix,iy,2:)) )
              ENDIF
            ENDDO
          ELSE
            DO iv=1,seam(ibl)%nvert
              ix=seam(ibl)%segment(iv)%intxys(1)
              iy=seam(ibl)%segment(iv)%intxys(2)
              IF (seam(ibl)%segment(iv)%h_side) THEN
                dprod=dprod+2._r8*(seam(ibl)%segment(iv)%ave_factor-1)*
     $                SUM(CONJG(v1(ibl)%arrh(:,:,ix,iy,:))*
     $                          v2(ibl)%arrh(:,:,ix,iy,:))
              ELSE
                dprod=dprod+2._r8*(seam(ibl)%segment(iv)%ave_factor-1)*
     $                SUM(CONJG(v1(ibl)%arrv(:,:,ix,iy,:))*
     $                          v2(ibl)%arrv(:,:,ix,iy,:))
              ENDIF
            ENDDO
          ENDIF
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     sum result over processors.
c-----------------------------------------------------------------------
      IF (nprocs>1) THEN
        CALL mpi_allreduce(dprod,tmp,1,mpi_nim_real,mpi_sum,
     $       mpi_comm_world,ierror)
        dprod=tmp
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_dot
c-----------------------------------------------------------------------
c     subprogram 5. iter_3d_cg_err. 
c     find the norm of a vector
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_err(er,nrb,ntotb,poly_deg,vec,norm_type)
      USE seam_storage_mod

      REAL(r8), INTENT(OUT) :: er
      INTEGER(i4), INTENT(IN) :: nrb,ntotb,poly_deg
      TYPE(cvector_type), DIMENSION(:), INTENT(IN) :: vec
      CHARACTER(*), INTENT(IN), OPTIONAL :: norm_type

      REAL(r8) :: tmp
      INTEGER(i4) :: ibl,ierror
c-----------------------------------------------------------------------
c     the optional norm_type parameter can be used to set symm, which
c     facilitates external use of this routine.
c-----------------------------------------------------------------------
      IF (PRESENT(norm_type)) THEN
        IF (norm_type=='2-norm') THEN
          symm=.false.
        ELSE
          symm=.true.
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     determine the norm of the passed vector.
c     for consistency with older computations, CG still uses the
c     infinity norm, but GMRES uses the 2-norm.  the inf option may
c     be removed at some point.
c-----------------------------------------------------------------------
      IF (symm) THEN
        er=0
        DO ibl=1,ntotb
          er=MAX(MAXVAL(ABS(REAL(vec(ibl)%arr,r8))),
     $           MAXVAL(ABS(AIMAG(vec(ibl)%arr))),er)
          IF (poly_deg>1) THEN
            er=MAX(MAXVAL(ABS(REAL(vec(ibl)%arrh,r8))),
     $             MAXVAL(ABS(AIMAG(vec(ibl)%arrh))),er)
            er=MAX(MAXVAL(ABS(REAL(vec(ibl)%arrv,r8))),
     $             MAXVAL(ABS(AIMAG(vec(ibl)%arrv))),er)
            er=MAX(MAXVAL(ABS(REAL(vec(ibl)%arri,r8))),
     $             MAXVAL(ABS(AIMAG(vec(ibl)%arri))),er)
          ENDIF
          IF (ASSOCIATED(vec(ibl)%arrtmp)) THEN
            er=MAX(MAXVAL(ABS(REAL(vec(ibl)%arrtmp,r8))),
     $             MAXVAL(ABS(AIMAG(vec(ibl)%arrtmp))),er)
          ENDIF
        ENDDO
c-----------------------------------------------------------------------
c       find the max over all processors.
c-----------------------------------------------------------------------
        IF (nprocs>1) THEN
          CALL mpi_allreduce(er,tmp,1,mpi_nim_real,mpi_max,
     $         mpi_comm_world,ierror)
          er=tmp
        ENDIF
c-----------------------------------------------------------------------
c     2-norm computation:
c-----------------------------------------------------------------------
      ELSE
        CALL iter_3d_cg_dot(tmp,vec,vec,seam,ntotb)
        er=SQRT(tmp)
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_err
c-----------------------------------------------------------------------
c     subprogram 6. iter_3d_cg_m2v. 
c     perform a standard 2d matrix-vector multiplication for each
c     Fourier component to permit computing fewer terms in the fe
c     computation.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_m2v(mat2d,vec,prod,c2vec,c2prd,ntotb,nfour,
     $                          nqty)
      USE matrix_mod

      TYPE(complex_matrix_type), DIMENSION(:), INTENT(IN) :: mat2d
      TYPE(cvector_type), DIMENSION(:), INTENT(IN) :: vec
      TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: prod
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2vec,c2prd
      INTEGER(i4), INTENT(IN) :: ntotb,nfour,nqty

      INTEGER(i4) :: ibl,ifour
c-----------------------------------------------------------------------
c     for each Fourier component, transfer the data to a 2D structure
c     and perform the matrix-vector multiplication.
c-----------------------------------------------------------------------
      DO ifour=1,nfour
        DO ibl=1,ntotb
          CALL cvector_2D_assign_cvec(c2vec(ibl),vec(ibl),ifour)
        ENDDO
        CALL matvec(mat2d(ifour),c2vec,c2prd,nqty)
        DO ibl=1,ntotb
          prod(ibl)%arr(:,:,:,ifour)=prod(ibl)%arr(:,:,:,ifour)+
     $                               c2prd(ibl)%arr(:,:,:)
          IF (ASSOCIATED(prod(ibl)%arrh)) THEN
            prod(ibl)%arrh(:,:,:,:,ifour)=prod(ibl)%arrh(:,:,:,:,ifour)+
     $                                    c2prd(ibl)%arrh(:,:,:,:)
          ENDIF
          IF (ASSOCIATED(prod(ibl)%arrv)) THEN
            prod(ibl)%arrv(:,:,:,:,ifour)=prod(ibl)%arrv(:,:,:,:,ifour)+
     $                                    c2prd(ibl)%arrv(:,:,:,:)
          ENDIF
          IF (ASSOCIATED(prod(ibl)%arri)) THEN
            prod(ibl)%arri(:,:,:,:,ifour)=prod(ibl)%arri(:,:,:,:,ifour)+
     $                                    c2prd(ibl)%arri(:,:,:,:)
          ENDIF
          IF (ASSOCIATED(prod(ibl)%arrtmp)) THEN
            prod(ibl)%arrtmp(:,:,:,:,ifour)=
     $        prod(ibl)%arrtmp(:,:,:,:,ifour)+
     $        c2prd(ibl)%arrtmp(:,:,:,:)
          ENDIF
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_m2v
c-----------------------------------------------------------------------
c     subprogram 7. iter_3d_cg_pre. 
c     collect the preconditioning steps here for convenience.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_cg_pre(mat_str,fac_str,res,ctmp,c2r,c2z,c2t,
     $                          c2p,poly_deg,nqty,nbdsc,nqdsc,nrb,ntotb,
     $                          nfour,precon,dot_routine,dir,
     $                          minus_str,plus_str)

      USE matrix_mod
      USE seam_storage_mod
      USE edge
      USE iter_cg
      USE time

      TYPE(complex_matrix_type), DIMENSION(:), INTENT(IN) :: mat_str
      TYPE(complex_factor_type), DIMENSION(:), INTENT(IN) :: fac_str
      TYPE(complex_matrix_type), DIMENSION(:,:), INTENT(IN), OPTIONAL ::
     $                           minus_str,plus_str
      TYPE(cvector_type), DIMENSION(:) :: res,ctmp,dir
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2r,c2z,c2t,c2p
      INTEGER(i4), INTENT(IN) :: poly_deg,nfour,nqty,nrb,ntotb
      INTEGER(i4), INTENT(IN) :: nbdsc,nqdsc  !  for discontin. bases
      CHARACTER(*), INTENT(IN) :: precon

      INTEGER(i4) :: imode,ibl,ix,iy,iv
      INTEGER(i4) :: ipoln,npolmax
      REAL(r8) :: polscale=0.6666667_r8

      LOGICAL :: goingup,useterm,rbfdone
      INTEGER(i4) :: nlimit,ioff,im,jm,jmt,ibf,ipass
c-----------------------------------------------------------------------
c     interface block for the external subroutine that calls get_rhs
c     for the matrix-free dot-product computation.
c-----------------------------------------------------------------------
      INTERFACE
        SUBROUTINE dot_routine(oper,prod,bc_oper)
        USE vector_type_mod
        USE local
        TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: oper,prod
        LOGICAL, INTENT(IN) :: bc_oper
        END SUBROUTINE dot_routine
      END INTERFACE
c-----------------------------------------------------------------------
c     copy the residual. return if there is no preconditioner.
c-----------------------------------------------------------------------
      DO ibl=1,ntotb
        ctmp(ibl)=res(ibl)
      ENDDO
      IF (precon=='no prec') RETURN
c-----------------------------------------------------------------------
c     this outer loop applies polynomial approximation to improve
c     the preconditioning.  defining N=I-P**-1.A, where P is the
c     preconditioning matrix, A**-1=(I-N)**-1.P**-1.  considering N to
c     be 'small' in some sense, (I+N+N**2+N**3+...).P**-1 is an
c     an approximation of A**-1.  starting with z_1=P**-1.r, applying
c     the approximation to the residual r is found recursively from
c        z_i+1=z_i+P**-1.(r-A.z_i)
c-----------------------------------------------------------------------
      IF (symm) THEN
        npolmax=0
      ELSE
        npolmax=npoln
      ENDIF
      poln_loop: DO ipoln=0,npolmax
        IF (ipoln>0) THEN
          DO ibl=1,ntotb
            dir(ibl)=ctmp(ibl)
          ENDDO
          CALL dot_routine(dir,ctmp,.false.)
          DO ibl=1,ntotb
            CALL vector_add(ctmp(ibl),res(ibl),v1fac=-polscale)
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       apply the preconditioning operation over the poloidal plane
c       for each Fourier component.
c
c       when SOR loops are used for off diagonal in Fourier component
c       matrices, copy the current right side into cvec.
c-----------------------------------------------------------------------
        IF (noff>0) THEN
          nlimit=nsym_pass*nfour
          DO ibl=1,ntotb
            cvec(ibl)=ctmp(ibl)
            ctmp(ibl)=0._r8
            ovec(ibl)=0._r8
          ENDDO
          DO im=1,noff
            IF (ASSOCIATED(offhi)) offhi(im)%datapass=0
            IF (ASSOCIATED(offlo)) offlo(im)%datapass=0
          ENDDO
        ELSE
          nlimit=nfour
        ENDIF
        mode_loop: DO im=1,nlimit
          IF (MODULO((im-1_i4)/nfour,2_i4)==0) THEN
            imode=MODULO(im-1,nfour)+1
            goingup=.true.
          ELSE
            imode=nfour-MODULO(im-1,nfour)
            goingup=.false.
          ENDIF
          ipass=(im-1_i4)/nfour+1
c-----------------------------------------------------------------------
c         carry out the matrix-vector multiplications for the off
c         diagonal in Fourier contributions.
c-----------------------------------------------------------------------
          IF (noff>0) THEN
            DO ibl=1,ntotb
              CALL cvector_2D_assign_cvec(c2r(ibl),cvec(ibl),imode)
            ENDDO
c-----------------------------------------------------------------------
c           passes through Fourier components need separate coding for
c           going up vs. going down.  However, during the first pass,
c           when the iterate is 0, Jacobi uses no off-diagonal data.
c
c           caution:  the order of parallel sends and receives 
c           is carefully worked into the relaxation computation.
c           when going up, data is sent in the order of increasing
c           Fourier index, and the ioff loops are arranged to post
c           receives in that same order.
c-----------------------------------------------------------------------
            pass_type: IF (goingup) THEN
              up_lo: DO ioff=noff,1,-1
                useterm=.true.
                jm=imode-ioff
                jmt=jm+mode_lo-1
                IF ((ioff==noff.OR.imode==nfour).AND.jmt<mode_lo) THEN
                  rbfdone=.true.  !  recv lo buff is done before matvec
                ELSE
                  rbfdone=.false.
                ENDIF
                IF (jmt>=mode_lo) THEN
                  DO ibl=1,ntotb
                    IF (nsym_rtype=='Gauss Seidel') THEN
                      CALL cvector_2D_assign_cvec(c2z(ibl),ovec(ibl),jm)
                    ELSE IF (ipass>1) THEN
                      CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),jm)
                    ELSE
                      useterm=.false.
                    ENDIF
                  ENDDO
                ELSE IF (jmt>=1) THEN
                  ibf=mode_lo-jmt
                  IF (ipass==1) THEN
                    useterm=.false.
                  ELSE
                    CALL iter_3d_offunld(offlo(ibf),c2z,ntotb,
     $                                   jmt,ipass,nsym_pass)
                  ENDIF
                ELSE		!	jmt < 1 (negative F-comp)
                  jmt=2-jmt
                  jm=jmt-mode_lo+1
                  rbfdone=.false.  !  buff used after wrapping
                  IF (jmt<mode_lo) THEN
                    ibf=mode_lo-jmt
                    IF (ipass==1) THEN
                      useterm=.false.
                    ELSE
                      CALL iter_3d_offunld(offlo(ibf),c2z,ntotb,
     $                                     jmt,ipass,nsym_pass)
                    ENDIF
                  ELSE IF (jmt<=mode_hi) THEN
                    DO ibl=1,ntotb
                      IF (imode>jm.AND.nsym_rtype=='Gauss Seidel') THEN
                        CALL cvector_2D_assign_cvec(c2z(ibl),ovec(ibl),
     $                                              jm)
                      ELSE IF (ipass>1) THEN
                        CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),
     $                                              jm)
                      ELSE
                        useterm=.false.
                      ENDIF
                    ENDDO
                  ELSE
                    ibf=jmt-mode_hi
                    IF (ipass==1) THEN
                      useterm=.false.
                    ELSE
                      CALL iter_3d_offunld(offhi(ibf),c2z,ntotb,
     $                                   jmt,ipass,nsym_pass)
                    ENDIF
                  ENDIF
                  IF (useterm) THEN
                    DO ibl=1,ntotb
                      c2z(ibl)%arr=CONJG(c2z(ibl)%arr)
                      IF (ASSOCIATED(c2z(ibl)%arrh)) THEN
                        c2z(ibl)%arrh=CONJG(c2z(ibl)%arrh)
                        c2z(ibl)%arrv=CONJG(c2z(ibl)%arrv)
                        c2z(ibl)%arri=CONJG(c2z(ibl)%arri)
                      ENDIF
                    ENDDO
                  ENDIF
                ENDIF
                IF (rbfdone) THEN
                  ibf=mode_lo-jmt
                  CALL iter_3d_offrready(offlo(ibf),jmt,ipass,nsym_pass)
                ENDIF
                IF (useterm) THEN
                  CALL matvec(minus_str(imode,ioff),c2z,c2t,nqty)
                  DO ibl=1,ntotb
                    CALL vector_add(c2r(ibl),c2t(ibl),1._r8,-1._r8)
                  ENDDO
                ENDIF
              ENDDO up_lo

              up_hi: DO ioff=1,noff
                jm=imode+ioff
                jmt=jm+mode_lo-1
                IF (imode==nfour) THEN
                  rbfdone=.true.  !  recv hi buff is done before matvec
                ELSE
                  rbfdone=.false.
                ENDIF
                IF (jmt>nftotal) THEN
                  CYCLE up_hi
                ELSE IF (jmt<=mode_hi.AND.ipass>1) THEN
                  DO ibl=1,ntotb
                    CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),jm)
                  ENDDO
                ELSE IF (ipass>1) THEN
                  ibf=jmt-mode_hi
                  CALL iter_3d_offunld(offhi(ibf),c2z,ntotb,jmt,
     $                                 ipass,nsym_pass)
                ENDIF
                IF (rbfdone) THEN
                  ibf=jmt-mode_hi
                  CALL iter_3d_offrready(offhi(ibf),jmt,ipass,nsym_pass)
                ENDIF
                IF (ipass==1) CYCLE up_hi
                CALL matvec(plus_str(imode,ioff),c2z,c2t,nqty)
                DO ibl=1,ntotb
                  CALL vector_add(c2r(ibl),c2t(ibl),1._r8,-1._r8)
                ENDDO
              ENDDO up_hi
c-----------------------------------------------------------------------
c           offset order and G-S data for decreasing Fourier index:
c
c           data is sent in the order of decreasing Fourier index,
c           and the ioff loops post receives in that same order.
c           here, the wrapping cases need a separate loop so that
c           the receives are posted in order.
c-----------------------------------------------------------------------
            ELSE
              down_wrap: DO ioff=1,noff  !  wrap is separate going down
                jm=imode-ioff
                jmt=jm+mode_lo-1
                IF (jmt>=1) THEN
                  CYCLE down_wrap
                ELSE
                  jmt=2-jmt
                  jm=jmt-mode_lo+1
                  IF (jmt<mode_lo) THEN
                    ibf=mode_lo-jmt
                    CALL iter_3d_offunld(offlo(ibf),c2z,ntotb,jmt,
     $                                   ipass,nsym_pass)
                  ELSE IF (jmt<=mode_hi) THEN
                    DO ibl=1,ntotb
                      IF (imode<jm.AND.nsym_rtype=='Gauss Seidel') THEN
                        CALL cvector_2D_assign_cvec(c2z(ibl),ovec(ibl),
     $                                              jm)
                      ELSE
                        CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),
     $                                              jm)
                      ENDIF
                    ENDDO
                  ELSE
                    ibf=jmt-mode_hi
                    CALL iter_3d_offunld(offhi(ibf),c2z,ntotb,jmt,
     $                                   ipass,nsym_pass)
                  ENDIF
                  DO ibl=1,ntotb
                    c2z(ibl)%arr=CONJG(c2z(ibl)%arr)
                    IF (ASSOCIATED(c2z(ibl)%arrh)) THEN
                      c2z(ibl)%arrh=CONJG(c2z(ibl)%arrh)
                      c2z(ibl)%arrv=CONJG(c2z(ibl)%arrv)
                      c2z(ibl)%arri=CONJG(c2z(ibl)%arri)
                    ENDIF
                  ENDDO
                ENDIF
                CALL matvec(minus_str(imode,ioff),c2z,c2t,nqty)
                DO ibl=1,ntotb
                  CALL vector_add(c2r(ibl),c2t(ibl),1._r8,-1._r8)
                ENDDO
              ENDDO down_wrap

              down_hi: DO ioff=noff,1,-1
                jm=imode+ioff
                jmt=jm+mode_lo-1
                IF ((ioff==noff.OR.imode==1).AND.jmt>mode_hi) THEN
                  rbfdone=.true.  !  recv hi buff is done before matvec
                ELSE
                  rbfdone=.false.
                ENDIF
                IF (jmt>nftotal) THEN
                  CYCLE down_hi
                ELSE IF (jmt<=mode_hi) THEN
                  DO ibl=1,ntotb
                    IF (nsym_rtype=='Gauss Seidel') THEN
                      CALL cvector_2D_assign_cvec(c2z(ibl),ovec(ibl),jm)
                    ELSE
                      CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),jm)
                    ENDIF
                  ENDDO
                ELSE
                  ibf=jmt-mode_hi
                  CALL iter_3d_offunld(offhi(ibf),c2z,ntotb,jmt,
     $                                 ipass,nsym_pass)
                ENDIF
                IF (rbfdone) THEN
                  ibf=jmt-mode_hi
                  CALL iter_3d_offrready(offhi(ibf),jmt,ipass,nsym_pass)
                ENDIF
                CALL matvec(plus_str(imode,ioff),c2z,c2t,nqty)
                DO ibl=1,ntotb
                  CALL vector_add(c2r(ibl),c2t(ibl),1._r8,-1._r8)
                ENDDO
              ENDDO down_hi

              down_lo: DO ioff=1,noff
                jm=imode-ioff
                jmt=jm+mode_lo-1
                IF (imode==1) THEN
                  rbfdone=.true.  !  recv lo buff is done before matvec
                ELSE
                  rbfdone=.false.
                ENDIF
                IF (jmt>=mode_lo) THEN
                  DO ibl=1,ntotb
                    CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),jm)
                  ENDDO
                ELSE IF (jmt>=1) THEN
                  ibf=mode_lo-jmt
                  CALL iter_3d_offunld(offlo(ibf),c2z,ntotb,jmt,
     $                                 ipass,nsym_pass)
                ELSE
                  CYCLE down_lo
                ENDIF
                IF (rbfdone) THEN
                  ibf=mode_lo-jmt
                  CALL iter_3d_offrready(offlo(ibf),jmt,ipass,nsym_pass)
                ENDIF
                CALL matvec(minus_str(imode,ioff),c2z,c2t,nqty)
                DO ibl=1,ntotb
                  CALL vector_add(c2r(ibl),c2t(ibl),1._r8,-1._r8)
                ENDDO
              ENDDO down_lo
            ENDIF pass_type
c-----------------------------------------------------------------------
c           complete off-diagonal operations.
c-----------------------------------------------------------------------
c-TEST no deal
            IF (imode==nfour.AND..NOT.it3deal) THEN
              DO ibl=1,ntotb
                c2r(ibl)%arr=REAL(c2r(ibl)%arr,r8)
                IF (ASSOCIATED(c2r(ibl)%arrh)) THEN
                  c2r(ibl)%arrh=REAL(c2r(ibl)%arrh,r8)
                  c2r(ibl)%arrv=REAL(c2r(ibl)%arrv,r8)
                  c2r(ibl)%arri=REAL(c2r(ibl)%arri,r8)
                ENDIF
              ENDDO
            ENDIF

            IF (poly_deg>1) THEN
              DO ibl=1,ntotb
                c2z(ibl)=c2r(ibl)
              ENDDO
            ENDIF
          ENDIF
c-----------------------------------------------------------------------
c         transfer the ctmp to a cvector_2D_type, and eliminate 
c         interior elements.  first multiply block-border coefficients
c         by their weights to mimic finite_elements operations.
c-----------------------------------------------------------------------
          IF (poly_deg>1) THEN
            DO ibl=1,ntotb
              IF (noff<=0)
     $          CALL cvector_2D_assign_cvec(c2z(ibl),ctmp(ibl),imode)
              DO iv=1,seam(ibl)%nvert
                ix=seam(ibl)%vertex(iv)%intxy(1)
                iy=seam(ibl)%vertex(iv)%intxy(2)
                c2z(ibl)%arr(:,ix,iy)=c2z(ibl)%arr(:,ix,iy)
     $                                 *seam(ibl)%vertex(iv)%ave_factor
                ix=seam(ibl)%segment(iv)%intxys(1)
                iy=seam(ibl)%segment(iv)%intxys(2)
                IF (seam(ibl)%segment(iv)%h_side) THEN
                  c2z(ibl)%arrh(:,:,ix,iy)=c2z(ibl)%arrh(:,:,ix,iy)
     $                                 *seam(ibl)%segment(iv)%ave_factor
                ELSE
                  c2z(ibl)%arrv(:,:,ix,iy)=c2z(ibl)%arrv(:,:,ix,iy)
     $                                 *seam(ibl)%segment(iv)%ave_factor
                ENDIF
              ENDDO
c-----------------------------------------------------------------------
c             if there are discontinuous fields, pack their coefficients
c             together with the interior coefficients of the continuous
c             fields.
c-----------------------------------------------------------------------
              IF (nqdsc>0.AND.ibl<=nrb) THEN
                CALL cvector_2D_pack_cvec2(c2z(ibl),c2p(ibl),nqty,
     $                                     (poly_deg-1_i4)**2,nqdsc,
     $                                     nbdsc)
                c2p(ibl)%arrh=>c2z(ibl)%arri ! holds original arri mem
                c2p(ibl)%arrv=>c2r(ibl)%arri
                c2z(ibl)%arri=>c2p(ibl)%arri ! flat arrays for packing
                c2r(ibl)%arri=>c2p(ibl)%arrtmp
              ENDIF
            ENDDO
c-----------------------------------------------------------------------
c           eliminate interior (and discontinuous) coefficients.
c-----------------------------------------------------------------------
            CALL timer(timestart)
            CALL matelim_presolve(mat_str(imode),c2z,c2r,nqty)
            CALL timer(timeend)
            time_stcon=time_stcon+timeend-timestart
c-----------------------------------------------------------------------
c           combine contributions across block borders.
c-----------------------------------------------------------------------
            DO ibl=1,ntotb
c-----------------------------------------------------------------------
c             unpack the discontinuous fields into the c2r%arrtmp
c             storage.  the first copy overwrites the old packed data.
c-----------------------------------------------------------------------
              IF (nqdsc>0.AND.ibl<=nrb) THEN
                c2p(ibl)%arri(:,:,:,:)=c2p(ibl)%arrtmp(:,:,:,:)
                c2r(ibl)%arri=>c2p(ibl)%arrv   ! original arri assigmnts
                c2z(ibl)%arri=>c2p(ibl)%arrh
                CALL cvector_2D_unpack_cvec2(c2p(ibl),c2r(ibl),nqty,
     $                                       (poly_deg-1_i4)**2,nqdsc,
     $                                       nbdsc,.true.)
                NULLIFY(c2p(ibl)%arrh,c2p(ibl)%arrv)
              ENDIF
              CALL edge_load_2D_carr(c2r(ibl),nqty,poly_deg-1_i4,
     $                               seam(ibl))
            ENDDO
            CALL edge_network(nqty,1_i4,poly_deg-1_i4,.false.)
            DO ibl=1,ntotb
              CALL edge_unload_2D_carr(c2r(ibl),nqty,
     $                                 poly_deg-1_i4,seam(ibl))
            ENDDO
          ELSE
            IF (noff<=0) THEN
              DO ibl=1,ntotb
                CALL cvector_2D_assign_cvec(c2r(ibl),ctmp(ibl),imode)
              ENDDO
            ENDIF
          ENDIF
c-----------------------------------------------------------------------
c         call 2D preconditioner.
c-----------------------------------------------------------------------
          CALL iter_pre(mat_str(imode),fac_str(imode),c2r,c2z,
     $                  c2t,nqty,nrb,ntotb,precon)
c-----------------------------------------------------------------------
c         determine the interior data if eliminated for the matrix
c         solve.
c-----------------------------------------------------------------------
          IF (poly_deg>1) THEN
            DO ibl=1,ntotb
c-----------------------------------------------------------------------
c             if there are discontinuous fields, information from their
c             elimination is in c2r%arrtmp.  pack this together with
c             information from eliminating the interior coefficients
c             (in c2r%arri) before the postsolve step.
c-----------------------------------------------------------------------
              IF (nqdsc>0.AND.ibl<=nrb) THEN
                CALL cvector_2D_pack_cvec2(c2r(ibl),c2p(ibl),nqty,
     $                                     (poly_deg-1_i4)**2,nqdsc,
     $                                     nbdsc)
                c2p(ibl)%arrh=>c2z(ibl)%arri ! hold original arri mem
                c2p(ibl)%arrv=>c2r(ibl)%arri
                c2r(ibl)%arri=>c2p(ibl)%arri ! flat packed array for c2r
                c2z(ibl)%arri=>c2p(ibl)%arrtmp ! flat array for c2z
              ENDIF
            ENDDO
c-----------------------------------------------------------------------
c           reassemble the coefficients that had been eliminated.
c-----------------------------------------------------------------------
            CALL timer(timestart)
            CALL matelim_postsolve(mat_str(imode),c2z,c2r,nqty)
            CALL timer(timeend)
            time_stcon=time_stcon+timeend-timestart
            DO ibl=1,ntotb
c-----------------------------------------------------------------------
c             reset pointers to their original memory assignments, then
c             unpack the discontinuous fields into the c2z%arrtmp
c             storage.
c-----------------------------------------------------------------------
              IF (nqdsc>0.AND.ibl<=nrb) THEN
                c2z(ibl)%arri=>c2p(ibl)%arrh   ! original arri assigmnts
                c2r(ibl)%arri=>c2p(ibl)%arrv
                CALL cvector_2D_unpack_cvec2(c2p(ibl),c2z(ibl),nqty,
     $                                       (poly_deg-1_i4)**2,nqdsc,
     $                                       nbdsc,.true.)
                NULLIFY(c2p(ibl)%arrh,c2p(ibl)%arrv)
              ELSE
                c2z(ibl)%arri(:,:,:,:)=c2r(ibl)%arri(:,:,:,:)
              ENDIF
            ENDDO
          ENDIF
c-----------------------------------------------------------------------
c         make sure that the n=0 part is real before moving to the next
c         component.
c-----------------------------------------------------------------------
          IF (mat_str(imode)%fcomp==0) THEN
            DO ibl=1,ntotb
              c2z(ibl)%arr=REAL(c2z(ibl)%arr,r8)
              IF (ASSOCIATED(c2z(ibl)%arrh)) THEN
                c2z(ibl)%arrh=REAL(c2z(ibl)%arrh,r8)
                c2z(ibl)%arrv=REAL(c2z(ibl)%arrv,r8)
                c2z(ibl)%arri=REAL(c2z(ibl)%arri,r8)
              ENDIF
            ENDDO
          ENDIF
c-----------------------------------------------------------------------
c         combine the result of this preconditioning step as needed for
c         Gauss-Seidel or Jacobi iteration.
c
c         also start sending data as soon as it is available.
c-----------------------------------------------------------------------
          IF (noff>0) THEN
            IF (nsym_rtype=='Gauss Seidel') THEN
              DO ibl=1,ntotb
                CALL cvector_assign_cvec2(ovec(ibl),c2z(ibl),imode)
                CALL cvector_2D_assign_cvec(c2t(ibl),ctmp(ibl),imode)
                CALL vector_add(c2z(ibl),c2t(ibl),nsym_rfac,
     $                          1._r8-nsym_rfac)
                CALL cvector_assign_cvec2(ctmp(ibl),c2z(ibl),imode)
              ENDDO
            ELSE
              DO ibl=1,ntotb
                IF (im<=nfour) THEN
                  CALL cvector_assign_cvec2(ctmp(ibl),c2z(ibl),imode)
                ELSE
                  CALL cvector_assign_cvec2(ovec(ibl),c2z(ibl),imode)
                  IF (goingup.AND.imode<=noff.OR.
     $                .NOT.goingup.AND.nfour+1-imode<=noff) THEN
                    CALL cvector_2D_assign_cvec(c2t(ibl),ctmp(ibl),
     $                                          imode)
                    CALL vector_add(c2z(ibl),c2t(ibl),nsym_rfac,
     $                              1._r8-nsym_rfac)
                  ENDIF
                  IF (MODULO(im,nfour)==0) THEN
                    CALL vector_add(ctmp(ibl),ovec(ibl),
     $                              1._r8-nsym_rfac,nsym_rfac)
                  ENDIF
                ENDIF
              ENDDO
            ENDIF
            IF (imode<=noff.AND.ASSOCIATED(offlo)) THEN
              ioff=noff+1-imode
              CALL iter_3d_offload(offlo(ioff),c2z,ntotb,
     $                             imode+mode_lo-1,ipass,nsym_pass)
            ENDIF
            IF (nfour+1-imode<=noff.AND.ASSOCIATED(offhi)) THEN
              ioff=noff+imode-nfour
              CALL iter_3d_offload(offhi(ioff),c2z,ntotb,
     $                             imode+mode_lo-1,ipass,nsym_pass)
            ENDIF
          ELSE
            DO ibl=1,ntotb
              CALL cvector_assign_cvec2(ctmp(ibl),c2z(ibl),imode)
            ENDDO
          ENDIF
        ENDDO mode_loop
c-----------------------------------------------------------------------
c       update the polynomial approximation.
c-----------------------------------------------------------------------
        IF (ipoln>0) THEN
          DO ibl=1,ntotb
            CALL vector_add(ctmp(ibl),dir(ibl))
          ENDDO
        ENDIF
      ENDDO poln_loop
c-----------------------------------------------------------------------
c     scaling if required.
c-----------------------------------------------------------------------
      IF (npolmax>0.AND.polscale/=1._r8) THEN
        DO ibl=1,ntotb
          CALL vector_mult(ctmp(ibl),polscale)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_cg_pre
c-----------------------------------------------------------------------
c     subprogram 8. iter_3d_ortho.
c     orthogonalize with modified Gram-Schmidt.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_ortho(jout,nbl,ep_n)
      USE seam_storage_mod

      INTEGER(i4), INTENT(IN) :: jout,nbl
      TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: ep_n

      INTEGER(i4) :: ibl,ii,ierror
      REAL(r8) :: rmag,rmag0,rtmp

      REAL(r8), PARAMETER :: small=1.e-14,orthtol=1.e-2
c-----------------------------------------------------------------------
c     compute the norm of A.v_jout prior to orthogonalization.
c-----------------------------------------------------------------------
      CALL iter_3d_cg_dot(rmag0,ep_n,ep_n,seam,nbl)
      rmag0=SQRT(rmag0)
c-----------------------------------------------------------------------
c     loop over existing basis vectors and save the inner products
c     in the hess array.
c-----------------------------------------------------------------------
      DO ii=1,jout
        CALL iter_3d_cg_dot(hess(ii,jout),ep_d(ii)%base,ep_n,seam,nbl)
        DO ibl=1,nbl
          CALL vector_add(ep_n(ibl),ep_d(ii)%base(ibl),
     $                    v1fac=1._r8,v2fac=-hess(ii,jout))
        ENDDO
      ENDDO
      CALL iter_3d_cg_dot(rmag,ep_n,ep_n,seam,nbl)
      rmag=SQRT(rmag)
c-----------------------------------------------------------------------
c     repeat orthogonalization of this vector if its magnitude is
c     reduced by more than a factor of orthtol.
c-----------------------------------------------------------------------
      IF (rmag/rmag0<orthtol) THEN
        DO ii=1,jout
          CALL iter_3d_cg_dot(rtmp,ep_d(ii)%base,ep_n,seam,nbl)
          hess(ii,jout)=hess(ii,jout)+rtmp
          DO ibl=1,nbl
            CALL vector_add(ep_n(ibl),ep_d(ii)%base(ibl),
     $                      v1fac=1._r8,v2fac=-rtmp)
          ENDDO
        ENDDO
        CALL iter_3d_cg_dot(rmag,ep_n,ep_n,seam,nbl)
        rmag=SQRT(rmag)
      ENDIF
c-----------------------------------------------------------------------
c     if there is nothing left of the new direction, the solution should
c     already be obtained.
c-----------------------------------------------------------------------
      IF (rmag<small*MAXVAL(ABS(hess(1:jout,jout)))) THEN
        n_gm=jout
        novec=.true.
        hess(jout+1,jout)=0._r8
        IF (node==0)
     $    WRITE(nim_wr,'(a,i4)') '   No new vector at step ',jout
      ELSE
        n_gm=jout
        hess(jout+1,jout)=rmag
        DO ibl=1,nbl
          CALL vector_mult(ep_n(ibl),1._r8/rmag)
          ep_d(jout+1)%base(ibl)=ep_n(ibl)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_ortho
c-----------------------------------------------------------------------
c     subprogram 9. iter_3d_grot.
c     find and apply the Givens rotation to remove below-diagonal
c     entries from the Hessenberg matrix and update the error vector.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_grot(jout)

      INTEGER(i4), INTENT(IN) :: jout

      INTEGER(i4) :: ii,jj
      REAL(r8) :: gden
      REAL(r8) :: gs,gc,rtmp
c-----------------------------------------------------------------------
c     Apply existing rotations to the new column.
c-----------------------------------------------------------------------
      DO ii=1,jout-1
        rtmp=givn(1,ii)*hess(ii,jout)+
     $       givn(2,ii)*hess(ii+1,jout)
        hess(ii+1,jout)=givn(1,ii)*hess(ii+1,jout)-
     $                  givn(2,ii)*hess(ii,jout)
        hess(ii,jout)=rtmp
      ENDDO
c-----------------------------------------------------------------------
c     Compute the new Givens sine and cosine coefficients.
c-----------------------------------------------------------------------
      gden=hess(jout,jout)*hess(jout,jout)+
     $     hess(jout+1,jout)*hess(jout+1,jout)
      gden=SQRT(gden)
      givn(1,jout)=hess(jout,jout)/gden
      givn(2,jout)=hess(jout+1,jout)/gden
c-----------------------------------------------------------------------
c     Apply the new rotation to the new Hessenberg column and the
c     error vector.
c-----------------------------------------------------------------------
      gc=givn(1,jout)
      gs=givn(2,jout)
      rtmp=gc*hess(jout,jout)+gs*hess(jout+1,jout)
      hess(jout+1,jout)=gc*hess(jout+1,jout)-gs*hess(jout,jout)
      hess(jout,jout)=rtmp

      errvec(jout+1)=-gs*errvec(jout)
      errvec(jout)=gc*errvec(jout)
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_grot
c-----------------------------------------------------------------------
c     subprogram 10. iter_3d_min.
c     find the minimizing vector of coefficients for the orthogonal
c     basis vectors.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_min
      USE math_tran

      INTEGER(i4) :: ii,jj
      LOGICAL :: singular

      REAL(r8) :: rtmp,gc,gs
c-----------------------------------------------------------------------
c     if the usefom flag is set, apply Arnoldi's full orthogonalization
c     method instead of minimizing the residual.  when the fixed
c     iteration count flag is not set, FOM is accomplished
c     by undoing the last Givens rotation and solving the truncated
c     system.
c-----------------------------------------------------------------------
      IF (usefom.AND..NOT.fixm) THEN
        gc=givn(1,n_gm)
        gs=givn(2,n_gm)
        rtmp=gc*hess(n_gm,n_gm)
        hess(n_gm+1,n_gm)=gs*hess(n_gm,n_gm)
        hess(n_gm,n_gm)=rtmp
        rtmp=gc*errvec(n_gm)-gs*errvec(n_gm+1)
        errvec(n_gm+1)=gs*errvec(n_gm)+gc*errvec(n_gm+1)
        errvec(n_gm)=rtmp
      ENDIF
c-----------------------------------------------------------------------
c     when the fixed iteration count flag is set, the Hessenberg matrix
c     is not reduced at this stage.  with FOM, we solve
c
c       H_m.minvec = errvec(1:n_gm)
c
c     and with GMRES we minimize including the last direction vector,
c
c       Hbar_m^T.Hbar_m.minvec = errvec(1:n_gm+1)
c
c     where rows 1:n_gm of Hbar_m are H_m, and the n_gm+1 row of Hbar_m
c     is zero except for the last entry.
c-----------------------------------------------------------------------
      IF (fixm) THEN
        IF (usefom) THEN
          DO ii=2,n_gm
            rtmp=hess(ii,ii-1)/hess(ii-1,ii-1)
            hess(ii,:)=hess(ii,:)-rtmp*hess(ii-1,:)
            errvec(ii)=errvec(ii)-rtmp*errvec(ii-1)
          ENDDO
        ELSE
          errvec(1:n_gm)=
     $      MATMUL(TRANSPOSE(hess(1:n_gm+1,1:n_gm)),errvec(1:n_gm+1))
          hess(1:n_gm,1:n_gm)=
     $      MATMUL(TRANSPOSE(hess(1:n_gm+1,1:n_gm)),
     $                       hess(1:n_gm+1,1:n_gm))
          CALL math_solve_q1_sym(n_gm,1_i4,hess(1:n_gm,1:n_gm),
     $                           minvec(1:n_gm),errvec(1:n_gm),'both',
     $                           singular)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     finish solving the system including the standard case where
c     usefom and fixm are both false.
c-----------------------------------------------------------------------
      IF (usefom.OR..NOT.fixm) THEN
        minvec(n_gm)=errvec(n_gm)/hess(n_gm,n_gm)
        DO ii=n_gm-1,1,-1
          minvec(ii)=errvec(ii)
          DO jj=ii+1,n_gm
            minvec(ii)=minvec(ii)-hess(ii,jj)*minvec(jj)
          ENDDO
          minvec(ii)=minvec(ii)/hess(ii,ii)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_min
c-----------------------------------------------------------------------
c     subprogram 11. iter_3d_hcheck.
c     recompute the Vm^T.A.Vm matrix as a test after performing
c     the standard orthogonalization steps.  this should produce a
c     Hessenberg matrix if the orthogonalization is accurate.  
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_hcheck(dot_routine,mat_str,fac_str,ctmp,
     $                          dir,nqty,poly_deg,nqdsc,nbdsc,nrb,ntotb,
     $                          nfour,precon,c2zee,c2rhs,c2tmp,c2pck,
     $                          minus_str,plus_str)
      USE matrix_mod
      USE seam_storage_mod
      USE edge
      USE iter_cg

      TYPE(complex_matrix_type), DIMENSION(:), INTENT(IN) :: mat_str
      TYPE(complex_matrix_type), DIMENSION(:,:), INTENT(IN), OPTIONAL ::
     $                           minus_str,plus_str
      TYPE(complex_factor_type), DIMENSION(:), INTENT(IN) :: fac_str
      TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: ctmp,dir
      INTEGER(i4), INTENT(IN) :: nqty,poly_deg,nrb,ntotb,nfour
      INTEGER(i4), INTENT(IN) :: nqdsc,nbdsc  !  for discontin. bases
      CHARACTER(*), INTENT(IN) :: precon
      TYPE(cvector_2D_type), DIMENSION(:), POINTER :: c2zee,c2rhs,c2tmp,
     $                       c2pck

      INTEGER(i4) :: ii,jj,ibl,mxb,myb
      REAL(r8), DIMENSION(:,:), ALLOCATABLE :: hnew
      TYPE(cvector_type), DIMENSION(:), ALLOCATABLE :: ctmp2
c-----------------------------------------------------------------------
c     interface block for the external subroutine that calls get_rhs
c     for the matrix-free dot-product computation.
c-----------------------------------------------------------------------
      INTERFACE
        SUBROUTINE dot_routine(oper,prod,bc_oper)
        USE vector_type_mod
        USE local
        TYPE(cvector_type), DIMENSION(:), INTENT(INOUT) :: oper,prod
        LOGICAL, INTENT(IN) :: bc_oper
        END SUBROUTINE dot_routine
      END INTERFACE
c-----------------------------------------------------------------------
c     allocate storage for the inner products, then loop
c     over the exiting basis vectors and compute A.v_jj for each.
c-----------------------------------------------------------------------
      ALLOCATE(hnew(n_gm,n_gm))
 
      ALLOCATE(ctmp2(ntotb))
      DO ibl=1,nrb
        mxb=SIZE(dir(ibl)%arr,2)-1
        myb=SIZE(dir(ibl)%arr,3)-1
        IF (nqdsc>0) THEN
          CALL vector_type_alloc(ctmp2(ibl),poly_deg,mxb,myb,nqty,nfour,
     $                           nbdsc,nqdsc)
        ELSE
          CALL vector_type_alloc(ctmp2(ibl),poly_deg,mxb,myb,nqty,nfour)
        ENDIF
      ENDDO
      DO ibl=nrb+1,ntotb
        mxb=SIZE(dir(ibl)%arr,2)-1
        CALL vector_type_alloc(ctmp2(ibl),1_i4,mxb,0_i4,nqty,nfour)
      ENDDO
c-----------------------------------------------------------------------
c     with left preconditioning, A->P^-1.A, and with right precondition-
c     ing, A->A.P^-1, where P is the approximation of A.
c-----------------------------------------------------------------------

      DO jj=1,n_gm
        IF (leftprec) THEN
          DO ibl=1,ntotb
            dir(ibl)=ep_d(jj)%base(ibl)
          ENDDO
        ELSE
          DO ibl=1,ntotb
            ctmp(ibl)=ep_d(jj)%base(ibl)
          ENDDO
          CALL iter_3d_cg_pre(mat_str,fac_str,ctmp,ctmp2,c2rhs,c2zee,
     $                        c2tmp,c2pck,poly_deg,nqty,nbdsc,nqdsc,nrb,
     $                        ntotb,nfour,precon,dot_routine,dir,
     $                        minus_str=minus_str,plus_str=plus_str)
          DO ibl=1,ntotb
            dir(ibl)=ctmp2(ibl)
          ENDDO
        ENDIF

        CALL dot_routine(dir,ctmp,.false.)
       
        IF (leftprec) THEN
          DO ibl=1,ntotb
            ctmp2(ibl)=ctmp(ibl)
          ENDDO
          CALL iter_3d_cg_pre(mat_str,fac_str,ctmp,ctmp2,c2rhs,c2zee,
     $                        c2tmp,c2pck,poly_deg,nqty,nbdsc,nqdsc,nrb,
     $                        ntotb,nfour,precon,dot_routine,dir,
     $                        minus_str=minus_str,plus_str=plus_str)
        ENDIF
c-----------------------------------------------------------------------
c       find the inner product of each basis in Vm with A.v_jj.
c-----------------------------------------------------------------------
        DO ii=1,n_gm
          CALL iter_3d_cg_dot(hnew(ii,jj),ep_d(ii)%base,ctmp,seam,ntotb)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     write the values of the inner products to a file.
c-----------------------------------------------------------------------
      WRITE(nim_wr,'(a,/)') " Hessenberg matrix Vm^T.A.Vm:"
      DO jj=1,n_gm
        DO ii=1,n_gm
          WRITE(nim_wr,'(2(a,i4),es15.8)')
     $      "   row ",ii,"  col ",jj,hnew(ii,jj)
        ENDDO
      ENDDO
      WRITE(nim_wr,'(/)')

      DO ibl=1,ntotb
        CALL vector_type_dealloc(ctmp2(ibl))
      ENDDO
      DEALLOCATE(ctmp2)
      DEALLOCATE(hnew)
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_hcheck
c-----------------------------------------------------------------------
c     subprogram 12. iter_3d_offload. 
c     load a cvector_2D_type into a 1D array that will be used for
c     communication buffering.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_offload(offstr,cvec2,nbl,icomp,ipass,npass)

      TYPE(iter_3d_off_type), INTENT(INOUT) :: offstr
      TYPE(cvector_2D_type), DIMENSION(:), INTENT(IN) :: cvec2
      INTEGER(i4), INTENT(IN) :: nbl,icomp,ipass,npass

      INTEGER(i4) :: ii,ibl,iq,ib,ix,iy,nq,nb,nx,ny,isnd,ierror
c-----------------------------------------------------------------------
c     if this is not the first pass, check that the previously loaded
c     data has been sent before overwriting the buffer.
c-----------------------------------------------------------------------
      IF (ipass>1) THEN
        CALL mpi_waitall(offstr%nsend,offstr%sendreq(1),mpistat(1,1),
     $                   ierror)
      ENDIF
c-----------------------------------------------------------------------
c     if this is the last pass, passing the above check is all that is
c     needed.
c-----------------------------------------------------------------------
      IF (ipass==npass) RETURN
c-----------------------------------------------------------------------
c     loop over the grid blocks, determine array sizes, and load the
c     buffer.
c-----------------------------------------------------------------------
      ii=1
      DO ibl=1,nbl
        nq=SIZE(cvec2(ibl)%arr,1)
        nx=SIZE(cvec2(ibl)%arr,2)-1
        ny=SIZE(cvec2(ibl)%arr,3)-1
        DO iy=0,ny
          DO ix=0,nx
            DO iq=1,nq
              offstr%offsend(ii)=cvec2(ibl)%arr(iq,ix,iy)
              ii=ii+1
            ENDDO
          ENDDO
        ENDDO
        IF (ASSOCIATED(cvec2(ibl)%arrh)) THEN
          nb=SIZE(cvec2(ibl)%arrh,2)
          DO iy=0,ny
            DO ix=1,nx
              DO ib=1,nb
                DO iq=1,nq
                  offstr%offsend(ii)=cvec2(ibl)%arrh(iq,ib,ix,iy)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
        IF (ASSOCIATED(cvec2(ibl)%arrv)) THEN
          nb=SIZE(cvec2(ibl)%arrv,2)
          DO iy=1,ny
            DO ix=0,nx
              DO ib=1,nb
                DO iq=1,nq
                  offstr%offsend(ii)=cvec2(ibl)%arrv(iq,ib,ix,iy)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
        IF (ASSOCIATED(cvec2(ibl)%arri)) THEN
          nb=SIZE(cvec2(ibl)%arri,2)
          DO iy=1,ny
            DO ix=1,nx
              DO ib=1,nb
                DO iq=1,nq
                  offstr%offsend(ii)=cvec2(ibl)%arri(iq,ib,ix,iy)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     start a set of asynchronous sends to all processors that need
c     the loaded data.
c-----------------------------------------------------------------------
      DO isnd=1,offstr%nsend
        CALL mpi_isend(offstr%offsend(1),offstr%size,mpi_nim_comp,
     $                 offstr%sendpr(isnd),icomp,mpi_comm_world,
     $                 offstr%sendreq(isnd),ierror)
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_offload
c-----------------------------------------------------------------------
c     subprogram 13. iter_3d_offunld. 
c     unload a 1D array for communication buffering into a
c     cvector_2D_type.
c
c     the parallel communication call used at the top of the routine
c     tests if the data in the buffer is current wrt a Jacobi step
c     (data from the previous pass).
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_offunld(offstr,cvec2,nbl,icomp,ipass,npass)

      TYPE(iter_3d_off_type), INTENT(INOUT) :: offstr
      TYPE(cvector_2D_type), DIMENSION(:), INTENT(INOUT) :: cvec2
      INTEGER(i4), INTENT(IN) :: nbl,icomp,ipass,npass

      INTEGER(i4) :: ii,ibl,iq,ib,ix,iy,nq,nb,nx,ny,ierror
c-----------------------------------------------------------------------
c     check if data from the previous pass is available.  note that
c     datapass is initialized at 0 before the first pass, when there
c     is nothing to receive.
c-----------------------------------------------------------------------
      IF (offstr%datapass<ipass-1) THEN
        CALL mpi_wait(offstr%recvreq,mpistat(1,1),ierror)
        offstr%datapass=ipass-1
      ENDIF
c-----------------------------------------------------------------------
c     loop over the grid blocks, determine array sizes, and unload the
c     buffer.
c-----------------------------------------------------------------------
      ii=1
      DO ibl=1,nbl
        nq=SIZE(cvec2(ibl)%arr,1)
        nx=SIZE(cvec2(ibl)%arr,2)-1
        ny=SIZE(cvec2(ibl)%arr,3)-1
        DO iy=0,ny
          DO ix=0,nx
            DO iq=1,nq
              cvec2(ibl)%arr(iq,ix,iy)=offstr%offrecv(ii)
              ii=ii+1
            ENDDO
          ENDDO
        ENDDO
        IF (ASSOCIATED(cvec2(ibl)%arrh)) THEN
          nb=SIZE(cvec2(ibl)%arrh,2)
          DO iy=0,ny
            DO ix=1,nx
              DO ib=1,nb
                DO iq=1,nq
                  cvec2(ibl)%arrh(iq,ib,ix,iy)=offstr%offrecv(ii)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
        IF (ASSOCIATED(cvec2(ibl)%arrv)) THEN
          nb=SIZE(cvec2(ibl)%arrv,2)
          DO iy=1,ny
            DO ix=0,nx
              DO ib=1,nb
                DO iq=1,nq
                  cvec2(ibl)%arrv(iq,ib,ix,iy)=offstr%offrecv(ii)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
        IF (ASSOCIATED(cvec2(ibl)%arri)) THEN
          nb=SIZE(cvec2(ibl)%arri,2)
          DO iy=1,ny
            DO ix=1,nx
              DO ib=1,nb
                DO iq=1,nq
                  cvec2(ibl)%arri(iq,ib,ix,iy)=offstr%offrecv(ii)
                  ii=ii+1
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_offunld
c-----------------------------------------------------------------------
c     subprogram 14. iter_3d_offrready. 
c     indicate that the receive buffer is available for new data.
c-----------------------------------------------------------------------
      SUBROUTINE iter_3d_offrready(offstr,icomp,ipass,npass)

      TYPE(iter_3d_off_type), INTENT(INOUT) :: offstr
      INTEGER(i4), INTENT(IN) :: icomp,ipass,npass

      INTEGER(i4) :: ierror
c-----------------------------------------------------------------------
c     if this is not the last pass, post the receive to get the next
c     set of data.
c-----------------------------------------------------------------------
      IF (ipass<npass) THEN
        CALL mpi_irecv(offstr%offrecv(1),offstr%size,mpi_nim_comp,
     $                 offstr%recvpr,icomp,mpi_comm_world,
     $                 offstr%recvreq,ierror)
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE iter_3d_offrready

c-----------------------------------------------------------------------
c     close module.
c-----------------------------------------------------------------------
      END MODULE iter_3d_cg_mod
