c-----------------------------------------------------------------------
c     file nimeq_ints.f
c     module that includes integrand routines used in the diagnostic
c     finite element integrations in nimplot.
c
c     note that nimplot does not use quadrature point storage, so the
c     interpolations are made on the fly with generic_all_eval instead
c     of setting pointers like in nimrod's integrands.f.  this also
c     means that the phi component of the interpolated equilibrium B is
c     R*B_eq_phi, while the phi component of equil J is J_eq_phi/R.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     code organization.
c-----------------------------------------------------------------------
c     1.  get_mass.
c     2.  mag_op.
c     3.  delstar_op.
c     4.  delstlj_op.
c     5.  delstar_rhs.
c     6.  gs_rhs.
c     7.  err_gs.
c     8.  get_B.
c     9.  jphi_rhs.
c     10. delstarlj_dot.
c-----------------------------------------------------------------------
c     declarations.
c-----------------------------------------------------------------------
      MODULE nimeq_ints

      USE local
      USE rblock_type_mod
      USE tblock_type_mod
      USE lagr_quad_mod
      USE tri_linear
      USE generic_evals
      USE math_tran
      USE physdat
      USE input
      USE input_eq
      USE global
      USE fft_mod
      IMPLICIT NONE

      CONTAINS

c-----------------------------------------------------------------------
c     subprogram 1. get_mass.
c     compute the integrand used for the mass matrix.
c-----------------------------------------------------------------------
      SUBROUTINE get_mass(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,5)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,0_i4,poly_degree)
c-----------------------------------------------------------------------
c     find alpha(jv)*alpha(iv).
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM(alpha(:,:,jv)*alpha(:,:,iv),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE get_mass
c-----------------------------------------------------------------------
c     subprogram 2. mag_op.
c     this routine provides a mass matrix for a 3-vector, but the
c     first two components are multiplied by R.
c-----------------------------------------------------------------------
      SUBROUTINE mag_op(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,5)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,0_i4,poly_degree)
c-----------------------------------------------------------------------
c     find alpha(jv)*alpha(iv), multiplied by R for the first two
c     vector components.
c-----------------------------------------------------------------------
      int=0._r8
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM(alpha(:,:,jv)*alpha(:,:,iv)*bigr,1)
          int(2,2,:,jv,iv)=SUM(alpha(:,:,jv)*alpha(:,:,iv)*bigr,1)
          int(3,3,:,jv,iv)=SUM(alpha(:,:,jv)*alpha(:,:,iv),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE mag_op
c-----------------------------------------------------------------------
c     subprogram 3. delstar_op.
c     the delstar operator acting on a field psi can be written as
c
c       div[ R**2 grad( psi/R**2 ) ]
c
c     the operator computed here is used to find psi/R**2 (not psi)
c     and in weak form after integration by parts, the integrand is
c
c       R**2*grad(test_function).grad(basis_function)
c
c     and the minus sign is moved to the right side of the equation.
c-----------------------------------------------------------------------
      SUBROUTINE delstar_op(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,5)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     find R^2* grad(alpha(jv)).grad(alpha(iv)).
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM((dalpdr(:,:,jv)*dalpdr(:,:,iv)+
     $           dalpdz(:,:,jv)*dalpdz(:,:,iv))*bigr(:,:)**2,1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE delstar_op
c-----------------------------------------------------------------------
c     subprogram 4. delstlj_op.
c     this version of the operator is used to solve a 2-vector system,
c     where the first component is lambda=psi/R**2 and the second is
c     mu0*J_phi/R.
c-----------------------------------------------------------------------
      SUBROUTINE delstlj_op(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,5)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     find R^2* grad(alpha(jv)).grad(alpha(iv)) for the first diagonal
c     entry and R*alpha(jv)*alpha(iv) for the second.
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM((dalpdr(:,:,jv)*dalpdr(:,:,iv)+
     $           dalpdz(:,:,jv)*dalpdz(:,:,iv))*bigr(:,:)**2,1)
          int(2,1,:,jv,iv)=0._r8
          int(1,2,:,jv,iv)=0._r8
          int(2,2,:,jv,iv)=SUM(alpha(:,:,jv)*alpha(:,:,iv)*
     $                         bigr(:,:)**2,1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE delstlj_op
c-----------------------------------------------------------------------
c     subprogram 5. delstar_rhs.
c     compute the rhs integrand used for the del_star solve.
c     this is -alpha*mu0*Jphi*R, and the minus sign appears due to an
c     integration-by-parts on the lhs.
c-----------------------------------------------------------------------
      SUBROUTINE delstar_rhs(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: ja,
     $     bez,ber,be
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: ja_eq
      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: psi,
     $     psiz,psir
      INTEGER(i4) :: iv,nv,im
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,3)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      CALL generic_all_eval(rb%ja_eq,tb%ja_eq,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      ja_eq,dv,dv,0_i4)
      CALL generic_all_eval(rb%be,tb%be,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      be,ber,bez,1_i4)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1._r8)
      CALL generic_all_eval(rb%rwork1,tb%rwork1,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      psi,psir,psiz,1_i4)
c-----------------------------------------------------------------------
c     convert the phi components of equilibrium fields to cylindrical.
c-----------------------------------------------------------------------
      ja_eq(3,:,:)=mu0*ja_eq(3,:,:)*bigr
      DO im=1,nmodes
        IF (keff(im)==0) THEN
          ja(:,:,:,im)=REAL(ja(:,:,:,im))
          IF (nonlinear) ja_eq=ja_eq+ja(:,:,:,im)
          EXIT
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     find alpha*mu*Jphi*R + grad(alpha)*grad(psi_s/R^2)*R^2, where
c     psi_s is from the surface only.
c-----------------------------------------------------------------------
      DO iv=1,nv
        int(1,:,iv)=SUM(-(alpha(:,:,iv)*ja_eq(3,:,:)*bigr
     $                 +(dalpdr(:,:,iv)*psir(1,:,:)+
     $                   dalpdz(:,:,iv)*psiz(1,:,:))*bigr(:,:)**2),1)
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE delstar_rhs
c-----------------------------------------------------------------------
c     subprogram 6. gs_rhs.
c     compute the rhs integrand used for the gs solve,
c       -alpha ( F*F' + mu0*R**2*P' )
c-----------------------------------------------------------------------
      SUBROUTINE gs_rhs(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: f,press,dp
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: psib,
     $     psizb,psirb,psi
      REAL(r8), DIMENSION(1,1,1) :: dv
      INTEGER(i4) :: iv,nv,im
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,3)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      IF (gs_type/='free')
     $  CALL generic_all_eval(rb%rwork1,tb%rwork1,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,psib,psirb,psizb,1_i4)
      CALL generic_ptr_set(rb%qrwork1,tb%qrwork1,tb%tgeom,
     $                     inode,f,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                     inode,press,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find alpha*(F*dF/dpsi+mu0*R**2*dP/dpsi) -
c     grad(alpha)*grad(psi_s/R^2)*R^2, where the second term is from
c     the surface flux.  if this is a free-boundary computation, there
c     is no separate inhomogeneous surface term.
c
c     note that press holds mu0*dP/dpsi at this point, not pressure, and
c     f has F*dF/dpsi.
c
c     the second quantity is without the surface-flux term for obtaining
c     a projection of mu0*J_phi/R, directly, during free-boundary
c     computations.
c-----------------------------------------------------------------------
      IF (SIZE(int,1)==1) THEN
        DO iv=1,nv
          int(1,:,iv)=SUM(
     $          alpha(:,:,iv)*(f(1,:,:)+bigr(:,:)**2*press(1,:,:))
     $       -(dalpdr(:,:,iv)*psirb(1,:,:)+
     $         dalpdz(:,:,iv)*psizb(1,:,:))*bigr(:,:)**2,1)
        ENDDO
      ELSE
        DO iv=1,nv
          int(1,:,iv)=SUM(
     $          alpha(:,:,iv)*(f(1,:,:)+bigr(:,:)**2*press(1,:,:)),1)
          int(2,:,iv)=-int(1,:,iv)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE gs_rhs
c-----------------------------------------------------------------------
c     subprogram 7. err_gs.
c     compute the integrand for rhs-delstar(psi).  this routine is
c     used for error estimates
c-----------------------------------------------------------------------
      SUBROUTINE err_gs(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz    
      REAL(r8), DIMENSION(:,:,:), POINTER :: f,press,dp
      INTEGER(i4) :: iv,nv,im,i,j
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: psi,
     $     psiz,psir,psib,psizb,psirb
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: jaeq
      REAL(r8), DIMENSION(1,1,1) :: dv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,3)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      CALL generic_all_eval(rb%rwork2,tb%rwork2,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      psi,psir,psiz,1_i4)
      CALL generic_ptr_set(rb%qrwork1,tb%qrwork1,tb%tgeom,
     $                     inode,f,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                     inode,press,dp,dp,0_i4)
      IF (gs_type=='free') THEN
        CALL generic_all_eval(rb%ja_eq,tb%ja_eq,rb%dxdr,rb%dydr,rb%dxdz,
     $                        rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                        jaeq,dv,dv,0_i4)
      ELSE
        CALL generic_all_eval(rb%rwork1,tb%rwork1,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,psib,psirb,psizb,1_i4)
c-----------------------------------------------------------------------
c       combine interior and surface contributions to psi/R**2
c-----------------------------------------------------------------------
        psi  = psi  + psib
        psir = psir + psirb
        psiz = psiz + psizb
      ENDIF
c-----------------------------------------------------------------------
c     find grad(alpha)*grad(psi/R^2)*R^2
c       -alpha*(F*dF/dpsi+mu0*R**2*dP/dpsi)
c
c     note that press holds mu0*dP/dpsi at this point, not pressure, and
c     f has F*dF/dpsi.
c
c     when there is a second vector component, mu0*J_phi/R, from the
c     linear relation with delstar(psi) for use in free-boundary solves.
c-----------------------------------------------------------------------
      IF (integrand_flag=="all terms") THEN
        IF (SIZE(int,1)==1) THEN
          DO iv=1,nv
            int(1,:,iv) = SUM(
     $        bigr(:,:)**2*(dalpdr(:,:,iv)*psir(1,:,:)+
     $                      dalpdz(:,:,iv)*psiz(1,:,:))-
     $        (f(1,:,:)+bigr(:,:)**2*press(1,:,:))*alpha(:,:,iv),1)
          ENDDO
        ELSE
          DO iv=1,nv
            int(1,:,iv) = SUM(
     $        bigr(:,:)**2*(dalpdr(:,:,iv)*psir(1,:,:)+
     $                      dalpdz(:,:,iv)*psiz(1,:,:))-
     $        (f(1,:,:)+bigr(:,:)**2*press(1,:,:))*alpha(:,:,iv),1)
            int(2,:,iv) = SUM(
     $        alpha(:,:,iv)*(bigr(:,:)**2*(jaeq(3,:,:)+press(1,:,:))+
     $                       f(1,:,:)),1)
          ENDDO
        ENDIF
      ELSE          !    just delstar part
        IF (SIZE(int,1)==1) THEN
          DO iv=1,nv
            int(1,:,iv) = SUM(
     $        bigr(:,:)**2*(dalpdr(:,:,iv)*psir(1,:,:)+
     $                      dalpdz(:,:,iv)*psiz(1,:,:)),1)
          ENDDO
        ELSE
          DO iv=1,nv
            int(1,:,iv) = SUM(
     $        bigr(:,:)**2*(dalpdr(:,:,iv)*psir(1,:,:)+
     $                      dalpdz(:,:,iv)*psiz(1,:,:)),1)
            int(2,:,iv)=0._r8
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE err_gs
c-----------------------------------------------------------------------
c     subprogram 8. get_B
c     calculate the RHS used to determine B from psi and F.
c       RBr = -dpsi/dz
c       RBz = dpsi/dr
c       RBp = F
c-----------------------------------------------------------------------
      SUBROUTINE get_B(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: psi,psiz,psir,
     $          rbp
      REAL(r8), DIMENSION(1,1,1) :: dv
      INTEGER(i4) :: iv,im,nv,ix,iy,nx,ny
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and evaluate the
c     incoming vector.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,0_i4,poly_degree)
      CALL generic_all_eval(rb%rwork2,tb%rwork2,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      psi,psir,psiz,1_i4)
      CALL generic_all_eval(rb%rwork3,tb%rwork3,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      rbp,dv,dv,0_i4)
c-----------------------------------------------------------------------
c     B-poloidal is grad(phi)Xgrad(psi), but use the expansion for
c     psi/R**2 that includes surface contributions at this point
c     (local array psi):
c
c     R*Bpol=phi-hat X (R**2*grad(psi/R**2)+2*R*(psi/R**2)*R_hat)
c
c     the third component is R*B_phi = F.
c-----------------------------------------------------------------------
      nv=SIZE(int,3)

      IF (geom=='tor') THEN
        DO iv=1,nv
          int(1,:,iv)=SUM(-alpha(:,:,iv)* psiz(1,:,:)*bigr**2,1)
          int(2,:,iv)=SUM( alpha(:,:,iv)*(psir(1,:,:)*bigr**2+
     $                         2._r8*bigr*psi (1,:,:)),1)
          int(3,:,iv)=SUM( alpha(:,:,iv)*rbp(1,:,:),1)
        ENDDO
      ELSE
        DO iv=1,nv
          int(1,:,iv)=SUM(-alpha(:,:,iv)*psiz(1,:,:),1)
          int(2,:,iv)=SUM( alpha(:,:,iv)*psir(1,:,:),1)
          int(3,:,iv)=SUM( alpha(:,:,iv)*rbp(1,:,:),1)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE get_B
c-----------------------------------------------------------------------
c     subprogram 9. jphi_rhs.
c     compute mu0*Jphi/R from the rhs of the GS equation.
c-----------------------------------------------------------------------
      SUBROUTINE jphi_rhs(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz    
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: psi,rlen
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: bigr2,psival,rlenval
      REAL(r8), EXTERNAL :: f_func,p_func
      INTEGER(i4) :: iv,nv,ix,iy,ncx,ncy
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.  rwork2 now has the
c     full psi/R**2 distribution across the interior and at the wall.
c
c     fllen holds the relative length of the field, as used to check
c     topology.
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      CALL generic_all_eval(rb%rwork2,tb%rwork2,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      psi,dv,dv,0_i4)
      CALL generic_all_eval(rb%fllen,tb%fllen,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      rlen,dv,dv,0_i4)
c-----------------------------------------------------------------------
c     find -alpha*(F*dF/dpsi/R**2+mu0*dP/dpsi).  here, we use the
c     external functions f_func and p_func which provide the user-
c     specified distributions according to psi.
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO iy=1,ncy
          int(1,iy,iv)=0._r8
          DO ix=1,ncx
            bigr2=bigr(ix,iy)**2
            psival=psi(1,ix,iy)*bigr2
            rlenval=rlen(1,ix,iy)
            int(1,iy,iv)=int(1,iy,iv)-alpha(ix,iy,iv)*(
     $        f_func(psival,0_i4,rlenval)*
     $        f_func(psival,1_i4,rlenval)/bigr2+
     $        p_func(psival,1_i4,rlenval) )
          ENDDO
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE jphi_rhs
c-----------------------------------------------------------------------
c     subprogram 10. delstarlj_dot.
c     compute the effect of the delstar operator applied to one 
c     component of a direction vector from a matrix-free computation,
c     and apply a mass matrix to the second.
c-----------------------------------------------------------------------
      SUBROUTINE delstarlj_dot(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: laj,
     $          lajr,lajz
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: bigr2
      INTEGER(i4) :: iv,nv,im
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,3)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     the direction vector from the solve is in the 2nd and 3rd 
c     components of ja_eq.  the 2nd component is lambda=psi/R**2,
c     and the third is mu0*J_phi/R.
c-----------------------------------------------------------------------
      CALL generic_all_eval(rb%ja_eq,tb%ja_eq,rb%dxdr,rb%dydr,rb%dxdz,
     $                      rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      laj,lajr,lajz,1_i4)
c-----------------------------------------------------------------------
c     grad(alpha)*grad(lambda)*R^2 and alpha*mu0*J_phi*R:
c-----------------------------------------------------------------------
      bigr2=bigr(:,:)**2
      DO iv=1,nv
        int(1,:,iv)=SUM((dalpdr(:,:,iv)*lajr(2,:,:)+
     $                   dalpdz(:,:,iv)*lajz(2,:,:))*bigr2(:,:),1)
        int(2,:,iv)=SUM(alpha(:,:,iv)*bigr2(:,:)*laj(3,:,:),1)
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE delstarlj_dot

c-----------------------------------------------------------------------
c     close module.
c-----------------------------------------------------------------------
      END MODULE nimeq_ints
