c-----------------------------------------------------------------------
c     file integrands.f
c     module that includes integrand routines used in the finite-
c     element integrations for advancing the solution.
c
c     integrand routines for matrices and rhs have been separated into
c     separate modules to facilitate compilation.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     code organization.
c-----------------------------------------------------------------------
c     module integrands_rhs
c     subprogram 1. get_vol.
c     subprogram 2. curl
c     subprogram 3. brhs_mhd.
c     subprogram 3.1 brhs_hyp.
c     subprogram 4. b_3deta_dot.
c     subprogram 5. brhs_hmhd.
c     subprogram 6. hall_cor.
c     subprogram 7. b_hmhd_dot.
c     subprogram 8. tirhs.
c     subprogram 9. terhs.
c     subprogram 10. t_aniso_dot.
c     subprogram 11. vrhs.
c     subprogram 12. advect_cor.
c     subprogram 13. v_aniso_dot.
c     subprogram 13.1 v_3dsi_dot.
c     subprogram 14. divb_rhs.
c     subprogram 15. ndrhs.
c     subprogram 16. cont_dot.
c     module integrands_mat
c     subprogram 17. get_mass.
c     subprogram 18. curl_de_iso.
c     subprogram 19. curl_de_ciso.
c     subprogram 20. b_hmhd_op.
c     subprogram 21. b_hmhd_cpop.
c     subprogram 21.1 b_hyp_op.
c     subprogram 22. n_iso_op.
c     subprogram 23. t_aniso_op.
c     subprogram 24. t_aniso_cpop.
c     subprogram 25. vec_lap_op.
c     subprogram 26. v_aniso_op.
c     subprogram 27. grad_div.
c     subprogram 28. cont_op.
c     module integrands
c-----------------------------------------------------------------------
c     module containing integrands for the rhs of equations.
c-----------------------------------------------------------------------
      MODULE integrands_rhs

      USE local
      USE rblock_type_mod
      USE tblock_type_mod
      USE lagr_quad_mod
      USE tri_linear
      USE generic_evals
      USE math_tran
      USE physdat
      USE input
      USE global
      USE fft_mod
      IMPLICIT NONE

      CONTAINS
c-----------------------------------------------------------------------
c     subprogram 1. get_vol.
c     compute the cell volume less a factor of 2*pi.
c-----------------------------------------------------------------------
      SUBROUTINE get_vol(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: wj
c-----------------------------------------------------------------------
c     put dVol and dArea in the vertex index used for cell-centered
c     integrals.  dArea/r is used for the computation of F.
c
c     the sum of the test functions produces the quadrature weight times
c     the Jacobian.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,0_i4,poly_degree)
      wj=SUM(alpha,3)

      int(1,:,1)=SUM(wj,1)
      int(2,:,1)=SUM(wj/bigr,1)
      int(3,:,1)=SUM(wj/bigr**2,1)
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE get_vol
c-----------------------------------------------------------------------
c     subprogram 2. curl
c     form the integrand for a curl operation.
c     the incoming vector and its derivatives have been evaluated at
c     the quadrature points and stored in the qwork1 structure.
c-----------------------------------------------------------------------
      SUBROUTINE curl(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: tor_fac
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: vec,dvecr,dvecz
      INTEGER(i4) :: iv,im,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and evaluate the
c     incoming vector.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,0_i4,poly_degree)
      CALL generic_ptr_set(rb%qwork1,tb%qwork1,tb%tgeom,inode,
     $                     vec,dvecr,dvecz,1_i4)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     set factor to collect extra terms according to geom.
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        tor_fac=1
      ELSE
        tor_fac=0
      ENDIF
c-----------------------------------------------------------------------
c     construct alpha.curl(vector) in cylindrical components
c     (poloidal plane).
c     note:  vector storage is (x,y,z) or (r,z,phi) for each complex
c     Fourier component.
c-----------------------------------------------------------------------
      DO im=1,nmodes
        DO iv=1,nv
          int(1,:,iv,im)=SUM( alpha(:,:,iv)*( dvecz(3,:,:,im)
     $                          -(0,1)*keff(im)*vec(2,:,:,im)/bigr ),1)
          int(2,:,iv,im)=SUM(-alpha(:,:,iv)*( dvecr(3,:,:,im)
     $      +(tor_fac*vec(3,:,:,im)-
     $        (0,1)*keff(im)*vec(1,:,:,im))/bigr ),1)
          int(3,:,iv,im)=SUM(alpha(:,:,iv)*( dvecr(2,:,:,im)
     $                                      -dvecz(1,:,:,im) ),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE curl
c-----------------------------------------------------------------------
c     subprogram 3. brhs_mhd.
c     compute the integrand used in the rhs of the semi-implicit
c     magnetic field equation for the mhd time-split.
c-----------------------------------------------------------------------
      SUBROUTINE brhs_mhd(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: aldis,dalddr,dalddz
      REAL(r8), DIMENSION(:,:,:), POINTER :: be_eq,ve_eq,e_a,e_ar,
     $          e_az,elecd_phi,elecd_n0,elecd_eq,ja_eq,tele_eq,dp,ds2
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve,real_be,real_eexp
      REAL(r8), DIMENSION(:,:,:), ALLOCATABLE :: real_j
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: diff
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: eleq_1d
      REAL(r8), DIMENSION(3,SIZE(bigr,1)*SIZE(bigr,2)) :: jaeq_1d 
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez,ve,tele,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             ja_old,eexp
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2),nmodes) :: divb
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: auxb
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,jq,nv,npol,ncx,ncy,ip,nvc,nvd
      REAL(r8) :: disdbfac,fhyp,fhdb
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and set pointers to
c     the velocity and magnetic field at the beginning of the time
c     split.  current density for resistive diffusion must be based on
c     this magnetic field.
c
c     if poly_divb is non-negative, divergence error is controlled with
c     an auxiliary field that is discontinuous at element borders, and
c     these bases are also needed.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      IF (poly_divb>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',aldis,
     $                          dalddr,dalddz,0_i4,poly_divb,
     $                          polydmin=poly_divb_min,
     $                          polydmax=poly_divb_max)
        nvd=SIZE(aldis,3)
      ENDIF

      CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,ve,dcp,dcp,0_i4)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,ber,bez,1_i4)
      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
      IF ((elecd>0.OR.(hyp_eta>0.OR.hyp_dbd>0).AND.
     $    .NOT.split_hypeta).AND.
     $    integrand_flag(1:3)=='mhd') THEN
        CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja_old,1/mu0)
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_eq,tb%qelecd_eq,tb%tgeom,
     $                         inode,elecd_eq,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                         inode,ja_eq,dp,dp,0_i4)
        ENDIF
        IF (.NOT.nonlinear.AND.eta_model=='eta full') THEN
          CALL generic_ptr_set(rb%qtele_eq,tb%qtele_eq,tb%tgeom,
     $                         inode,tele_eq,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,
     $                         inode,tele,dcp,dcp,0_i4)
        ELSE IF (threedeta) THEN
          CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                         inode,elecd_phi,dp,dp,0_i4)
          ALLOCATE(real_j(3,mpseudo,nphi))
          eleq_1d=RESHAPE(elecd_eq,(/ncx*ncy/))
          jaeq_1d=RESHAPE( ja_eq,(/3,ncx*ncy/))
        ELSE IF (eta_model=='eta n=0 only') THEN
          CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                         inode,elecd_n0,dp,dp,0_i4)
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          diff=elecd*MAX(ds2(1,:,:),0._r8)
        ELSE
          diff=elecd
        ENDIF
      ENDIF
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find div(b) from the old B if cleaning is not split.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb.OR.poly_divb>=0) THEN
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            divb(:,:,imode)=ber(1,:,:,imode)+bez(2,:,:,imode)
     $                    +((0,1)*keff(imode)*be(3,:,:,imode)
     $                                       +be(1,:,:,imode))/bigr
          ELSE
            divb(:,:,imode)=ber(1,:,:,imode)+bez(2,:,:,imode)
     $                     +(0,1)*keff(imode)*be(3,:,:,imode)/bigr
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     reset magnetic field to the fb_vxb-centered predicted level if it
c     is needed for an advection corrector step.
c-----------------------------------------------------------------------
      IF ((nonlinear.OR.eq_flow/='none').AND.
     $    integrand_flag/='mhd_predict'.AND..NOT.impladv) THEN
        CALL generic_ptr_set(rb%qwork1,tb%qwork1,tb%tgeom,inode,
     $                       be,ber,bez,1_i4)
      ENDIF
c-----------------------------------------------------------------------
c     find the explicit nonlinear contributions to the electric field.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
c-----------------------------------------------------------------------
c       first find b_perturbed and the perturbed velocity in real space.
c-----------------------------------------------------------------------
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,be,real_be,
     $               dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $               dealiase)
c-----------------------------------------------------------------------
c       find dt * b_perturbed X perturbed velocity in real space
c-----------------------------------------------------------------------
        CALL math_cart_cross(real_eexp,real_be,real_ve,dt)
c-----------------------------------------------------------------------
c       add the explicit contribution to resistive dissipation, if
c       3D resistivity is used.
c-----------------------------------------------------------------------
        IF (elecd>0.AND.integrand_flag(1:3)=='mhd'.AND.threedeta) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,
     $                 ja_old,real_j,dealiase)
          DO ip=1,nphi
            DO jq=1,3
              real_eexp(jq,:,ip)=real_eexp(jq,:,ip)+
     $          dt*mu0*( elecd_phi(1,:,ip)*real_j(jq,:,ip)
     $                 +(elecd_phi(1,:,ip)-eleq_1d(ipseust:ipseuen))
     $                  *jaeq_1d(jq,ipseust:ipseuen))
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       transform to Fourier space.
c-----------------------------------------------------------------------
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,eexp,real_eexp,
     $               dealiase)
      ELSE
        eexp=0
      ENDIF
c-----------------------------------------------------------------------
c     begin the mode loop for linear contributions.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
c-----------------------------------------------------------------------
c       find the explicit linear contributions to the electric field.
c-----------------------------------------------------------------------
        CALL math_cadd_cross(eexp(:,:,:,imode),be_eq,ve(:,:,:,imode),dt)
        IF (eq_flow/='none')
     $    CALL math_cadd_cross(eexp(:,:,:,imode),
     $                         be(:,:,:,imode),ve_eq,dt)
c-----------------------------------------------------------------------
c       add the explicit contribution to resistive dissipation, where
c       the electrical diffusivity is 2D.
c-----------------------------------------------------------------------
        IF (elecd>0.AND.integrand_flag(1:3)=='mhd') THEN
          IF (eta_model=='eta n=0 only') THEN
            DO jq=1,3
              eexp(jq,:,:,imode)=eexp(jq,:,:,imode)
     $          +dt*mu0*elecd_n0(1,:,:)*ja_old(jq,:,:,imode)
            ENDDO
            IF (nonlinear.AND.keff(imode)==0) THEN
              DO jq=1,3
                eexp(jq,:,:,imode)=eexp(jq,:,:,imode)+
     $            dt*mu0*(elecd_n0(1,:,:)-elecd_eq(1,:,:))*ja_eq(jq,:,:)
              ENDDO
            ENDIF
          ELSE IF (eta_model=='eta full'.OR.eta_model=='chodura') THEN
            IF (.NOT.nonlinear.AND.eta_model=='eta full') THEN
              DO jq=1,3
                eexp(jq,:,:,imode)=eexp(jq,:,:,imode)+
     $            dt*mu0*elecd_eq(1,:,:)*(ja_old(jq,:,:,imode)-
     $               1.5*ja_eq(jq,:,:)*tele(1,:,:,imode)/tele_eq(1,:,:))
              ENDDO
            ENDIF
          ELSE
            DO jq=1,3
              eexp(jq,:,:,imode)=eexp(jq,:,:,imode)
     $          +dt*mu0*diff*ja_old(jq,:,:,imode)
            ENDDO
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       construct -curl(alpha^*).E_explicit  [ -div(alpha)*div(B) ].
c-----------------------------------------------------------------------
        IF (divbd>0.AND..NOT.split_divb) THEN
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(   dalpdz(:,:,iv)*eexp(3,:,:,imode)
     $        +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(2,:,:,imode)
     $           -divbd*dt*dalpdrc(:,:,iv)*divb(:,:,imode),1)
            int(2,:,iv,imode)=SUM(  -dalpdr(:,:,iv)*eexp(3,:,:,imode)
     $        -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(1,:,:,imode)
     $        -divbd*dt*dalpdz (:,:,iv)*divb(:,:,imode),1)
            int(3,:,iv,imode)=SUM(  -dalpdz (:,:,iv)*eexp(1,:,:,imode)
     $                              +dalpdrc(:,:,iv)*eexp(2,:,:,imode)
     $                 +divbd*dt*(0,1)*alpha(:,:,iv)*keff(imode)
     $                          *divb(:,:,imode)/bigr,1)
          ENDDO
        ELSE
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(   dalpdz(:,:,iv)*eexp(3,:,:,imode)
     $        +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(2,:,:,imode),1)
            int(2,:,iv,imode)=SUM(  -dalpdr(:,:,iv)*eexp(3,:,:,imode)
     $        -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(1,:,:,imode),1)
            int(3,:,iv,imode)=SUM(  -dalpdz(:,:,iv)*eexp(1,:,:,imode)
     $                             +dalpdrc(:,:,iv)*eexp(2,:,:,imode),1)
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       find the integrand for the discontinuous scalar equation,
c       disdbfac*aldis*div(b), where aldis is the value of the
c       discontinuous basis function.
c-----------------------------------------------------------------------
        IF (poly_divb>=0) THEN
          disdbfac=SQRT(disc_dbd*dt/fdivb)
          DO iv=nvc+1,nv
            int(1,:,iv,imode)=
     $        disdbfac*SUM(aldis(:,:,iv-nvc)*divb(:,:,imode),1)
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       add -alpha*curl(E_applied) to the n=0 mode for cases without
c       an equilibrium field.  note:  the third component of e_applied
c       is 1/(2*pi) or 1/per_length at the surface.
c-----------------------------------------------------------------------
c-PRE   IF () THEN
c         CALL generic_ptr_set(rb%qe_applied,tb%qe_applied,tb%tgeom,
c    $                         inode,e_a,e_ar,e_az,1_i4)
c         DO iv=1,nv
c           int(1,:,iv,imode)=int(1,:,iv,imode)
c    $        -SUM(alpha(:,:,iv)*dt*volt*e_az(3,:,:)/bigr,1)
c           int(2,:,iv,imode)=int(2,:,iv,imode)
c    $        +SUM(alpha(:,:,iv)*dt*volt*e_ar(3,:,:)/bigr,1)
c           int(3,:,iv,imode)=int(3,:,iv,imode)
c    $        -SUM(alpha(:,:,iv)*dt*e_vert*e_ar(2,:,:),1)
c         ENDDO
c       ENDIF
      ENDDO mode_loop
c-----------------------------------------------------------------------
c     if hyper-resistivity is used without a split step, an auxiliary
c     equation is solved simultaneously with dB.  the integrand for the
c     right side of this equation is
c
c       -SQRT(dt*hyp_eta/fhyp_eta)*curl(C^*).curl(B_old)
c       -SQRT(dt*hyp_dbd/fhyp_dbd)*div(C^*).div(B_old)
c
c     where C is the test vector for the auxiliary field.  note that
c     ja_old and divb are modified at this point.
c-----------------------------------------------------------------------
      IF ((hyp_eta>0.OR.hyp_dbd>0).AND..NOT.split_hypeta) THEN
        fhyp=SQRT(dt*hyp_eta/fhyp_eta)*mu0
        fhdb=SQRT(dt*hyp_dbd/fhyp_dbd)
        DO imode=1,nmodes
          ja_old(:,:,:,imode)=ja_old(:,:,:,imode)*fhyp
          divb(:,:,imode)=divb(:,:,imode)*fhdb
          DO iv=1,nvc
            int(4,:,iv,imode)=
     $        SUM( dalpdz(:,:,iv)*ja_old(3,:,:,imode)+
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*
     $             ja_old(2,:,:,imode)-
     $             dalpdrc(:,:,iv)*divb(:,:,imode),1)
            int(5,:,iv,imode)=
     $        SUM(-dalpdr(:,:,iv)*ja_old(3,:,:,imode)-
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*
     $             ja_old(1,:,:,imode)-
     $             dalpdz(:,:,iv)*divb(:,:,imode),1)
            int(6,:,iv,imode)=
     $        SUM(-dalpdz (:,:,iv)*ja_old(1,:,:,imode)+
     $             dalpdrc(:,:,iv)*ja_old(2,:,:,imode)+
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*
     $             divb(:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     deallocate arrays as needed.
c-----------------------------------------------------------------------
      IF (integrand_flag(1:3)=='mhd'.AND.threedeta) DEALLOCATE(real_j)
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE brhs_mhd
c-----------------------------------------------------------------------
c     subprogram 3.1 brhs_hyp.
c     compute the integrand used in the rhs of the hyper-resistivity
c     time split.
c-----------------------------------------------------------------------
      SUBROUTINE brhs_hyp(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: ja
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: divb
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,nvc
      REAL(r8) :: fhyp,fhdb
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and set pointers to
c     the magnetic field at the beginning of the time split.  
c     current density for resistive diffusion is based on
c     this magnetic field.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,ber,bez,1_i4)
c-----------------------------------------------------------------------
c     convenience parameters and curl(B):
c-----------------------------------------------------------------------
      nvc=SIZE(alpha,3)
      fhyp=SQRT(dt*hyp_eta/fhyp_eta)
      fhdb=SQRT(dt*hyp_dbd/fhyp_dbd)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,fhyp)
c-----------------------------------------------------------------------
c     start the mode loop with the divergence of this component.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        IF (geom=='tor') THEN
          divb(:,:)=
     $      fhdb*( ber(1,:,:,imode)+bez(2,:,:,imode)
     $           +((0,1)*keff(imode)*be(3,:,:,imode)
     $                              +be(1,:,:,imode))/bigr)
        ELSE
          divb(:,:)=fhdb*(ber(1,:,:,imode)+bez(2,:,:,imode)
     $                   +(0,1)*keff(imode)*be(3,:,:,imode)/bigr)
        ENDIF
c-----------------------------------------------------------------------
c       if hyper-resistivity is used without a split step, an auxiliary
c       equation is solved simultaneously with dB.  the integrand for
c       the right side of this equation is
c
c         -SQRT(dt*hyp_eta/fhyp_eta)*curl(C^*).curl(B_old)
c         -SQRT(dt*hyp_dbd/fhyp_dbd)*div(C^*).div(B_old)
c
c       where C is the test vector for the auxiliary field.  note that
c-----------------------------------------------------------------------
        DO iv=1,nvc
          int(1:3,:,iv,imode)=0._r8
          int(4,:,iv,imode)=
     $      SUM( dalpdz(:,:,iv)*ja(3,:,:,imode)+
     $           (0,1)*alpha(:,:,iv)*keff(imode)/bigr*ja(2,:,:,imode)-
     $           dalpdrc(:,:,iv)*divb,1)
          int(5,:,iv,imode)=
     $      SUM(-dalpdr(:,:,iv)*ja(3,:,:,imode)-
     $           (0,1)*alpha(:,:,iv)*keff(imode)/bigr*ja(1,:,:,imode)-
     $           dalpdz(:,:,iv)*divb,1)
          int(6,:,iv,imode)=
     $      SUM(-dalpdz (:,:,iv)*ja(1,:,:,imode)+
     $           dalpdrc(:,:,iv)*ja(2,:,:,imode)+
     $           (0,1)*alpha(:,:,iv)*keff(imode)/bigr*divb,1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE brhs_hyp
c-----------------------------------------------------------------------
c     subprogram 4. b_3deta_dot.
c     compute the action of the 3d resistive diffusion matrix (due to
c     3d resistivity) on a vector stored in work4.
c-----------------------------------------------------------------------
      SUBROUTINE b_3deta_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: elecd_phi,dp
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_e
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: be,
     $             ber,bez,dtcurlb
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2),nmodes) :: divb
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,jq,nv,npol,ncx,ncy,b_derivs
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the magnetic
c     field iterate (from work4).  multiply the curl by feta*dt but not
c     1/mu0.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      CALL generic_all_eval(rb%work4,tb%work4,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      be,ber,bez,1_i4)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,dtcurlb,feta*dt)
      CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                     inode,elecd_phi,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find div(b) from the old B if cleaning is not split.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb) THEN
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            divb(:,:,imode)=fdivb*divbd*dt*
     $        ( ber(1,:,:,imode)+bez(2,:,:,imode)
     $        +((0,1)*keff(imode)*be(3,:,:,imode)+be(1,:,:,imode))/bigr)
          ELSE
            divb(:,:,imode)=fdivb*divbd*dt*
     $        ( ber(1,:,:,imode)+bez(2,:,:,imode)
     $         +(0,1)*keff(imode)*be(3,:,:,imode)/bigr )
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     compute the resistive electric field.  dtcurlb becomes 
c     f*dt*eta*j.
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,dtcurlb,real_e,
     $             dealiase)
      DO jq=1,3
        real_e(jq,:,:)=elecd_phi(1,:,:)*real_e(jq,:,:)
      ENDDO
      CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,dtcurlb,real_e,
     $             dealiase)
c-----------------------------------------------------------------------
c     find the contributions that multiply the basis vector itself.
c     part of this is just the iterate (from the d/dt term).
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb) THEN
        DO imode=1,nmodes
          be(1,:,:,imode)=be(1,:,:,imode)-
     $                    (0,1)*keff(imode)/bigr*dtcurlb(2,:,:,imode)
          be(2,:,:,imode)=be(2,:,:,imode)+
     $                    (0,1)*keff(imode)/bigr*dtcurlb(1,:,:,imode)
          be(3,:,:,imode)=be(3,:,:,imode)-
     $                    (0,1)*keff(imode)*divb(:,:,imode)/bigr
        ENDDO
      ELSE
        DO imode=1,nmodes
          be(1,:,:,imode)=be(1,:,:,imode)-
     $                    (0,1)*keff(imode)/bigr*dtcurlb(2,:,:,imode)
          be(2,:,:,imode)=be(2,:,:,imode)+
     $                    (0,1)*keff(imode)/bigr*dtcurlb(1,:,:,imode)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     construct B+dt*curl(alpha^*).E [ +dt*div(alpha)*div(B) ].
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb) THEN
        DO imode=1,nmodes
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(-dalpdz (:,:,iv)*dtcurlb(3,:,:,imode)
     $                            +alpha  (:,:,iv)*be(1,:,:,imode)
     $                            +dalpdrc(:,:,iv)*divb(:,:,imode),1)
            int(2,:,iv,imode)=SUM( dalpdr(:,:,iv)*dtcurlb(3,:,:,imode)
     $                            +alpha (:,:,iv)*be(2,:,:,imode)
     $                            +dalpdz(:,:,iv)*divb(:,:,imode),1)
            int(3,:,iv,imode)=SUM( dalpdz (:,:,iv)*dtcurlb(1,:,:,imode)
     $                            -dalpdrc(:,:,iv)*dtcurlb(2,:,:,imode)
     $                            +alpha  (:,:,iv)*be(3,:,:,imode),1)
          ENDDO
        ENDDO
      ELSE
        DO imode=1,nmodes
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(-dalpdz (:,:,iv)*dtcurlb(3,:,:,imode)
     $                            +alpha  (:,:,iv)*be(1,:,:,imode),1)
            int(2,:,iv,imode)=SUM( dalpdr(:,:,iv)*dtcurlb(3,:,:,imode)
     $                            +alpha (:,:,iv)*be(2,:,:,imode),1)
            int(3,:,iv,imode)=SUM( dalpdz (:,:,iv)*dtcurlb(1,:,:,imode)
     $                          -dalpdrc(:,:,iv)*dtcurlb(2,:,:,imode)
     $                            +alpha(:,:,iv)*be(3,:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE b_3deta_dot
c-----------------------------------------------------------------------
c     subprogram 5. brhs_hmhd.
c     compute the integrand used in the rhs of the semi-implicit
c     magnetic field equation for the combined Hall-MHD time-split.
c-----------------------------------------------------------------------
      SUBROUTINE brhs_hmhd(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: aldis,dalddr,dalddz
      REAL(r8), DIMENSION(:,:,:), POINTER :: be_eq,ja_eq,
     $          jaeq_r,jaeq_z,nd_eq,nd_n0,ve_eq,grdv_eq,
     $          real_ndptr,dp,ds2,elecd_n0,nd_eqr,nd_eqz,
     $          eq_force,elecd_eq,elecd_phi,tele_eq
      REAL(r8), DIMENSION(3,mpseudo,nphi), TARGET :: real_ja,real_be,
     $          real_eexp,real_ve
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_te
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: eleq_1d
      REAL(r8), DIMENSION(3,SIZE(bigr,1)*SIZE(bigr,2)) :: jaeq_1d 
      REAL(r8), DIMENSION(:,:,:), ALLOCATABLE :: real_divv
      REAL(r8), DIMENSION(:,:,:), POINTER :: real_gradv,real_gradj,
     $                            divv_eq
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: diff
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: hfac,pfac,dbdt,dtm,disdbfac,fhyp,fhdb

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez,ve,ver,vez,
     $             jafe,jar,jaz,nd,ndr,ndz,tele,teler,telez,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             jaore,eexp,ebxv,ja,grad_nd
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2),nmodes) :: divb
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: auxb
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: dbqd
      COMPLEX(r8), DIMENSION(:,:,:,:), ALLOCATABLE :: jap,vep,divv
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,iq,jq,nv,ncx,ncy,in0,ip,nvc,nvd
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      hfac=(1._r8-meomi)/(elementary_q*(1._r8+meomi))
      pfac=hfac*kboltz
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the 0-th
c     order magnetic field.  evaluate the old current density
c     from the gradients of the old magnetic fields for resitive
c     dissipation.  for the mhd part, evaluate flow velocities.
c
c     if poly_divb is non-negative, divergence error is controlled with
c     an auxiliary field that is discontinuous at element borders, and
c     these bases are also needed.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      IF (poly_divb>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',aldis,
     $                          dalddr,dalddz,0_i4,poly_divb,
     $                          polydmin=poly_divb_min,
     $                          polydmax=poly_divb_max)
        nvd=SIZE(aldis,3)
      ENDIF
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,
     $                     be,ber,bez,1_i4)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1/mu0)
      CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                     inode,ja_eq,jaeq_r,jaeq_z,1_i4)
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,nd_eqr,nd_eqz,1_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,
     $                     nd,ndr,ndz,1_i4)
      IF (beta>0) THEN
        CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,inode,
     $                       tele,teler,telez,1_i4)
        CALL generic_ptr_set(rb%qtele_eq,tb%qtele_eq,tb%tgeom,
     $                       inode,tele_eq,dp,dp,0_i4)
      ENDIF
      CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,ve,ver,vez,1_i4)
      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     dissipation coefficients.
c-----------------------------------------------------------------------
      dbdt=dt*divbd
      dtm=dt*mu0
      IF (elecd>0) THEN
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_eq,tb%qelecd_eq,tb%tgeom,
     $                         inode,elecd_eq,dp,dp,0_i4)
        ENDIF
        IF (.NOT.nonlinear.AND.eta_model=='eta full') THEN
          CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,
     $                         inode,tele,dcp,dcp,0_i4)
        ELSE IF (threedeta) THEN
          CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                         inode,elecd_phi,dp,dp,0_i4)
          eleq_1d=RESHAPE(elecd_eq,(/ncx*ncy/))
          jaeq_1d=RESHAPE( ja_eq,(/3,ncx*ncy/))
        ELSE IF (eta_model=='eta n=0 only') THEN
          CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                         inode,elecd_n0,dp,dp,0_i4)
          diff=dtm*elecd_n0(1,:,:)
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          diff=dtm*elecd*MAX(ds2(1,:,:),0._r8)
        ELSE
          diff=dtm*elecd
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     find div(b) for cleaning if not split.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb.OR.poly_divb>=0) THEN
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            divb(:,:,imode)=(ber(1,:,:,imode)+bez(2,:,:,imode)
     $                     +((0,1)*keff(imode)*be(3,:,:,imode)
     $                                        +be(1,:,:,imode))/bigr)
          ELSE
            divb(:,:,imode)=(ber(1,:,:,imode)+bez(2,:,:,imode)
     $                      +(0,1)*keff(imode)*be(3,:,:,imode)/bigr)
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find grad(n) for the grad(pe) term.  this uses the time-averaged
c     number density.
c-----------------------------------------------------------------------
      IF (beta>0) THEN
        DO imode=1,nmodes
          grad_nd(1,:,:,imode)=pfac*ndr(1,:,:,imode)
          grad_nd(2,:,:,imode)=pfac*ndz(1,:,:,imode)
          grad_nd(3,:,:,imode)=(0,1)*pfac*keff(imode)*
     $                               nd(1,:,:,imode)/bigr
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the explicit nonlinear contributions to the electric field.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
c-----------------------------------------------------------------------
c       first find perturbed b, and j in real space.
c-----------------------------------------------------------------------
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,be,real_be,
     $               dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,real_ja,
     $               dealiase)
c-----------------------------------------------------------------------
c       find j_perturbed X b_perturbed / e in real space,
c-----------------------------------------------------------------------
        CALL math_cart_cross(real_eexp,real_ja,real_be,hfac)
c-----------------------------------------------------------------------
c       include the -kboltz*Te*grad(n) term for finite pressure.  the
c       grad(Te) part does not contribute to dB/dt and is not included.
c       real_ve is used as temporary storage.
c-----------------------------------------------------------------------
        IF (beta>0) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $                 real_ve,dealiase)
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,tele,
     $                 real_te,dealiase)
          real_eexp(1,:,:)=real_eexp(1,:,:)-
     $                       real_ve(1,:,:)*real_te(1,:,:)
          real_eexp(2,:,:)=real_eexp(2,:,:)-
     $                       real_ve(2,:,:)*real_te(1,:,:)
          real_eexp(3,:,:)=real_eexp(3,:,:)-
     $                       real_ve(3,:,:)*real_te(1,:,:)
        ENDIF
c-----------------------------------------------------------------------
c       transform to Fourier space.
c-----------------------------------------------------------------------
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,eexp,real_eexp,
     $               dealiase)
c-----------------------------------------------------------------------
c       create the nonlinear ideal mhd electric field.  these terms
c       are kept separate because they are not divided by n later.
c-----------------------------------------------------------------------
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $               dealiase)
        CALL math_cart_cross(real_eexp,real_be,real_ve,dt)
c-----------------------------------------------------------------------
c       add the explicit contribution to resistive dissipation if
c       3D resistivity is used.
c-----------------------------------------------------------------------
        IF (elecd>0.AND.threedeta) THEN
          DO ip=1,nphi
            DO jq=1,3
              real_eexp(jq,:,ip)=real_eexp(jq,:,ip)+
     $          dtm*( elecd_phi(1,:,ip)*real_ja(jq,:,ip)
     $              +(elecd_phi(1,:,ip)-eleq_1d(ipseust:ipseuen))
     $                *jaeq_1d(jq,ipseust:ipseuen))
            ENDDO
          ENDDO
        ENDIF
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,ebxv,real_eexp,
     $               dealiase)
      ELSE
        ebxv=0
        eexp=0
      ENDIF
c-----------------------------------------------------------------------
c     begin the mode loop for linear contributions.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
c-----------------------------------------------------------------------
c       find the explicit linear contributions to the electric field.
c-----------------------------------------------------------------------
        CALL math_cadd_cross(eexp(:,:,:,imode),
     $                       ja(:,:,:,imode),be_eq,hfac)
        CALL math_cadd_cross(eexp(:,:,:,imode),
     $                       ja_eq,be(:,:,:,imode),hfac)
c-----------------------------------------------------------------------
c       subtract the linear kbolz*grad(n)*Te terms.  the grad(Te)
c       should not contribute to dB/dt and is not included.
c-----------------------------------------------------------------------
        IF (beta>0) THEN
          eexp(1,:,:,imode)=eexp(1,:,:,imode)-
     $                   grad_nd(1,:,:,imode)*tele_eq(1,:,:)-
     $                 pfac*tele(1,:,:,imode)* nd_eqr(1,:,:)
          eexp(2,:,:,imode)=eexp(2,:,:,imode)-
     $                   grad_nd(2,:,:,imode)*tele_eq(1,:,:)-
     $                 pfac*tele(1,:,:,imode)* nd_eqz(1,:,:)
          eexp(3,:,:,imode)=eexp(3,:,:,imode)-
     $                   grad_nd(3,:,:,imode)*tele_eq(1,:,:)
        ENDIF
c-----------------------------------------------------------------------
c       find the linear contributions to the ideal mhd electric field.
c-----------------------------------------------------------------------
        CALL math_cadd_cross(ebxv(:,:,:,imode),be_eq,ve(:,:,:,imode),dt)
        IF (eq_flow/='none')
     $    CALL math_cadd_cross(ebxv(:,:,:,imode),
     $                         be(:,:,:,imode),ve_eq,dt)
      ENDDO mode_loop
c-----------------------------------------------------------------------
c     add current advection terms ~ div(JV+VJ).  first
c     allocate and evaluate the needed fields, including the
c     finite element representation of J.
c-----------------------------------------------------------------------
      IF (ohms=='2fl'.AND.advect=='all') THEN
        ALLOCATE(vep(3,ncx,ncy,nmodes),divv(1,ncx,ncy,nmodes),
     $           real_divv(1,mpseudo,nphi),jap(3,ncx,ncy,nmodes))
        CALL generic_ptr_set(rb%qja,tb%qja,tb%tgeom,inode,
     $                       jafe,jar,jaz,1_i4)
        CALL generic_ptr_set(rb%qdvv_eq,tb%qdvv_eq,tb%tgeom,inode,
     $                       divv_eq,dp,dp,0_i4)
        IF (eq_flow/='none') THEN
          CALL generic_ptr_set(rb%qgrdveq,tb%qgrdveq,tb%tgeom,
     $                         inode,grdv_eq,dp,dp,0_i4)
        ENDIF
c-----------------------------------------------------------------------
c       evaluate div(v) and the effective d(v)/dphi and d(j)/dphi.
c-----------------------------------------------------------------------
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            divv(1,:,:,imode)=ver(1,:,:,imode)+vez(2,:,:,imode)
     $                      +((0,1)*keff(imode)*ve(3,:,:,imode)
     $                                         +ve(1,:,:,imode))/bigr
          ELSE
            divv(1,:,:,imode)=ver(1,:,:,imode)+vez(2,:,:,imode)
     $                       +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr
          ENDIF
          vep(:,:,:,imode)=(0,1)*keff(imode)*ve(:,:,:,imode)
          jap(:,:,:,imode)=(0,1)*keff(imode)*jafe(:,:,:,imode)
          IF (geom=='tor') THEN
            vep(1,:,:,imode)=(vep(1,:,:,imode)-ve(3,:,:,imode))/bigr
            vep(2,:,:,imode)= vep(2,:,:,imode)/bigr
            vep(3,:,:,imode)=(vep(3,:,:,imode)+ve(1,:,:,imode))/bigr
            jap(1,:,:,imode)=(jap(1,:,:,imode)-jafe(3,:,:,imode))/bigr
            jap(2,:,:,imode)= jap(2,:,:,imode)/bigr
            jap(3,:,:,imode)=(jap(3,:,:,imode)+jafe(1,:,:,imode))/bigr
          ENDIF
        ENDDO
c-----------------------------------------------------------------------
c       nonlinear products:
c-----------------------------------------------------------------------
        IF (nonlinear) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,jafe,real_ja,
     $                 dealiase)
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,divv,
     $                 real_divv,dealiase)
c-----------------------------------------------------------------------
c         div(v)*j
c-----------------------------------------------------------------------
          DO iq=1,3
            real_eexp(iq,:,:)=real_divv(1,:,:)*real_ja(iq,:,:)
          ENDDO
c-----------------------------------------------------------------------
c         j.grad(v)--d(V)/dr, d(V)/dz and d(V)/dphi contributions.
c         pointers are used to save a little memory.
c-----------------------------------------------------------------------
          real_gradv=>real_be
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ver,
     $                 real_gradv,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ja(1,:,:)*real_gradv(iq,:,:)
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,vez,
     $                 real_gradv,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ja(2,:,:)*real_gradv(iq,:,:)
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,vep,
     $                 real_gradv,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ja(3,:,:)*real_gradv(iq,:,:)
          ENDDO
          NULLIFY(real_gradv)
c-----------------------------------------------------------------------
c         v.grad(j)--d(J)/dr, d(J)/dz and d(J)/dphi contributions.
c-----------------------------------------------------------------------
          real_gradj=>real_be
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,jar,
     $                 real_gradj,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ve(1,:,:)*real_gradj(iq,:,:)
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,jaz,
     $                 real_gradj,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ve(2,:,:)*real_gradj(iq,:,:)
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,jap,
     $                 real_gradj,dealiase)
          DO iq=1,3
            real_eexp(iq,:,:)=real_eexp(iq,:,:)
     $                +real_ve(3,:,:)*real_gradj(iq,:,:)
          ENDDO
          NULLIFY(real_gradj)
c-----------------------------------------------------------------------
c         transform the nonlinear contributions
c         (here, jaore is just a temporary array).
c-----------------------------------------------------------------------
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,jaore,
     $                 real_eexp,dealiase)
        ELSE
          jaore=0
        ENDIF

        cur_adv: DO imode=1,nmodes
c-----------------------------------------------------------------------
c         div(v)*J_eq+div(V_eq)*j
c-----------------------------------------------------------------------
          DO jq=1,3
            jaore(jq,:,:,imode)=jaore(jq,:,:,imode)
     $                       +divv(1,:,:,imode)*ja_eq(jq,:,:)
     $                       +divv_eq(1,:,:)*jafe(jq,:,:,imode)
          ENDDO
c-----------------------------------------------------------------------
c         J_eq.grad(v)+j.grad(V_eq)
c-----------------------------------------------------------------------
          DO jq=1,3
            jaore(jq,:,:,imode)=jaore(jq,:,:,imode)
     $                         +ja_eq(1,:,:)*ver(jq,:,:,imode)
     $                         +ja_eq(2,:,:)*vez(jq,:,:,imode)
     $                         +ja_eq(3,:,:)*vep(jq,:,:,imode)
          ENDDO
          IF (eq_flow/='none') THEN
            jaore(1,:,:,imode)=jaore(1,:,:,imode)+
     $                          jafe(1,:,:,imode)*grdv_eq(1,:,:)+
     $                          jafe(2,:,:,imode)*grdv_eq(2,:,:)+
     $                          jafe(3,:,:,imode)*grdv_eq(3,:,:) 
            jaore(2,:,:,imode)=jaore(2,:,:,imode)+
     $                          jafe(1,:,:,imode)*grdv_eq(4,:,:)+
     $                          jafe(2,:,:,imode)*grdv_eq(5,:,:) 
            jaore(3,:,:,imode)=jaore(3,:,:,imode)+
     $                          jafe(1,:,:,imode)*grdv_eq(7,:,:)+
     $                          jafe(2,:,:,imode)*grdv_eq(8,:,:)+
     $                          jafe(3,:,:,imode)*grdv_eq(9,:,:) 
          ENDIF
c-----------------------------------------------------------------------
c         V_eq.grad(j)+v.grad(J_eq)
c-----------------------------------------------------------------------
          IF (eq_flow/='none') THEN
            DO jq=1,3
              jaore(jq,:,:,imode)=jaore(jq,:,:,imode)
     $                           +ve_eq(1,:,:)*jar(jq,:,:,imode)
     $                           +ve_eq(2,:,:)*jaz(jq,:,:,imode)
     $                           +ve_eq(3,:,:)*jap(jq,:,:,imode)
            ENDDO
          ENDIF
          DO jq=1,3
            jaore(jq,:,:,imode)=jaore(jq,:,:,imode)
     $                         +ve(1,:,:,imode)*jaeq_r(jq,:,:)
     $                         +ve(2,:,:,imode)*jaeq_z(jq,:,:)
          ENDDO
          IF (geom=='tor') THEN
            jaore(1,:,:,imode)=jaore(1,:,:,imode)
     $                        -ve(3,:,:,imode)*ja_eq(3,:,:)/bigr
            jaore(3,:,:,imode)=jaore(2,:,:,imode)
     $                        +ve(3,:,:,imode)*ja_eq(1,:,:)/bigr
          ENDIF
c-----------------------------------------------------------------------
c         multiply by me/e**2 and add to other terms.
c-----------------------------------------------------------------------
          DO jq=1,3
            eexp(jq,:,:,imode)=ms(1)*jaore(jq,:,:,imode)
     $                        /((1+meomi)*elementary_q**2)
     $                        +eexp(jq,:,:,imode)
          ENDDO
        ENDDO cur_adv
      ENDIF
c-----------------------------------------------------------------------
c     all of the preceeding terms must be divided by number density.
c     the n that is used depends on the continuity input parameter.
c-----------------------------------------------------------------------
      IF (continuity=='none'.OR.
     $    nonlinear.AND.continuity=='fix profile') THEN
        DO imode=1,nmodes
          DO jq=1,3
            eexp(jq,:,:,imode)=dt*eexp(jq,:,:,imode)/nd_eq(1,:,:)
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c     when a perturbed n is used in the denominator, there is also
c     a term associated with equilibrium quantities in the numerator.
c     it is now computed once and saved.
c-----------------------------------------------------------------------
      ELSE
        CALL generic_ptr_set(rb%qeq_force,tb%qeq_force,tb%tgeom,
     $                       inode,eq_force,dp,dp,0_i4)
c-----------------------------------------------------------------------
c       linear density perturbation.
c-----------------------------------------------------------------------
        IF (.NOT.nonlinear) THEN
          DO imode=1,nmodes
            DO jq=1,3
              eexp(jq,:,:,imode)=dt/nd_eq(1,:,:)*
     $          (eexp(jq,:,:,imode)-
     $           eq_force(jq,:,:)*nd(1,:,:,imode)/nd_eq(1,:,:))
            ENDDO
          ENDDO
c-----------------------------------------------------------------------
c       use symmetric part of evolving density.
c-----------------------------------------------------------------------
        ELSE IF (continuity=='n=0 only') THEN
          CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                         nd_n0,dp,dp,0_i4)
          DO imode=1,nmodes
            IF (keff(imode)==0) THEN
              DO jq=1,3
                eexp(jq,:,:,imode)=dt*(eexp(jq,:,:,imode)/
     $                             (nd_eq(1,:,:)+nd_n0(1,:,:))
     $              +eq_force(jq,:,:)*(1/(nd_eq(1,:,:)+nd_n0(1,:,:))
     $                                -1/ nd_eq(1,:,:)))
              ENDDO
            ELSE
              DO jq=1,3
                eexp(jq,:,:,imode)=dt*eexp(jq,:,:,imode)/
     $            (nd_eq(1,:,:)+nd_n0(1,:,:))
              ENDDO
            ENDIF
          ENDDO
c-----------------------------------------------------------------------
c       use 3D evolving density.
c-----------------------------------------------------------------------
        ELSE
          CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                         inode,real_ndptr,dp,dp,0_i4)
          DO imode=1,nmodes
            IF (keff(imode)==0) THEN
              DO jq=1,3
                eexp(jq,:,:,imode)=eexp(jq,:,:,imode)+eq_force(jq,:,:)
              ENDDO
            ENDIF
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,
     $                 eexp,real_eexp,dealiase)
          DO jq=1,3
            real_eexp(jq,:,:)=dt*real_eexp(jq,:,:)/real_ndptr(1,:,:)
          ENDDO
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,
     $                 eexp,real_eexp,dealiase)
          DO imode=1,nmodes
            IF (keff(imode)==0) THEN
              DO jq=1,3
                eexp(jq,:,:,imode)=eexp(jq,:,:,imode)-
     $                             dt*eq_force(jq,:,:)/nd_eq(1,:,:)
              ENDDO
            ENDIF
          ENDDO
        ENDIF
      ENDIF
      IF (ohms=='2fl'.AND.advect=='all')
     $  DEALLOCATE(vep,divv,real_divv,jap)
c-----------------------------------------------------------------------
c     add the explicit contribution to resistive dissipation, where
c     the electrical diffusivity is 2D.
c-----------------------------------------------------------------------
      IF (elecd>0) THEN
        IF (.NOT.(eta_model=='eta full'.OR.eta_model=='chodura')) THEN
          DO imode=1,nmodes
            eexp(1,:,:,imode)=eexp(1,:,:,imode)+diff*ja(1,:,:,imode)
            eexp(2,:,:,imode)=eexp(2,:,:,imode)+diff*ja(2,:,:,imode)
            eexp(3,:,:,imode)=eexp(3,:,:,imode)+diff*ja(3,:,:,imode)
          ENDDO
          IF (eta_model=='eta n=0 only'.AND.nonlinear) THEN
            DO imode=1,nmodes
              IF (keff(imode)/=0) CYCLE
              DO jq=1,3
                eexp(jq,:,:,imode)=eexp(jq,:,:,imode)+
     $            dtm*(elecd_n0(1,:,:)-elecd_eq(1,:,:))*ja_eq(jq,:,:)
              ENDDO
            ENDDO
          ENDIF
        ELSE IF (.NOT.nonlinear.AND.eta_model=='eta full') THEN
          DO imode=1,nmodes
            DO jq=1,3
              eexp(jq,:,:,imode)=eexp(jq,:,:,imode)+
     $          dtm*elecd_eq(1,:,:)*(ja(jq,:,:,imode)-
     $             1.5*ja_eq(jq,:,:)*tele(1,:,:,imode)/tele_eq(1,:,:))
            ENDDO
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     construct -curl(alpha).E_explicit [ -div(alpha)*div(B) ].
c-----------------------------------------------------------------------
      eexp=eexp+ebxv
      IF (divbd>0.AND..NOT.split_divb) THEN
        DO imode=1,nmodes
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM( dalpdz(:,:,iv)*eexp(3,:,:,imode)
     $         +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(2,:,:,imode)
     $             -dbdt*dalpdrc(:,:,iv)*divb(:,:,imode),1)
            int(2,:,iv,imode)=SUM(-dalpdr(:,:,iv)*eexp(3,:,:,imode)
     $         -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(1,:,:,imode)
     $             -dbdt*dalpdz (:,:,iv)*divb(:,:,imode),1)
            int(3,:,iv,imode)=SUM(-dalpdz (:,:,iv)*eexp(1,:,:,imode)
     $                          +dalpdrc(:,:,iv)*eexp(2,:,:,imode)
     $                      +(0,1)*alpha(:,:,iv)*keff(imode)/bigr
     $                             *dbdt*divb(:,:,imode),1)
          ENDDO
        ENDDO
      ELSE
        DO imode=1,nmodes
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM( dalpdz(:,:,iv)*eexp(3,:,:,imode)
     $        +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(2,:,:,imode),1)
            int(2,:,iv,imode)=SUM(-dalpdr(:,:,iv)*eexp(3,:,:,imode)
     $        -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(1,:,:,imode),1)
            int(3,:,iv,imode)=SUM(-dalpdz (:,:,iv)*eexp(1,:,:,imode)
     $                            +dalpdrc(:,:,iv)*eexp(2,:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     if hyper-resistivity is used without a split step, an auxiliary
c     equation is solved simultaneously with dB.  the integrand for the
c     right side of this equation is
c
c       -SQRT(dt*hyp_eta/fhyp_eta)*curl(C^*).curl(B_old)
c       -SQRT(dt*hyp_dbd/fhyp_dbd)*div(C^*).div(B_old)
c
c     where C is the test vector for the auxiliary field.  note that
c     ja is modified at this point.
c-----------------------------------------------------------------------
      IF ((hyp_eta>0.OR.hyp_dbd>0).AND..NOT.split_hypeta) THEN
        fhyp=SQRT(dt*hyp_eta/fhyp_eta)*mu0
        fhdb=SQRT(dt*hyp_dbd/fhyp_dbd)
        DO imode=1,nmodes
          ja(:,:,:,imode)=ja(:,:,:,imode)*fhyp
          dbqd=fhdb*divb(:,:,imode)
          DO iv=1,nvc
            int(4,:,iv,imode)=
     $        SUM( dalpdz(:,:,iv)*ja(3,:,:,imode)+
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*ja(2,:,:,imode)-
     $             dalpdrc(:,:,iv)*dbqd,1)
            int(5,:,iv,imode)=
     $        SUM(-dalpdr(:,:,iv)*ja(3,:,:,imode)-
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*ja(1,:,:,imode)-
     $             dalpdz(:,:,iv)*dbqd,1)
            int(6,:,iv,imode)=
     $        SUM(-dalpdz (:,:,iv)*ja(1,:,:,imode)+
     $             dalpdrc(:,:,iv)*ja(2,:,:,imode)+
     $             (0,1)*alpha(:,:,iv)*keff(imode)/bigr*dbqd,1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the integrand for the discontinuous scalar equation,
c     disdbfac*aldis*div(b), where aldis is the value of the
c     discontinuous basis function.
c-----------------------------------------------------------------------
      IF (poly_divb>=0) THEN
        disdbfac=SQRT(disc_dbd*dt/fdivb)
        DO imode=1,nmodes
          DO iv=nvc+1,nv
            int(1,:,iv,imode)=
     $        disdbfac*SUM(aldis(:,:,iv-nvc)*divb(:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE brhs_hmhd
c-----------------------------------------------------------------------
c     subprogram 6. hall_cor.
c     find the correction term for the rhs of the bhmhd advance
c     when nonlinear iteration is used for dJxdB.
c
c     here, db is all of the change from the start of the time-step, and
c     because it's also used in the dot routine, the correction to the
c     rhs has the opposite sign of the explicit Hall term, i.e.
c
c       +0.25*dt*curl(test vector).(dJXdB)/ne
c-----------------------------------------------------------------------
      SUBROUTINE hall_cor(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,nd_n0,real_ndptr,dp
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ja,real_be,real_eexp
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: hfac

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             eexp,ja
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,iq,nvc,ncx,ncy
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      hfac=-0.25_r8*dt*(1._r8-meomi)/(elementary_q*(1._r8+meomi))
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the 0-th
c     order magnetic field.  evaluate the old current density
c     from the gradients of the old magnetic fields for resitive
c     dissipation.  for the mhd part, evaluate flow velocities.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      CALL generic_ptr_set(rb%qwork1,tb%qwork1,tb%tgeom,inode,
     $                     be,ber,bez,1_i4)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1._r8/mu0)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find the explicit nonlinear contributions to the electric field.
c
c     first find perturbed b, and j in real space.
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,be,real_be,
     $             dealiase)
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,real_ja,
     $             dealiase)
c-----------------------------------------------------------------------
c     find j_perturbed X b_perturbed / e in real space,
c-----------------------------------------------------------------------
      CALL math_cart_cross(real_eexp,real_ja,real_be,hfac)
c-----------------------------------------------------------------------
c     transform to Fourier space.
c-----------------------------------------------------------------------
      IF (continuity/='full') THEN
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,eexp,real_eexp,
     $               dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     all of the preceeding terms must be divided by number density.
c     the n that is used depends on the continuity input parameter.
c-----------------------------------------------------------------------
      IF (continuity=='none'.OR.continuity=='fix profile') THEN
        DO imode=1,nmodes
          eexp(1,:,:,imode)=eexp(1,:,:,imode)/nd_eq(1,:,:)
          eexp(2,:,:,imode)=eexp(2,:,:,imode)/nd_eq(1,:,:)
          eexp(3,:,:,imode)=eexp(3,:,:,imode)/nd_eq(1,:,:)
        ENDDO
c-----------------------------------------------------------------------
c       use symmetric part of evolving density.
c-----------------------------------------------------------------------
      ELSE IF (continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                       nd_n0,dp,dp,0_i4)
        DO imode=1,nmodes
          eexp(1,:,:,imode)=eexp(1,:,:,imode)/
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))
          eexp(2,:,:,imode)=eexp(2,:,:,imode)/
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))
          eexp(3,:,:,imode)=eexp(3,:,:,imode)/
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))
        ENDDO
c-----------------------------------------------------------------------
c     use 3D evolving density.
c-----------------------------------------------------------------------
      ELSE
        CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)
        real_eexp(1,:,:)=real_eexp(1,:,:)/real_ndptr(1,:,:)
        real_eexp(2,:,:)=real_eexp(2,:,:)/real_ndptr(1,:,:)
        real_eexp(3,:,:)=real_eexp(3,:,:)/real_ndptr(1,:,:)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,
     $               eexp,real_eexp,dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     construct -curl(alpha).delta(E)
c
c     if hyper-resitivity is used without splitting, ensure that the
c     rows for the auxiliary field are 0.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        DO iv=1,nvc
          int(1,:,iv,imode)=SUM( dalpdz(:,:,iv)*eexp(3,:,:,imode)
     $       +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(2,:,:,imode),1)
          int(2,:,iv,imode)=SUM(-dalpdr(:,:,iv)*eexp(3,:,:,imode)
     $       -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*eexp(1,:,:,imode),1)
          int(3,:,iv,imode)=SUM(-dalpdz (:,:,iv)*eexp(1,:,:,imode)
     $                          +dalpdrc(:,:,iv)*eexp(2,:,:,imode),1)
        ENDDO
        IF ((hyp_eta>0._r8.OR.hyp_dbd>0._r8).AND..NOT.split_hypeta) THEN
          DO iv=1,nvc
            int(4:6,:,iv,imode)=0._r8
          ENDDO
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE hall_cor
c-----------------------------------------------------------------------
c     subprogram 7. b_hmhd_dot.
c     compute the action of the implicit HMHD operator on the iterate
c     for the change in magnetic field (work4).
c
c     the terms that appear in this routine are only the ones that 
c     depend on changes in B over the time-split.
c
c     this has all such terms, including those that do not couple
c     Fourier components.
c
c       de=[dj X (Beq+b) + (Jeq+j) X db]/ne - (Veq+v) X db + eta*dj
c
c-----------------------------------------------------------------------
      SUBROUTINE b_hmhd_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: almod,dalmr,dalmz
      REAL(r8), DIMENSION(:,:,:), POINTER :: real_be,real_ja,
     $          nd_eq,nd_n0,real_ndptr,real_ve,ds2,elecd_n0,elecd_phi,dp
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: diff
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_dbe,real_de,real_tmp
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: hfac,hdt,fdb,fe,felin,disdbfac

      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: aux
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             de,dja,dbe,dber,dbez,crlv2
      COMPLEX(r8), DIMENSION(6,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             d6v,d6vr,d6vz
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2),nmodes) :: divb
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: divbim,dv2
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,iq,jq,nv,ncx,ncy,in0,nvc,nvm
      REAL(r8) :: fhyp1,fhyp2,fhdb
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the 0-th
c     order magnetic field.  evaluate the magnetic field and current
c     density at the start of the time step, and evaluate the change
c     in each associated with the iterate of the matrix-free solve.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      IF (hyp_eta<=0._r8.AND.hyp_dbd<=0._r8.OR.split_hypeta) THEN
        CALL generic_all_eval(rb%work4,tb%work4,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,dbe,dber,dbez,1_i4)
      ELSE
        CALL generic_all_eval(rb%w6v2,tb%w6v2,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,d6v,d6vr,d6vz,1_i4)
        fhyp1=SQRT(hyp_eta*fhyp_eta*dt)
        fhyp2=fhyp1*mu0
        fhdb=SQRT(hyp_dbd*fhyp_dbd*dt)
        dbe =d6v (4:6,:,:,:)
        dber=d6vr(4:6,:,:,:)
        dbez=d6vz(4:6,:,:,:)
        CALL math_curl(nmodes,keff,geom,bigr,dbe,dber,dbez,
     $                 crlv2,fhyp1)
        dbe =d6v (1:3,:,:,:)
        dber=d6vr(1:3,:,:,:)
        dbez=d6vz(1:3,:,:,:)
      ENDIF
      CALL math_curl(nmodes,keff,geom,bigr,dbe,dber,dbez,dja,1/mu0)

      CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,inode,
     $                     real_ve,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qja_tot,tb%qja_tot,tb%tgeom,inode,
     $                     real_ja,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,inode,
     $                     real_be,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     dissipation coefficients.
c-----------------------------------------------------------------------
      IF (elecd>0) THEN
        IF (threedeta) THEN
          CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                         inode,elecd_phi,dp,dp,0_i4)
          fe=dt*feta*mu0
        ELSE IF (eta_model=='eta n=0 only') THEN
          CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                         inode,elecd_n0,dp,dp,0_i4)
          diff=dt*feta*mu0*elecd_n0(1,:,:)
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          diff=dt*feta*mu0*elecd*MAX(ds2(1,:,:),0._r8)
        ELSE
          diff=dt*feta*mu0*elecd
        ENDIF
      ENDIF
      IF (split_divb.OR.divbd<0) THEN
        fdb=0._r8
      ELSE
        fdb=dt*fdivb*divbd
      ENDIF
c-----------------------------------------------------------------------
c     find div(b) for cleaning if not split.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb.OR.poly_divb>=0) THEN
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            divb(:,:,imode)=(dber(1,:,:,imode)+dbez(2,:,:,imode)
     $                     +((0,1)*keff(imode)* dbe(3,:,:,imode)
     $                                         +dbe(1,:,:,imode))/bigr)
          ELSE
            divb(:,:,imode)=(dber(1,:,:,imode)+dbez(2,:,:,imode)
     $                      +(0,1)*keff(imode)* dbe(3,:,:,imode)/bigr)
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the implicit contributions to the electric field that
c     couple Fourier components.
c
c     first find perturbed db and dj in real space, then compute
c     0.5*dt*(dj X B + J X db)/e, and transform back to Fourier
c     coefficients.  note that B and J include the equilibrium and
c     old-time-step perturbed fields.
c
c     also add the partial(J)/partial(t) term here.
c-----------------------------------------------------------------------
      hdt=0.5*dt
      hfac=hdt*(1-meomi)/(elementary_q*(1+meomi))
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,dbe,real_dbe,
     $             dealiase)

      IF (ohms/='mhd') THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,dja,real_tmp,
     $               dealiase)
        CALL math_cart_cross(real_de,real_tmp,real_be,hfac)
        CALL math_cadd_cross(real_de,real_ja,real_dbe,hfac)
        IF (ohms=='2fl') THEN
          felin=1._r8/(qs(1)**2/ms(1)+qs(2)**2/(zeff*ms(2)))
          real_de=real_de+real_tmp*felin
        ENDIF
c-----------------------------------------------------------------------
c       all of the preceeding terms must be divided by number density.
c       the n that is used depends on the continuity input parameter.
c-----------------------------------------------------------------------
        IF (continuity/='full')
     $    CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,de,real_de,
     $                 dealiase)
        CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                       inode,nd_eq,dp,dp,0_i4)
        IF (continuity=='none'.OR.continuity=='fix profile') THEN
          DO imode=1,nmodes
            DO jq=1,3
              de(jq,:,:,imode)=de(jq,:,:,imode)/nd_eq(1,:,:)
            ENDDO
          ENDDO
c-----------------------------------------------------------------------
c       use symmetric part of evolving density.
c-----------------------------------------------------------------------
        ELSE IF (continuity=='n=0 only') THEN
          CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                         nd_n0,dp,dp,0_i4)
          DO imode=1,nmodes
            DO jq=1,3
              de(jq,:,:,imode)=de(jq,:,:,imode)/
     $                           (nd_eq(1,:,:)+nd_n0(1,:,:))
            ENDDO
          ENDDO
c-----------------------------------------------------------------------
c       use 3D evolving density but do not forward transform until the
c       ideal term is added.  note that this no longer goes
c       back and forth to Fourier representation between the JxB
c       product and the division by n.
c-----------------------------------------------------------------------
        ELSE
          CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                         inode,real_ndptr,dp,dp,0_i4)
          DO jq=1,3
            real_de(jq,:,:)=real_de(jq,:,:)/real_ndptr(1,:,:)
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     construct terms that do not depend on number density starting
c     with the ideal electric field, 0.5*dt * db X V.
c     [if continuity is not full, dber is used for temporary space.]
c     real_ve includes any equilibrium COM flow.
c
c     3D resistivity is also here to minimize the number of FFTs.
c-----------------------------------------------------------------------
      IF (ohms=='mhd') THEN
        IF (threedeta) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,dja,real_tmp,
     $                 dealiase)
          real_tmp(1,:,:)=real_tmp(1,:,:)*elecd_phi(1,:,:)*fe
          real_tmp(2,:,:)=real_tmp(2,:,:)*elecd_phi(1,:,:)*fe
          real_tmp(3,:,:)=real_tmp(3,:,:)*elecd_phi(1,:,:)*fe
          CALL math_cadd_cross(real_tmp,real_dbe,real_ve,hdt)
        ELSE
          CALL math_cart_cross(real_tmp,real_dbe,real_ve,hdt)
        ENDIF
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,de,real_tmp,
     $               dealiase)

      ELSE IF (continuity=='full') THEN
        IF (threedeta) THEN
          real_de(1,:,:)=real_de(1,:,:)+
     $                   real_tmp(1,:,:)*elecd_phi(1,:,:)*fe
          real_de(2,:,:)=real_de(2,:,:)+
     $                   real_tmp(2,:,:)*elecd_phi(1,:,:)*fe
          real_de(3,:,:)=real_de(3,:,:)+
     $                   real_tmp(3,:,:)*elecd_phi(1,:,:)*fe
        ENDIF
        CALL math_cadd_cross(real_de,real_dbe,real_ve,hdt)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,de,real_de,
     $               dealiase)

      ELSE
        IF (threedeta) THEN
          real_tmp(1,:,:)=real_tmp(1,:,:)*elecd_phi(1,:,:)*fe
          real_tmp(2,:,:)=real_tmp(2,:,:)*elecd_phi(1,:,:)*fe
          real_tmp(3,:,:)=real_tmp(3,:,:)*elecd_phi(1,:,:)*fe
          CALL math_cadd_cross(real_tmp,real_dbe,real_ve,hdt)
        ELSE
          CALL math_cart_cross(real_tmp,real_dbe,real_ve,hdt)
        ENDIF
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,dber,real_tmp,
     $               dealiase)
        de=de+dber
      ENDIF
c-----------------------------------------------------------------------
c     resistive term for 2D eta_models.
c-----------------------------------------------------------------------
      IF (elecd>0.AND..NOT.threedeta) THEN
        DO imode=1,nmodes
          de(1,:,:,imode)=de(1,:,:,imode)+diff*dja(1,:,:,imode)
          de(2,:,:,imode)=de(2,:,:,imode)+diff*dja(2,:,:,imode)
          de(3,:,:,imode)=de(3,:,:,imode)+diff*dja(3,:,:,imode)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the contribution for the discontinuous scalar equation,
c     disdbfac*almod*div(delta-b), where almod is the value of the
c     discontinuous basis function and mwork2 holds div(delta-b).
c-----------------------------------------------------------------------
      IF (poly_divb>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',almod,dalmr,
     $                          dalmz,0_i4,poly_divb,
     $                          polydmin=poly_divb_min,
     $                          polydmax=poly_divb_max)
        nvm=SIZE(almod,3)
        CALL generic_all_eval(rb%mwork2,tb%mwork2,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,aux,dc,dc,0_i4)

        disdbfac=SQRT(disc_dbd*dt*fdivb)
      ELSE
        aux=0._r8
        disdbfac=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     construct +curl(A*).de_implicit plus div(dB) corrections
c-----------------------------------------------------------------------
      IF (hyp_eta<=0._r8.AND.hyp_dbd<=0._r8.OR.split_hypeta) THEN
        DO imode=1,nmodes
          divbim=fdb*divb(:,:,imode)+disdbfac*aux(1,:,:,imode)
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(   -dalpdz(:,:,iv)*de(3,:,:,imode)
     $         -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*de(2,:,:,imode)
     $             +dalpdrc(:,:,iv)*divbim
     $         +alpha(:,:,iv)*dbe(1,:,:,imode),1)
            int(2,:,iv,imode)=SUM(    dalpdr(:,:,iv)*de(3,:,:,imode)
     $         +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*de(1,:,:,imode)
     $             +dalpdz (:,:,iv)*divbim
     $         +alpha(:,:,iv)*dbe(2,:,:,imode),1)
            int(3,:,iv,imode)=SUM( dalpdz (:,:,iv)*de(1,:,:,imode)
     $                            -dalpdrc(:,:,iv)*de(2,:,:,imode)
     $                      -(0,1)*alpha(:,:,iv)*keff(imode)/bigr
     $                             *divbim
     $         +alpha(:,:,iv)*dbe(3,:,:,imode),1)
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c     duplicate version with contributions from the hyper-resistivity
c     auxiliary vector.  note that dja is multiplied by fhyp2 and
c     the hyper-resistive term is combined with de.  the divergence
c     of the auxiliary vector is computed one Fourier component at a
c     time.
c-----------------------------------------------------------------------
      ELSE
        DO imode=1,nmodes
          divbim=fdb*divb(:,:,imode)+disdbfac*aux(1,:,:,imode)
          dja(:,:,:,imode)=dja(:,:,:,imode)*fhyp2
          de(:,:,:,imode) =de (:,:,:,imode)-crlv2(:,:,:,imode)

          IF (geom=='tor') THEN
            divbim=fdb*divb(:,:,imode)+disdbfac*aux(1,:,:,imode)-
     $        fhdb*(d6vr(4,:,:,imode)+d6vz(5,:,:,imode)
     $             +((0,1)*keff(imode)*d6v(6,:,:,imode)
     $                                +d6v(4,:,:,imode))/bigr)
          ELSE
            divbim=fdb*divb(:,:,imode)+disdbfac*aux(1,:,:,imode)-
     $        fhdb*(d6vr(4,:,:,imode)+d6vz(5,:,:,imode)
     $             +(0,1)*keff(imode)* d6v(6,:,:,imode)/bigr)
          ENDIF

          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(   -dalpdz(:,:,iv)*de(3,:,:,imode)
     $         -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*de(2,:,:,imode)
     $             +dalpdrc(:,:,iv)*divbim
     $         +alpha(:,:,iv)*d6v(1,:,:,imode),1)
            int(2,:,iv,imode)=SUM(    dalpdr(:,:,iv)*de(3,:,:,imode)
     $         +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*de(1,:,:,imode)
     $             +dalpdz (:,:,iv)*divbim
     $         +alpha(:,:,iv)*d6v(2,:,:,imode),1)
            int(3,:,iv,imode)=SUM( dalpdz (:,:,iv)*de(1,:,:,imode)
     $                            -dalpdrc(:,:,iv)*de(2,:,:,imode)
     $         -(0,1)*alpha(:,:,iv)*keff(imode)*divbim/bigr
     $         +alpha(:,:,iv)*d6v(3,:,:,imode),1)

            int(4,:,iv,imode)=SUM(alpha(:,:,iv)*d6v(4,:,:,imode)-
     $         dalpdz(:,:,iv)*dja(3,:,:,imode)
     $        -(0,1)*alpha(:,:,iv)*keff(imode)/bigr*dja(2,:,:,imode)
     $        +fhdb*dalpdrc(:,:,iv)*divb(:,:,imode),1)
            int(5,:,iv,imode)=SUM(alpha(:,:,iv)*d6v(5,:,:,imode)+
     $         dalpdr(:,:,iv)*dja(3,:,:,imode)
     $        +(0,1)*alpha(:,:,iv)*keff(imode)/bigr*dja(1,:,:,imode)
     $        +fhdb*dalpdz(:,:,iv)*divb(:,:,imode),1)
            int(6,:,iv,imode)=SUM(alpha(:,:,iv)*d6v(6,:,:,imode)+
     $          dalpdz (:,:,iv)*dja(1,:,:,imode)
     $         -dalpdrc(:,:,iv)*dja(2,:,:,imode)
     $         -fhdb*(0,1)*keff(imode)*alpha(:,:,iv)*
     $               divb(:,:,imode)/bigr,1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     complete the integrand for the discontinuous scalar equation.
c-----------------------------------------------------------------------
      IF (poly_divb>=0) THEN
        divb=disdbfac*divb
        DO imode=1,nmodes
          DO iv=nvc+1,nvc+nvm
            int(1,:,iv,imode)=SUM(almod(:,:,iv-nvc)*
     $                            (aux(1,:,:,imode)-divb(:,:,imode)),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE b_hmhd_dot
c-----------------------------------------------------------------------
c     subprogram 8. tirhs.
c     compute the integrand used for the rhs of the ion temperature (or
c     temperature, if separate_pe=F) change equation with implicit
c     diffusion.
c
c     dt*(gamma-1)*(D(Ti_n)-ni*Ti*div(Vi))-dt*ni*Vi.grad(Ti)
c
c     where Ti_n is the temperature at the beginning of the time step,
c     Ti is either Ti_n or the predicted temperature, and D is the
c     diffusion operator.
c-----------------------------------------------------------------------
      SUBROUTINE tirhs(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: ti_eq,ti_eqr,ti_eqz,
     $          ve_eq,be_eq,ja_eq,b0,q_a,nd_eq,nd_eqr,
     $          nd_eqz,nd_n0,real_bptr,real_ndptr,elecd_t,elecd_eq,dp,
     $          ds2,kappli,kaprpi,ti_b2,bcrgti,divv_eq,real_vptr,
     $          real_jptr,upwc
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: bmagsq
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve,real_work,bbgradt
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal,real_divv
      REAL(r8), DIMENSION(mpseudo,nphi) :: magBsq3D
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: eleq_1d,elt_1d
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: tqr_1d,tqz_1d
      REAL(r8), DIMENSION(3,SIZE(bigr,1)*SIZE(bigr,2)) :: jaeq_1d 
      REAL(r8), DIMENSION(3,mpseudo) :: rvec
      REAL(r8), DIMENSION(  mpseudo) :: jaeq2,rdot
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: kdiff
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ve,dvr,dvz,be,ber,bez,
     $             ti,tir,tiz,nd,ndr,ndz,vheat,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             grad_ti,be_tot,grad_nd,vel,ja
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             vdgr_ti,divv,tdivv,bdgr_ti,ohm_en
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: b_cr_gt
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: cr_coef
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      INTEGER(i4) :: nv,iv,imode,iq,jq,ncx,ncy,ip
      REAL(r8) :: jfac,fupw
      LOGICAL :: told_set
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      told_set=.false.
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and evaluate velocity.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     temperature and grad(T) at the beginning of this time split for
c     predictor steps, or at the centered prediction for corrector
c     steps.  we assume ja_eq/=0, so if separate_pe=T, this integrand
c     is used with predictor/corrector steps.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qtion_eq,tb%qtion_eq,tb%tgeom,
     $                     inode,ti_eq,ti_eqr,ti_eqz,1_i4)
      IF (nonlinear.OR.eq_flow/='none'.OR.separate_pe) THEN
        IF (integrand_flag=='all ti terms pre'.OR.
     $      separate_pe.AND.k_cross>0.OR.impladv) THEN
          CALL generic_ptr_set(rb%qtion,tb%qtion,tb%tgeom,inode,
     $                         ti,tir,tiz,1_i4)
          told_set=.true.
        ELSE
          CALL generic_ptr_set(rb%qwork3,tb%qwork3,tb%tgeom,inode,
     $                         ti,tir,tiz,1_i4)
        ENDIF
        DO imode=1,nmodes
          grad_ti(1,:,:,imode)=tir(1,:,:,imode)
          grad_ti(2,:,:,imode)=tiz(1,:,:,imode)
          grad_ti(3,:,:,imode)=(0,1)*keff(imode)*ti(1,:,:,imode)/bigr
        ENDDO
      ENDIF
      IF (visc_heat) THEN
        CALL generic_ptr_set(rb%qvisc,tb%qvisc,tb%tgeom,inode,vheat,
     $                       dcp,dcp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     evaluate ion (if separate_pe=T) or COM velocity.
c
c     The ion velocity is computed from the COM velocity
c     and the total current density as
c
c          vel = v + j * nu / ( n e (nu+1) )
c
c     where "n" is the electron number density, "e" the electron
c     charge and "nu" is the electron-ion mass ratio times the atomic
c     number "Z":
c                  nu = Z me / mi
c
c     we also need div(v)=d(Vr)/dR + Vr/R + d(Vz)dZ + 1/R d(Vth)/dTh
c     here we assume that Jeq.grad(nd_eq)=0.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,ve,dvr,dvz,1_i4)
      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,nd,ndr,ndz,1_i4)
      IF (nonlinear.AND.continuity=='n=0 only')
     $  CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,nd_eqr,nd_eqz,1_i4)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,
     $                     be,ber,bez,1_i4)
      IF (separate_pe.OR.ohm_heat) THEN
        CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1/mu0)
        CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                       inode,ja_eq,dp,dp,0_i4)
        jfac=-meomi/(elementary_q*(meomi+1))
      ENDIF
c-----------------------------------------------------------------------
c     extra pointers are needed for the upwinding-like artificial
c     diffusion.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,real_vptr,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qupti_phi,tb%qupti_phi,tb%tgeom,
     $                       inode,upwc,dp,dp,0_i4)
        IF (separate_pe)
     $    CALL generic_ptr_set(rb%qja_tot,tb%qja_tot,tb%tgeom,
     $                         inode,real_jptr,dp,dp,0_i4)
        fupw=t_dart_upw*dt**2
        tqr_1d=RESHAPE(ti_eqr,(/ncx*ncy/))
        tqz_1d=RESHAPE(ti_eqz,(/ncx*ncy/))
      ENDIF
c-----------------------------------------------------------------------
c     div(v-COM) and grad(n)
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        IF (geom=='tor') THEN
          divv(1,:,:,imode)=dvr(1,:,:,imode)+dvz(2,:,:,imode)
     $                    +((0,1)*keff(imode)*ve(3,:,:,imode)
     $                                       +ve(1,:,:,imode))/bigr
        ELSE
          divv(1,:,:,imode)=dvr(1,:,:,imode)+dvz(2,:,:,imode)
     $                     +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr
        ENDIF
        grad_nd(1,:,:,imode)=ndr(1,:,:,imode)
        grad_nd(2,:,:,imode)=ndz(1,:,:,imode)
        grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
      ENDDO

      IF (nonlinear)
     $  CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)

      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qdvv_eq,tb%qdvv_eq,tb%tgeom,inode,
     $                       divv_eq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find the Ohmic heating energy before the ja array is modified.
c     only 1-pe_frac of it will be applied in the ion energy equation
c     when separate_pe=false; otherwise, none of it is applied.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.(separate_pe.OR.ohm_heat))
     $  CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,real_ve,
     $               dealiase)
      IF (ohm_heat.AND..NOT.separate_pe) THEN
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_eq,tb%qelecd_eq,tb%tgeom,
     $                         inode,elecd_eq,dp,dp,0_i4)
          eleq_1d=RESHAPE(elecd_eq,(/ncx*ncy/))
          jaeq_1d=RESHAPE( ja_eq,(/3,ncx*ncy/))
          DO iq=1,mpseudo
            jaeq2(iq)=SUM(jaeq_1d(:,iq-1+ipseust)**2)
          ENDDO
        ENDIF
        DO imode=1,nmodes
          ohm_en(1,:,:,imode)=2*SUM(ja_eq*ja(:,:,:,imode),1)
        ENDDO
        IF (nonlinear) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,
     $                 ohm_en,real_scal,dealiase)
          real_scal(1,:,:)=real_scal(1,:,:)+SUM(real_ve**2,1)
          IF (threedeta) THEN
            CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            DO ip=1,nphi
              DO iq=1,mpseudo
                real_scal(1,iq,ip)=real_scal(1,iq,ip)*elecd_t(1,iq,ip)+
     $            (elecd_t(1,iq,ip)-eleq_1d(iq-1+ipseust))*jaeq2(iq)
              ENDDO
            ENDDO
          ELSE IF (eta_model=='eta n=0 only') THEN
            CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            elt_1d=RESHAPE(elecd_t,(/ncx*ncy/))
            DO iq=1,mpseudo
              jaeq2(iq)=jaeq2(iq)*
     $              (elt_1d(iq-1+ipseust)-eleq_1d(iq-1+ipseust))
            ENDDO
            DO ip=1,nphi
              DO iq=1,mpseudo
                real_scal(1,iq,ip)=
     $            real_scal(1,iq,ip)*elt_1d(iq-1+ipseust)+jaeq2(iq)
              ENDDO
            ENDDO
          ENDIF
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                 ohm_en,real_scal,dealiase)
        ENDIF
        IF (nonlinear.AND.threedeta) THEN  ! done
        ELSE IF (eta_model=='eta n=0 only') THEN  ! done if nonlinear
          IF (.NOT.nonlinear) THEN
            CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            DO imode=1,nmodes
              ohm_en(:,:,:,imode)=ohm_en(:,:,:,imode)*elecd_t
            ENDDO
          ENDIF
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          DO imode=1,nmodes
            ohm_en(:,:,:,imode)=
     $        ohm_en(:,:,:,imode)*elecd*MAX(ds2,0._r8)
          ENDDO
        ELSE
          DO imode=1,nmodes
            ohm_en(:,:,:,imode)=ohm_en(:,:,:,imode)*elecd
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     if separate_pe=T, convert v and div(v) to ion velocity.  the
c     O(me/mi) terms are retained in case me/mi is modified.
c     j.grad(n) (saved in vdgr_ti) for nonlinear, full continuity
c     computations.  [keep div(j)=0 in mind here]
c-----------------------------------------------------------------------
      IF (separate_pe) THEN
        IF (nonlinear) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $                 real_work,dealiase)
          real_scal(1,:,:)=SUM(real_ve*real_work,1)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $                 real_scal,dealiase)
        ELSE
          vdgr_ti=0
        ENDIF

        IF (.NOT.nonlinear.OR.continuity=='fix profile') THEN
          DO imode=1,nmodes
            DO jq=1,3
              vel(jq,:,:,imode)=ve(jq,:,:,imode)
     $          -ja(jq,:,:,imode)*jfac/nd_eq(1,:,:)
            ENDDO
            divv(1,:,:,imode)=divv(1,:,:,imode)+
     $        jfac*(SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $              +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $              +ja(2,:,:,imode)*nd_eqz(1,:,:)
     $              +vdgr_ti(1,:,:,imode))/(nd_eq(1,:,:)**2)
          ENDDO
        ELSE IF (continuity=='n=0 only') THEN
          bmagsq=1/(nd_eq(1,:,:)+nd_n0(1,:,:))
          DO imode=1,nmodes
            DO jq=1,3
              vel(jq,:,:,imode)=ve(jq,:,:,imode)
     $          -ja(jq,:,:,imode)*jfac*bmagsq
            ENDDO
            divv(1,:,:,imode)=divv(1,:,:,imode)+
     $        jfac*(SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $              +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $              +ja(2,:,:,imode)*nd_eqz(1,:,:)
     $              +vdgr_ti(1,:,:,imode))*bmagsq**2
          ENDDO
        ELSE  !  continuity is full  (ja will be left as ja/n).
          DO imode=1,nmodes
            vdgr_ti(1,:,:,imode)=jfac*(vdgr_ti(1,:,:,imode)+
     $        SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $        +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $        +ja(2,:,:,imode)*nd_eqz(1,:,:))
          ENDDO
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $                 real_scal,dealiase)
          DO jq=1,3
            real_ve(jq,:,:)=real_ve(jq,:,:)/real_ndptr(1,:,:)
          ENDDO
          real_scal=real_scal/real_ndptr**2
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,ja,
     $                 real_ve,dealiase)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $                 real_scal,dealiase)
          vel=ve-jfac*ja
          divv=divv+vdgr_ti
        ENDIF
      ELSE
        vel=ve
      ENDIF
c-----------------------------------------------------------------------
c     find the linear v_tilde.grad(Teq) for each mode--needed here
c     to complete the upwinding-like diffusion at the next step.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        vdgr_ti(1,:,:,imode)=vel(1,:,:,imode)*ti_eqr(1,:,:)+
     $                       vel(2,:,:,imode)*ti_eqz(1,:,:)
      ENDDO
c-----------------------------------------------------------------------
c     find the nonlinear contributions to v.grad(t) and t div(v)
c
c     transform v_perturbed and grad(t_perturbed) to real space,
c     find v.grad(t), then transform to Fourier space.  tdivv is
c     just used as temporary storage.
c
c     the upwinding-like diffusion term is also computed here.  the
c     predetermined coefficient is multiplied by VV.grad(T).
c     ** NOTE ** real_ve is now used to save this coefficient from
c     this point, so V_tilde(phi) array is lost.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,vel,real_ve,
     $               dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_ti,
     $               real_work,dealiase)
        real_scal(1,:,:)=SUM(real_ve*real_work,1)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tdivv,
     $               real_scal,dealiase)

        IF (impladv.AND.t_dart_upw>0) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $                 real_scal,dealiase)
          IF (separate_pe) THEN
            DO ip=1,nphi
              rvec(1,:)=real_vptr(1,:,ip)-
     $             jfac*real_jptr(1,:,ip)/real_ndptr(1,:,ip)
              rvec(2,:)=real_vptr(2,:,ip)-
     $             jfac*real_jptr(2,:,ip)/real_ndptr(1,:,ip)
              rvec(3,:)=real_vptr(3,:,ip)-
     $             jfac*real_jptr(3,:,ip)/real_ndptr(1,:,ip)
              rdot=fupw*upwc(1,:,ip)*upw_aniso*
     $          ( real_scal(1,:,ip) + SUM(rvec*real_work(:,:,ip),1) ) /
     $          ( SUM(rvec*rvec,1) + smallnum )
              jaeq2=fupw*upwc(1,:,ip)*(1._r8-upw_aniso)
              real_ve(1,:,ip)=rvec(1,:)*rdot+
     $          (real_work(1,:,ip)+tqr_1d(ipseust:ipseuen))*jaeq2
              real_ve(2,:,ip)=rvec(2,:)*rdot+
     $          (real_work(2,:,ip)+tqz_1d(ipseust:ipseuen))*jaeq2
              real_ve(3,:,ip)=rvec(3,:)*rdot+real_work(3,:,ip)*jaeq2
            ENDDO
          ELSE
            DO ip=1,nphi
              rdot=fupw*upwc(1,:,ip)*upw_aniso * ( real_scal(1,:,ip) +
     $            SUM(real_vptr(:,:,ip)*real_work(:,:,ip),1) ) /
     $          ( SUM(real_vptr(:,:,ip)*real_vptr(:,:,ip),1) + smallnum)
              jaeq2=fupw*upwc(1,:,ip)*(1._r8-upw_aniso)
              real_ve(1,:,ip)=real_vptr(1,:,ip)*rdot+
     $          (real_work(1,:,ip)+tqr_1d(ipseust:ipseuen))*jaeq2
              real_ve(2,:,ip)=real_vptr(2,:,ip)*rdot+
     $          (real_work(2,:,ip)+tqz_1d(ipseust:ipseuen))*jaeq2
              real_ve(3,:,ip)=real_vptr(3,:,ip)*rdot+
     $           real_work(3,:,ip)*jaeq2
            ENDDO
          ENDIF
        ENDIF
      ELSE
        tdivv=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     subtract the rest of the linear V.grad(T) for each mode.
c-----------------------------------------------------------------------
      IF (eq_flow/='none') THEN
        DO imode=1,nmodes
          vdgr_ti(1,:,:,imode)=-dt*( vdgr_ti(1,:,:,imode)+
     $      tdivv(1,:,:,imode)+SUM(ve_eq*grad_ti(:,:,:,imode),1) )
        ENDDO
      ELSE
        vdgr_ti=-dt*(vdgr_ti+tdivv)
      ENDIF
      IF (separate_pe) THEN
        DO imode=1,nmodes
          vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)+
     $       dt*jfac*SUM(ja_eq*grad_ti(:,:,:,imode),1)/nd_eq(1,:,:)
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     multiply t_perturbed by div(v), then transform the
c     product to Fourier space.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.p_model/='isothermal') THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,ti,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,divv,
     $               real_divv,dealiase)
        real_scal=real_scal*real_divv
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tdivv,
     $               real_scal,dealiase)
      ELSE
        tdivv=0
      ENDIF
c-----------------------------------------------------------------------
c     subtract the linear (gamma-1)*T*div(V) for this mode.
c     also form the n_tilde*Teq*div(Veq) term and save in tdivv.
c-----------------------------------------------------------------------
      IF (p_model/='isothermal') THEN
        IF (eq_flow/='none') THEN
          DO imode=1,nmodes
            vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)
     $                         -dt*gamm1*(tdivv(1,:,:,imode)
     $                              +ti_eq(1,:,:)*divv(1,:,:,imode)
     $                              +ti(1,:,:,imode)*divv_eq(1,:,:))
            tdivv(1,:,:,imode)=-dt*gamm1*nd(1,:,:,imode)*
     $                               ti_eq(1,:,:)*divv_eq(1,:,:)/zeff
          ENDDO
        ELSE
          DO imode=1,nmodes
            vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)
     $                         -dt*gamm1*(tdivv(1,:,:,imode)
     $                              +ti_eq(1,:,:)*divv(1,:,:,imode))
            tdivv(1,:,:,imode)=0._r8
          ENDDO
        ENDIF
      ELSE
        tdivv=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     set temperature and grad(T) with T from the beginning of this
c     time split for thermal conduction if not already done.  
c-----------------------------------------------------------------------
      IF (p_model=='adiabat'.OR.p_model=='isothermal') THEN
        grad_ti=0._r8
      ELSE
        IF (.NOT.told_set) THEN
          CALL generic_ptr_set(rb%qtion,tb%qtion,tb%tgeom,inode,
     $                         ti,tir,tiz,1_i4)
          DO imode=1,nmodes
            grad_ti(1,:,:,imode)=tir(1,:,:,imode)
            grad_ti(2,:,:,imode)=tiz(1,:,:,imode)
            grad_ti(3,:,:,imode)=(0,1)*keff(imode)*ti(1,:,:,imode)/bigr
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c-PRE Heat soure term.
c     q_applied is set in q_applied_init.f and is normally zero
c-----------------------------------------------------------------------
c     DO imode=1,nmodes
c       IF (keff(imode)==0) THEN
c         CALL generic_ptr_set(rb%qq_applied,tb%qq_applied
c    $                         tb%tgeom,inode,q_a,dp,dp,0_i4)
c         vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)
c    $                         +dt*gamm1*q_a/(kboltz*bigr)
c       ENDIF
c     ENDDO
c-----------------------------------------------------------------------
c     For isotropic diffusion, multiply grad_ti by dt*diffusivity.
c     assume k_perp has gamma-1 factor in it.
c-----------------------------------------------------------------------
      pmodelif: IF (p_model=='isotropic') THEN
        grad_ti=grad_ti*dt*k_perpi
      ELSE IF (p_model(1:5)=='aniso') THEN
c-----------------------------------------------------------------------
c     for anisotropic thermal conduction, get the
c     magnetic field first.  note that we assume k_pll and k_perp have
c     the factor of (gamma-1) in them already.  for single-fluid
c     temperature, use the dominant k_plle.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      IF (nonlinear.AND.(p_model=="aniso_plltdep".OR.
     $                   p_model=="aniso_tdep")) THEN
        CALL generic_ptr_set(rb%qkappli_phi,tb%qkappli_phi,tb%tgeom,
     $                       inode,kappli,dp,dp,0_i4)

        IF (.NOT.closure_n0_only) THEN
          CALL generic_ptr_set(rb%qkaprpi_phi,tb%qkaprpi_phi,tb%tgeom,
     $                         inode,kaprpi,dp,dp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qkaprpi_n0,tb%qkaprpi_n0,tb%tgeom,
     $                         inode,kaprpi,dp,dp,0_i4)
          elt_1d=RESHAPE(kaprpi,(/ncx*ncy/))
        ENDIF
      ELSE
        IF (separate_pe) THEN
          kdiff=k_plli-k_perpi
        ELSE
          kdiff=k_plle-k_perpi
        ENDIF
      ENDIF
      IF (separate_pe.AND.k_cross>0) THEN
        CALL generic_ptr_set(rb%qti_b2,tb%qti_b2,tb%tgeom,
     $                       inode,ti_b2,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qbcrgti,tb%qbcrgti,tb%tgeom,
     $                       inode,bcrgti,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     For nonlinear calculations, the parallel tensor is 
c     B_tot B_tot / B^{2}. 
c     note that beq dot grad Teq is assumed to be zero, a slightly
c     stronger assumption than just a steady state relation.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                       inode,b0,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                       inode,real_bptr,dp,dp,0_i4)
        DO imode=1,nmodes
          bdgr_ti(1,:,:,imode)=be(1,:,:,imode)*ti_eqr(1,:,:)
     $                        +be(2,:,:,imode)*ti_eqz(1,:,:)
        ENDDO

        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,bdgr_ti,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_ti,
     $               real_work,dealiase)
c-----------------------------------------------------------------------
c       Build the product B (B dot grad T).  we are adding 
c       (be_eq+be).grad(T) to be.grad(T_eq).
c
c       if the thermal diffusivity is temperature-dependent,
c       multiply by its distribution in phi here.
c-----------------------------------------------------------------------
        IF (p_model=="aniso_plltdep".OR.p_model=="aniso_tdep") THEN
          IF (.NOT.closure_n0_only) THEN
            DO ip=1,nphi
              real_scal(1,:,ip)=(real_scal(1,:,ip)+
     $                           real_bptr(1,:,ip)*real_work(1,:,ip)+
     $                           real_bptr(2,:,ip)*real_work(2,:,ip)+
     $                           real_bptr(3,:,ip)*real_work(3,:,ip))*
     $            MAX(0._r8,kappli(1,:,ip)-kaprpi(1,:,ip))
            ENDDO
          ELSE
            DO ip=1,nphi
              real_scal(1,:,ip)=(real_scal(1,:,ip)+
     $                           real_bptr(1,:,ip)*real_work(1,:,ip)+
     $                           real_bptr(2,:,ip)*real_work(2,:,ip)+
     $                           real_bptr(3,:,ip)*real_work(3,:,ip))*
     $            MAX(0._r8,kappli(1,:,ip)-elt_1d(ipseust:ipseuen))
            ENDDO
          ENDIF
        ELSE
          real_scal(1,:,:)=
     $      kdiff*(real_scal(1,:,:)+SUM(real_bptr*real_work,1))
        ENDIF
        bbgradt(1,:,:)=real_bptr(1,:,:)*real_scal(1,:,:)
        bbgradt(2,:,:)=real_bptr(2,:,:)*real_scal(1,:,:)
        bbgradt(3,:,:)=real_bptr(3,:,:)*real_scal(1,:,:)
c-----------------------------------------------------------------------
c	Convert the above vector to the Fourier representation
c	and store in be_tot.  the denominator, B**2 is approximated
c       as (be_n0+be_eq)**2.
c
c       include fully 3D temperature_dependent perpendicular
c       thermal conduction if specified.
c-----------------------------------------------------------------------
        IF (.NOT.closure_n0_only) THEN
          magBsq3D(:,:)=SUM(real_bptr(:,:,:)**2,1)
          bbgradt(1,:,:)=bbgradt(1,:,:)/magBsq3D(:,:)
     $         +kaprpi(1,:,:)*real_work(1,:,:)
          bbgradt(2,:,:)=bbgradt(2,:,:)/magBsq3D(:,:)
     $         +kaprpi(1,:,:)*real_work(2,:,:)
          bbgradt(3,:,:)=bbgradt(3,:,:)/magBsq3D(:,:)
     $         +kaprpi(1,:,:)*real_work(3,:,:)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $         bbgradt,dealiase)
        ELSE
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $                 bbgradt,dealiase)
          bmagsq=SUM((b0+be_eq)**2,1)
          DO imode=1,nmodes
            be_tot(1,:,:,imode)=be_tot(1,:,:,imode)/bmagsq
            be_tot(2,:,:,imode)=be_tot(2,:,:,imode)/bmagsq
            be_tot(3,:,:,imode)=be_tot(3,:,:,imode)/bmagsq
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c     For linear calculations, the parallel tensor is divided 
c     into linear contributions with B^{2} approximated as Beq^{2}.
c-----------------------------------------------------------------------
      ELSE
        bmagsq=SUM(be_eq*be_eq,1)
        DO imode=1,nmodes
          bdgr_ti(1,:,:,imode)=kdiff*
     $      (be(1,:,:,imode)*ti_eqr(1,:,:)
     $      +be(2,:,:,imode)*ti_eqz(1,:,:)
     $      +SUM(grad_ti(:,:,:,imode)*be_eq,1))/bmagsq
          be_tot(1,:,:,imode)=be_eq(1,:,:)*bdgr_ti(1,:,:,imode)
          be_tot(2,:,:,imode)=be_eq(2,:,:)*bdgr_ti(1,:,:,imode)
          be_tot(3,:,:,imode)=be_eq(3,:,:)*bdgr_ti(1,:,:,imode)
c-----------------------------------------------------------------------
c         add the Be_eqXgrad(Ti) and Be_eqXgrad(Ti_eq) terms here.
c-----------------------------------------------------------------------
          IF (separate_pe.AND.k_cross>0) THEN
            CALL math_cart_cross(b_cr_gt,be_eq,grad_ti(:,:,:,imode),
     $                           -1._r8)
            b_cr_gt(1,:,:)=b_cr_gt(1,:,:)+be(3,:,:,imode)*ti_eqz(1,:,:)
            b_cr_gt(2,:,:)=b_cr_gt(2,:,:)-be(3,:,:,imode)*ti_eqr(1,:,:)
            b_cr_gt(3,:,:)=b_cr_gt(3,:,:)-be(1,:,:,imode)*ti_eqz(1,:,:)+
     $                                    be(2,:,:,imode)*ti_eqr(1,:,:)
            cr_coef=ti_b2(1,:,:)*(2*SUM(be_eq*be(:,:,:,imode),1)/bmagsq-
     $                            nd(1,:,:,imode)/nd_eq(1,:,:)-
     $                            ti(1,:,:,imode)/ti_eq(1,:,:))
            be_tot(1,:,:,imode)=be_tot(1,:,:,imode)+
     $                          b_cr_gt(1,:,:)*ti_b2(1,:,:)+
     $                          bcrgti (1,:,:)*cr_coef
            be_tot(2,:,:,imode)=be_tot(2,:,:,imode)+
     $                          b_cr_gt(2,:,:)*ti_b2(1,:,:)+
     $                          bcrgti (2,:,:)*cr_coef
            be_tot(3,:,:,imode)=be_tot(3,:,:,imode)+
     $                          b_cr_gt(3,:,:)*ti_b2(1,:,:)+
     $                          bcrgti (3,:,:)*cr_coef
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     if the perpendicular thermal diffusivity is temperature-dependent,
c     multiply grad_ti by its distribution.  also add parallel
c     contributions.
c
c     nonlinear temperature-dependent perpendicular thermal conduction
c     is coded with the parallel contribution above.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.p_model=="aniso_tdep") THEN
        IF (.NOT.closure_n0_only) THEN
          grad_ti=dt*be_tot
        ELSE
          DO imode=1,nmodes
            grad_ti(1,:,:,imode)=
     $        dt*(kaprpi(1,:,:)*grad_ti(1,:,:,imode)+
     $        be_tot(1,:,:,imode))
            grad_ti(2,:,:,imode)=
     $        dt*(kaprpi(1,:,:)*grad_ti(2,:,:,imode)+
     $            be_tot(2,:,:,imode))
            grad_ti(3,:,:,imode)=
     $        dt*(kaprpi(1,:,:)*grad_ti(3,:,:,imode)+
     $            be_tot(3,:,:,imode))
          ENDDO
        ENDIF
      ELSE
        grad_ti=dt*(k_perpi*grad_ti+be_tot)
      ENDIF

      ENDIF pmodelif
c-----------------------------------------------------------------------
c     also add the upwinding-like diffusion.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $               real_ve,dealiase)
        grad_ti=grad_ti+be_tot
      ENDIF
c-----------------------------------------------------------------------
c     multiply by number density.
c-----------------------------------------------------------------------
      IF (.NOT.nonlinear.OR.continuity=='fix profile') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_ti(jq,:,:,imode)=
     $        nd_eq(1,:,:)/zeff*grad_ti(jq,:,:,imode)
          ENDDO
          vdgr_ti(1,:,:,imode)=nd_eq(1,:,:)/zeff*vdgr_ti(1,:,:,imode)
        ENDDO
      ELSE IF (continuity=='n=0 only') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_ti(jq,:,:,imode)=
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))/zeff*grad_ti(jq,:,:,imode)
          ENDDO
          vdgr_ti(1,:,:,imode)=(nd_eq(1,:,:)+nd_n0(1,:,:))/zeff*
     $                         vdgr_ti(1,:,:,imode)
        ENDDO
      ELSE ! continuity = full
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_ti,
     $               real_work,dealiase)
        real_scal=real_scal*real_ndptr/zeff
        DO jq=1,3
          real_work(jq,:,:)=real_work(jq,:,:)*real_ndptr(1,:,:)/zeff
        ENDDO
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_ti,
     $               real_scal,dealiase)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,grad_ti,
     $               real_work,dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     add the divv_eq term that was multiplied by n-tilde earlier, and
c     form the contributions to each basis function.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        IF (visc_heat) THEN
          IF (separate_pe) THEN
            vdgr_ti(:,:,:,imode)=vdgr_ti(:,:,:,imode)+
     $        gamm1*vheat(:,:,:,imode)/kboltz
          ELSE
            vdgr_ti(:,:,:,imode)=vdgr_ti(:,:,:,imode)+
     $        (1-pe_frac)*gamm1*vheat(:,:,:,imode)/kboltz
          ENDIF
        ENDIF
        IF (ohm_heat.AND..NOT.separate_pe) THEN
          vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)+
     $       (0,1)*keff(imode)*grad_ti(3,:,:,imode)/bigr+
     $       dt*gamm1*mu0*ohm_en(1,:,:,imode)*(1-pe_frac)/kboltz+
     $       tdivv(1,:,:,imode)
        ELSE
          vdgr_ti(1,:,:,imode)=vdgr_ti(1,:,:,imode)+
     $       (0,1)*keff(imode)*grad_ti(3,:,:,imode)/bigr+
     $       tdivv(1,:,:,imode)
        ENDIF
        DO iv=1,nv
          int(1,:,iv,imode)=SUM(  alpha(:,:,iv)*vdgr_ti(1,:,:,imode)
     $                          -dalpdr(:,:,iv)*grad_ti(1,:,:,imode)
     $                          -dalpdz(:,:,iv)*grad_ti(2,:,:,imode),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE tirhs
c-----------------------------------------------------------------------
c     subprogram 9. terhs.
c     compute the integrand used for the rhs of the electron temperature
c     change equation with implicit diffusion.
c
c     dt*(gamma-1)*(D(Te_n)-ni*Te*div(Ve))-dt*ni*Ve.grad(Te)
c
c     where Te_n is the temperature at the beginning of the time step,
c     Te is either Te_n or the predicted temperature, and D is the
c     diffusion operator.
c-----------------------------------------------------------------------
      SUBROUTINE terhs(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: te_eq,te_eqr,te_eqz,
     $          ve_eq,be_eq,ja_eq,b0,q_a,nd_eq,nd_eqr,
     $          nd_eqz,nd_n0,real_bptr,real_ndptr,elecd_t,elecd_eq,dp,
     $          ds2,kapple,kaprpe,te_b2,bcrgte,divv_eq,real_vptr,
     $          real_jptr,upwc
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: bmagsq
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve,real_work,bbgradt
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal,real_divv
      REAL(r8), DIMENSION(mpseudo,nphi) :: magBsq3D
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: eleq_1d,elt_1d
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: tqr_1d,tqz_1d
      REAL(r8), DIMENSION(3,SIZE(bigr,1)*SIZE(bigr,2)) :: jaeq_1d 
      REAL(r8), DIMENSION(3,mpseudo) :: rvec
      REAL(r8), DIMENSION(  mpseudo) :: jaeq2,rdot
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: kdiff
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ve,dvr,dvz,be,ber,bez,
     $             te,ter,tez,nd,ndr,ndz,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             grad_te,be_tot,grad_nd,vel,ja
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             vdgr_te,divv,tdivv,bdgr_te,ohm_en
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: b_cr_gt
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: cr_coef
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      INTEGER(i4) :: nv,iv,imode,iq,jq,ncx,ncy,ip
      REAL(r8) :: jfac,fupw
      LOGICAL :: told_set
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      told_set=.false.
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     temperature and grad(T) at the beginning of this time split for
c     predictor steps, or at the centered prediction for corrector
c     steps.  we assume ja_eq/=0, so cases that call this integrand
c     always do predictor/corrector steps.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qtele_eq,tb%qtele_eq,tb%tgeom,
     $                     inode,te_eq,te_eqr,te_eqz,1_i4)
      IF (integrand_flag=='all te terms pre'.OR.
     $    k_cross>0.OR.impladv) THEN
        CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,inode,
     $                       te,ter,tez,1_i4)
        told_set=.true.
      ELSE
        CALL generic_ptr_set(rb%qwork3,tb%qwork3,tb%tgeom,inode,
     $                       te,ter,tez,1_i4)
      ENDIF
      DO imode=1,nmodes
        grad_te(1,:,:,imode)=ter(1,:,:,imode)
        grad_te(2,:,:,imode)=tez(1,:,:,imode)
        grad_te(3,:,:,imode)=(0,1)*keff(imode)*te(1,:,:,imode)/bigr
      ENDDO
c-----------------------------------------------------------------------
c     evaluate electron velocity.
c
c     The electron velocity is computed from the COM velocity 
c     and the total current density as
c
c          vel = v - j / ( n e (nu+1) )
c
c     where "n" is the electron number density, "e" the electron
c     charge and "nu" is the electron-ion mass ratio times the atomic
c     number "Z":
c                  nu = Z me / mi
c
c     we also need div(v)=d(Vr)/dR + Vr/R + d(Vz)dZ + 1/R d(Vth)/dTh
c     here we assume that Jeq.grad(nd_eq)=0.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,ve,dvr,dvz,1_i4)
      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,nd,ndr,ndz,1_i4)
      IF (nonlinear.AND.continuity=='n=0 only')
     $  CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,nd_eqr,nd_eqz,1_i4)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,ber,bez,1_i4)
      CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1/mu0)
      CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                     inode,ja_eq,dp,dp,0_i4)
      jfac=1/(elementary_q*(meomi+1))
c-----------------------------------------------------------------------
c     extra pointers are needed for the upwinding-like artificial
c     diffusion.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,real_vptr,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qupte_phi,tb%qupte_phi,tb%tgeom,
     $                       inode,upwc,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qja_tot,tb%qja_tot,tb%tgeom,
     $                       inode,real_jptr,dp,dp,0_i4)
        fupw=t_dart_upw*dt**2
        tqr_1d=RESHAPE(te_eqr,(/ncx*ncy/))
        tqz_1d=RESHAPE(te_eqz,(/ncx*ncy/))
      ENDIF
c-----------------------------------------------------------------------
c     div(v-COM) and grad(n)
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        IF (geom=='tor') THEN
          divv(1,:,:,imode)=dvr(1,:,:,imode)+dvz(2,:,:,imode)
     $                    +((0,1)*keff(imode)*ve(3,:,:,imode)
     $                                       +ve(1,:,:,imode))/bigr
        ELSE
          divv(1,:,:,imode)=dvr(1,:,:,imode)+dvz(2,:,:,imode)
     $                     +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr
        ENDIF
        grad_nd(1,:,:,imode)=ndr(1,:,:,imode)
        grad_nd(2,:,:,imode)=ndz(1,:,:,imode)
        grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
      ENDDO

      IF (nonlinear)
     $  CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)

      IF (eq_flow/='none')
     $  CALL generic_ptr_set(rb%qdvv_eq,tb%qdvv_eq,tb%tgeom,inode,
     $                       divv_eq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     find the Ohmic heating energy before the ja array is modified.
c-----------------------------------------------------------------------
      IF (nonlinear)
     $  CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,real_ve,
     $               dealiase)
      IF (ohm_heat) THEN
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_eq,tb%qelecd_eq,tb%tgeom,
     $                         inode,elecd_eq,dp,dp,0_i4)
          eleq_1d=RESHAPE(elecd_eq,(/ncx*ncy/))
          jaeq_1d=RESHAPE( ja_eq,(/3,ncx*ncy/))
          DO iq=1,mpseudo
            jaeq2(iq)=SUM(jaeq_1d(:,iq-1+ipseust)**2)
          ENDDO
        ENDIF
        DO imode=1,nmodes
          ohm_en(1,:,:,imode)=2*SUM(ja_eq*ja(:,:,:,imode),1)
        ENDDO
        IF (nonlinear) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,
     $                 ohm_en,real_scal,dealiase)
          real_scal(1,:,:)=real_scal(1,:,:)+SUM(real_ve**2,1)
          IF (threedeta) THEN
            CALL generic_ptr_set(rb%qelecd_phi,tb%qelecd_phi,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            DO ip=1,nphi
              DO iq=1,mpseudo
                real_scal(1,iq,ip)=real_scal(1,iq,ip)*elecd_t(1,iq,ip)+
     $            (elecd_t(1,iq,ip)-eleq_1d(iq-1+ipseust))*jaeq2(iq)
              ENDDO
            ENDDO
          ELSE IF (eta_model=='eta n=0 only') THEN
            CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            elt_1d=RESHAPE(elecd_t,(/ncx*ncy/))
            DO iq=1,mpseudo
              jaeq2(iq)=jaeq2(iq)*
     $              (elt_1d(iq-1+ipseust)-eleq_1d(iq-1+ipseust))
            ENDDO
            DO ip=1,nphi
              DO iq=1,mpseudo
                real_scal(1,iq,ip)=
     $            real_scal(1,iq,ip)*elt_1d(iq-1+ipseust)+jaeq2(iq)
              ENDDO
            ENDDO
          ENDIF
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                 ohm_en,real_scal,dealiase)
        ENDIF
        IF (nonlinear.AND.threedeta) THEN  ! done
        ELSE IF (eta_model=='eta n=0 only') THEN  ! done if nonlinear
          IF (.NOT.nonlinear) THEN
            CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                           inode,elecd_t,dp,dp,0_i4)
            DO imode=1,nmodes
              ohm_en(:,:,:,imode)=ohm_en(:,:,:,imode)*elecd_t
            ENDDO
          ENDIF
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          DO imode=1,nmodes
            ohm_en(:,:,:,imode)=
     $        ohm_en(:,:,:,imode)*elecd*MAX(ds2,0._r8)
          ENDDO
        ELSE
          DO imode=1,nmodes
            ohm_en(:,:,:,imode)=ohm_en(:,:,:,imode)*elecd
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     j.grad(n) (saved in vdgr_te) for nonlinear, full continuity
c     computations.  [keep div(j)=0 in mind here]
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $               real_work,dealiase)
        real_scal(1,:,:)=SUM(real_ve*real_work,1)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $               real_scal,dealiase)
      ELSE
        vdgr_te=0
      ENDIF

      IF (.NOT.nonlinear.OR.continuity=='fix profile') THEN
        DO imode=1,nmodes
          DO jq=1,3
            vel(jq,:,:,imode)=ve(jq,:,:,imode)
     $        -ja(jq,:,:,imode)*jfac/nd_eq(1,:,:)
          ENDDO
          divv(1,:,:,imode)=divv(1,:,:,imode)+
     $      jfac*(SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $            +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $            +ja(2,:,:,imode)*nd_eqz(1,:,:)
     $            +vdgr_te(1,:,:,imode))/(nd_eq(1,:,:)**2)
        ENDDO
      ELSE IF (continuity=='n=0 only') THEN
        bmagsq=1/(nd_eq(1,:,:)+nd_n0(1,:,:))
        DO imode=1,nmodes
          DO jq=1,3
            vel(jq,:,:,imode)=ve(jq,:,:,imode)
     $        -ja(jq,:,:,imode)*jfac*bmagsq
          ENDDO
          divv(1,:,:,imode)=divv(1,:,:,imode)+
     $      jfac*(SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $            +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $            +ja(2,:,:,imode)*nd_eqz(1,:,:)
     $            +vdgr_te(1,:,:,imode))*bmagsq**2
        ENDDO
      ELSE  !  continuity is full  (ja will be left as ja/n).
        DO imode=1,nmodes
          vdgr_te(1,:,:,imode)=jfac*(vdgr_te(1,:,:,imode)+
     $      SUM(ja_eq*grad_nd(:,:,:,imode),1)
     $      +ja(1,:,:,imode)*nd_eqr(1,:,:)
     $      +ja(2,:,:,imode)*nd_eqz(1,:,:))
        ENDDO
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $               real_scal,dealiase)
        DO jq=1,3
          real_ve(jq,:,:)=real_ve(jq,:,:)/real_ndptr(1,:,:)
        ENDDO
        real_scal=real_scal/real_ndptr**2
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,ja,
     $               real_ve,dealiase)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $               real_scal,dealiase)
        vel=ve-jfac*ja
        divv=divv+vdgr_te
      ENDIF
c-----------------------------------------------------------------------
c     find the linear v_tilde.grad(Teq) for each mode--needed here
c     to complete the upwinding-like diffusion at the next step.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        vdgr_te(1,:,:,imode)=vel(1,:,:,imode)*te_eqr(1,:,:)+
     $                       vel(2,:,:,imode)*te_eqz(1,:,:)
      ENDDO
c-----------------------------------------------------------------------
c     find the nonlinear contributions to v.grad(t) and t div(v)
c
c     transform v_perturbed and grad(t_perturbed) to real space,
c     find v.grad(t), then transform to Fourier space.  tdivv is
c     just used as temporary storage.
c
c     the upwinding-like diffusion term is also computed here.  the
c     predetermined coefficient is multiplied by VV.grad(T).
c     ** NOTE ** real_ve is now used to save this coefficient from
c     this point, so V_tilde(phi) array is lost.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,vel,real_ve,
     $               dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_te,
     $               real_work,dealiase)
        real_scal(1,:,:)=SUM(real_ve*real_work,1)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tdivv,
     $               real_scal,dealiase)

        IF (impladv.AND.t_dart_upw>0) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $                 real_scal,dealiase)
          DO ip=1,nphi
            rvec(1,:)=real_vptr(1,:,ip)-
     $           jfac*real_jptr(1,:,ip)/real_ndptr(1,:,ip)
            rvec(2,:)=real_vptr(2,:,ip)-
     $           jfac*real_jptr(2,:,ip)/real_ndptr(1,:,ip)
            rvec(3,:)=real_vptr(3,:,ip)-
     $           jfac*real_jptr(3,:,ip)/real_ndptr(1,:,ip)
            rdot=fupw*upwc(1,:,ip)*upw_aniso*
     $        ( real_scal(1,:,ip) + SUM(rvec*real_work(:,:,ip),1) ) /
     $        ( SUM(rvec*rvec,1) + smallnum )
            jaeq2=fupw*upwc(1,:,ip)*(1._r8-upw_aniso)
            real_ve(1,:,ip)=rvec(1,:)*rdot+
     $        (real_work(1,:,ip)+tqr_1d(ipseust:ipseuen))*jaeq2
            real_ve(2,:,ip)=rvec(2,:)*rdot+
     $        (real_work(2,:,ip)+tqz_1d(ipseust:ipseuen))*jaeq2
            real_ve(3,:,ip)=rvec(3,:)*rdot+real_work(3,:,ip)*jaeq2
          ENDDO
        ENDIF
      ELSE
        tdivv=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     subtract the rest of the linear V.grad(T) for each mode.
c-----------------------------------------------------------------------
      IF (eq_flow/='none') THEN
        DO imode=1,nmodes
          vdgr_te(1,:,:,imode)=-dt*( vdgr_te(1,:,:,imode)+
     $      tdivv(1,:,:,imode)+SUM(ve_eq*grad_te(:,:,:,imode),1)-
     $      jfac*SUM(ja_eq*grad_te(:,:,:,imode),1)/nd_eq(1,:,:) )
        ENDDO
      ELSE
        DO imode=1,nmodes
          vdgr_te(1,:,:,imode)=-dt*( vdgr_te(1,:,:,imode)+
     $                                 tdivv(1,:,:,imode)-
     $      jfac*SUM(ja_eq*grad_te(:,:,:,imode),1)/nd_eq(1,:,:) )
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     multiply t_perturbed by div(v), then transform the
c     product to Fourier space.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.p_model/='isothermal') THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,te,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,divv,
     $               real_divv,dealiase)
        real_scal=real_scal*real_divv
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tdivv,
     $               real_scal,dealiase)
      ELSE
        tdivv=0
      ENDIF
c-----------------------------------------------------------------------
c     subtract the linear (gamma-1)*T*div(V) for this mode.
c     also form the n_tilde*Teq*div(Veq) term and save in tdivv.
c-----------------------------------------------------------------------
      IF (p_model/='isothermal') THEN
        IF (eq_flow/='none') THEN
          DO imode=1,nmodes
            vdgr_te(1,:,:,imode)=vdgr_te(1,:,:,imode)
     $                         -dt*gamm1*(tdivv(1,:,:,imode)
     $                              +te_eq(1,:,:)*divv(1,:,:,imode)
     $                              +te(1,:,:,imode)*divv_eq(1,:,:))
            tdivv(1,:,:,imode)=-dt*gamm1*nd(1,:,:,imode)*
     $                               te_eq(1,:,:)*divv_eq(1,:,:)
          ENDDO
        ELSE
          DO imode=1,nmodes
            vdgr_te(1,:,:,imode)=vdgr_te(1,:,:,imode)
     $                         -dt*gamm1*(tdivv(1,:,:,imode)
     $                              +te_eq(1,:,:)*divv(1,:,:,imode))
            tdivv(1,:,:,imode)=0._r8
          ENDDO
        ENDIF
      ELSE
        tdivv=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     set temperature and grad(T) with T from the beginning of this
c     time split for thermal conduction if not already done.  
c-----------------------------------------------------------------------
      IF (p_model=='adiabat'.OR.p_model=='isothermal') THEN
        grad_te=0._r8
      ELSE
        IF (.NOT.told_set) THEN
          CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,inode,
     $                         te,ter,tez,1_i4)
          DO imode=1,nmodes
            grad_te(1,:,:,imode)=ter(1,:,:,imode)
            grad_te(2,:,:,imode)=tez(1,:,:,imode)
            grad_te(3,:,:,imode)=(0,1)*keff(imode)*te(1,:,:,imode)/bigr
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c-PRE Heat soure term.
c     q_applied is set in q_applied_init.f and is normally zero
c-----------------------------------------------------------------------
c     DO imode=1,nmodes
c       IF (keff(imode)==0) THEN
c         CALL generic_ptr_set(rb%qq_applied,tb%qq_applied
c    $                         tb%tgeom,inode,q_a,dp,dp,0_i4)
c         vdgr_te(1,:,:,imode)=vdgr_te(1,:,:,imode)
c    $                         +dt*gamm1*q_a/(kboltz*bigr)
c       ENDIF
c     ENDDO
c-----------------------------------------------------------------------
c     For isotropic diffusion, add the grad k_perp grad T old with
c     the adiabatic terms.  assume k_perp has gamma-1 factor in it.
c-----------------------------------------------------------------------
      pmodelif: IF (p_model=='isotropic') THEN
        grad_te=grad_te*dt*k_perpe
      ELSE IF (p_model(1:5)=='aniso') THEN
c-----------------------------------------------------------------------
c     for anisotropic thermal conduction, get the
c     magnetic field first.  note that we assume k_pll and k_perp have
c     the factor of (gamma-1) in them already.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      IF (nonlinear.AND.(p_model=="aniso_plltdep".OR.
     $                   p_model=="aniso_tdep")) THEN
        CALL generic_ptr_set(rb%qkapple_phi,tb%qkapple_phi,tb%tgeom,
     $                       inode,kapple,dp,dp,0_i4)
        IF (.NOT.closure_n0_only) THEN
          CALL generic_ptr_set(rb%qkaprpe_phi,tb%qkaprpe_phi,tb%tgeom,
     $                         inode,kaprpe,dp,dp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qkaprpe_n0,tb%qkaprpe_n0,tb%tgeom,
     $                         inode,kaprpe,dp,dp,0_i4)
          elt_1d=RESHAPE(kaprpe,(/ncx*ncy/))
        ENDIF
      ELSE
        kdiff=k_plle-k_perpe
      ENDIF
      IF (k_cross>0) THEN
        CALL generic_ptr_set(rb%qte_b2,tb%qte_b2,tb%tgeom,
     $                       inode,te_b2,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qbcrgte,tb%qbcrgte,tb%tgeom,
     $                       inode,bcrgte,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     For nonlinear calculations, the parallel tensor is 
c     B_tot B_tot / B^{2}. 
c     note that beq dot grad Teq is assumed to be zero, a slightly
c     stronger assumption than just a steady state relation.
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                       inode,b0,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                       inode,real_bptr,dp,dp,0_i4)
        DO imode=1,nmodes
          bdgr_te(1,:,:,imode)=be(1,:,:,imode)*te_eqr(1,:,:)
     $                        +be(2,:,:,imode)*te_eqz(1,:,:)
        ENDDO

        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,bdgr_te,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_te,
     $               real_work,dealiase)
c-----------------------------------------------------------------------
c       Build the product B (B dot grad T).  we are adding 
c       (be_eq+be).grad(T) to be.grad(T_eq).
c
c       if the thermal conductivity is temperature-dependent
c       multiply by its distribution in phi here.
c-----------------------------------------------------------------------
        IF (p_model=="aniso_plltdep".OR.p_model=="aniso_tdep") THEN
          IF (.NOT.closure_n0_only) THEN
            DO ip=1,nphi
              real_scal(1,:,ip)=(real_scal(1,:,ip)+
     $                           real_bptr(1,:,ip)*real_work(1,:,ip)+
     $                           real_bptr(2,:,ip)*real_work(2,:,ip)+
     $                           real_bptr(3,:,ip)*real_work(3,:,ip))*
     $            MAX(0._r8,kapple(1,:,ip)-kaprpe(1,:,ip))
            ENDDO
          ELSE
            DO ip=1,nphi
              real_scal(1,:,ip)=(real_scal(1,:,ip)+
     $                           real_bptr(1,:,ip)*real_work(1,:,ip)+
     $                           real_bptr(2,:,ip)*real_work(2,:,ip)+
     $                           real_bptr(3,:,ip)*real_work(3,:,ip))*
     $            MAX(0._r8,kapple(1,:,ip)-elt_1d(ipseust:ipseuen))
            ENDDO
          ENDIF
        ELSE
          real_scal(1,:,:)=
     $      kdiff*(real_scal(1,:,:)+SUM(real_bptr*real_work,1))
        ENDIF
        bbgradt(1,:,:)=real_bptr(1,:,:)*real_scal(1,:,:)
        bbgradt(2,:,:)=real_bptr(2,:,:)*real_scal(1,:,:)
        bbgradt(3,:,:)=real_bptr(3,:,:)*real_scal(1,:,:)
c-----------------------------------------------------------------------
c	Convert the above vector to the Fourier representation
c	and store in be_tot.  the denominator, B**2 is approximated
c       as (be_n0+be_eq)**2.
c
c       include fully 3D temperature_dependent perpendicular thermal
c       conduction if specified.
c-----------------------------------------------------------------------
        IF (.NOT.closure_n0_only) THEN
          magBsq3D(:,:)=SUM(real_bptr(:,:,:)**2,1)
          bbgradt(1,:,:)=bbgradt(1,:,:)/magBsq3D(:,:)
     $        +kaprpe(1,:,:)*real_work(1,:,:)
          bbgradt(2,:,:)=bbgradt(2,:,:)/magBsq3D(:,:)
     $        +kaprpe(1,:,:)*real_work(2,:,:)
          bbgradt(3,:,:)=bbgradt(3,:,:)/magBsq3D(:,:)
     $        +kaprpe(1,:,:)*real_work(3,:,:)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $        bbgradt,dealiase)
        ELSE
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $                 bbgradt,dealiase)
          bmagsq=SUM((b0+be_eq)**2,1)
          DO imode=1,nmodes
            be_tot(1,:,:,imode)=be_tot(1,:,:,imode)/bmagsq
            be_tot(2,:,:,imode)=be_tot(2,:,:,imode)/bmagsq
            be_tot(3,:,:,imode)=be_tot(3,:,:,imode)/bmagsq
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c     For linear calculations, the parallel tensor is divided 
c     into linear contributions with B^{2} approximated as Beq^{2}.
c-----------------------------------------------------------------------
      ELSE
        bmagsq=SUM(be_eq*be_eq,1)
        DO imode=1,nmodes
          bdgr_te(1,:,:,imode)=kdiff*
     $      (be(1,:,:,imode)*te_eqr(1,:,:)
     $      +be(2,:,:,imode)*te_eqz(1,:,:)
     $      +SUM(grad_te(:,:,:,imode)*be_eq,1))/bmagsq
          be_tot(1,:,:,imode)=be_eq(1,:,:)*bdgr_te(1,:,:,imode)
          be_tot(2,:,:,imode)=be_eq(2,:,:)*bdgr_te(1,:,:,imode)
          be_tot(3,:,:,imode)=be_eq(3,:,:)*bdgr_te(1,:,:,imode)
c-----------------------------------------------------------------------
c         add the Be_eqXgrad(Te) and Be_eqXgrad(Te_eq) terms here.
c-----------------------------------------------------------------------
          IF (k_cross>0) THEN
            CALL math_cart_cross(b_cr_gt,be_eq,grad_te(:,:,:,imode),
     $                           1._r8)
            b_cr_gt(1,:,:)=b_cr_gt(1,:,:)-be(3,:,:,imode)*te_eqz(1,:,:)
            b_cr_gt(2,:,:)=b_cr_gt(2,:,:)+be(3,:,:,imode)*te_eqr(1,:,:)
            b_cr_gt(3,:,:)=b_cr_gt(3,:,:)+be(1,:,:,imode)*te_eqz(1,:,:)-
     $                                    be(2,:,:,imode)*te_eqr(1,:,:)
            cr_coef=te_b2(1,:,:)*(nd(1,:,:,imode)/nd_eq(1,:,:)+
     $                            te(1,:,:,imode)/te_eq(1,:,:)-
     $                            2*SUM(be_eq*be(:,:,:,imode),1)/bmagsq)
            be_tot(1,:,:,imode)=be_tot(1,:,:,imode)+
     $                          b_cr_gt(1,:,:)*te_b2(1,:,:)+
     $                          bcrgte (1,:,:)*cr_coef
            be_tot(2,:,:,imode)=be_tot(2,:,:,imode)+
     $                          b_cr_gt(2,:,:)*te_b2(1,:,:)+
     $                          bcrgte (2,:,:)*cr_coef
            be_tot(3,:,:,imode)=be_tot(3,:,:,imode)+
     $                          b_cr_gt(3,:,:)*te_b2(1,:,:)+
     $                          bcrgte (3,:,:)*cr_coef
          ENDIF
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     if the perpendicular thermal diffusivity is temperature-dependent,
c     multiply grad_te by its distribution.  also add parallel
c     contributions.
c
c     nonlinear temperature-dependent perpendicular thermal conduction
c     is now above.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.p_model=="aniso_tdep") THEN
        IF (.NOT.closure_n0_only) THEN
          grad_te=dt*be_tot
        ELSE
          DO imode=1,nmodes
            grad_te(1,:,:,imode)=
     $         dt*(kaprpe(1,:,:)*grad_te(1,:,:,imode)+
     $             be_tot(1,:,:,imode))
            grad_te(2,:,:,imode)=
     $         dt*(kaprpe(1,:,:)*grad_te(2,:,:,imode)+
     $             be_tot(2,:,:,imode))
            grad_te(3,:,:,imode)=
     $         dt*(kaprpe(1,:,:)*grad_te(3,:,:,imode)+
     $             be_tot(3,:,:,imode))
          ENDDO
        ENDIF
      ELSE
        grad_te=dt*(k_perpe*grad_te+be_tot)
      ENDIF

      ENDIF pmodelif
c-----------------------------------------------------------------------
c     also add the upwinding-like diffusion.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $               real_ve,dealiase)
        grad_te=grad_te+be_tot
      ENDIF
c-----------------------------------------------------------------------
c     multiply by number density.
c-----------------------------------------------------------------------
      IF (.NOT.nonlinear.OR.continuity=='fix profile') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_te(jq,:,:,imode)=nd_eq(1,:,:)*grad_te(jq,:,:,imode)
          ENDDO
          vdgr_te(1,:,:,imode)=nd_eq(1,:,:)*vdgr_te(1,:,:,imode)
        ENDDO
      ELSE IF (continuity=='n=0 only') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_te(jq,:,:,imode)=
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))*grad_te(jq,:,:,imode)
          ENDDO
          vdgr_te(1,:,:,imode)=(nd_eq(1,:,:)+nd_n0(1,:,:))*
     $                         vdgr_te(1,:,:,imode)
        ENDDO
      ELSE ! continuity = full
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $               real_scal,dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_te,
     $               real_work,dealiase)
        real_scal=real_scal*real_ndptr
        DO jq=1,3
          real_work(jq,:,:)=real_work(jq,:,:)*real_ndptr(1,:,:)
        ENDDO
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,vdgr_te,
     $               real_scal,dealiase)
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,grad_te,
     $               real_work,dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     add the divv_eq term that was multiplied by n-tilde earlier, and
c     form the contributions to each basis function.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        IF (ohm_heat) THEN
          vdgr_te(1,:,:,imode)=vdgr_te(1,:,:,imode)+
     $       (0,1)*keff(imode)*grad_te(3,:,:,imode)/bigr+
     $       dt*gamm1*mu0*ohm_en(1,:,:,imode)/kboltz+
     $       tdivv(1,:,:,imode)
        ELSE
          vdgr_te(1,:,:,imode)=vdgr_te(1,:,:,imode)+
     $       (0,1)*keff(imode)*grad_te(3,:,:,imode)/bigr+
     $       tdivv(1,:,:,imode)
        ENDIF
        DO iv=1,nv
          int(1,:,iv,imode)=SUM(  alpha(:,:,iv)*vdgr_te(1,:,:,imode)
     $                          -dalpdr(:,:,iv)*grad_te(1,:,:,imode)
     $                          -dalpdz(:,:,iv)*grad_te(2,:,:,imode),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE terhs
c-----------------------------------------------------------------------
c     subprogram 10. t_aniso_dot.
c
c     compute the action of the anisotropic thermal conduction operator
c     (plus identity term) on a vector:
c
c     (ns*I-dt*f*D).dTs
c
c     where D is the conduction operator, f is a centering parameter,
c     dTs is an iterate, and n is number density.
c
c     when implicit advection is used, the operator is
c
c     (ns*I + 0.5*dt*ns*( V.grad(I)+(gamma-1)*div(V)*I ) - dt*f*D).dTs
c
c     this is now used for nonlinear computations only.
c-----------------------------------------------------------------------
      SUBROUTINE t_aniso_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: be_eq,b0,nd_eq,nd_n0,
     $          real_bptr,real_ndptr,kappl,kaprp,real_jptr,real_vptr,
     $          ndqr,ndqz,real_grdv,upwc,dp
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: bmagsq
      REAL(r8), DIMENSION(  SIZE(bigr,1)*SIZE(bigr,2)) :: kprp_1d
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_vect,bbgradt,real_grn
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scl,real_t
      REAL(r8), DIMENSION(mpseudo,nphi) :: magBsq3D
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: kdiff,zz,kpll,kprp,jfac,fdt,hdt,fupw,fupwi,gm1
      REAL(r8), DIMENSION(3) :: rvec
      REAL(r8) :: rscl2,rdot

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,nd,ndr,ndz,dc
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             grad_tie,grad_nd,be_tot
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: tie,
     $             tier,tiez,bdgr_tie

      INTEGER(i4) :: nv,iv,imode,iq,jq,ncx,ncy,ip,ix
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      fdt=fthc*dt
      hdt=0.5*dt
      IF (integrand_flag(1:6)=='all ti') THEN
        zz=zeff
        IF (separate_pe) THEN
          kpll=k_plli
          jfac=hdt*meomi/(zeff*elementary_q*(1+meomi))
        ELSE
          kpll=k_plle
          jfac=0
        ENDIF
        kprp=k_perpi
      ELSE
        zz=1
        kpll=k_plle
        kprp=k_perpe
        jfac=-hdt/(elementary_q*(1+meomi))
      ENDIF
      IF (p_model=='isothermal') THEN
        gm1=0
      ELSE
        gm1=gamm1
      ENDIF
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     number densities as required.
c-PRE faster to form J.grad(ne) at quadrature points.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,ndqr,ndqz,1_i4)
      CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                     inode,real_ndptr,dp,dp,0_i4)
      IF (continuity=='n=0 only')
     $  CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,dp,dp,0_i4)
      IF (impladv.AND.separate_pe) THEN
        CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,
     $                       inode,nd,ndr,ndz,1_i4)
        DO imode=1,nmodes
          grad_nd(1,:,:,imode)=ndr(1,:,:,imode)
          grad_nd(2,:,:,imode)=ndz(1,:,:,imode)
          grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
          IF (keff(imode)==0) THEN
            grad_nd(1,:,:,imode)=grad_nd(1,:,:,imode)+ndqr(1,:,:)
            grad_nd(2,:,:,imode)=grad_nd(2,:,:,imode)+ndqz(1,:,:)
          ENDIF
        ENDDO
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $               real_grn,dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     set dT_s and grad(dT_s) with the iterate stored in work2.  it
c     is used only once, so quadrature storage is not used--call
c     all_eval not ptr_set.
c-----------------------------------------------------------------------
      CALL generic_all_eval(rb%work2,tb%work2,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      tie,tier,tiez,1_i4)
      DO imode=1,nmodes
        grad_tie(1,:,:,imode)=tier(1,:,:,imode)
        grad_tie(2,:,:,imode)=tiez(1,:,:,imode)
        grad_tie(3,:,:,imode)=(0,1)*keff(imode)*tie(1,:,:,imode)/bigr
      ENDDO
c-----------------------------------------------------------------------
c     terms associated with implicit advection and the n*dT/dt term
c     are here.  note that tie becomes
c
c     n_s*( dT_s + dt/2*(V_s.grad(dT_s) + (gamma-1)*n_s*div(V_s)*dT_s) )
c
c     time-centered implicit advection, n_s*V_s and n_s*div(V_s) for
c     each species s.  the Jeq.grad(neq)=0 approximation is used here.
c
c     for ions, we need
c
c       (dt/2)*n_i*V_i = (dt/2)*(n_e*V/z + J*meomi/(z*e*(1+meomi)) )
c
c     where meomi = z*me/mi.  jfac is the full coefficient for J.
c     we also need
c
c       (dt/2)*n_i*div(V_i) =
c         (dt/2)*(n_e*div(V)/z - J.grad(n_e)*meomi/(n_e*z*e*(1+meomi))
c
c     the electron terms have z->1 and the meomi in the numerator
c     replaced by -1.
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,tie,real_t,
     $             dealiase)
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_tie,
     $             real_vect,dealiase)

      IF (impladv) THEN
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,real_vptr,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                       inode,real_grdv,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qja_tot,tb%qja_tot,tb%tgeom,
     $                       inode,real_jptr,dp,dp,0_i4)

        IF (continuity=='fix profile') THEN
          real_scl(1,:,:)=
     $      hdt*(SUM(real_vptr*real_vect,1)+
     $           gm1*real_t(1,:,:)*(real_grdv(1,:,:)+real_grdv(5,:,:)+
     $                              real_grdv(9,:,:)) )
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                 bdgr_tie,real_scl,dealiase)
          DO imode=1,nmodes
            tie(:,:,:,imode)=
     $        (tie(:,:,:,imode)+bdgr_tie(:,:,:,imode))*nd_eq/zz
          ENDDO
          IF (separate_pe) THEN
            real_scl(1,:,:)=jfac*(
     $        SUM(real_jptr*real_vect,1)-
     $        SUM(real_jptr*real_grn ,1)*
     $          gm1*real_t(1,:,:)/real_ndptr(1,:,:) )
            CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                   bdgr_tie,real_scl,dealiase)
            tie=tie+bdgr_tie
          ENDIF

        ELSE IF (continuity=='n=0 only') THEN
          real_scl(1,:,:)=
     $      hdt*(SUM(real_vptr*real_vect,1)+
     $           gm1*real_t(1,:,:)*(real_grdv(1,:,:)+real_grdv(5,:,:)+
     $                              real_grdv(9,:,:)) )
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                 bdgr_tie,real_scl,dealiase)
          DO imode=1,nmodes
            tie(:,:,:,imode)=
     $        (tie(:,:,:,imode)+bdgr_tie(:,:,:,imode))*(nd_eq+nd_n0)/zz
          ENDDO
          IF (separate_pe) THEN
            real_scl(1,:,:)=jfac*(
     $        SUM(real_jptr*real_vect,1)-
     $        SUM(real_jptr*real_grn ,1)*
     $          gm1*real_t(1,:,:)/real_ndptr(1,:,:) )
            CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,
     $                   bdgr_tie,real_scl,dealiase)
            tie=tie+bdgr_tie
          ENDIF

        ELSE ! continuity = full
          real_scl(1,:,:)=real_ndptr(1,:,:)*( real_t(1,:,:)+
     $      hdt*(SUM(real_vptr*real_vect,1)+
     $           gm1*real_t(1,:,:)*(real_grdv(1,:,:)+real_grdv(5,:,:)+
     $                              real_grdv(9,:,:)) ) )/zz
          IF (separate_pe) THEN
            real_scl(1,:,:)=real_scl(1,:,:) + jfac*(
     $        SUM(real_jptr*real_vect,1)-
     $        SUM(real_jptr*real_grn ,1)*
     $          gm1*real_t(1,:,:)/real_ndptr(1,:,:) )
          ENDIF
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tie,real_scl,
     $                 dealiase)
        ENDIF

      ELSE
        IF (continuity=='fix profile') THEN
          DO imode=1,nmodes
            tie(:,:,:,imode)=tie(:,:,:,imode)*nd_eq/zz
          ENDDO
        ELSE IF (continuity=='n=0 only') THEN
          DO imode=1,nmodes
            tie(:,:,:,imode)=tie(:,:,:,imode)*(nd_eq+nd_n0)/zz
          ENDDO
        ELSE ! continuity = full
          real_t=real_t*real_ndptr/zz
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,tie,real_t,
     $                 dealiase)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     magnetic field.  note that we assume k_pll and k_perp have
c     the factor of (gamma-1) in them already.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,dc,dc,0_i4)
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      IF (p_model=="aniso_plltdep".OR.p_model=="aniso_tdep") THEN
        IF (integrand_flag(1:6)=='all ti') THEN
          CALL generic_ptr_set(rb%qkappli_phi,tb%qkappli_phi,tb%tgeom,
     $                         inode,kappl,dp,dp,0_i4)
          IF (.NOT.closure_n0_only) THEN
            CALL generic_ptr_set(rb%qkaprpi_phi,tb%qkaprpi_phi,tb%tgeom,
     $                           inode,kaprp,dp,dp,0_i4)
          ELSE
            CALL generic_ptr_set(rb%qkaprpi_n0,tb%qkaprpi_n0,tb%tgeom,
     $                           inode,kaprp,dp,dp,0_i4)
            kprp_1d=RESHAPE(kaprp,(/ncx*ncy/))
          ENDIF
        ELSE
          CALL generic_ptr_set(rb%qkapple_phi,tb%qkapple_phi,tb%tgeom,
     $                         inode,kappl,dp,dp,0_i4)
          IF (.NOT.closure_n0_only) THEN
            CALL generic_ptr_set(rb%qkaprpe_phi,tb%qkaprpe_phi,tb%tgeom,
     $                           inode,kaprp,dp,dp,0_i4)
          ELSE
            CALL generic_ptr_set(rb%qkaprpe_n0,tb%qkaprpe_n0,tb%tgeom,
     $                           inode,kaprp,dp,dp,0_i4)
            kprp_1d=RESHAPE(kaprp,(/ncx*ncy/))
          ENDIF
        ENDIF
      ELSE
        kdiff=kpll-kprp
      ENDIF
c-----------------------------------------------------------------------
c     For nonlinear calculations, the parallel tensor is 
c     B_tot B_tot / B^{2}. 
c     note that beq dot grad Teq is assumed to be zero.
c-----------------------------------------------------------------------
      IF (p_model(1:5)=='aniso') THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                       inode,b0,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                       inode,real_bptr,dp,dp,0_i4)
c-----------------------------------------------------------------------
c       Build the product B (B dot grad T), (be_eq+be).grad(T) first.
c
c       if the thermal conductivity is temperature-dependent
c       multiply by its distribution in phi here.
c-----------------------------------------------------------------------
        IF (p_model=="aniso_plltdep".OR.p_model=="aniso_tdep") THEN
          IF (.NOT.closure_n0_only) THEN
            DO ip=1,nphi
              real_scl(1,:,ip)=(real_bptr(1,:,ip)*real_vect(1,:,ip)+
     $                          real_bptr(2,:,ip)*real_vect(2,:,ip)+
     $                          real_bptr(3,:,ip)*real_vect(3,:,ip))*
     $            MAX(0._r8,kappl(1,:,ip)-kaprp(1,:,ip))
            ENDDO
          ELSE
            DO ip=1,nphi
              real_scl(1,:,ip)=(real_bptr(1,:,ip)*real_vect(1,:,ip)+
     $                          real_bptr(2,:,ip)*real_vect(2,:,ip)+
     $                          real_bptr(3,:,ip)*real_vect(3,:,ip))*
     $            MAX(0._r8,kappl(1,:,ip)-kprp_1d(ipseust:ipseuen))
            ENDDO
          ENDIF
        ELSE
          real_scl(1,:,:)=kdiff*SUM(real_bptr*real_vect,1)
        ENDIF
        bbgradt(1,:,:)=real_bptr(1,:,:)*real_scl(1,:,:)
        bbgradt(2,:,:)=real_bptr(2,:,:)*real_scl(1,:,:)
        bbgradt(3,:,:)=real_bptr(3,:,:)*real_scl(1,:,:)
c-----------------------------------------------------------------------
c	Convert the above vector to the Fourier representation
c	and store in be_tot.  B**2 is approximated as (b_n0+be_eq)**2.
c
c       include fully 3D temperature_dependent perpendicular
c       thermal conduction if specified.
c-----------------------------------------------------------------------
        IF (.NOT.closure_n0_only) THEN
          magBsq3D(:,:)=SUM(real_bptr(:,:,:)**2,1)
          bbgradt(1,:,:)=bbgradt(1,:,:)/magBsq3D(:,:)
     $         +kaprp(1,:,:)*real_vect(1,:,:)
          bbgradt(2,:,:)=bbgradt(2,:,:)/magBsq3D(:,:)
     $         +kaprp(1,:,:)*real_vect(2,:,:)
          bbgradt(3,:,:)=bbgradt(3,:,:)/magBsq3D(:,:)
     $         +kaprp(1,:,:)*real_vect(3,:,:)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $         bbgradt,dealiase)
        ELSE
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $                 bbgradt,dealiase)
          bmagsq=SUM((b0+be_eq)**2,1)
          DO imode=1,nmodes
            be_tot(1,:,:,imode)=be_tot(1,:,:,imode)/bmagsq
            be_tot(2,:,:,imode)=be_tot(2,:,:,imode)/bmagsq
            be_tot(3,:,:,imode)=be_tot(3,:,:,imode)/bmagsq
          ENDDO
        ENDIF
      ELSE
        be_tot=0
      ENDIF
c-----------------------------------------------------------------------
c     if the perpendicular thermal diffusivity is temperature-dependent,
c     multiply grad_t by its distribution.  also add parallel
c     contributions.
c
c     nonlinear temperature-dependent perpendicular thermal conduction
c     is now above.
c-----------------------------------------------------------------------
      IF (p_model=="aniso_tdep") THEN
        IF (.NOT.closure_n0_only) THEN
          grad_tie=dt*be_tot
        ELSE
          DO imode=1,nmodes
            grad_tie(1,:,:,imode)=fdt*
     $        (kaprp(1,:,:)*grad_tie(1,:,:,imode)+be_tot(1,:,:,imode))
            grad_tie(2,:,:,imode)=fdt*
     $        (kaprp(1,:,:)*grad_tie(2,:,:,imode)+be_tot(2,:,:,imode))
            grad_tie(3,:,:,imode)=fdt*
     $        (kaprp(1,:,:)*grad_tie(3,:,:,imode)+be_tot(3,:,:,imode))
          ENDDO
        ENDIF
      ELSE IF (p_model(1:5)=='aniso') THEN
        grad_tie=fdt*(kprp*grad_tie+be_tot)
      ELSE IF (p_model=='isotropic') THEN
        grad_tie=fdt*kprp*grad_tie
      ELSE
        grad_tie=0
      ENDIF
c-----------------------------------------------------------------------
c     find the upwinding-like artificial diffusion.  at this point
c     real_vect still holds grad_tie as a function of phi.  the
c     bbgradt array is used for temporary storage of species total
c     velocity and then the VV.grad(T) vector.
c
c-PRE reuse V.grad(T) computation earlier when continuity is unified.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        IF (integrand_flag(1:6)=='all ti') THEN
          CALL generic_ptr_set(rb%qupti_phi,tb%qupti_phi,tb%tgeom,
     $                         inode,upwc,dp,dp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qupte_phi,tb%qupte_phi,tb%tgeom,
     $                         inode,upwc,dp,dp,0_i4)
        ENDIF

        IF (separate_pe) THEN
          fupw =dt**2*t_dart_upw*upw_aniso
          fupwi=dt**2*t_dart_upw*(1._r8-upw_aniso)
          DO ip=1,nphi
            DO ix=1,mpseudo
              rvec(1)=hdt*real_vptr(1,ix,ip)+
     $               jfac*real_jptr(1,ix,ip)/real_ndptr(1,ix,ip)
              rvec(2)=hdt*real_vptr(2,ix,ip)+
     $               jfac*real_jptr(2,ix,ip)/real_ndptr(1,ix,ip)
              rvec(3)=hdt*real_vptr(3,ix,ip)+
     $               jfac*real_jptr(3,ix,ip)/real_ndptr(1,ix,ip)
              rdot=fupw*upwc(1,ix,ip)*
     $             SUM(rvec*real_vect(:,ix,ip)) /
     $           ( SUM(rvec*rvec) + smallnum )
              rscl2=fupwi*upwc(1,ix,ip)
              bbgradt(:,ix,ip)=rvec*rdot+real_vect(:,ix,ip)*rscl2
            ENDDO
          ENDDO
        ELSE
          fupw =dt**2*t_dart_upw*upw_aniso
          fupwi=dt**2*t_dart_upw*(1._r8-upw_aniso)
          DO ip=1,nphi
            DO ix=1,mpseudo
              rvec=real_vptr(:,ix,ip)
              rdot=fupw*upwc(1,ix,ip)*
     $             SUM(rvec*real_vect(:,ix,ip)) /
     $           ( SUM(rvec*rvec) + smallnum )
              rscl2=fupwi*upwc(1,ix,ip)
              bbgradt(:,ix,ip)=rvec*rdot+real_vect(:,ix,ip)*rscl2
            ENDDO
          ENDDO
        ENDIF

        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,be_tot,
     $               bbgradt,dealiase)
        grad_tie=grad_tie+be_tot
      ENDIF
c-----------------------------------------------------------------------
c     multiply by number density for all p_models.
c     the identity term also needs a factor of n in these cases.
c-----------------------------------------------------------------------
      IF (continuity=='fix profile') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_tie(jq,:,:,imode)=
     $        nd_eq(1,:,:)/zz*grad_tie(jq,:,:,imode)
          ENDDO
        ENDDO
      ELSE IF (continuity=='n=0 only') THEN
        DO imode=1,nmodes
          DO jq=1,3
            grad_tie(jq,:,:,imode)=
     $        (nd_eq(1,:,:)+nd_n0(1,:,:))/zz*grad_tie(jq,:,:,imode)
          ENDDO
        ENDDO
      ELSE ! continuity = full
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_tie,
     $               real_vect,dealiase)
        DO jq=1,3
          real_vect(jq,:,:)=real_vect(jq,:,:)*real_ndptr(1,:,:)/zz
        ENDDO
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,grad_tie,
     $               real_vect,dealiase)
      ENDIF
c-----------------------------------------------------------------------
c     sum thermal diffusion from isotropic and anisotropic terms,
c     then complete all contributions to each basis function.
c     the identity term and the gradt_phi term are combined before the
c     basis function loop for efficiency.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        grad_tie(3,:,:,imode)=tie(1,:,:,imode)
     $    -(0,1)*keff(imode)*grad_tie(3,:,:,imode)/bigr
        DO iv=1,nv
          int(1,:,iv,imode)=SUM( dalpdr(:,:,iv)*grad_tie(1,:,:,imode)
     $                          +dalpdz(:,:,iv)*grad_tie(2,:,:,imode)
     $                          + alpha(:,:,iv)*grad_tie(3,:,:,imode),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE t_aniso_dot
c-----------------------------------------------------------------------
c     subprogram 11. vrhs.
c     form the integrand for the rhs of the velocity equation
c     at a single gauss quadrature point for all cells in a block.
c-----------------------------------------------------------------------
      SUBROUTINE vrhs(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: almod,dalmr,dalmz
      REAL(r8), DIMENSION(:,:,:), POINTER :: be_eq,ja_eq,nd_eq,
     $          ve_eq,veqten,nd_n0,real_ve,ti_eq,te_eq,
     $          real_gradv,real_ndptr,real_bptr,real_vten,
     $          pi_veq,pareq,gyreq,kappli,dp,beqr,beqz,be0,be0r,be0z,
     $          peq,p0,real_titot,ds2
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal
      REAL(r8), DIMENSION(3,mpseudo,nphi), TARGET :: real_ja,real_be,
     $          real_force
      REAL(r8), DIMENSION(9,mpseudo,nphi), TARGET :: real_pten
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: diff,dtmp
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: teff_eq
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: eq_vadv,bes,
     $          besr,besz
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8), DIMENSION(3,3) :: vtmpr,ptmpr,btmp,wdotr,bc_wdotr
      REAL(r8) :: divtmpr,dtmt,rscal,rb2,vcoef,gcoef,g,dtpv,dtdv
      REAL(r8) :: third=1._r8/3._r8,twot=2._r8/3._r8

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez,ve,ve_r,
     $             ve_z,nd,vheat,tion,tele,prsptr,dcp
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             pres,teff
      COMPLEX(r8), DIMENSION(2,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: aux
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: ja,
     $             force
      COMPLEX(r8), DIMENSION(9,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             vten,piten,parten
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      COMPLEX(r8), DIMENSION(3,3) :: vtmp,ptmp,wdot,b_cr_wdot,btmpc
      COMPLEX(r8) :: divtmp,cscal

      INTEGER(i4) :: iv,imode,iq,jt,ivind,nv,ncx,ncy,ix,iy,ip,im
      INTEGER(i4) :: i1,i2,nvc,nvm,nvhyp
      LOGICAL :: ve_set,nd_set,aux_set
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      ve_set=.false.
      nd_set=.false.
      aux_set=.false.
      dtmt=dt*mtot
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the perturbed
c     and 0-th order magnetic fields.  the current density is evaluated
c     from the gradients of the perturbed magnetic field
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,inode,
     $                     be_eq,beqr,beqz,1_i4)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,ber,bez,1_i4)
      IF (integrand_flag(1:3)=='all'.OR.integrand_flag(1:3)=='mhd') THEN
        CALL math_curl(nmodes,keff,geom,bigr,be,ber,bez,ja,1/mu0)
        CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                       inode,ja_eq,dp,dp,0_i4)
        IF (beta>0) THEN
          CALL generic_ptr_set(rb%qtion,tb%qtion,tb%tgeom,inode,
     $                         tion,dcp,dcp,0_i4)
          CALL generic_ptr_set(rb%qtele,tb%qtele,tb%tgeom,inode,
     $                         tele,dcp,dcp,0_i4)
          CALL generic_ptr_set(rb%qtion_eq,tb%qtion_eq,tb%tgeom,
     $                         inode,ti_eq,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qtele_eq,tb%qtele_eq,tb%tgeom,
     $                         inode,te_eq,dp,dp,0_i4)
          teff=dt*kboltz*(tele+tion/zeff)
          teff_eq=dt*kboltz*(te_eq+ti_eq/zeff)
          IF (p_computation=='at nodes') THEN
            CALL generic_ptr_set(rb%qpres,tb%qpres,tb%tgeom,inode,
     $                           prsptr,dcp,dcp,0_i4)
            pres=dt*prsptr
          ENDIF
        ELSE
          pres=0
        ENDIF
      ELSE
        pres=0
      ENDIF
      IF (eq_flow/='none') THEN
        CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qgrdveq,tb%qgrdveq,tb%tgeom,
     $                       inode,veqten,dp,dp,0_i4)
      ENDIF
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,
     $                     nd,dcp,dcp,0_i4)
      IF (continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                       nd_n0,dp,dp,0_i4)
      ENDIF
      IF (nonlinear.AND.continuity/='none') THEN
        CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     if poly_divv is non-negative, auxiliary fields are used to
c     stabilize the V-advance at the limit of element resolution.  the
c     modal bases for this field are then needed.
c-----------------------------------------------------------------------
      nvm=0
      IF (poly_divv>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',almod,dalmr,
     $                          dalmz,0_i4,poly_divv,
     $                          polydmin=poly_divv_min,
     $                          polydmax=poly_divv_max)
        nvm=SIZE(almod,3)
      ENDIF
c-----------------------------------------------------------------------
c     choose mhd forces, viscous & advective forces, or both according
c     to the time splitting.
c-----------------------------------------------------------------------
      IF (integrand_flag(1:3)=='mhd'.OR.integrand_flag(1:3)=='all') THEN
c-----------------------------------------------------------------------
c       find the nonlinear contributions to the force density.
c-----------------------------------------------------------------------
        IF (nonlinear) THEN
c-----------------------------------------------------------------------
c         first find b_perturbed and j_perturbed in real space.
c-----------------------------------------------------------------------
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,be,real_be,
     $                 dealiase)
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,real_ja,
     $                 dealiase)
c-----------------------------------------------------------------------
c         find j_perturbed X b_perturbed in real space,
c         then transform to Fourier space.
c-----------------------------------------------------------------------
          CALL math_cart_cross(real_force,real_ja,real_be,dt)
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,force,
     $                 real_force,dealiase)
c-----------------------------------------------------------------------
c         find pressure from the product of n and Teff=Te+Ti/zeff.
c         the real_ndptr storage has nd_eq already added.  the
c         real_scal array is used for temporary storage.
c-----------------------------------------------------------------------
          IF (beta>0.AND.p_computation/='at nodes') THEN
            CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,teff,
     $                   real_scal,dealiase)
            real_scal=real_scal*real_ndptr
            CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,1_i4,pres,
     $                   real_scal,dealiase)
          ENDIF
        ELSE
          force=0
        ENDIF
c-----------------------------------------------------------------------
c       construct JXB and P separately
c-----------------------------------------------------------------------
        mode_loop: DO imode=1,nmodes
          CALL math_cadd_cross(force(:,:,:,imode),
     $                         ja(:,:,:,imode),be_eq,dt)
          CALL math_cadd_cross(force(:,:,:,imode),
     $                         ja_eq,be(:,:,:,imode),dt)
          IF (beta>0.AND.p_computation/='at nodes') THEN
            IF (nonlinear) THEN
              pres(:,:,:,imode)=pres(:,:,:,imode)+
     $                            nd(:,:,:,imode)*teff_eq
            ELSE
              pres(:,:,:,imode)=teff(:,:,:,imode)*nd_eq+
     $                            nd(:,:,:,imode)*teff_eq
            ENDIF
          ENDIF
        ENDDO mode_loop
      ELSE
        force=0
      ENDIF
c-----------------------------------------------------------------------
c     explicit gravitational acceleration.
c-----------------------------------------------------------------------
      IF (MAXVAL(ABS(gravity))>0._r8) THEN
        force(1,:,:,:)=force(1,:,:,:)+nd(1,:,:,:)*dtmt*gravity(1)
        force(2,:,:,:)=force(2,:,:,:)+nd(1,:,:,:)*dtmt*gravity(2)
        force(3,:,:,:)=force(3,:,:,:)+nd(1,:,:,:)*dtmt*gravity(3)
      ENDIF
c-----------------------------------------------------------------------
c     the vten grad(V) tensor is in Fourier representation and is left
c     without equilibrium contributions.  real_vten is in real space
c     with equilibrium contributions.
c-----------------------------------------------------------------------
      IF ((advect=='V only'.OR.advect=='all') .AND.
     $    (nonlinear.OR.eq_flow/='none') .AND.
     $    (integrand_flag(1:3)=='all'.OR.integrand_flag(9:12)=='vdgv')
     $   .OR.poly_divv>=0) THEN
        IF (integrand_flag(14:16)=='pre'.OR.
     $      integrand_flag(12:14)=='pre'.OR.impladv) THEN
          CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,
     $                         inode,ve,ve_r,ve_z,1_i4)
          CALL math_grad(nmodes,keff,3_i4,geom,ve,ve_r,ve_z,vten,bigr)
        ELSE     !     data for corrector step of p/c advection
          CALL generic_ptr_set(rb%qwork1,tb%qwork1,tb%tgeom,
     $                         inode,ve,ve_r,ve_z,1_i4)
          CALL math_grad(nmodes,keff,3_i4,geom,ve,ve_r,ve_z,vten,bigr)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     subtract alpha_i*rho*V.grad(V) when momentum density advection
c     is requested using the appropriate velocity (either old or
c     predicted, stored in work1). the copies are done here to reuse
c     memory from the viscous term, and pointers are used just to
c     rename the configuration-space variables.
c
c     real_vten is in real space with equilibrium contributions.
c-----------------------------------------------------------------------
      IF ((advect=='V only'.OR.advect=='all') .AND.
     $    (nonlinear.OR.eq_flow/='none') .AND.
     $    (integrand_flag(1:3)=='all'.OR.integrand_flag(9:12)=='vdgv')
     $   ) THEN
        IF (integrand_flag(14:16)=='pre'.OR.
     $      integrand_flag(12:14)=='pre'.OR.impladv) THEN
          IF (nonlinear) THEN
            CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                           inode,real_vten,dp,dp,0_i4)
          ENDIF
          ve_set=.true.
        ELSE     !     data for corrector step of p/c advection
          IF (nonlinear) THEN
            IF (eq_flow/='none') THEN
              DO imode=1,nmodes
                IF (keff(imode)==0)
     $            vten(:,:,:,imode)=vten(:,:,:,imode)+veqten
              ENDDO
            ENDIF
            real_vten=>real_pten
            CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,9_i4,vten,
     $                   real_vten,dealiase)
            IF (eq_flow/='none') THEN
              DO imode=1,nmodes
                IF (keff(imode)==0)
     $            vten(:,:,:,imode)=vten(:,:,:,imode)-veqten
              ENDDO
            ENDIF
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       evolving density requires a V0.grad(V0) term.
c-----------------------------------------------------------------------
        IF (eq_flow/='none'.AND.continuity/='none') THEN
          eq_vadv(1,:,:)=dtmt*SUM(ve_eq*veqten(1:3,:,:),1)
          eq_vadv(2,:,:)=dtmt*SUM(ve_eq*veqten(4:6,:,:),1)
          eq_vadv(3,:,:)=dtmt*SUM(ve_eq*veqten(7:9,:,:),1)
        ENDIF
c-----------------------------------------------------------------------
c       putting the linear advection loop first helps reduce the number
c       of FFTs.
c
c       compute linear terms V_eq.grad(V) and V.grad(V_eq) for linear
c       cases or just the former for nonlinear cases, since grad(V_eq)
c       is part of real_vten.
c
c       ja is used as temporary space.
c-----------------------------------------------------------------------
        IF (eq_flow/='none') THEN
          DO imode=1,nmodes
            ja(1,:,:,imode)=SUM(ve_eq*vten(1:3,:,:,imode),1)
            ja(2,:,:,imode)=SUM(ve_eq*vten(4:6,:,:,imode),1)
            ja(3,:,:,imode)=SUM(ve_eq*vten(7:9,:,:,imode),1)
          ENDDO
          IF (nonlinear) THEN
            CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ja,
     $                   real_force,dealiase)
          ELSE
            DO imode=1,nmodes
              ja(1,:,:,imode)=dtmt*(ja(1,:,:,imode)
     $                   +SUM(ve(:,:,:,imode)*veqten(1:3,:,:),1))
              ja(2,:,:,imode)=dtmt*(ja(2,:,:,imode)
     $                   +SUM(ve(:,:,:,imode)*veqten(4:6,:,:),1))
              ja(3,:,:,imode)=dtmt*(ja(3,:,:,imode)
     $                   +SUM(ve(:,:,:,imode)*veqten(7:9,:,:),1))
            ENDDO
          ENDIF
        ELSE
          IF (nonlinear) real_force=0._r8
        ENDIF
c-----------------------------------------------------------------------
c       nonlinear terms:  determine V in configuration space, then
c       find V.grad(V+V_eq)
c-----------------------------------------------------------------------
        IF (nonlinear) THEN
          real_ve=>real_ja
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $                 dealiase)
          IF (continuity=='full') THEN
            DO ip=1,nphi
              DO ix=1,mpseudo
                rscal=dtmt*real_ndptr(1,ix,ip)
                real_force(1,ix,ip)=rscal*( real_force(1,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(1,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(2,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(3,ix,ip))
                real_force(2,ix,ip)=rscal*( real_force(2,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(4,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(5,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(6,ix,ip))
                real_force(3,ix,ip)=rscal*( real_force(3,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(7,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(8,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(9,ix,ip))
              ENDDO
            ENDDO
          ELSE
            DO ip=1,nphi
              DO ix=1,mpseudo
                real_force(1,ix,ip)=dtmt*( real_force(1,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(1,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(2,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(3,ix,ip))
                real_force(2,ix,ip)=dtmt*( real_force(2,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(4,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(5,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(6,ix,ip))
                real_force(3,ix,ip)=dtmt*( real_force(3,ix,ip)+
     $                              real_ve(1,ix,ip)*real_vten(7,ix,ip)+
     $                              real_ve(2,ix,ip)*real_vten(8,ix,ip)+
     $                              real_ve(3,ix,ip)*real_vten(9,ix,ip))
              ENDDO
            ENDDO
          ENDIF
          IF (.NOT.ve_set) NULLIFY(real_vten)
c-----------------------------------------------------------------------
c         transform to Fourier space (ja is used as a temporary array).
c-----------------------------------------------------------------------
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,ja,
     $                 real_force,dealiase)
        ENDIF
c-----------------------------------------------------------------------
c       multiply by mass density here for cases where it doesn't give
c       further Fourier-component coupling.
c-----------------------------------------------------------------------
        IF (.NOT.nonlinear.OR.continuity=='none'.OR.
     $      continuity=='fix profile') THEN
          IF (.NOT.nonlinear.AND.continuity/='none') THEN
            DO im=1,nmodes
              force(1,:,:,im)=force(1,:,:,im)-ja(1,:,:,im)*nd_eq(1,:,:)
     $                        -eq_vadv(1,:,:)*nd(1,:,:,im)
              force(2,:,:,im)=force(2,:,:,im)-ja(2,:,:,im)*nd_eq(1,:,:)
     $                        -eq_vadv(2,:,:)*nd(1,:,:,im)
              force(3,:,:,im)=force(3,:,:,im)-ja(3,:,:,im)*nd_eq(1,:,:)
     $                        -eq_vadv(3,:,:)*nd(1,:,:,im)
            ENDDO
          ELSE
            DO im=1,nmodes
              force(1,:,:,im)=force(1,:,:,im)-ja(1,:,:,im)*nd_eq(1,:,:)
              force(2,:,:,im)=force(2,:,:,im)-ja(2,:,:,im)*nd_eq(1,:,:)
              force(3,:,:,im)=force(3,:,:,im)-ja(3,:,:,im)*nd_eq(1,:,:)
            ENDDO
          ENDIF
        ELSE IF (continuity=='n=0 only') THEN
          DO im=1,nmodes
            force(1,:,:,im)=force(1,:,:,im)-
     $         ja(1,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
            force(2,:,:,im)=force(2,:,:,im)-
     $         ja(2,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
            force(3,:,:,im)=force(3,:,:,im)-
     $         ja(3,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
            IF (keff(im)==0.AND.eq_flow/='none') THEN
             force(1,:,:,im)=force(1,:,:,im)-eq_vadv(1,:,:)*nd_n0(1,:,:)
             force(2,:,:,im)=force(2,:,:,im)-eq_vadv(2,:,:)*nd_n0(1,:,:)
             force(3,:,:,im)=force(3,:,:,im)-eq_vadv(3,:,:)*nd_n0(1,:,:)
            ENDIF
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       complete the advective term with 3D mass density.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.continuity=='full') THEN
          IF (eq_flow/='none') THEN
            DO imode=1,nmodes
              force(1,:,:,imode)=force(1,:,:,imode)-ja(1,:,:,imode)
     $                              -eq_vadv(1,:,:)*nd(1,:,:,imode)
              force(2,:,:,imode)=force(2,:,:,imode)-ja(2,:,:,imode)
     $                              -eq_vadv(2,:,:)*nd(1,:,:,imode)
              force(3,:,:,imode)=force(3,:,:,imode)-ja(3,:,:,imode)
     $                              -eq_vadv(3,:,:)*nd(1,:,:,imode)
            ENDDO
          ELSE
            force=force-ja
          ENDIF
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     evaluations for viscous dissipation:
c-----------------------------------------------------------------------
      IF ((kin_visc>0.OR.iso_visc>0.OR.par_visc>0.OR.gyr_visc>0).AND.
     $ (integrand_flag(1:4)=='visc'.OR.integrand_flag(1:3)=='all')) THEN
        IF (.NOT.ve_set) THEN
          CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,
     $                         ve,ve_r,ve_z,1_i4)
          CALL math_grad(nmodes,keff,3_i4,geom,ve,ve_r,ve_z,vten,bigr)
        ENDIF
        IF (ds_use=='kin_visc'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          diff=dtmt*MAX(ds2(1,:,:),0._r8)
        ELSE
          diff=dtmt
        ENDIF
        IF (eq_flow/='none')
     $    CALL generic_ptr_set(rb%qpi_veq,tb%qpi_veq,tb%tgeom,
     $                         inode,pi_veq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c       the grad(V) tensor is in real space and includes any Veq terms.
c       find B+B_eq in real space.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.(par_visc>0.OR.gyr_visc>0)) THEN
          CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                         inode,real_bptr,dp,dp,0_i4)
          IF (.NOT.ve_set)
     $      CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                           inode,real_vten,dp,dp,0_i4)
        ELSE IF (nonlinear.AND.visc_heat.AND..NOT.ve_set) THEN
          CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                         inode,real_vten,dp,dp,0_i4)
        ENDIF
c-----------------------------------------------------------------------
c       the present strategy for the viscous dissipation terms in
c       nonlinear computations is to find the full stress, including
c       all equilibrium terms, then subtracting the purely equilibrium
c       part.  here, the steady-state is not considerably
c       larger than the perturbed contribution, so there should be
c       no loss of accuracy.
c  
c       kinematic and isotropic stresses are modulated by the shape
c       function diff.
c
c       add grad(V_eq) to the grad(V) tensor.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.eq_flow/='none') THEN
          DO imode=1,nmodes
            IF (keff(imode)==0)
     $        vten(:,:,:,imode)=vten(:,:,:,imode)+veqten
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       stress tensor for kinetematic and isotropic viscosity--now
c       in an external subroutine.
c-----------------------------------------------------------------------
        IF (iso_visc>0.OR.kin_visc>0) THEN
          CALL iso_stress(ncx,ncy,nmodes,kin_visc,iso_visc,
     $                    vten,diff,piten)
        ELSE
          piten=0._r8
        ENDIF
c-----------------------------------------------------------------------
c       transform piten from kinematic and isotropic contributions.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.(par_visc>0.OR.gyr_visc>0.OR.
     $                     continuity=='full')) THEN
          IF (iso_visc>0.OR.kin_visc>0) THEN
            CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,9_i4,
     $                   piten,real_pten,dealiase)
          ELSE
            real_pten=0._r8
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       stress tensor for parallel viscosity
c
c       the steady-state term is subtracted after forming the 
c       complete stress. 
c
c       the stored equilibrium stress does not include the factor of dt.
c-----------------------------------------------------------------------
        IF (par_visc>0) THEN
          vcoef=-3*dtmt*par_visc
          IF (eq_flow/='none')
     $      CALL generic_ptr_set(rb%qpi_pareq,tb%qpi_pareq,tb%tgeom,
     $                           inode,pareq,dp,dp,0_i4)
c-----------------------------------------------------------------------
c         nonlinear stress--now from the external subroutine
c         par_stress_nl.
c
c         find (B.grad(V).B - B**2*div(V))/B**4 and multiply the result
c         by (BB - I*B**2/3).
c
c         the Braginskii temperature dependence is the same as the
c         parallel ion thermal diffusivity, so use that data for
c         that coefficient if specified.
c-----------------------------------------------------------------------
          IF (nonlinear) THEN
            IF (parvisc_model=='plltdep') THEN
              CALL generic_ptr_set(rb%qkappli_phi,tb%qkappli_phi,
     $                             tb%tgeom,inode,kappli,dp,dp,0_i4)
              real_scal=vcoef*kappli/k_plli
            ELSE
              real_scal=vcoef
            ENDIF
            CALL par_stress_nl(mpseudo,nphi,smallnum,real_scal,
     $                         real_vten,real_bptr,real_pten)
c-----------------------------------------------------------------------
c         linear version from the external subroutine par_stress_lin.
c-----------------------------------------------------------------------
          ELSE
            CALL par_stress_lin(ncx,ncy,nmodes,vcoef,smallnum,
     $                          vten,be_eq,parten)
c-----------------------------------------------------------------------
c           V_eq contributions to the linear parallel stress from
c           the external subroutine par_stress_veq.
c-PRE
c           Braginskii parallel does not have density.
c-----------------------------------------------------------------------
            IF (eq_flow/='none') THEN
              CALL par_stress_veq(ncx,ncy,nmodes,vcoef,dt,smallnum,vten,
     $                            be,be_eq,veqten,pareq,dtmp,parten)
            ENDIF
            piten=piten+parten
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       Braginskii gyroviscosity is 0.5*eta_3*{bxW.(I+3bb)-(I+3bb).Wxb}
c       or
c
c       0.25*[n_e*m_i*Ti/(Z**2*B**4)]*{ [(BxW.(B**2+3BB)]+transp[] }
c
c       leave the number density for later.  note that the div(V) terms
c       cancel analytically and are left out of the W used here.
c
c       reuse the parten array to save a little memory; this
c       section is strictly gyroviscosity, however.
c
c       the stored equilibrium part does not include the factor of dt.
c-----------------------------------------------------------------------
        IF (gyr_visc>0) THEN
          vcoef=0.25*gyr_visc*dt*ms(2)*kboltz/(zeff**2*elementary_q)
          IF (eq_flow/='none') THEN
            CALL generic_ptr_set(rb%qpi_gyreq,tb%qpi_gyreq,tb%tgeom,
     $                           inode,gyreq,dp,dp,0_i4)
          ENDIF
c-----------------------------------------------------------------------
c         nonlinear gyroviscous stress including grad(V_eq)
c         contributions if there are any.
c-----------------------------------------------------------------------
          IF (nonlinear) THEN
            CALL generic_ptr_set(rb%qti_tot,tb%qti_tot,tb%tgeom,
     $                           inode,real_titot,dp,dp,0_i4)
            DO ip=1,nphi
              DO ix=1,mpseudo
                rb2=SUM(real_bptr(:,ix,ip)**2)+smallnum
                gcoef=vcoef*real_titot(1,ix,ip)/rb2**2
                vtmpr(1:3,1)=real_vten(1:3,ix,ip)+real_vten(1:7:3,ix,ip)
                vtmpr(1:3,2)=real_vten(4:6,ix,ip)+real_vten(2:8:3,ix,ip)
                vtmpr(1:3,3)=real_vten(7:9,ix,ip)+real_vten(3:9:3,ix,ip)
                btmp(1:3,1)=3._r8*real_bptr(1:3,ix,ip)*
     $                            real_bptr(1  ,ix,ip)
                btmp(1:3,2)=3._r8*real_bptr(1:3,ix,ip)*
     $                            real_bptr(2  ,ix,ip)
                btmp(1:3,3)=3._r8*real_bptr(1:3,ix,ip)*
     $                            real_bptr(3  ,ix,ip)
                btmp(1,1)=rb2+btmp(1,1)
                btmp(2,2)=rb2+btmp(2,2)
                btmp(3,3)=rb2+btmp(3,3)
                DO i2=1,3
                  wdotr(1,i2)=SUM(vtmpr(1,:)*btmp(:,i2))
                  wdotr(2,i2)=SUM(vtmpr(2,:)*btmp(:,i2))
                  wdotr(3,i2)=SUM(vtmpr(3,:)*btmp(:,i2))
                ENDDO
                bc_wdotr(1,:)=real_bptr(2,ix,ip)*wdotr(3,:)-
     $                        real_bptr(3,ix,ip)*wdotr(2,:)
                bc_wdotr(2,:)=real_bptr(3,ix,ip)*wdotr(1,:)-
     $                        real_bptr(1,ix,ip)*wdotr(3,:)
                bc_wdotr(3,:)=real_bptr(1,ix,ip)*wdotr(2,:)-
     $                        real_bptr(2,ix,ip)*wdotr(1,:)
                real_pten(1:3,ix,ip)=real_pten(1:3,ix,ip)+
     $                               gcoef*(bc_wdotr(:,1)+bc_wdotr(1,:))
                real_pten(4:6,ix,ip)=real_pten(4:6,ix,ip)+
     $                               gcoef*(bc_wdotr(:,2)+bc_wdotr(2,:))
                real_pten(7:9,ix,ip)=real_pten(7:9,ix,ip)+
     $                               gcoef*(bc_wdotr(:,3)+bc_wdotr(3,:))
              ENDDO
            ENDDO
c-----------------------------------------------------------------------
c         stress tensor for linear gyroviscosity.
c-----------------------------------------------------------------------
          ELSE
            IF (eq_flow=='none') THEN
              DO im=1,nmodes
                DO iy=1,ncy
                  DO ix=1,ncx
                    rb2=SUM(be_eq(:,ix,iy)**2)+smallnum
                    gcoef=vcoef*ti_eq(1,ix,iy)/rb2**2
                    vtmp(1:3,1)=vten(1:3,ix,iy,im)+vten(1:7:3,ix,iy,im)
                    vtmp(1:3,2)=vten(4:6,ix,iy,im)+vten(2:8:3,ix,iy,im)
                    vtmp(1:3,3)=vten(7:9,ix,iy,im)+vten(3:9:3,ix,iy,im)
                    btmp(1:3,1)=3._8*be_eq(1:3,ix,iy)*be_eq(1,ix,iy)
                    btmp(1:3,2)=3._8*be_eq(1:3,ix,iy)*be_eq(2,ix,iy)
                    btmp(1:3,3)=3._8*be_eq(1:3,ix,iy)*be_eq(3,ix,iy)
                    btmp(1,1)=rb2+btmp(1,1)
                    btmp(2,2)=rb2+btmp(2,2)
                    btmp(3,3)=rb2+btmp(3,3)
                    DO i2=1,3
                      wdot(1,i2)=SUM(vtmp(1,:)*btmp(:,i2))
                      wdot(2,i2)=SUM(vtmp(2,:)*btmp(:,i2))
                      wdot(3,i2)=SUM(vtmp(3,:)*btmp(:,i2))
                    ENDDO
                    b_cr_wdot(1,:)=be_eq(2,ix,iy)*wdot(3,:)-
     $                             be_eq(3,ix,iy)*wdot(2,:)
                    b_cr_wdot(2,:)=be_eq(3,ix,iy)*wdot(1,:)-
     $                             be_eq(1,ix,iy)*wdot(3,:)
                    b_cr_wdot(3,:)=be_eq(1,ix,iy)*wdot(2,:)-
     $                             be_eq(2,ix,iy)*wdot(1,:)
                    piten(1:3,ix,iy,im)=piten(1:3,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,1)+b_cr_wdot(1,:))
                    piten(4:6,ix,iy,im)=piten(4:6,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,2)+b_cr_wdot(2,:))
                    piten(7:9,ix,iy,im)=piten(7:9,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,3)+b_cr_wdot(3,:))
                  ENDDO
                ENDDO
              ENDDO
c-----------------------------------------------------------------------
c           extra linear gyroviscous terms for equilibrium flow--not
c           for those with a weak stomach.  like the related parallel
c           stress, there are three terms here: 1) b-tilde in the cross
c           product with Weq, 2) perturbation of B**2+3BB, and 3)
c           perturbation of the Ti/B**4 coefficient.
c-----------------------------------------------------------------------
            ELSE
              DO im=1,nmodes
                DO iy=1,ncy
                  DO ix=1,ncx
                    rb2=SUM(be_eq(:,ix,iy)**2)+smallnum
                    gcoef=vcoef*ti_eq(1,ix,iy)/rb2**2
                    vtmp(1:3,1)=vten(1:3,ix,iy,im)+vten(1:7:3,ix,iy,im)
                    vtmp(1:3,2)=vten(4:6,ix,iy,im)+vten(2:8:3,ix,iy,im)
                    vtmp(1:3,3)=vten(7:9,ix,iy,im)+vten(3:9:3,ix,iy,im)
                    vtmpr(1:3,1)=veqten(1:3,ix,iy)+veqten(1:7:3,ix,iy)
                    vtmpr(1:3,2)=veqten(4:6,ix,iy)+veqten(2:8:3,ix,iy)
                    vtmpr(1:3,3)=veqten(7:9,ix,iy)+veqten(3:9:3,ix,iy)

                    btmp(1:3,1)=3._8*be_eq(1:3,ix,iy)*be_eq(1,ix,iy)
                    btmp(1:3,2)=3._8*be_eq(1:3,ix,iy)*be_eq(2,ix,iy)
                    btmp(1:3,3)=3._8*be_eq(1:3,ix,iy)*be_eq(3,ix,iy)
                    btmp(1,1)=rb2+btmp(1,1)
                    btmp(2,2)=rb2+btmp(2,2)
                    btmp(3,3)=rb2+btmp(3,3)
                    divtmp=2._r8*SUM(be_eq(:,ix,iy)*be(:,ix,iy,im))
                    btmpc(1:3,1)=3._8*(be_eq(1:3,ix,iy)*be(1,ix,iy,im)+
     $                                 be(1:3,ix,iy,im)*be_eq(1,ix,iy))
                    btmpc(1:3,2)=3._8*(be_eq(1:3,ix,iy)*be(2,ix,iy,im)+
     $                                 be(1:3,ix,iy,im)*be_eq(2,ix,iy))
                    btmpc(1:3,3)=3._8*(be_eq(1:3,ix,iy)*be(3,ix,iy,im)+
     $                                 be(1:3,ix,iy,im)*be_eq(3,ix,iy))
                    btmpc(1,1)=divtmp+btmpc(1,1)
                    btmpc(2,2)=divtmp+btmpc(2,2)
                    btmpc(3,3)=divtmp+btmpc(3,3)

                    DO i2=1,3
                      wdot(1,i2)=SUM(vtmp (1,:)*btmp (:,i2)+
     $                               vtmpr(1,:)*btmpc(:,i2))
                      wdot(2,i2)=SUM(vtmp (2,:)*btmp (:,i2)+
     $                               vtmpr(2,:)*btmpc(:,i2))
                      wdot(3,i2)=SUM(vtmp (3,:)*btmp (:,i2)+
     $                               vtmpr(3,:)*btmpc(:,i2))
                      wdotr(1,i2)=SUM(vtmpr(1,:)*btmp(:,i2))
                      wdotr(2,i2)=SUM(vtmpr(2,:)*btmp(:,i2))
                      wdotr(3,i2)=SUM(vtmpr(3,:)*btmp(:,i2))
                    ENDDO
                    b_cr_wdot(1,:)=be_eq(2,ix,iy)*wdot (3,:)-
     $                             be_eq(3,ix,iy)*wdot (2,:)+
     $                             be(2,ix,iy,im)*wdotr(3,:)-
     $                             be(3,ix,iy,im)*wdotr(2,:)
                    b_cr_wdot(2,:)=be_eq(3,ix,iy)*wdot (1,:)-
     $                             be_eq(1,ix,iy)*wdot (3,:)+
     $                             be(3,ix,iy,im)*wdotr(1,:)-
     $                             be(1,ix,iy,im)*wdotr(3,:)
                    b_cr_wdot(3,:)=be_eq(1,ix,iy)*wdot (2,:)-
     $                             be_eq(2,ix,iy)*wdot (1,:)+
     $                             be(1,ix,iy,im)*wdotr(2,:)-
     $                             be(2,ix,iy,im)*wdotr(1,:)

                    cscal=dt*(tion(1,ix,iy,im)/ti_eq(1,ix,iy)-
     $                        2._r8*divtmp/rb2)
                    piten(1:3,ix,iy,im)=piten(1:3,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,1)+b_cr_wdot(1,:))+
     $                cscal*gyreq(1:3,ix,iy)
                    piten(4:6,ix,iy,im)=piten(4:6,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,2)+b_cr_wdot(2,:))+
     $                cscal*gyreq(4:6,ix,iy)
                    piten(7:9,ix,iy,im)=piten(7:9,ix,iy,im)+
     $                gcoef*(b_cr_wdot(:,3)+b_cr_wdot(3,:))+
     $                cscal*gyreq(7:9,ix,iy)
                  ENDDO
                ENDDO
              ENDDO
            ENDIF
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       now multiply by number density.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.continuity=='full') THEN
          DO ip=1,nphi
            DO ix=1,mpseudo
              real_pten(:,ix,ip)=real_pten(:,ix,ip)*real_ndptr(1,ix,ip)
            ENDDO
          ENDDO
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,9_i4,
     $                 piten,real_pten,dealiase)
        ELSE IF (nonlinear.AND.(par_visc>0.OR.gyr_visc>0)) THEN
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,9_i4,
     $                 piten,real_pten,dealiase)
        ENDIF

        IF (nonlinear.AND.continuity=='n=0 only') THEN
          DO imode=1,nmodes
            DO iy=1,ncy
              DO ix=1,ncx
                piten(:,ix,iy,imode)=piten(:,ix,iy,imode)*
     $                               (nd_eq(1,ix,iy)+nd_n0(1,ix,iy))
              ENDDO
            ENDDO
          ENDDO
        ELSE IF (.NOT.nonlinear.OR.continuity/='full') THEN
          DO imode=1,nmodes
            DO iy=1,ncy
              DO ix=1,ncx
                piten(:,ix,iy,imode)=piten(:,ix,iy,imode)*nd_eq(1,ix,iy)
              ENDDO
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       now eliminate the n_eq*Pi(V_eq) term from the nonlinear stress
c       and add n_tilde*Pi(V0) to the linear stress if needed.  the
c       stored equilibrium stress now includes nd_eq.
c-----------------------------------------------------------------------
        IF (eq_flow/='none') THEN
          IF (nonlinear) THEN
            DO imode=1,nmodes
              IF (keff(imode)/=0) CYCLE
              DO iy=1,ncy
                DO ix=1,ncx
                  piten(:,ix,iy,imode)=piten(:,ix,iy,imode)-
     $              dt*pi_veq(:,ix,iy)
                ENDDO
              ENDDO
              IF (par_visc>0) THEN
                DO iy=1,ncy
                  DO ix=1,ncx
                    piten(:,ix,iy,imode)=piten(:,ix,iy,imode)-
     $                dt*pareq(:,ix,iy)
                  ENDDO
                ENDDO
              ENDIF
              IF (gyr_visc>0) THEN
                DO iy=1,ncy
                  DO ix=1,ncx
                    piten(:,ix,iy,imode)=piten(:,ix,iy,imode)-
     $                dt*gyreq(:,ix,iy)
                  ENDDO
                ENDDO
              ENDIF
            ENDDO
          ELSE IF (continuity=='full') THEN
            DO imode=1,nmodes
              DO iy=1,ncy
                DO ix=1,ncx
                  piten(:,ix,iy,imode)=piten(:,ix,iy,imode)+
     $              dt*nd(1,ix,iy,imode)*pi_veq(:,ix,iy)
                ENDDO
              ENDDO
              IF (par_visc>0) THEN
                DO iy=1,ncy
                  DO ix=1,ncx
                    piten(:,ix,iy,imode)=piten(:,ix,iy,imode)+
     $                dt*nd(1,ix,iy,imode)*pareq(:,ix,iy)
                  ENDDO
                ENDDO
              ENDIF
              IF (gyr_visc>0) THEN
                DO iy=1,ncy
                  DO ix=1,ncx
                    piten(:,ix,iy,imode)=piten(:,ix,iy,imode)+
     $                dt*nd(1,ix,iy,imode)*gyreq(:,ix,iy)
                  ENDDO
                ENDDO
              ENDIF
            ENDDO
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       add total pressure to the diagonal of the stress tensor.
c-----------------------------------------------------------------------
        IF (beta>0.AND.(integrand_flag(1:3)=='mhd'.OR.
     $                  integrand_flag(1:3)=='all')) THEN
          DO imode=1,nmodes
            DO iy=1,ncy
              DO ix=1,ncx
                piten(1,ix,iy,imode)=piten(1,ix,iy,imode)+
     $                                pres(1,ix,iy,imode)
                piten(5,ix,iy,imode)=piten(5,ix,iy,imode)+
     $                                pres(1,ix,iy,imode)
                piten(9,ix,iy,imode)=piten(9,ix,iy,imode)+
     $                                pres(1,ix,iy,imode)
              ENDDO
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       add conjg(grad(alpha))^T : Pi(V)  transfer phi
c       derivatives to force before scattering contributions to the
c       nodes.
c-----------------------------------------------------------------------
        DO imode=1,nmodes
          IF (geom=='tor') THEN
            force(1,:,:,imode)=force(1,:,:,imode)
     $        +(                  piten(9,:,:,imode)
     $         -(0,1)*keff(imode)*piten(3,:,:,imode))/bigr
            force(2,:,:,imode)=force(2,:,:,imode)
     $         -(0,1)*keff(imode)*piten(6,:,:,imode) /bigr
            force(3,:,:,imode)=force(3,:,:,imode)
     $        -(                  piten(3,:,:,imode)
     $         +(0,1)*keff(imode)*piten(9,:,:,imode))/bigr
          ELSE
            force(1,:,:,imode)=force(1,:,:,imode)
     $         -(0,1)*keff(imode)*piten(3,:,:,imode)
            force(2,:,:,imode)=force(2,:,:,imode)
     $         -(0,1)*keff(imode)*piten(6,:,:,imode)
            force(3,:,:,imode)=force(3,:,:,imode)
     $         -(0,1)*keff(imode)*piten(9,:,:,imode)
          ENDIF
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(  alpha(:,:,iv)*force(1,:,:,imode)
     $                            +dalpdr(:,:,iv)*piten(1,:,:,imode)
     $                            +dalpdz(:,:,iv)*piten(2,:,:,imode),1)
            int(2,:,iv,imode)=SUM(  alpha(:,:,iv)*force(2,:,:,imode)
     $                            +dalpdr(:,:,iv)*piten(4,:,:,imode)
     $                            +dalpdz(:,:,iv)*piten(5,:,:,imode),1)
            int(3,:,iv,imode)=SUM(  alpha(:,:,iv)*force(3,:,:,imode)
     $                            +dalpdr(:,:,iv)*piten(7,:,:,imode)
     $                            +dalpdz(:,:,iv)*piten(8,:,:,imode),1)
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c     or just accumulate the force vertex contibutions.
c-----------------------------------------------------------------------
      ELSE
        DO imode=1,nmodes
          force(3,:,:,imode)=force(3,:,:,imode)
     $       -(0,1)*keff(imode)*pres(1,:,:,imode)/bigr
          DO iv=1,nvc
            int(1,:,iv,imode)=SUM(  alpha(:,:,iv)*force(1,:,:,imode)+
     $                            dalpdrc(:,:,iv)* pres(1,:,:,imode),1)
            int(2,:,iv,imode)=SUM(  alpha(:,:,iv)*force(2,:,:,imode)+
     $                             dalpdz(:,:,iv)* pres(1,:,:,imode),1)
            int(3,:,iv,imode)=SUM(alpha(:,:,iv)*force(3,:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     if auxiliary fields are used, find their contribution to the
c     velocity bases and the rhs of its own evolution equation.  the
c     equations for div(V) are expressed directly in weak form:
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ sqrt(ddivv*nu*dt*f)*auxv*div(A^*) ]
c       int[(ups^*)*auxv]-int[sqrt(ddivv*nu*dt*f)*(ups^*)*div(delta(V))]
c         = int[ sqrt(ddivv*nu*dt/f)*(ups^*)*div(V_old)]
c
c     where f is the implicit centering, nu is a numerical viscosity
c     coefficient, nu=dt*cma**2, cma is sqrt(rho) times the
c     magneto-acoustic speed, and ups is the scalar test field for the
c     auxiliary field auxv.  pres and teff_eq are used as temporary
c     storage.
c
c     the auxiliary field to stabilize vorticity (auxw) is expressed
c     directly in weak form as
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ dt*sqrt(dpvrt*f/mu0)*auxw*B.curl(A^*) ]
c       int[(lam^*)*auxw]
c        - dt*sqrt(dpvrt*f/mu0)*int[(lam^*)*B.curl(delta(V))]
c        = dt*sqrt(dpvrt/(f*mu0))*int[(lam^*)*B.curl(V_old)]
c
c     where A is the test vector for V, lam is the test function for
c     the discontinuous scalar field auxw, f is the centering
c     coefficient.  here, the effective viscosity is dt*v_Alven**2.
c
c     teff is used for temporary space, and ja becomes
c     dt*cpvrt*curl(V)/sqrt(mu0).
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        g=1
      ELSE
        g=0
      ENDIF
      IF (poly_divv>=0) THEN
        CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                       inode,peq,dp,dp,0_i4)
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                         inode,be0,be0r,be0z,1_i4)
          bes=be_eq+be0
          besr=beqr+be0r
          besz=beqz+be0z
          IF (beta>0) THEN
            CALL generic_ptr_set(rb%qpres_n0,tb%qpres_n0,tb%tgeom,
     $                           inode,p0,dp,dp,0_i4)
            teff_eq(1,:,:)=SQRT(gamma*(peq(1,:,:)+p0(1,:,:))+
     $                          SUM(bes**2,1)/mu0)
          ELSE
            teff_eq(1,:,:)=SQRT(SUM(bes**2,1)/mu0)
          ENDIF
        ELSE
          bes=be_eq
          besr=beqr
          besz=beqz
          IF (beta>0) THEN
            teff_eq(1,:,:)=SQRT(gamma*peq(1,:,:)+SUM(bes**2,1)/mu0)
          ELSE
            teff_eq(1,:,:)=SQRT(SUM(bes**2,1)/mu0)
          ENDIF
        ENDIF

        dtdv=SQRT(ddivv/fdivv)*dt        !  for diffusive correction
        dtpv=dt*SQRT(dpvrt/(mu0*fpvrt))  !  for diffusive correction
        CALL math_curl(nmodes,keff,geom,bigr,ve,ve_r,ve_z,ja,dtpv)

        DO imode=1,nmodes
          IF (geom=='tor') THEN
            pres(1,:,:,imode)=(ve_r(1,:,:,imode)+ve_z(2,:,:,imode)
     $                         +((0,1)*keff(imode)*ve(3,:,:,imode)
     $                         +ve(1,:,:,imode))/bigr)*teff_eq(1,:,:)*
     $                          dtdv
          ELSE
            pres(1,:,:,imode)=(ve_r(1,:,:,imode)+ve_z(2,:,:,imode)
     $                         +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr)
     $                        *teff_eq(1,:,:)*dtdv
          ENDIF
          teff(1,:,:,imode)=SUM(bes*ja(:,:,:,imode),1)

          DO iv=nvc+1,nv  !  auxiliary equation
            int(1,:,iv,imode)=SUM(almod(:,:,iv-nvc)*pres(1,:,:,imode),1)
            int(2,:,iv,imode)=
     $        SUM(almod(:,:,iv-nvc)*teff(1,:,:,imode),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE vrhs
c-----------------------------------------------------------------------
c     subprogram 12. advect_cor.
c     this integrand is just the correction needed on the rhs
c     for iterating advection in a nonlinear implicit advance of V.
c-----------------------------------------------------------------------
      SUBROUTINE advect_cor(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,nd_n0,
     $          real_gradv,real_ndptr,dp
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve,real_force
      REAL(r8), DIMENSION(9,mpseudo,nphi) :: real_vten
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: dtmt,rscal

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ve,ve_r,ve_z,nd,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) ::
     $             force
      COMPLEX(r8), DIMENSION(9,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: vten
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,imode,iq,nvc,ncx,ncy,ix,iy,ip,im
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      dtmt=0.25_r8*dt*mtot
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the perturbed
c     and 0-th order magnetic fields.  the current density is evaluated
c     from the gradients of the perturbed magnetic field
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,
     $                     nd,dcp,dcp,0_i4)
      IF (continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                       nd_n0,dp,dp,0_i4)
      ENDIF
      IF (continuity/='none') THEN
        CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     find 0.25*alpha_i*rho*dV.grad(dV) for nonlinear corrector steps
c     for implicit advection.  the dV vector is the difference between
c     the last nonlinear iteration and the beginning of the time-step,
c     and it is stored in work1.
c
c     the vten grad(V) tensor is in Fourier representation and is left
c     without equilibrium contributions.  real_vten is in real space
c     with equilibrium contributions.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qwork1,tb%qwork1,tb%tgeom,
     $                     inode,ve,ve_r,ve_z,1_i4)
      CALL math_grad(nmodes,keff,3_i4,geom,ve,ve_r,ve_z,vten,bigr)
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,9_i4,vten,
     $             real_vten,dealiase)
c-----------------------------------------------------------------------
c     determine dV in configuration space, then find dV.grad(dV).
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $             dealiase)
      IF (continuity=='full') THEN
        DO ip=1,nphi
          DO ix=1,mpseudo
            rscal=dtmt*real_ndptr(1,ix,ip)
            real_force(1,ix,ip)=rscal*(
     $                          real_ve(1,ix,ip)*real_vten(1,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(2,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(3,ix,ip))
            real_force(2,ix,ip)=rscal*(
     $                          real_ve(1,ix,ip)*real_vten(4,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(5,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(6,ix,ip))
            real_force(3,ix,ip)=rscal*(
     $                          real_ve(1,ix,ip)*real_vten(7,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(8,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(9,ix,ip))
          ENDDO
        ENDDO
      ELSE
        DO ip=1,nphi
          DO ix=1,mpseudo
            real_force(1,ix,ip)=dtmt*(
     $                          real_ve(1,ix,ip)*real_vten(1,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(2,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(3,ix,ip))
            real_force(2,ix,ip)=dtmt*(
     $                          real_ve(1,ix,ip)*real_vten(4,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(5,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(6,ix,ip))
            real_force(3,ix,ip)=dtmt*(
     $                          real_ve(1,ix,ip)*real_vten(7,ix,ip)+
     $                          real_ve(2,ix,ip)*real_vten(8,ix,ip)+
     $                          real_ve(3,ix,ip)*real_vten(9,ix,ip))
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     transform to Fourier space.
c-----------------------------------------------------------------------
      CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,force,
     $             real_force,dealiase)
c-----------------------------------------------------------------------
c     multiply by mass density here for cases where it doesn't
c     involve Fourier-component coupling.
c-----------------------------------------------------------------------
      IF (continuity=='none'.OR.continuity=='fix profile') THEN
        DO im=1,nmodes
          force(1,:,:,im)=force(1,:,:,im)*nd_eq(1,:,:)
          force(2,:,:,im)=force(2,:,:,im)*nd_eq(1,:,:)
          force(3,:,:,im)=force(3,:,:,im)*nd_eq(1,:,:)
        ENDDO
      ELSE IF (continuity=='n=0 only') THEN
        DO im=1,nmodes
          force(1,:,:,im)=force(1,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
          force(2,:,:,im)=force(2,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
          force(3,:,:,im)=force(3,:,:,im)*(nd_eq(1,:,:)+nd_n0(1,:,:))
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     accumulate the force vertex contibutions.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        DO iv=1,nvc
          int(1,:,iv,imode)=SUM(alpha(:,:,iv)*force(1,:,:,imode),1)
          int(2,:,iv,imode)=SUM(alpha(:,:,iv)*force(2,:,:,imode),1)
          int(3,:,iv,imode)=SUM(alpha(:,:,iv)*force(3,:,:,imode),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE advect_cor
c-----------------------------------------------------------------------
c     subprogram 13. v_aniso_dot.
c     find the action of the (mass density*I + viscous) operator dotted
c     into a vector (work4) for conjugate gradient iterations.  this is
c     used when mass density has 3d variations and continuity is
c     set to 'full,' when parallel viscosity is used in a nonlinear
c     problem, or for nonlinear implicit advection.  In these cases, the
c     matrix couples Fourier components.
c
c     the semi-implicit terms are also computed here for the first
c     time since nimrod3_1.  the cost of finding matrices has exceeded
c     that of the extra dot computations.
c-----------------------------------------------------------------------
      SUBROUTINE v_aniso_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: almod,dalmr,dalmz
      REAL(r8), DIMENSION(:,:,:), POINTER :: real_ndptr,real_bptr,
     $          real_oldv,real_oldgv,beq,beqr,beqz,b0,b0r,b0z,nd_eq,
     $          nd_n0,j0,nl_pres,peq,peqr,peqz,p0,p0r,p0z,ti_eq,kappli,
     $          dp,real_titot,ds2
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be,ber,bez,ja
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: pres,presr,
     $          presz
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: diff,n0,bigr2,
     $          iso,grd,bmagsq,sqptot
      REAL(r8), DIMENSION(3,3,SIZE(bigr,1),SIZE(bigr,2),ncontb) :: rcrl
      REAL(r8), DIMENSION(9,mpseudo,nphi) :: real_gvten
      REAL(r8), DIMENSION(6,mpseudo,nphi) :: real_pten
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve
      REAL(r8), DIMENSION(3,3) :: vtmpr,btmp,wdotr,bc_wdotr
      REAL(r8), DIMENSION(3) :: tmp1,tmp2

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: tion,dcp
      COMPLEX(r8), DIMENSION(2,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: aux
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: ve,
     $             ver,vez,force
      COMPLEX(r8), DIMENSION(9,SIZE(bigr,1),SIZE(bigr,2),nmodes) ::
     $             gvten
      COMPLEX(r8), DIMENSION(6,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: pten
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: dbe,
     $             ja_cross_dbe,ve_cross_ja,forck
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: divv,
     $             ve_gpr,sqpaux,bvort
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: baux
      COMPLEX(r8), DIMENSION(3,3) :: vtmp,crl
      COMPLEX(r8) :: div

      INTEGER(i4) :: ncx,ncy,iv,imode,jq,jt,ix,iy,ip,im,i2,nvc,nvm
      REAL(r8) :: kin_coef,iso_coef,par_coef,gcoef,vcoef,dloc
      REAL(r8) :: rb2,divtmpr,rscal,hdt,g,ani,cj0,dt2,rr,dtdv,dtpv
      REAL(r8), PARAMETER :: third=1._r8/3._r8,twoth=2._r8/3._r8
      LOGICAL :: use_gv
c-----------------------------------------------------------------------
c     convenience
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      hdt=0.5_r8*dt
      bigr2=bigr**2
      dt2=dt*dt
      IF ((kin_visc>0.OR.iso_visc>0.OR.par_visc>0.OR.gyr_visc>0).AND.
     $ (integrand_flag(1:4)=='visc'.OR.integrand_flag(1:3)=='all')) THEN
        use_gv=.true.
      ELSE
        use_gv=.false.
      ENDIF
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the
c     diffusivity shape function.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
      IF ((ds_use=='kin_visc'.OR.ds_use=='both').AND.
     $    (kin_visc>0.OR.iso_visc>0)) THEN
        CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                       inode,ds2,dp,dp,0_i4)
        diff=MAX(ds2(1,:,:),0._r8)
      ELSE
        diff=1
      ENDIF
      IF (par_visc>0.OR.gyr_visc>0) THEN
        CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                       inode,real_bptr,dp,dp,0_i4)
        par_coef=3._r8*dt*fvsc*mtot*par_visc
      ENDIF
      IF (gyr_visc>0) THEN
        CALL generic_ptr_set(rb%qtion,tb%qtion,tb%tgeom,
     $                       inode,tion,dcp,dcp,0_i4)
        CALL generic_ptr_set(rb%qtion_eq,tb%qtion_eq,tb%tgeom,
     $                       inode,ti_eq,dp,dp,0_i4)
      ENDIF
      kin_coef=dt*fvsc*mtot*kin_visc
      iso_coef=dt*fvsc*mtot*iso_visc
c-----------------------------------------------------------------------
c     evaluate the current direction vector for velocity change and
c     the number density.
c-----------------------------------------------------------------------
      CALL generic_all_eval(rb%work4,tb%work4,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      ve,ver,vez,1_i4)

      IF (continuity=='full') THEN
        CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)
      ELSE
        CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                       inode,nd_eq,dp,dp,0_i4)
      ENDIF
      IF (continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,dp,dp,0_i4)
        n0=nd_eq(1,:,:)+nd_n0(1,:,:)
      ELSE IF (continuity/='full') THEN
        n0=nd_eq(1,:,:)
      ENDIF
      IF (impladv) THEN
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,real_oldv,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                       inode,real_oldgv,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     find rho*dve product for the d/dt term.
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $             dealiase)
      IF (continuity=='full') THEN
        DO jq=1,3
          real_ve(jq,:,:)=real_ve(jq,:,:)*real_ndptr(1,:,:)*mtot
        ENDDO
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,force,real_ve,
     $               dealiase)
      ELSE 
        DO imode=1,nmodes
          DO jq=1,3
            force(jq,:,:,imode)=ve(jq,:,:,imode)*n0*mtot
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     evaluations for viscous dissipation:
c     first find and transform the nd*grad(V) tensor.
c-----------------------------------------------------------------------
      IF ((kin_visc>0.OR.iso_visc>0.OR.par_visc>0.OR.impladv).AND.
     $ (integrand_flag(1:4)=='visc'.OR.integrand_flag(1:3)=='all')) THEN

        CALL math_grad(nmodes,keff,3_i4,geom,ve,ver,vez,gvten,bigr)
        IF (continuity/='full') THEN
          DO imode=1,nmodes
            gvten(1,:,:,imode)=n0*gvten(1,:,:,imode)
            gvten(2,:,:,imode)=n0*gvten(2,:,:,imode)
            gvten(3,:,:,imode)=n0*gvten(3,:,:,imode)
            gvten(4,:,:,imode)=n0*gvten(4,:,:,imode)
            gvten(5,:,:,imode)=n0*gvten(5,:,:,imode)
            gvten(6,:,:,imode)=n0*gvten(6,:,:,imode)
            gvten(7,:,:,imode)=n0*gvten(7,:,:,imode)
            gvten(8,:,:,imode)=n0*gvten(8,:,:,imode)
            gvten(9,:,:,imode)=n0*gvten(9,:,:,imode)
          ENDDO
        ENDIF
        IF (continuity=='full'.OR.par_visc>0.OR.gyr_visc>0.OR.
     $      impladv) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,9_i4,
     $                 gvten,real_gvten,dealiase)
        ENDIF
        IF (continuity=='full') THEN
          DO jt=1,9
            real_gvten(jt,:,:)=real_gvten(jt,:,:)*real_ndptr(1,:,:)
          ENDDO
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,9_i4,
     $                 gvten,real_gvten,dealiase)
        ENDIF
c-----------------------------------------------------------------------
c       for nonlinear parallel viscosity, find the unique components of
c
c         [ B.grad(V).B - B**2*div(V) ] * (BB - I*B**2/3)
c
c       also divide by B**4 to make unit vectors.
c-----------------------------------------------------------------------
        IF (par_visc>0) THEN
          DO ip=1,nphi
            DO ix=1,mpseudo
              rb2=SUM(real_bptr(:,ix,ip)*real_bptr(:,ix,ip))+smallnum
              divtmpr=-third*rb2*(real_gvten(1,ix,ip)+
     $                real_gvten(5,ix,ip)+real_gvten(9,ix,ip))
              rscal=par_coef*(divtmpr+real_bptr(1,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(1,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(2,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(3,ix,ip))+
     $              real_bptr(2,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(4,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(5,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(6,ix,ip))+
     $              real_bptr(3,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(7,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(8,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(9,ix,ip)))/rb2**2
              rb2=-third*rb2
              real_pten(1,ix,ip)=rscal*
     $          (real_bptr(1,ix,ip)*real_bptr(1,ix,ip)+rb2)
              real_pten(2,ix,ip)=rscal*
     $           real_bptr(1,ix,ip)*real_bptr(2,ix,ip)
              real_pten(3,ix,ip)=rscal*
     $           real_bptr(1,ix,ip)*real_bptr(3,ix,ip)
              real_pten(4,ix,ip)=rscal*
     $          (real_bptr(2,ix,ip)*real_bptr(2,ix,ip)+rb2)
              real_pten(5,ix,ip)=rscal*
     $           real_bptr(2,ix,ip)*real_bptr(3,ix,ip)
              real_pten(6,ix,ip)=rscal*
     $          (real_bptr(3,ix,ip)*real_bptr(3,ix,ip)+rb2)
            ENDDO
          ENDDO
c-----------------------------------------------------------------------
c         multiply by the temperature-dependent profile for Braginskii
c         parallel viscosity if desired.
c-----------------------------------------------------------------------
          IF (parvisc_model=='plltdep') THEN
            CALL generic_ptr_set(rb%qkappli_phi,tb%qkappli_phi,
     $                           tb%tgeom,inode,kappli,dp,dp,0_i4)
            DO ip=1,nphi
              DO ix=1,mpseudo
                rscal=kappli(1,ix,ip)/k_plli
                real_pten(1,ix,ip)=real_pten(1,ix,ip)*rscal
                real_pten(2,ix,ip)=real_pten(2,ix,ip)*rscal
                real_pten(3,ix,ip)=real_pten(3,ix,ip)*rscal
                real_pten(4,ix,ip)=real_pten(4,ix,ip)*rscal
                real_pten(5,ix,ip)=real_pten(5,ix,ip)*rscal
                real_pten(6,ix,ip)=real_pten(6,ix,ip)*rscal
              ENDDO
            ENDDO
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       for nonlinear gyroviscosity, find the unique components of
c
c         (n_e*m_i*T_i/Z**2*B**4)*{ [BxW.(B**2+3BB) ] + transpose[] }
c
c-----------------------------------------------------------------------
        IF (gyr_visc>0) THEN
          vcoef=-0.125*gyr_visc*dt*ms(2)*kboltz/(zeff**2*elementary_q)
          CALL generic_ptr_set(rb%qti_tot,tb%qti_tot,tb%tgeom,
     $                         inode,real_titot,dp,dp,0_i4)
          IF (par_visc<=0) real_pten=0._r8
          DO ip=1,nphi
            DO ix=1,mpseudo
              rb2=SUM(real_bptr(:,ix,ip)**2)+smallnum
              gcoef=vcoef*real_titot(1,ix,ip)/rb2**2
              vtmpr(1:3,1)=real_gvten(1:3,ix,ip)+real_gvten(1:7:3,ix,ip)
              vtmpr(1:3,2)=real_gvten(4:6,ix,ip)+real_gvten(2:8:3,ix,ip)
              vtmpr(1:3,3)=real_gvten(7:9,ix,ip)+real_gvten(3:9:3,ix,ip)
              btmp(1:3,1)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(1  ,ix,ip)
              btmp(1:3,2)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(2  ,ix,ip)
              btmp(1:3,3)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(3  ,ix,ip)
              btmp(1,1)=rb2+btmp(1,1)
              btmp(2,2)=rb2+btmp(2,2)
              btmp(3,3)=rb2+btmp(3,3)
              DO i2=1,3
                wdotr(1,i2)=SUM(vtmpr(1,:)*btmp(:,i2))
                wdotr(2,i2)=SUM(vtmpr(2,:)*btmp(:,i2))
                wdotr(3,i2)=SUM(vtmpr(3,:)*btmp(:,i2))
              ENDDO
              bc_wdotr(1,:)=real_bptr(2,ix,ip)*wdotr(3,:)-
     $                      real_bptr(3,ix,ip)*wdotr(2,:)
              bc_wdotr(2,:)=real_bptr(3,ix,ip)*wdotr(1,:)-
     $                      real_bptr(1,ix,ip)*wdotr(3,:)
              bc_wdotr(3,:)=real_bptr(1,ix,ip)*wdotr(2,:)-
     $                      real_bptr(2,ix,ip)*wdotr(1,:)
              real_pten(1:3,ix,ip)=real_pten(1:3,ix,ip)+
     $                           gcoef*(bc_wdotr(:,1)+bc_wdotr(1,:))
              real_pten(4  ,ix,ip)=real_pten(4  ,ix,ip)+
     $                           gcoef*(bc_wdotr(2,2)+bc_wdotr(2,2))
              real_pten(5:6,ix,ip)=real_pten(5:6,ix,ip)+
     $                           gcoef*(bc_wdotr(3,2:3)+bc_wdotr(2:3,3))
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       transform the anisotropic stresses.
c-----------------------------------------------------------------------
        IF (par_visc>0.OR.gyr_visc>0) THEN
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,6_i4,
     $                 pten,real_pten,dealiase)
        ENDIF
c-----------------------------------------------------------------------
c       complete -Pi(rho,V) and save it in gvten.
c-----------------------------------------------------------------------
        DO im=1,nmodes
          DO iy=1,ncy
            DO ix=1,ncx
              dloc=diff(ix,iy)
              vtmp(1:3,1)=gvten(1:3,ix,iy,im)
              vtmp(1:3,2)=gvten(4:6,ix,iy,im)
              vtmp(1:3,3)=gvten(7:9,ix,iy,im)
              div=-twoth*(vtmp(1,1)+vtmp(2,2)+vtmp(3,3))
              gvten(1,ix,iy,im)=dloc*((kin_coef+2*iso_coef)*vtmp(1,1)+
     $                          iso_coef*div)
              gvten(2,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(2,1)+
     $                          iso_coef*vtmp(1,2))
              gvten(3,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(3,1)+
     $                          iso_coef*vtmp(1,3))
              gvten(4,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(1,2)+
     $                          iso_coef*vtmp(2,1))
              gvten(5,ix,iy,im)=dloc*((kin_coef+2*iso_coef)*vtmp(2,2)+
     $                          iso_coef*div)
              gvten(6,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(3,2)+
     $                          iso_coef*vtmp(2,3))
              gvten(7,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(1,3)+
     $                          iso_coef*vtmp(3,1))
              gvten(8,ix,iy,im)=dloc*((kin_coef+  iso_coef)*vtmp(2,3)+
     $                          iso_coef*vtmp(3,2))
              gvten(9,ix,iy,im)=dloc*((kin_coef+2*iso_coef)*vtmp(3,3)+
     $                          iso_coef*div)
            ENDDO
          ENDDO
        ENDDO
        IF (par_visc>0.OR.gyr_visc>0) THEN
          gvten(1:3,:,:,:)=gvten(1:3,:,:,:)+pten(1:3,:,:,:)
          gvten(4,:,:,:)=gvten(4,:,:,:)+pten(2,:,:,:)
          gvten(5,:,:,:)=gvten(5,:,:,:)+pten(4,:,:,:)
          gvten(6,:,:,:)=gvten(6,:,:,:)+pten(5,:,:,:)
          gvten(7,:,:,:)=gvten(7,:,:,:)+pten(3,:,:,:)
          gvten(8,:,:,:)=gvten(8,:,:,:)+pten(5,:,:,:)
          gvten(9,:,:,:)=gvten(9,:,:,:)+pten(6,:,:,:)
        ENDIF
c-----------------------------------------------------------------------
c       terms for implicit advection:
c
c          0.5*dt*rho*( dV.grad(V_old) + V_old.grad(dV) )
c
c       where dV is the iterate from work4.
c
c       note that V_old already includes any equilibrium flow.  Also,
c       real_gvten still holds n*grad(dV) in real space, force still 
c       holds mtot*n*dV by Fourier component, and real_ve holds 
c       mtot*n*dV in real space if continuity is full.
c-----------------------------------------------------------------------
        IF (impladv.AND.(advect=='V only'.OR.advect=='all')) THEN
          IF (.NOT.continuity=='full') CALL fft_nim('inverse',ncx*ncy,
     $      mpseudo,lphi,3_i4,force,real_ve,dealiase)
          DO ip=1,nphi
            DO ix=1,mpseudo
               tmp1=hdt*real_ve(:,ix,ip)
               tmp2=hdt*mtot*real_oldv(:,ix,ip)
               real_ve(1,ix,ip)=real_ve(1,ix,ip)+
     $                   tmp1(1)*real_oldgv(1,ix,ip)+
     $                   tmp1(2)*real_oldgv(2,ix,ip)+
     $                   tmp1(3)*real_oldgv(3,ix,ip)+
     $                   tmp2(1)*real_gvten(1,ix,ip)+
     $                   tmp2(2)*real_gvten(2,ix,ip)+
     $                   tmp2(3)*real_gvten(3,ix,ip)
               real_ve(2,ix,ip)=real_ve(2,ix,ip)+
     $                   tmp1(1)*real_oldgv(4,ix,ip)+
     $                   tmp1(2)*real_oldgv(5,ix,ip)+
     $                   tmp1(3)*real_oldgv(6,ix,ip)+
     $                   tmp2(1)*real_gvten(4,ix,ip)+
     $                   tmp2(2)*real_gvten(5,ix,ip)+
     $                   tmp2(3)*real_gvten(6,ix,ip)
               real_ve(3,ix,ip)=real_ve(3,ix,ip)+
     $                   tmp1(1)*real_oldgv(7,ix,ip)+
     $                   tmp1(2)*real_oldgv(8,ix,ip)+
     $                   tmp1(3)*real_oldgv(9,ix,ip)+
     $                   tmp2(1)*real_gvten(7,ix,ip)+
     $                   tmp2(2)*real_gvten(8,ix,ip)+
     $                   tmp2(3)*real_gvten(9,ix,ip)
            ENDDO
          ENDDO
          CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,
     $                 force,real_ve,dealiase)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     evaluate fields for the semi-implicit operator.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,beq,beqr,beqz,1_i4)
      CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                     inode,j0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,inode,
     $                     b0,b0r,b0z,1_i4)
      ja(1,:,:)=j0(1,:,:)+ b0z(3,:,:)/mu0
      ja(2,:,:)=j0(2,:,:)- b0r(3,:,:)/mu0
      ja(3,:,:)=j0(3,:,:)+(b0r(2,:,:)-b0z(1,:,:))/mu0
      IF (geom=='tor') ja(2,:,:)=ja(2,:,:)-b0(3,:,:)/(mu0*bigr)
      be=b0+beq
      ber=b0r+beqr
      bez=b0z+beqz
      CALL generic_ptr_set(rb%qsi_nl_pres,tb%qsi_nl_pres,tb%tgeom,
     $                     inode,nl_pres,dp,dp,0_i4)
      bmagsq=SUM(be*be,1)
      IF (beta>0) THEN
        CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                       inode,peq,peqr,peqz,1_i4)
        CALL generic_ptr_set(rb%qpres_n0,tb%qpres_n0,tb%tgeom,
     $                       inode,p0,p0r,p0z,1_i4)
        pres=p0+peq
        presr=p0r+peqr
        presz=p0z+peqz
      ELSE
        pres=0._r8
        presr=0._r8
        presz=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     find the spatial functions used as coefficients for the isotropic
c     and the anisotropic parts of the  operator.
c-----------------------------------------------------------------------
      ani=0.25_r8*si_fac_mhd* (1._r8-mhd_si_iso)*dt2/mu0
      cj0=0.125_r8*si_fac_j0*dt2
      IF (beta>0) THEN
        iso=0.25_r8*dt2*( mhd_si_iso*
     $    (si_fac_mhd*bmagsq/mu0+si_fac_pres*gamma*pres(1,:,:))+
     $      si_fac_nl*(nl_pres(1,:,:)/mu0+gamma*nl_pres(2,:,:)) )
        grd=0.25_r8*si_fac_pres*(1._r8-mhd_si_iso)*dt2*
     $              gamma*pres(1,:,:)
      ELSE
        iso=0.25_r8*dt2*( mhd_si_iso*si_fac_mhd*bmagsq+
     $                    si_fac_nl*nl_pres(1,:,:) )/mu0
        grd=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     the anisotropic operator has a term of the form
c     curl(alpha_iXB0).curl(alpha_jXB0) after the integration by parts,
c     and each curl is broken into B0.grad(alpha)-alpha.grad(B0)
c     -B0(div(alpha)).  B0 is in cylindrical components.
c     the first two indices of the temporary crl matrix are defined as
c     ( curl vector index, operand vector index ), where the vector
c     indices runs from 1 to 3 as (r,z,phi), and the last index is
c     the basis function index.
c
c     dbe is the ideal change in b, curl(veXB0).
c
c     this part of the operation is diagonal in Fourier component.
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        g=1
      ELSE
        g=0
      ENDIF
c-----------------------------------------------------------------------
c     this part of crl is k-independent.
c-PRE and could be saved for each step.
c-----------------------------------------------------------------------
      DO iv=1,nvc
        rcrl(1,1,:,:,iv)=
     $    dalpdz(:,:,iv)*  be(2,:,:)
     $    -alpha(:,:,iv)*(be(1,:,:)*g/bigr+ber(1,:,:))
        rcrl(1,2,:,:,iv)=
     $    -(dalpdz(:,:,iv)*be(1,:,:)+alpha(:,:,iv)*bez(1,:,:))
        rcrl(1,3,:,:,iv)=0
        rcrl(2,1,:,:,iv)=
     $    -(dalpdr(:,:,iv)* be(2,:,:)
     $      +alpha(:,:,iv)*(be(2,:,:)*g/bigr+ber(2,:,:)))
        rcrl(2,2,:,:,iv)=
     $    dalpdr(:,:,iv)*be(1,:,:)-alpha(:,:,iv)*bez(2,:,:)
        rcrl(2,3,:,:,iv)=0
        rcrl(3,1,:,:,iv)=
     $    -(dalpdr(:,:,iv)*be(3,:,:)+alpha(:,:,iv)*ber(3,:,:))
        rcrl(3,2,:,:,iv)=
     $    -(dalpdz(:,:,iv)*be(3,:,:)+alpha(:,:,iv)*bez(3,:,:))
        rcrl(3,3,:,:,iv)=
     $    dalpdr(:,:,iv)*be(1,:,:)+dalpdz(:,:,iv)*be(2,:,:)
     $    -alpha(:,:,iv)*be(1,:,:)*g/bigr
      ENDDO
c-----------------------------------------------------------------------
c     if auxiliary fields are used, find their contribution to the
c     velocity bases and the lhs of its own evolution equation.  the
c     equations for div(V) are expressed directly in weak form:
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ sqrt(ddivv*nu*dt*f)*auxv*div(A^*) ]
c       int[(ups^*)*auxv]-int[sqrt(ddivv*nu*dt*f)*(ups^*)*div(delta(V))]
c         = int[ sqrt(ddivv*nu*dt/f)*(ups^*)*div(V_old)]
c
c     where f is the implicit centering, nu is a numerical viscosity
c     coefficient, nu=dt*cma**2, cma is sqrt(rho) times the
c     magneto-acoustic speed, and ups is the scalar test field for the
c     auxiliary field auxv.  note that the real pres array becomes
c     sqrt(gamma*P+B^2/mu0) and that mwork4 holds the div(delta-V) and
c     B.curl(delta-V) projections.
c
c     the auxiliary field to stabilize vorticity (auxw) is expressed
c     directly in weak form as
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ dt*sqrt(dpvrt*f/mu0)*auxw*B.curl(A^*) ]
c       int[(lam^*)*auxw]
c        - dt*sqrt(dpvrt*f/mu0)*int[(lam^*)*B.curl(delta(V))]
c        = dt*sqrt(dpvrt/(f*mu0))*int[(lam^*)*B.curl(V_old)]
c
c     where A is the test vector for V, lam is the test function for
c     the discontinuous scalar field auxw, where f is the centering
c     coefficient.  here, the effective viscosity is dt*v_Alven**2.
c
c     divv is again used for temporary space, and force becomes
c     dt*cpvrt*curl(V)/sqrt(mu0).
c-----------------------------------------------------------------------
      IF (poly_divv>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',almod,dalmr,
     $                          dalmz,0_i4,poly_divv,
     $                          polydmin=poly_divv_min,
     $                          polydmax=poly_divv_max)
        nvm=SIZE(almod,3)
        CALL generic_all_eval(rb%mwork4,tb%mwork4,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,aux,dcp,dcp,0_i4)

        IF (beta>0) THEN
          sqptot=SQRT(gamma*pres(1,:,:)+SUM(be**2,1)/mu0)
        ELSE
          sqptot=SQRT(SUM(be**2,1)/mu0)
        ENDIF
        dtdv=SQRT(ddivv*fdivv)*dt        !  for diffusive correction
        dtpv=dt*SQRT(dpvrt*fpvrt/(mu0))  !  for diffusive correction
      ELSE
        dtdv=0._r8
        dtpv=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     start mode loop and find dbe.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
        dbe(1,:,:)=
     $      vez(1,:,:,imode)*  be(2,:,:)
     $      -ve(1,:,:,imode)*((be(1,:,:)*g-(0,1)*keff(imode)*be(3,:,:))
     $                      /bigr+ber(1,:,:))
     $    -(vez(2,:,:,imode)*be(1,:,:)+ve(2,:,:,imode)*bez(1,:,:))
     $      -ve(3,:,:,imode)*(0,1)*keff(imode)*be(1,:,:)/bigr
        dbe(2,:,:)=
     $    -(ver(1,:,:,imode)* be(2,:,:)
     $      +ve(1,:,:,imode)*(be(2,:,:)*g/bigr+ber(2,:,:)))
     $    + ver(2,:,:,imode)*be(1,:,:)
     $    +  ve(2,:,:,imode)*((0,1)*keff(imode)*be(3,:,:)/bigr
     $                       -bez(2,:,:))
     $    -  ve(3,:,:,imode)*(0,1)*keff(imode)*be(2,:,:)/bigr
        dbe(3,:,:)=
     $    -(ver(1,:,:,imode)*be(3,:,:)+ ve(1,:,:,imode)*ber(3,:,:)
     $     +vez(2,:,:,imode)*be(3,:,:)+ ve(2,:,:,imode)*bez(3,:,:))
     $    + ver(3,:,:,imode)*be(1,:,:)+vez(3,:,:,imode)*be(2,:,:)
     $    -  ve(3,:,:,imode)*be(1,:,:)*g/bigr
        CALL math_cart_cross(ja_cross_dbe,ja,dbe,1._r8)
        CALL math_cart_cross(ve_cross_ja,ve(:,:,:,imode),ja,1._r8)
c-----------------------------------------------------------------------
c       isotropic terms that multiply alpha only:
c-----------------------------------------------------------------------
        IF (geom=='tor') THEN
          forck(1,:,:)=force(1,:,:,imode)+
     $      iso*((1+k2ef(imode))*ve(1,:,:,imode)+
     $        2*(0,1)*keff(imode)*ve(3,:,:,imode))/bigr2
          forck(2,:,:)=force(2,:,:,imode)+
     $      iso*k2ef(imode)*ve(2,:,:,imode)/bigr2
          forck(3,:,:)=force(3,:,:,imode)+
     $      iso*((1+k2ef(imode))*ve(3,:,:,imode)-
     $        2*(0,1)*keff(imode)*ve(1,:,:,imode))/bigr2
          divv=ver(1,:,:,imode)+ve(1,:,:,imode)/bigr+vez(2,:,:,imode)
     $        +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr
        ELSE
          DO jq=1,3
            forck(jq,:,:)=force(jq,:,:,imode)+
     $        iso*k2ef(imode)*ve(jq,:,:,imode)/bigr2
          ENDDO
          divv=ver(1,:,:,imode)+vez(2,:,:,imode)
     $        +(0,1)*keff(imode)*ve(3,:,:,imode)/bigr
        ENDIF
        IF (use_gv) THEN
          IF (geom=='tor') THEN
            forck(1,:,:)=forck(1,:,:)
     $        +(                  gvten(9,:,:,imode)
     $         -(0,1)*keff(imode)*gvten(3,:,:,imode))/bigr
            forck(2,:,:)=forck(2,:,:)
     $         -(0,1)*keff(imode)*gvten(6,:,:,imode) /bigr
            forck(3,:,:)=forck(3,:,:)
     $        -(                  gvten(3,:,:,imode)
     $         +(0,1)*keff(imode)*gvten(9,:,:,imode))/bigr
          ELSE
            forck(1,:,:)=forck(1,:,:)
     $         -(0,1)*keff(imode)*gvten(3,:,:,imode)/bigr
            forck(2,:,:)=forck(2,:,:)
     $         -(0,1)*keff(imode)*gvten(6,:,:,imode)/bigr
            forck(3,:,:)=forck(3,:,:)
     $         -(0,1)*keff(imode)*gvten(9,:,:,imode)/bigr
          ENDIF
        ENDIF
c-----------------------------------------------------------------------
c       compute isotropic, anisotropic, and pressure contributions
c       together for effeciency.
c
c       the "ani" contributions are the dot product of the curls
c       from the different vertices around each cell and the
c
c       symmetrized contributions from the J0Xcurl(vxB0) term are
c       the "cj0" terms without pressure gradient coefficients.
c
c       grad(P0) terms include the div(alpha)*div() part of the operator
c       and the symmetrized div(alpha)*alpha.grad(P0) part.
c
c       the mass density term is with the sink term.
c-----------------------------------------------------------------------
        IF (beta>0) THEN
          ve_gpr=ve(1,:,:,imode)*presr(1,:,:)
     $          +ve(2,:,:,imode)*presz(1,:,:)
        ELSE
          ve_gpr=0._r8
        ENDIF
c-----------------------------------------------------------------------
c       find the contributions of P-total times div(dV) and B.curl(dV)
c       from the auxiliary discontinuous fields.
c-----------------------------------------------------------------------
        IF (poly_divv>0) THEN
          sqpaux=dtdv*sqptot*aux(1,:,:,imode)
          baux(1,:,:)=dtpv*be(1,:,:)*aux(2,:,:,imode)
          baux(2,:,:)=dtpv*be(2,:,:)*aux(2,:,:,imode)
          baux(3,:,:)=dtpv*be(3,:,:)*aux(2,:,:,imode)
          bvort=(be(1,:,:)*(vez(3,:,:,imode)
     $                     -(0,1)*keff(imode)*ve(2,:,:,imode)/bigr)+
     $           be(2,:,:)*(((0,1)*keff(imode)*ve(1,:,:,imode)-
     $                      g*ve(3,:,:,imode))/bigr-ver(3,:,:,imode))+
     $           be(3,:,:)*(ver(2,:,:,imode)-vez(1,:,:,imode)))*dtpv
        ELSE
          sqpaux=0._r8
          baux=0._r8
        ENDIF
c-----------------------------------------------------------------------
c       accumulate all contributions, including viscous terms
c       -conjg(grad(alpha))^T : Pi(rho,V)
c
c       the k-dependent part of crl is completed here.  note that crl is
c       explicitly conjugated here since it is only used in conjugated
c       form.
c-----------------------------------------------------------------------
        IF (use_gv) THEN
          DO iv=1,nvc
            DO iy=1,ncy
              int(:,iy,iv,imode)=0._r8
              DO ix=1,ncx
                rr=bigr(ix,iy)
                crl(1,1)=rcrl(1,1,ix,iy,iv)
     $            -alpha(ix,iy,iv)*(0,1)*keff(imode)*be(3,ix,iy)/rr
                crl(2,1)=rcrl(2,1,ix,iy,iv)
                crl(3,1)=rcrl(3,1,ix,iy,iv)
                crl(1,2)=rcrl(1,2,ix,iy,iv)
                crl(2,2)=rcrl(2,2,ix,iy,iv)
     $            -alpha(ix,iy,iv)*(0,1)*keff(imode)*be(3,ix,iy)/rr
                crl(3,2)=rcrl(3,2,ix,iy,iv)
                crl(1,3)=rcrl(1,3,ix,iy,iv)
     $            +alpha(ix,iy,iv)*(0,1)*keff(imode)*be(1,ix,iy)/rr
                crl(2,3)=rcrl(2,3,ix,iy,iv)
     $            +alpha(ix,iy,iv)*(0,1)*keff(imode)*be(2,ix,iy)/rr
                crl(3,3)=rcrl(3,3,ix,iy,iv)
                div=grd(ix,iy)*divv(ix,iy)+sqpaux(ix,iy)

                int(1,iy,iv,imode)=int(1,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(1,ix,iy)+
     $                cj0*(divv(ix,iy)*presr(1,ix,iy)-
     $                          ja_cross_dbe(1,ix,iy))-
     $                baux(2,ix,iy)*(0,1)*keff(imode)/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,1))
     $            +dalpdrc(ix,iy,iv)*(div+cj0*ve_gpr(ix,iy))
     $            +ani*SUM(crl(:,1)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(1,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(1,ix,iy,imode))
     $            +dalpdr(ix,iy,iv)*gvten(1,ix,iy,imode)
     $            +dalpdz(ix,iy,iv)*(gvten(2,ix,iy,imode)-baux(3,ix,iy))
 
                int(2,iy,iv,imode)=int(2,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(2,ix,iy)+
     $                cj0*(divv(ix,iy)*presz(1,ix,iy)-
     $                          ja_cross_dbe(2,ix,iy))+
     $                baux(1,ix,iy)*(0,1)*keff(imode)/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,2))
     $            +dalpdz (ix,iy,iv)*(div+cj0*ve_gpr(ix,iy))
     $            +ani*SUM(crl(:,2)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(2,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(2,ix,iy,imode))
     $            +dalpdr(ix,iy,iv)*(gvten(4,ix,iy,imode)+baux(3,ix,iy))
     $            +dalpdz(ix,iy,iv)*gvten(5,ix,iy,imode)

                int(3,iy,iv,imode)=int(3,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(3,ix,iy)
     $               -cj0*ja_cross_dbe(3,ix,iy)
     $               -(0,1)*keff(imode)*(div+cj0*ve_gpr(ix,iy))/rr
     $               -baux(2,ix,iy)*g/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,3))
     $            +ani*SUM(crl(:,3)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(3,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(3,ix,iy,imode))
     $            +dalpdr(ix,iy,iv)*(gvten(7,ix,iy,imode)-baux(2,ix,iy))
     $            +dalpdz(ix,iy,iv)*(gvten(8,ix,iy,imode)+baux(1,ix,iy))
              ENDDO
            ENDDO
          ENDDO
c-----------------------------------------------------------------------
c       same without viscous terms.
c-----------------------------------------------------------------------
        ELSE
          DO iv=1,nvc
            DO iy=1,ncy
              int(:,iy,iv,imode)=0._r8
              DO ix=1,ncx
                rr=bigr(ix,iy)
                crl(1,1)=rcrl(1,1,ix,iy,iv)
     $            -alpha(ix,iy,iv)*(0,1)*keff(imode)*be(3,ix,iy)/rr
                crl(2,1)=rcrl(2,1,ix,iy,iv)
                crl(3,1)=rcrl(3,1,ix,iy,iv)
                crl(1,2)=rcrl(1,2,ix,iy,iv)
                crl(2,2)=rcrl(2,2,ix,iy,iv)
     $            -alpha(ix,iy,iv)*(0,1)*keff(imode)*be(3,ix,iy)/rr
                crl(3,2)=rcrl(3,2,ix,iy,iv)
                crl(1,3)=rcrl(1,3,ix,iy,iv)
     $            +alpha(ix,iy,iv)*(0,1)*keff(imode)*be(1,ix,iy)/rr
                crl(2,3)=rcrl(2,3,ix,iy,iv)
     $            +alpha(ix,iy,iv)*(0,1)*keff(imode)*be(2,ix,iy)/rr
                crl(3,3)=rcrl(3,3,ix,iy,iv)
                div=grd(ix,iy)*divv(ix,iy)+sqpaux(ix,iy)

                int(1,iy,iv,imode)=int(1,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(1,ix,iy)+
     $                cj0*(divv(ix,iy)*presr(1,ix,iy)-
     $                          ja_cross_dbe(1,ix,iy))-
     $                baux(2,ix,iy)*(0,1)*keff(imode)/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,1))
     $            +dalpdrc(ix,iy,iv)*(div+cj0*ve_gpr(ix,iy))
     $            +ani*SUM(crl(:,1)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(1,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(1,ix,iy,imode))
     $            -dalpdz(ix,iy,iv)*baux(3,ix,iy)
 
                int(2,iy,iv,imode)=int(2,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(2,ix,iy)+
     $                cj0*(divv(ix,iy)*presz(1,ix,iy)-
     $                          ja_cross_dbe(2,ix,iy))+
     $                baux(1,ix,iy)*(0,1)*keff(imode)/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,2))
     $            +dalpdz (ix,iy,iv)*(div+cj0*ve_gpr(ix,iy))
     $            +ani*SUM(crl(:,2)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(2,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(2,ix,iy,imode))
     $            +dalpdr(ix,iy,iv)*baux(3,ix,iy)

                int(3,iy,iv,imode)=int(3,iy,iv,imode)+
     $            alpha(ix,iy,iv)*(forck(3,ix,iy)
     $               -cj0*ja_cross_dbe(3,ix,iy)
     $               -(0,1)*keff(imode)*(div+cj0*ve_gpr(ix,iy))/rr
     $               -baux(2,ix,iy)*g/rr)
     $            -cj0*SUM(ve_cross_ja(:,ix,iy)*crl(:,3))
     $            +ani*SUM(crl(:,3)*dbe(:,ix,iy))
     $            +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(3,ix,iy,imode)
     $                        +dalpdz(ix,iy,iv)*vez(3,ix,iy,imode))
     $            -dalpdr(ix,iy,iv)*baux(2,ix,iy)
     $            +dalpdz(ix,iy,iv)*baux(1,ix,iy)
              ENDDO
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       complete the integrand for the discontinuous scalar equations.
c-----------------------------------------------------------------------
        IF (poly_divv>0) THEN
          divv=divv*sqptot*dtdv
          DO iv=nvc+1,nvc+nvm  !  auxiliary equation
            int(1,:,iv,imode)=
     $        SUM(almod(:,:,iv-nvc)*(aux(1,:,:,imode)-divv),1)
            int(2,:,iv,imode)=
     $        SUM(almod(:,:,iv-nvc)*(aux(2,:,:,imode)-bvort),1)
          ENDDO
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE v_aniso_dot
c-----------------------------------------------------------------------
c     subprogram 13.1 v_3dsi_dot.
c     find the action of the (mass density*I + viscous + semi-implicit)
c     operator dotted into a vector (work4) for CG or GMRES iterations.
c     this version is for computations with the 3D semi-implicit
c     operator, with or without implicit advection.  the matrix couples
c     Fourier components.
c
c     the semi-implicit terms are computed with 3D magnetic field,
c     current density, and pressure in this version.  note that after
c     integration by parts, the volumetric terms can be arranged into
c
c     A^*.[df-grad(B).db] + div(A^*)*dp + grad(A^*)^T:[B db], where
c
c     df=0.5*[div(dv)*grad(P)-JXcurl(dvXB)],
c     db=curl(dvXB)/mu0-0.5*dvXJ,
c     dp=gamma*P*div(dv)+0.5*dv.grad(P)-B.db, and
c     A^* is the complex conjugate of a test vector.
c
c     B, grad(B), J, P, and grad(P) have been stored as functions of
c     the periodic angle including steady-state parts.  the terms
c     that are multiplied by grad(A^*) are assembled in the stress
c     tensor before viscous effects are computed; this order of
c     operations differs from that in v_aniso_dot.
c-----------------------------------------------------------------------
      SUBROUTINE v_3dsi_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: almod,dalmr,dalmz
      REAL(r8), DIMENSION(:,:,:), POINTER :: real_ndptr,real_bptr,
     $          real_jptr,real_pptr,real_grdb,real_grdp,
     $          real_oldv,real_oldgv,beq,b0,nd_eq,nd_n0,nl_pres,peq,
     $          p0,kappli,dp,real_titot,ds2
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: pres
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: n0,bigr2,
     $          iso,bmagsq,sqptot
      REAL(r8), DIMENSION(SIZE(bigr,1)*SIZE(bigr,2)) :: n0_1d,df_1d
      REAL(r8), DIMENSION(9,mpseudo,nphi) :: real_gvten,real_pten
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_ve,real_vtmp
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal
      REAL(r8), DIMENSION(3,3) :: vtmpr,btmp,wdotr,bc_wdotr
      REAL(r8), DIMENSION(3) :: tmp1,tmp2,dbeff,dfeff

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: dcp
      COMPLEX(r8), DIMENSION(2,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: aux
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: ve,
     $             ver,vez,force
      COMPLEX(r8), DIMENSION(9,SIZE(bigr,1),SIZE(bigr,2),nmodes) ::
     $             gvten,pten
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: forck
      COMPLEX(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: divv,bvort
      COMPLEX(r8), DIMENSION(3) :: baux
      COMPLEX(r8) :: div,sqpaux

      INTEGER(i4) :: ncx,ncy,iv,imode,jq,jt,ix,iy,ip,im,i1,i2,nvc,nvm
      REAL(r8) :: kin_coef,iso_coef,par_coef,gcoef,vcoef,dloc
      REAL(r8) :: rb2,divtmpr,rscal,hdt,g,ani,cj0,dt2,cp0,rr,dtdv,dtpv
      REAL(r8) :: ten12,ten13,ten23,dpeff
      REAL(r8), PARAMETER :: third=1._r8/3._r8,twoth=2._r8/3._r8
c-----------------------------------------------------------------------
c     convenience
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      hdt=0.5_r8*dt
      bigr2=bigr*bigr
      dt2=dt*dt
      IF (geom=='tor') THEN
        g=1._r8
      ELSE
        g=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nvc=SIZE(alpha,3)
c-----------------------------------------------------------------------
c     evaluate the current direction vector for velocity change and
c     the number density.
c
c     also multiply the diffusivity shape by n0.
c-----------------------------------------------------------------------
      CALL generic_all_eval(rb%work4,tb%work4,rb%dxdr,rb%dydr,
     $                      rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,tb%ng,
     $                      ve,ver,vez,1_i4)
      CALL math_grad(nmodes,keff,3_i4,geom,ve,ver,vez,gvten,bigr)
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,9_i4,gvten,
     $             real_gvten,dealiase)
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_ve,
     $             dealiase)

      IF (continuity=='full') THEN
        CALL generic_ptr_set(rb%qnd_tot,tb%qnd_tot,tb%tgeom,
     $                       inode,real_ndptr,dp,dp,0_i4)
        real_scal=real_ndptr
      ELSE
        CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                       inode,nd_eq,dp,dp,0_i4)
        real_scal=1._r8
      ENDIF
      IF (continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,dp,dp,0_i4)
        n0=nd_eq(1,:,:)+nd_n0(1,:,:)
        n0_1d=RESHAPE(n0,(/ncx*ncy/))
      ELSE IF (continuity=='full') THEN
        n0=1._r8
      ELSE
        n0=nd_eq(1,:,:)
        n0_1d=RESHAPE(n0,(/ncx*ncy/))
      ENDIF

      IF (impladv) THEN
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,real_oldv,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qgrdv,tb%qgrdv,tb%tgeom,
     $                       inode,real_oldgv,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     collect information for diffusivity shapes and coefficients.
c-----------------------------------------------------------------------
      IF ((ds_use=='kin_visc'.OR.ds_use=='both').AND.
     $    (kin_visc>0.OR.iso_visc>0)) THEN
        CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                       inode,ds2,dp,dp,0_i4)
        df_1d=RESHAPE(MAX(ds2(1,:,:),0._r8)*n0,(/ncx*ncy/))
      ELSE
        IF (continuity=='full') THEN
          df_1d=1._r8
        ELSE
          df_1d=n0_1d
        ENDIF
      ENDIF

      kin_coef=dt*fvsc*mtot*kin_visc
      iso_coef=dt*fvsc*mtot*iso_visc
      par_coef=3._r8*dt*fvsc*mtot*par_visc
c-----------------------------------------------------------------------
c     set pointers to fields used in the semi-implicit operator.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_tot,tb%qbe_tot,tb%tgeom,
     $                     inode,real_bptr,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qgrdb,tb%qgrdb,tb%tgeom,
     $                     inode,real_grdb,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qja_tot,tb%qja_tot,tb%tgeom,
     $                     inode,real_jptr,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,beq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,inode,
     $                     b0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qsi_nl_pres,tb%qsi_nl_pres,tb%tgeom,
     $                     inode,nl_pres,dp,dp,0_i4)
      be=b0+beq
      bmagsq=SUM(be**2,1)
      IF (beta>0) THEN
        CALL generic_ptr_set(rb%qpr_tot,tb%qpr_tot,tb%tgeom,
     $                       inode,real_pptr,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qgrdp,tb%qgrdp,tb%tgeom,
     $                       inode,real_grdp,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                       inode,peq,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qpres_n0,tb%qpres_n0,tb%tgeom,
     $                       inode,p0,dp,dp,0_i4)
        pres=p0+peq
      ELSE
c-PRE always allocate and zero-out in nimrod_init.
        ALLOCATE(real_pptr(1,mpseudo,nphi))
        ALLOCATE(real_grdp(3,mpseudo,nphi))
        real_pptr=0._r8
        real_grdp=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     if auxiliary fields are used, they have contributions to the
c     velocity bases and to the lhs of its own evolution equation.  the
c     equations for div(V) contributions are expressed directly in weak
c     form:
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ sqrt(ddivv*nu*dt*f)*auxv*div(A^*) ]
c       int[(ups^*)*auxv]-int[sqrt(ddivv*nu*dt*f)*(ups^*)*div(delta(V))]
c         = int[ sqrt(ddivv*nu*dt/f)*(ups^*)*div(V_old)]
c
c     where f is the implicit centering, nu is a numerical viscosity
c     coefficient, nu=dt*cma**2, cma is sqrt(rho) times the
c     magneto-acoustic speed, and ups is the scalar test field for the
c     auxiliary field auxv.  note that the real pres array becomes
c     sqrt(gamma*P+B^2/mu0) and that mwork4 holds the div(delta-V) and
c     B.curl(delta-V) projections.
c
c     the auxiliary field to stabilize vorticity (auxw) is expressed
c     directly in weak form as
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ dt*sqrt(dpvrt*f/mu0)*auxw*B.curl(A^*) ]
c       int[(lam^*)*auxw]
c        - dt*sqrt(dpvrt*f/mu0)*int[(lam^*)*B.curl(delta(V))]
c        = dt*sqrt(dpvrt/(f*mu0))*int[(lam^*)*B.curl(V_old)]
c
c     where A is the test vector for V, lam is the test function for
c     the discontinuous scalar field auxw, where f is the centering
c     coefficient.  here, the effective viscosity is dt*v_Alven**2.
c
c     at this point, just evaluate the auxiliary fields and coefficients
c     used in the equations.

cc     divv is again used for temporary space, and force becomes
cc     dt*cpvrt*curl(V)/sqrt(mu0).
c-----------------------------------------------------------------------
      IF (poly_divv>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlrhs',almod,dalmr,
     $                          dalmz,0_i4,poly_divv,
     $                          polydmin=poly_divv_min,
     $                          polydmax=poly_divv_max)
        nvm=SIZE(almod,3)
        CALL generic_all_eval(rb%mwork4,tb%mwork4,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,aux,dcp,dcp,0_i4)

        IF (beta>0) THEN
          sqptot=SQRT(gamma*pres(1,:,:)+SUM(be**2,1)/mu0)
        ELSE
          sqptot=SQRT(SUM(be**2,1)/mu0)
        ENDIF
        dtdv=SQRT(ddivv*fdivv)*dt        !  for diffusive correction
        dtpv=dt*SQRT(dpvrt*fpvrt/mu0)  !  for diffusive correction
      ENDIF
c-----------------------------------------------------------------------
c     find the spatial functions used as coefficients for the isotropic
c     and the anisotropic parts of the  operator.
c-----------------------------------------------------------------------
      ani=0.25_r8*si_fac_mhd* (1._r8-mhd_si_iso)*dt2/mu0
      cp0=0.25_r8*si_fac_pres*(1._r8-mhd_si_iso)*dt2*gamma
      cj0=0.125_r8*si_fac_j0*dt2
      IF (beta>0) THEN
        iso=0.25_r8*dt2*( mhd_si_iso*
     $    (si_fac_mhd*bmagsq/mu0+si_fac_pres*gamma*pres(1,:,:))+
     $      si_fac_nl*(nl_pres(1,:,:)/mu0+gamma*nl_pres(2,:,:)) )
      ELSE
        iso=0.25_r8*dt2*( mhd_si_iso*si_fac_mhd*bmagsq+
     $                    si_fac_nl*nl_pres(1,:,:) )/mu0
      ENDIF
c-----------------------------------------------------------------------
c     evaluate contributions to the 3D part of the semi-implicit
c     operator as a dyad and a vector to be multiplied by grad(A^*)^T
c     and A^*, respectively.  this is completed before the viscous
c     computations where grad(dv) gets multiplied by n.
c
c     start by finding the effective delta-f, delta-b, and delta-p that
c     are defined in the introductory comments:
c       db=curl(dvXB)/mu0-0.5*dvXJ
c       dp=gamma*P*div(dv)+0.5*dv.grad(P)-B.db
c       df=0.5*[div(dv)*grad(P)-JXcurl(dvXB)]
c     note that dbeff initially holds just curl(dvxB)
c
c     since isotropic viscous diffusion is used in nearly all nonlinear
c     computations, it is added here to avoid extra looping later.  the
c     factor of n is either in real_scal or in df_1d.
c-----------------------------------------------------------------------
      DO ip=1,nphi
        i1=ipseust-1
        DO ix=1,mpseudo
          i1=i1+1
          divtmpr=real_gvten(1,ix,ip)+real_gvten(5,ix,ip)+
     $            real_gvten(9,ix,ip)
          dbeff(1)=SUM(real_bptr(:,ix,ip)*real_gvten(1:3,ix,ip))-
     $             SUM(real_ve  (:,ix,ip)*real_grdb (1:3,ix,ip))-
     $             real_bptr(1,ix,ip)*divtmpr
          dbeff(2)=SUM(real_bptr(:,ix,ip)*real_gvten(4:6,ix,ip))-
     $             SUM(real_ve  (:,ix,ip)*real_grdb (4:6,ix,ip))-
     $             real_bptr(2,ix,ip)*divtmpr
          dbeff(3)=SUM(real_bptr(:,ix,ip)*real_gvten(7:9,ix,ip))-
     $             SUM(real_ve  (:,ix,ip)*real_grdb (7:9,ix,ip))-
     $             real_bptr(3,ix,ip)*divtmpr

          dfeff(1)=cj0*(divtmpr*real_grdp(1,ix,ip)+
     $                  dbeff(2)*real_jptr(3,ix,ip)-
     $                  dbeff(3)*real_jptr(2,ix,ip))
          dfeff(2)=cj0*(divtmpr*real_grdp(2,ix,ip)+
     $                  dbeff(3)*real_jptr(1,ix,ip)-
     $                  dbeff(1)*real_jptr(3,ix,ip))
          dfeff(3)=cj0*(divtmpr*real_grdp(3,ix,ip)+
     $                  dbeff(1)*real_jptr(2,ix,ip)-
     $                  dbeff(2)*real_jptr(1,ix,ip))

          dbeff(1)=ani*dbeff(1)+
     $             cj0*(real_ve(3,ix,ip)*real_jptr(2,ix,ip)-
     $                  real_ve(2,ix,ip)*real_jptr(3,ix,ip))
          dbeff(2)=ani*dbeff(2)+
     $             cj0*(real_ve(1,ix,ip)*real_jptr(3,ix,ip)-
     $                  real_ve(3,ix,ip)*real_jptr(1,ix,ip))
          dbeff(3)=ani*dbeff(3)+
     $             cj0*(real_ve(2,ix,ip)*real_jptr(1,ix,ip)-
     $                  real_ve(1,ix,ip)*real_jptr(2,ix,ip))

          dpeff=cp0*real_pptr(1,ix,ip)*divtmpr+
     $          cj0*SUM(real_ve(:,ix,ip)*real_grdp(:,ix,ip))-
     $          SUM(real_bptr(:,ix,ip)*dbeff)

          real_vtmp(1,ix,ip)=dfeff(1)-SUM(real_grdb(1:7:3,ix,ip)*dbeff)
          real_vtmp(2,ix,ip)=dfeff(2)-SUM(real_grdb(2:8:3,ix,ip)*dbeff)
          real_vtmp(3,ix,ip)=dfeff(3)-SUM(real_grdb(3:9:3,ix,ip)*dbeff)

          divtmpr=-twoth*divtmpr
          rscal=df_1d(i1)*real_scal(1,ix,ip)
          real_pten(1,ix,ip)=real_bptr(1,ix,ip)*dbeff(1)+dpeff+
     $      ((kin_coef+2._r8*iso_coef)*real_gvten(1,ix,ip)+
     $       iso_coef*divtmpr)*rscal
          real_pten(2,ix,ip)=real_bptr(2,ix,ip)*dbeff(1)+
     $      ((kin_coef+      iso_coef)*real_gvten(2,ix,ip)+
     $       iso_coef*real_gvten(4,ix,ip))*rscal
          real_pten(3,ix,ip)=real_bptr(3,ix,ip)*dbeff(1)+
     $      ((kin_coef+      iso_coef)*real_gvten(3,ix,ip)+
     $       iso_coef*real_gvten(7,ix,ip))*rscal
          real_pten(4,ix,ip)=real_bptr(1,ix,ip)*dbeff(2)+
     $      ((kin_coef+      iso_coef)*real_gvten(4,ix,ip)+
     $       iso_coef*real_gvten(2,ix,ip))*rscal
          real_pten(5,ix,ip)=real_bptr(2,ix,ip)*dbeff(2)+dpeff+
     $      ((kin_coef+2._r8*iso_coef)*real_gvten(5,ix,ip)+
     $       iso_coef*divtmpr)*rscal
          real_pten(6,ix,ip)=real_bptr(3,ix,ip)*dbeff(2)+
     $      ((kin_coef+      iso_coef)*real_gvten(6,ix,ip)+
     $       iso_coef*real_gvten(8,ix,ip))*rscal
          real_pten(7,ix,ip)=real_bptr(1,ix,ip)*dbeff(3)+
     $      ((kin_coef+      iso_coef)*real_gvten(7,ix,ip)+
     $       iso_coef*real_gvten(3,ix,ip))*rscal
          real_pten(8,ix,ip)=real_bptr(2,ix,ip)*dbeff(3)+
     $      ((kin_coef+      iso_coef)*real_gvten(8,ix,ip)+
     $       iso_coef*real_gvten(6,ix,ip))*rscal
          real_pten(9,ix,ip)=real_bptr(3,ix,ip)*dbeff(3)+dpeff+
     $      ((kin_coef+2._r8*iso_coef)*real_gvten(9,ix,ip)+
     $       iso_coef*divtmpr)*rscal
        ENDDO
      ENDDO
      IF (beta<=0) DEALLOCATE(real_pptr,real_grdp)
c-----------------------------------------------------------------------
c     find rho*dve product for the d/dt term, and add to real_vtmp.
c-----------------------------------------------------------------------
      IF (continuity=='full') THEN
        DO ip=1,nphi
          DO ix=1,mpseudo
            real_ve(:,ix,ip)=real_ve(:,ix,ip)*real_ndptr(1,ix,ip)*mtot
            real_vtmp(:,ix,ip)=real_vtmp(:,ix,ip)+real_ve(:,ix,ip)
          ENDDO
        ENDDO
      ELSE
        DO ip=1,nphi
          DO ix=1,mpseudo
            real_ve(:,ix,ip)=real_ve(:,ix,ip)*n0_1d(ix+ipseust-1)*mtot
            real_vtmp(:,ix,ip)=real_vtmp(:,ix,ip)+real_ve(:,ix,ip)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     evaluations for viscous dissipation:
c     first multiply grad(dv) by number density.
c-----------------------------------------------------------------------
      IF ((kin_visc>0.OR.iso_visc>0.OR.par_visc>0.OR.impladv).AND.
     $ (integrand_flag(1:4)=='visc'.OR.integrand_flag(1:3)=='all')) THEN

        IF (continuity=='full') THEN
          DO ip=1,nphi
            DO ix=1,mpseudo
              real_gvten(:,ix,ip)=
     $          real_gvten(:,ix,ip)*real_ndptr(1,ix,ip)
            ENDDO
          ENDDO
        ELSE
          DO ip=1,nphi
            DO ix=1,mpseudo
              real_gvten(:,ix,ip)=
     $          real_gvten(:,ix,ip)*n0_1d(ix+ipseust-1)
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       for nonlinear parallel viscosity, find the unique components of
c
c         [ B.grad(V).B - B**2*div(V) ] * (BB - I*B**2/3)
c
c       also divide by B**4 to make unit vectors.
c
c-PRE this could be separated for efficiency.
c       multiply by the temperature-dependent profile for Braginskii
c       parallel viscosity if desired.
c-----------------------------------------------------------------------
        IF (par_visc>0) THEN
          IF (parvisc_model=='plltdep') THEN
            CALL generic_ptr_set(rb%qkappli_phi,tb%qkappli_phi,
     $                           tb%tgeom,inode,kappli,dp,dp,0_i4)
            real_scal=kappli/k_plli
          ELSE
            real_scal=1._r8
          ENDIF
          DO ip=1,nphi
            DO ix=1,mpseudo
              rb2=SUM(real_bptr(:,ix,ip)*real_bptr(:,ix,ip))+smallnum
              divtmpr=-third*rb2*(real_gvten(1,ix,ip)+
     $                real_gvten(5,ix,ip)+real_gvten(9,ix,ip))
              rscal=par_coef*real_scal(1,ix,ip)*(divtmpr+
     $              real_bptr(1,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(1,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(2,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(3,ix,ip))+
     $              real_bptr(2,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(4,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(5,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(6,ix,ip))+
     $              real_bptr(3,ix,ip)*(
     $                real_bptr(1,ix,ip)*real_gvten(7,ix,ip)+
     $                real_bptr(2,ix,ip)*real_gvten(8,ix,ip)+
     $                real_bptr(3,ix,ip)*real_gvten(9,ix,ip)))/rb2**2
              rb2=-third*rb2
              ten12=rscal*real_bptr(1,ix,ip)*real_bptr(2,ix,ip)
              ten13=rscal*real_bptr(1,ix,ip)*real_bptr(3,ix,ip)
              ten23=rscal*real_bptr(2,ix,ip)*real_bptr(3,ix,ip)
              real_pten(1,ix,ip)=real_pten(1,ix,ip)+rscal*
     $          (real_bptr(1,ix,ip)*real_bptr(1,ix,ip)+rb2)
              real_pten(2,ix,ip)=real_pten(2,ix,ip)+ten12
              real_pten(3,ix,ip)=real_pten(3,ix,ip)+ten13
              real_pten(4,ix,ip)=real_pten(4,ix,ip)+ten12
              real_pten(5,ix,ip)=real_pten(5,ix,ip)+rscal*
     $          (real_bptr(2,ix,ip)*real_bptr(2,ix,ip)+rb2)
              real_pten(6,ix,ip)=real_pten(6,ix,ip)+ten23
              real_pten(7,ix,ip)=real_pten(7,ix,ip)+ten13
              real_pten(8,ix,ip)=real_pten(8,ix,ip)+ten23
              real_pten(9,ix,ip)=real_pten(9,ix,ip)+rscal*
     $          (real_bptr(3,ix,ip)*real_bptr(3,ix,ip)+rb2)
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       for nonlinear gyroviscosity, find the unique components of
c
c         (n_e*m_i*T_i/Z**2*B**4)*{ [BxW.(B**2+3BB) ] + transpose[] }
c
c-----------------------------------------------------------------------
        IF (gyr_visc>0) THEN
          vcoef=-0.125*gyr_visc*dt*ms(2)*kboltz/(zeff**2*elementary_q)
          CALL generic_ptr_set(rb%qti_tot,tb%qti_tot,tb%tgeom,
     $                         inode,real_titot,dp,dp,0_i4)
          DO ip=1,nphi
            DO ix=1,mpseudo
              rb2=SUM(real_bptr(:,ix,ip)**2)+smallnum
              gcoef=vcoef*real_titot(1,ix,ip)/rb2**2
              vtmpr(1:3,1)=real_gvten(1:3,ix,ip)+real_gvten(1:7:3,ix,ip)
              vtmpr(1:3,2)=real_gvten(4:6,ix,ip)+real_gvten(2:8:3,ix,ip)
              vtmpr(1:3,3)=real_gvten(7:9,ix,ip)+real_gvten(3:9:3,ix,ip)
              btmp(1:3,1)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(1  ,ix,ip)
              btmp(1:3,2)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(2  ,ix,ip)
              btmp(1:3,3)=3._r8*real_bptr(1:3,ix,ip)*
     $                          real_bptr(3  ,ix,ip)
              btmp(1,1)=rb2+btmp(1,1)
              btmp(2,2)=rb2+btmp(2,2)
              btmp(3,3)=rb2+btmp(3,3)
              DO i2=1,3
                wdotr(1,i2)=SUM(vtmpr(1,:)*btmp(:,i2))
                wdotr(2,i2)=SUM(vtmpr(2,:)*btmp(:,i2))
                wdotr(3,i2)=SUM(vtmpr(3,:)*btmp(:,i2))
              ENDDO
              bc_wdotr(1,:)=real_bptr(2,ix,ip)*wdotr(3,:)-
     $                      real_bptr(3,ix,ip)*wdotr(2,:)
              bc_wdotr(2,:)=real_bptr(3,ix,ip)*wdotr(1,:)-
     $                      real_bptr(1,ix,ip)*wdotr(3,:)
              bc_wdotr(3,:)=real_bptr(1,ix,ip)*wdotr(2,:)-
     $                      real_bptr(2,ix,ip)*wdotr(1,:)
              ten12=gcoef*(bc_wdotr(2,1)+bc_wdotr(1,2))
              ten13=gcoef*(bc_wdotr(3,1)+bc_wdotr(1,3))
              ten23=gcoef*(bc_wdotr(3,2)+bc_wdotr(2,3))
              real_pten(1,ix,ip)=real_pten(1,ix,ip)+
     $                           gcoef*(bc_wdotr(1,1)+bc_wdotr(1,1))
              real_pten(2,ix,ip)=real_pten(2,ix,ip)+ten12
              real_pten(3,ix,ip)=real_pten(3,ix,ip)+ten13
              real_pten(4,ix,ip)=real_pten(4,ix,ip)+ten12
              real_pten(5,ix,ip)=real_pten(5,ix,ip)+
     $                           gcoef*(bc_wdotr(2,2)+bc_wdotr(2,2))
              real_pten(6,ix,ip)=real_pten(6,ix,ip)+ten23
              real_pten(7,ix,ip)=real_pten(7,ix,ip)+ten13
              real_pten(8,ix,ip)=real_pten(8,ix,ip)+ten23
              real_pten(9,ix,ip)=real_pten(9,ix,ip)+
     $                           gcoef*(bc_wdotr(3,3)+bc_wdotr(3,3))
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       transform the stress/semi-implicit tensor.
c-----------------------------------------------------------------------
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,9_i4,
     $               pten,real_pten,dealiase)
c-----------------------------------------------------------------------
c       terms for implicit advection:
c
c          0.5*dt*rho*( dV.grad(V_old) + V_old.grad(dV) )
c
c       where dV is the iterate from work4.
c
c       note that V_old already includes any equilibrium flow.  Also,
c       real_gvten still holds n*grad(dV) in real space, and real_ve
c       holds  mtot*n*dV in real space.
c
c       add this advective force to the vector semi-implicit
c       contribution real_vtmp, which also has mtot*n*dV.
c-----------------------------------------------------------------------
        IF (impladv.AND.(advect=='V only'.OR.advect=='all')) THEN
          DO ip=1,nphi
            DO ix=1,mpseudo
               tmp1=hdt*real_ve(:,ix,ip)
               tmp2=hdt*mtot*real_oldv(:,ix,ip)
               real_vtmp(1,ix,ip)=real_vtmp(1,ix,ip)+
     $                   tmp1(1)*real_oldgv(1,ix,ip)+
     $                   tmp1(2)*real_oldgv(2,ix,ip)+
     $                   tmp1(3)*real_oldgv(3,ix,ip)+
     $                   tmp2(1)*real_gvten(1,ix,ip)+
     $                   tmp2(2)*real_gvten(2,ix,ip)+
     $                   tmp2(3)*real_gvten(3,ix,ip)
               real_vtmp(2,ix,ip)=real_vtmp(2,ix,ip)+
     $                   tmp1(1)*real_oldgv(4,ix,ip)+
     $                   tmp1(2)*real_oldgv(5,ix,ip)+
     $                   tmp1(3)*real_oldgv(6,ix,ip)+
     $                   tmp2(1)*real_gvten(4,ix,ip)+
     $                   tmp2(2)*real_gvten(5,ix,ip)+
     $                   tmp2(3)*real_gvten(6,ix,ip)
               real_vtmp(3,ix,ip)=real_vtmp(3,ix,ip)+
     $                   tmp1(1)*real_oldgv(7,ix,ip)+
     $                   tmp1(2)*real_oldgv(8,ix,ip)+
     $                   tmp1(3)*real_oldgv(9,ix,ip)+
     $                   tmp2(1)*real_gvten(7,ix,ip)+
     $                   tmp2(2)*real_gvten(8,ix,ip)+
     $                   tmp2(3)*real_gvten(9,ix,ip)
            ENDDO
          ENDDO
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     transform the vector contribution from the semi-implicit operator,
c     the d/dt term, and implicit advection.
c-----------------------------------------------------------------------
      CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,
     $             force,real_vtmp,dealiase)
c-----------------------------------------------------------------------
c     the resultant is formed by finding
c
c       A^*.force + grad(A^*):pten plus isotropic contributions.
c
c     for each of the test vectors, one for each vector-direction /
c     spatial basis function.
c-----------------------------------------------------------------------
      DO imode=1,nmodes
c-----------------------------------------------------------------------
c       find the contributions of P-total times div(dV) and B.curl(dV)
c       from the auxiliary discontinuous fields.  add the contributions
c       to the pten tensor so that grad(A^*)^T:pten includes
c       the div(A^*) and B.curl(A^*)=grad(A^*)^T:(BxI)^T from the
c       auxiliary forcing.
c
c-PRE   this could be done in the viscous-dissipation loop.
c-----------------------------------------------------------------------
        IF (poly_divv>0) THEN
          DO iy=1,ncy
            DO ix=1,ncx
              sqpaux=dtdv*sqptot(ix,iy)*aux(1,ix,iy,imode)
              baux(1)=dtpv*be(1,ix,iy)*aux(2,ix,iy,imode)
              baux(2)=dtpv*be(2,ix,iy)*aux(2,ix,iy,imode)
              baux(3)=dtpv*be(3,ix,iy)*aux(2,ix,iy,imode)
              pten(1,ix,iy,imode)=pten(1,ix,iy,imode)+sqpaux
              pten(2,ix,iy,imode)=pten(2,ix,iy,imode)-baux(3)
              pten(3,ix,iy,imode)=pten(3,ix,iy,imode)+baux(2)
              pten(4,ix,iy,imode)=pten(4,ix,iy,imode)+baux(3)
              pten(5,ix,iy,imode)=pten(5,ix,iy,imode)+sqpaux
              pten(6,ix,iy,imode)=pten(6,ix,iy,imode)-baux(1)
              pten(7,ix,iy,imode)=pten(7,ix,iy,imode)-baux(2)
              pten(8,ix,iy,imode)=pten(8,ix,iy,imode)+baux(1)
              pten(9,ix,iy,imode)=pten(9,ix,iy,imode)+sqpaux
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       complete the factors that are dotted into the test vectors, A.
c-----------------------------------------------------------------------
        IF (geom=='tor') THEN
          forck(1,:,:)=force(1,:,:,imode)+
     $      iso*((1+k2ef(imode))*ve(1,:,:,imode)+
     $        2*(0,1)*keff(imode)*ve(3,:,:,imode))/bigr2
     $      +(                  pten(9,:,:,imode)
     $       -(0,1)*keff(imode)*pten(3,:,:,imode))/bigr
          forck(2,:,:)=force(2,:,:,imode)+
     $      iso*k2ef(imode)*ve(2,:,:,imode)/bigr2
     $       -(0,1)*keff(imode)*pten(6,:,:,imode) /bigr
          forck(3,:,:)=force(3,:,:,imode)+
     $      iso*((1+k2ef(imode))*ve(3,:,:,imode)-
     $        2*(0,1)*keff(imode)*ve(1,:,:,imode))/bigr2
     $      -(                  pten(3,:,:,imode)
     $       +(0,1)*keff(imode)*pten(9,:,:,imode))/bigr
        ELSE
          forck(1,:,:)=force(1,:,:,imode)+
     $                 iso*k2ef(imode)*ve(1,:,:,imode)-
     $                 (0,1)*keff(imode)*pten(3,:,:,imode)
          forck(2,:,:)=force(2,:,:,imode)+
     $                 iso*k2ef(imode)*ve(2,:,:,imode)-
     $                 (0,1)*keff(imode)*pten(6,:,:,imode)
          forck(3,:,:)=force(3,:,:,imode)+
     $                 iso*k2ef(imode)*ve(3,:,:,imode)-
     $                 (0,1)*keff(imode)*pten(9,:,:,imode)
        ENDIF
c-----------------------------------------------------------------------
c       accumulate all contributions.
c-----------------------------------------------------------------------
        DO iv=1,nvc
          DO iy=1,ncy
            int(:,iy,iv,imode)=0._r8
            DO ix=1,ncx
              int(1,iy,iv,imode)=int(1,iy,iv,imode)+
     $          alpha(ix,iy,iv)*forck(1,ix,iy)
     $          +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(1,ix,iy,imode)
     $                      +dalpdz(ix,iy,iv)*vez(1,ix,iy,imode))
     $          +dalpdr(ix,iy,iv)*pten(1,ix,iy,imode)
     $          +dalpdz(ix,iy,iv)*pten(2,ix,iy,imode)

              int(2,iy,iv,imode)=int(2,iy,iv,imode)+
     $          alpha(ix,iy,iv)*forck(2,ix,iy)
     $          +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(2,ix,iy,imode)
     $                      +dalpdz(ix,iy,iv)*vez(2,ix,iy,imode))
     $          +dalpdr(ix,iy,iv)*pten(4,ix,iy,imode)
     $          +dalpdz(ix,iy,iv)*pten(5,ix,iy,imode)

              int(3,iy,iv,imode)=int(3,iy,iv,imode)+
     $          alpha(ix,iy,iv)*forck(3,ix,iy)
     $          +iso(ix,iy)*(dalpdr(ix,iy,iv)*ver(3,ix,iy,imode)
     $                      +dalpdz(ix,iy,iv)*vez(3,ix,iy,imode))
     $          +dalpdr(ix,iy,iv)*pten(7,ix,iy,imode)
     $          +dalpdz(ix,iy,iv)*pten(8,ix,iy,imode)
            ENDDO
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c       complete the integrand for the discontinuous scalar equations.
c-----------------------------------------------------------------------
        IF (poly_divv>0) THEN
          divv=sqptot*dtdv*(ver(1,:,:,imode)+vez(2,:,:,imode)+
     $                      ((0,1)*keff(imode)*ve(3,:,:,imode)+
     $                                       g*ve(1,:,:,imode))/bigr)
          bvort=(be(1,:,:)*(vez(3,:,:,imode)
     $                     -(0,1)*keff(imode)*ve(2,:,:,imode)/bigr)+
     $           be(2,:,:)*(((0,1)*keff(imode)*ve(1,:,:,imode)-
     $                      g*ve(3,:,:,imode))/bigr-ver(3,:,:,imode))+
     $           be(3,:,:)*(ver(2,:,:,imode)-vez(1,:,:,imode)))*dtpv
          DO iv=nvc+1,nvc+nvm  !  auxiliary equation
            int(1,:,iv,imode)=
     $        SUM(almod(:,:,iv-nvc)*(aux(1,:,:,imode)-divv),1)
            int(2,:,iv,imode)=
     $        SUM(almod(:,:,iv-nvc)*(aux(2,:,:,imode)-bvort),1)
          ENDDO
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE v_3dsi_dot
c-----------------------------------------------------------------------
c     subprogram 14. divb_rhs.
c     compute grad_div(b) for the rhs of the div(b) diffuser.
c-----------------------------------------------------------------------
      SUBROUTINE divb_rhs(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: be,ber,bez
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: divb

      INTEGER(i4) :: iv,imode,iq,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the perturbed
c     b for each mode.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      CALL generic_ptr_set(rb%qbe,tb%qbe,tb%tgeom,inode,be,ber,bez,1_i4)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     begin mode loop, and find the real and imaginary parts of div(b)
c     for this mode.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
        IF (geom=='tor') THEN
          divb=ber(1,:,:,imode)+bez(2,:,:,imode)
     $            +((0,1)*keff(imode)*be(3,:,:,imode)
     $                               +be(1,:,:,imode))/bigr
        ELSE
          divb=ber(1,:,:,imode)+bez(2,:,:,imode)
     $             +(0,1)*keff(imode)*be(3,:,:,imode)/bigr
        ENDIF
c-----------------------------------------------------------------------
c       evaluate -div(alpha)*div(B).
c-----------------------------------------------------------------------
        DO iv=1,nv
          int(1,:,iv,imode)=SUM(
     $           -divbd*dt*dalpdrc(:,:,iv)*divb,1)
          int(2,:,iv,imode)=SUM(
     $           -divbd*dt*dalpdz (:,:,iv)*divb,1)
          int(3,:,iv,imode)=SUM(
     $           +(0,1)*divbd*dt*alpha(:,:,iv)*keff(imode)/bigr*divb,1)
        ENDDO
      ENDDO mode_loop
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE divb_rhs
c-----------------------------------------------------------------------
c     subprogram 15. ndrhs.
c     compute the integrand used in the rhs of the continuity equation.
c-----------------------------------------------------------------------
      SUBROUTINE ndrhs(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,ve_eq,dart,upwc,
     $                                       vetot,ndeqr,ndeqz,dp
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_vec,real_gn
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal,real_n
      REAL(r8), DIMENSION(  mpseudo) :: rdot,rscl
      REAL(r8), DIMENSION(SIZE(bigr,1)*SIZE(bigr,2)) :: ndeqr_1d,
     $          ndeqz_1d

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ve,nd,ndr,ndz,dcp
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             nvel,grad_nd

      INTEGER(i4) :: nv,iv,imode,iq,jq,ncx,ncy,ip
      REAL(r8) :: fupw,fhyp
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and evaluate velocity.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      CALL generic_ptr_set(rb%qve,tb%qve,tb%tgeom,inode,ve,dcp,dcp,0_i4)
      IF (eq_flow/='none') THEN
        CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,ve_eq,dp,dp,0_i4)
      ENDIF
c-----------------------------------------------------------------------
c     density
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,ndeqr,ndeqz,1_i4)
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,
     $                     nd,ndr,ndz,1_i4)
      IF (nd_diff>0.OR.nonlinear.AND.nd_dart_upw>0.AND.impladv.OR.
     $    nd_hypd>0.AND.impladv) THEN
        DO imode=1,nmodes
          grad_nd(1,:,:,imode)=ndr(1,:,:,imode)
          grad_nd(2,:,:,imode)=ndz(1,:,:,imode)
          grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
        ENDDO
        fhyp=SQRT(nd_hypd*dt)
      ENDIF
      IF (nonlinear.AND.nd_diff>0.AND.nd_floor>0)
     $  CALL generic_ptr_set(rb%qdart,tb%qdart,tb%tgeom,
     $                       inode,dart,dp,dp,0_i4)
      IF (nonlinear.AND.nd_dart_upw>0.AND.impladv) THEN
        CALL generic_ptr_set(rb%qupw_phi,tb%qupw_phi,tb%tgeom,
     $                       inode,upwc,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,
     $                       inode,vetot,dp,dp,0_i4)
        ndeqr_1d=RESHAPE(ndeqr,(/ncx*ncy/))
        ndeqz_1d=RESHAPE(ndeqz,(/ncx*ncy/))
        fupw=nd_dart_upw*dt**2
      ENDIF
      IF ((nonlinear.OR.eq_flow/='none').AND.
     $    integrand_flag/='n predict'.AND..NOT.impladv) THEN
        CALL generic_ptr_set(rb%qwork3,tb%qwork3,tb%tgeom,inode,
     $                       nd,ndr,ndz,1_i4)
      ENDIF
c-----------------------------------------------------------------------
c     find the nonlinear contributions to n*V.
c     note that the real arrays have three dimensions, where the
c     first is the vector component, the second covers the poloidal
c     plane, and the third is the phi index,
c-----------------------------------------------------------------------
      IF (nonlinear) THEN
c-----------------------------------------------------------------------
c       transform v_perturbed and n_perturbed to real space.
c-----------------------------------------------------------------------
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,ve,real_vec,
     $               dealiase)
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,nd,real_n,
     $               dealiase)
c-----------------------------------------------------------------------
c       apply an upwinding-like artificial diffusion with the
c       pre-determined diffusivity multiplied by VV.grad(n) and add
c       this to n*V.
c-----------------------------------------------------------------------
        IF (nd_dart_upw>0.AND.impladv) THEN
          CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $                 real_gn,dealiase)
          DO ip=1,nphi
            rdot=fupw*upwc(1,:,ip)*upw_aniso * ( 
     $          SUM(vetot(:,:,ip)*real_gn(:,:,ip),1) +
     $              real_vec(1,:,ip)*ndeqr_1d(ipseust:ipseuen) +
     $              real_vec(2,:,ip)*ndeqz_1d(ipseust:ipseuen) ) /
     $        ( SUM(vetot(:,:,ip)*vetot(:,:,ip),1) + smallnum )
            rscl=fupw*upwc(1,:,ip)*(1._r8-upw_aniso)
            real_vec(1,:,ip)=dt*real_vec(1,:,ip)*real_n(1,:,ip)-
     $        ( rdot*vetot(1,:,ip)+
     $          rscl*(real_gn(1,:,ip)+ndeqr_1d(ipseust:ipseuen)) )
            real_vec(2,:,ip)=dt*real_vec(2,:,ip)*real_n(1,:,ip)-
     $        ( rdot*vetot(2,:,ip)+
     $          rscl*(real_gn(2,:,ip)+ndeqz_1d(ipseust:ipseuen)) )
            real_vec(3,:,ip)=dt*real_vec(3,:,ip)*real_n(1,:,ip)-
     $        ( rdot*vetot(3,:,ip)+rscl*real_gn(3,:,ip) )
          ENDDO
c-----------------------------------------------------------------------
c       or just find n*V
c-----------------------------------------------------------------------
        ELSE
          real_vec(1,:,:)=dt*real_vec(1,:,:)*real_n(1,:,:)
          real_vec(2,:,:)=dt*real_vec(2,:,:)*real_n(1,:,:)
          real_vec(3,:,:)=dt*real_vec(3,:,:)*real_n(1,:,:)
        ENDIF
c-----------------------------------------------------------------------
c       transform flux to Fourier space
c-----------------------------------------------------------------------
        CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,nvel,real_vec,
     $               dealiase)
      ELSE
        nvel=0
      ENDIF
c-----------------------------------------------------------------------
c     begin mode loop.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
c-----------------------------------------------------------------------
c       find the linear n*V for this mode.
c-----------------------------------------------------------------------
        IF (eq_flow/='none') THEN
          DO iv=1,3
            nvel(iv,:,:,imode)=nvel(iv,:,:,imode)
     $        +dt*(ve(iv,:,:,imode)*nd_eq(1,:,:)
     $            +ve_eq(iv,:,:)*nd(1,:,:,imode))
          ENDDO
        ELSE
          DO iv=1,3
            nvel(iv,:,:,imode)=nvel(iv,:,:,imode)
     $        +dt*ve(iv,:,:,imode)*nd_eq(1,:,:)
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       add diffusive term if needed.
c-----------------------------------------------------------------------
        IF (nonlinear.AND.nd_floor>0.AND.nd_diff>0) THEN
          DO iv=1,3
            nvel(iv,:,:,imode)=nvel(iv,:,:,imode)-
     $        (dt*nd_diff+dart(1,:,:))*grad_nd(iv,:,:,imode)
          ENDDO
        ELSE IF (nd_diff>0) THEN
          nvel(:,:,:,imode)=nvel(:,:,:,imode)-
     $         dt*nd_diff*grad_nd(:,:,:,imode)
        ENDIF
c-----------------------------------------------------------------------
c       construct grad(alpha).( n*V )
c       with linear and nonlinear contributions, possibly including
c       the rhs for the hyper diffusivity auxiliary variable 
c       that is proportional to grad**2(n).
c-----------------------------------------------------------------------
        IF (nd_hypd>0.AND.impladv) THEN
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(
     $        +dalpdr(:,:,iv)*nvel(1,:,:,imode)
     $        +dalpdz(:,:,iv)*nvel(2,:,:,imode)
     $        -(0,1)*keff(imode)*alpha(:,:,iv)*nvel(3,:,:,imode)/bigr,1)
            int(2,:,iv,imode)=SUM(-fhyp*(
     $        dalpdr(:,:,iv)*grad_nd(1,:,:,imode)+
     $        dalpdz(:,:,iv)*grad_nd(2,:,:,imode)-
     $        (0,1)*keff(imode)*alpha(:,:,iv)*
     $                       grad_nd(3,:,:,imode)/bigr),1)
          ENDDO
        ELSE
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(
     $        +dalpdr(:,:,iv)*nvel(1,:,:,imode)
     $        +dalpdz(:,:,iv)*nvel(2,:,:,imode)
     $        -(0,1)*keff(imode)*alpha(:,:,iv)*nvel(3,:,:,imode)/bigr,1)
          ENDDO
        ENDIF
      ENDDO mode_loop
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE ndrhs
c-----------------------------------------------------------------------
c     subprogram 16. cont_dot.
c     compute the matrix-vector dot product for the continuity advance
c     with implicit advection.
c-----------------------------------------------------------------------
      SUBROUTINE cont_dot(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,real_ve,dart,upwc,dp
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: kap_art
      REAL(r8), DIMENSION(3,mpseudo,nphi) :: real_vec,real_gn
      REAL(r8), DIMENSION(1,mpseudo,nphi) :: real_scal

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ndold,dcp
      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: nd,
     $             ndr,ndz
      COMPLEX(r8), DIMENSION(2,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: v2,
     $             v2r,v2z
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),nmodes) :: 
     $             nvel,grad_nd

      INTEGER(i4) :: nv,iv,imode,iq,jq,ncx,ncy,ix,ip
      REAL(r8) :: hdt,fdt,dfac,fupw,fupwi,rdot,rscl,fhyp
c-----------------------------------------------------------------------
c     convenience parameters:
c-----------------------------------------------------------------------
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nv=SIZE(int,3)
      hdt=0.5*dt
      dfac=fthc*dt*nd_diff
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and evaluate velocity.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'rhs',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      CALL generic_ptr_set(rb%qve_tot,tb%qve_tot,tb%tgeom,inode,
     $                     real_ve,dp,dp,0_i4)
c-----------------------------------------------------------------------
c     density: the iterate for the change is put in nd,ndr,ndz, and
c     n at the beginning of the step is found with the ndold pointer.
c
c     if hyper-diffusivity is used, the iterate is a 2-vector.  number
c     density is the first quantity, and the auxiliary scalar, which is
c     proportional to grad**2(n) is the second quantity.
c-----------------------------------------------------------------------
      IF (nd_hypd>0) THEN
        fhyp=SQRT(nd_hypd*dt)
        CALL generic_all_eval(rb%work6,tb%work6,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,v2,v2r,v2z,2_i4)
        nd(1,:,:,:)=v2(1,:,:,:)
      ELSE
        CALL generic_all_eval(rb%work2,tb%work2,rb%dxdr,rb%dydr,
     $                        rb%dxdz,rb%dydz,rb%xg,rb%yg,tb%tgeom,
     $                        tb%ng,nd,ndr,ndz,1_i4)
      ENDIF
      CALL generic_ptr_set(rb%qnd,tb%qnd,tb%tgeom,inode,
     $                     ndold,dcp,dcp,0_i4)
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      IF (nd_floor>0.AND.nd_diff>0) THEN
        CALL generic_ptr_set(rb%qdart,tb%qdart,tb%tgeom,
     $                       inode,dart,dp,dp,0_i4)
        kap_art=dart(1,:,:)+dfac
      ENDIF
      IF (nd_dart_upw>0) THEN
        CALL generic_ptr_set(rb%qupw_phi,tb%qupw_phi,tb%tgeom,
     $                       inode,upwc,dp,dp,0_i4)
      ENDIF
      IF (nd_hypd>0) THEN
        DO imode=1,nmodes
          grad_nd(1,:,:,imode)=v2r(1,:,:,imode)
          grad_nd(2,:,:,imode)=v2z(1,:,:,imode)
          grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
        ENDDO
      ELSE IF (nd_diff>0.OR.nd_dart_upw>0) THEN
        DO imode=1,nmodes
          grad_nd(1,:,:,imode)=ndr(1,:,:,imode)
          grad_nd(2,:,:,imode)=ndz(1,:,:,imode)
          grad_nd(3,:,:,imode)=(0,1)*keff(imode)*nd(1,:,:,imode)/bigr
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the nonlinear contributions to n*V.
c     note that the real arrays have three dimensions, where the
c     first is the vector component, the second covers the poloidal
c     plane, and the third is the phi index,
c
c     real_ve already contains any equilibrium flow.
c-----------------------------------------------------------------------
      CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,1_i4,nd,real_scal,
     $             dealiase)
c-----------------------------------------------------------------------
c     include the upwinding diffusive flux that is proportional to
c     VV.grad(n).
c-----------------------------------------------------------------------
      IF (nd_dart_upw>0) THEN
        CALL fft_nim('inverse',ncx*ncy,mpseudo,lphi,3_i4,grad_nd,
     $               real_gn,dealiase)
        fupw =nd_dart_upw*dt**2*upw_aniso
        fupwi=nd_dart_upw*dt**2*(1._r8-upw_aniso)
        DO ip=1,nphi
          DO ix=1,mpseudo
            rdot=fupw*upwc(1,ix,ip)*
     $          SUM(real_ve(:,ix,ip)*real_gn(:,ix,ip)) /
     $        ( SUM(real_ve(:,ix,ip)*real_ve(:,ix,ip)) + smallnum )
            rscl=fupwi*upwc(1,ix,ip)
            real_vec(:,ix,ip)=hdt*real_ve(:,ix,ip)*real_scal(1,ix,ip)-
     $        ( rdot*real_ve(:,ix,ip)+rscl*real_gn(:,ix,ip) )
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c     or just V*dn
c-----------------------------------------------------------------------
      ELSE
        real_vec(1,:,:)=hdt*real_ve(1,:,:)*real_scal(1,:,:)
        real_vec(2,:,:)=hdt*real_ve(2,:,:)*real_scal(1,:,:)
        real_vec(3,:,:)=hdt*real_ve(3,:,:)*real_scal(1,:,:)
      ENDIF
c-----------------------------------------------------------------------
c     transform flux.
c-----------------------------------------------------------------------
      CALL fft_nim('forward',ncx*ncy,mpseudo,lphi,3_i4,nvel,real_vec,
     $             dealiase)
c-----------------------------------------------------------------------
c     begin mode loop.
c-----------------------------------------------------------------------
      mode_loop: DO imode=1,nmodes
c-----------------------------------------------------------------------
c       add the diffusive term -D*grad(n) and the hyper-diffusive
c       part of the flux +D_hyp*grad(f) if needed.
c-----------------------------------------------------------------------
        IF (nd_floor>0.AND.nd_diff>0) THEN
          nvel(1,:,:,imode)=nvel(1,:,:,imode)-
     $                kap_art*grad_nd(1,:,:,imode)
          nvel(2,:,:,imode)=nvel(2,:,:,imode)-
     $                kap_art*grad_nd(2,:,:,imode)
          nvel(3,:,:,imode)=nvel(3,:,:,imode)-
     $                kap_art*grad_nd(3,:,:,imode)
        ELSE IF (nd_diff>0) THEN
          nvel(1,:,:,imode)=nvel(1,:,:,imode)-dfac*grad_nd(1,:,:,imode)
          nvel(2,:,:,imode)=nvel(2,:,:,imode)-dfac*grad_nd(2,:,:,imode)
          nvel(3,:,:,imode)=nvel(3,:,:,imode)-dfac*grad_nd(3,:,:,imode)
        ENDIF
        IF (nd_hypd>0) THEN
          nvel(1,:,:,imode)=nvel(1,:,:,imode)+fhyp*v2r(2,:,:,imode)
          nvel(2,:,:,imode)=nvel(2,:,:,imode)+fhyp*v2z(2,:,:,imode)
          nvel(3,:,:,imode)=nvel(3,:,:,imode)+
     $                   fhyp*v2(2,:,:,imode)*(0,1)*keff(imode)/bigr
        ENDIF
c-----------------------------------------------------------------------
c       construct n-f*dt*grad(alpha*).( n*V - D*grad(n) ).
c
c       if hyper-diffusivity is used, also compute the off-diagonal
c       contributions for +fhyp*grad**2(aux-scalar) in the n rows and
c       -fhyp*grad**2(n) in the aux-scalar rows.
c-----------------------------------------------------------------------
        nd(1,:,:,imode)=  nd(1,:,:,imode)+
     $                  nvel(3,:,:,imode)*(0,1)*keff(imode)/bigr
        IF (nd_hypd>0) THEN
          v2(2,:,:,imode)=v2(2,:,:,imode)-
     $          fhyp*grad_nd(3,:,:,imode)*(0,1)*keff(imode)/bigr
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(alpha(:,:,iv)*  nd(1,:,:,imode)-
     $                           dalpdr(:,:,iv)*nvel(1,:,:,imode)-
     $                           dalpdz(:,:,iv)*nvel(2,:,:,imode),1)
            int(2,:,iv,imode)=SUM(alpha(:,:,iv)*     v2(2,:,:,imode)+
     $                     fhyp*(dalpdr(:,:,iv)*grad_nd(1,:,:,imode)+
     $                           dalpdz(:,:,iv)*grad_nd(2,:,:,imode)),1)
          ENDDO
        ELSE
          DO iv=1,nv
            int(1,:,iv,imode)=SUM(alpha(:,:,iv)*  nd(1,:,:,imode)-
     $                           dalpdr(:,:,iv)*nvel(1,:,:,imode)-
     $                           dalpdz(:,:,iv)*nvel(2,:,:,imode),1)
          ENDDO
        ENDIF
      ENDDO mode_loop
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE cont_dot
c-----------------------------------------------------------------------
c     close module.
c-----------------------------------------------------------------------
      END MODULE integrands_rhs


c-----------------------------------------------------------------------
c     module containing integrands for matrices on the lhs of equations.
c-----------------------------------------------------------------------
      MODULE integrands_mat

      USE local
      USE rblock_type_mod
      USE tblock_type_mod
      USE lagr_quad_mod
      USE tri_linear
      USE generic_evals
      USE math_tran
      USE physdat
      USE input
      USE global
      USE fft_mod
      IMPLICIT NONE

      CONTAINS
c-----------------------------------------------------------------------
c     subprogram 17. get_mass.
c     compute the integrand used for the mass matrix.
c
c     with separate polynomial degree specification for each field,
c     this is only used for the jfromb computation.
c-----------------------------------------------------------------------
      SUBROUTINE get_mass(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(1,1,1) :: dv
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      nv=SIZE(int,5)
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
c-----------------------------------------------------------------------
c     find alpha(iv)*alpha(jv)
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE get_mass
c-----------------------------------------------------------------------
c     subprogram 18. curl_de_iso.
c     compute the integrand for the linearized matrix on the lhs of the 
c     magnetic field equation, where the effective impedance tensor
c     is isotropic.  This is now used for mhd computations only.
c-----------------------------------------------------------------------
      SUBROUTINE curl_de_iso(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: elecd_n0,dp,ds2
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: ziso,bigr2,diff
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      IF (elecd>0) THEN
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                         inode,elecd_n0,dp,dp,0_i4)
          diff=elecd_n0(1,:,:)
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          diff=elecd*MAX(ds2(1,:,:),0._r8)
        ELSE
          diff=elecd
        ENDIF
      ENDIF
      nv=SIZE(int,5)
c-----------------------------------------------------------------------
c     find R**2.
c-----------------------------------------------------------------------
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     evaluate coefficients for the impedance matrix--now just
c     the implicit contribution to resistive dissipation, where the
c     electrical diffusivity is modulated by the shape function diff.
c-----------------------------------------------------------------------
      IF (elecd>0) THEN
        ziso=dt*feta*diff
      ELSE
        ziso=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     compute the eta/mu0Xcurl(alpha_iv).curl(alpha_jv) operator.
c     Components are cylindrical.
c     note:  vector storage is (real1,real2,-imag3) or
c     (imag1,imag2,real3) and jmode is set in the calling routine.
c
c     if div(b) cleaning is not time split, add the grad(div) operator,
c     too.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb) THEN
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                          ziso*(dalpdz(:,:,iv)*dalpdz(:,:,jv)
     $              +  alpha(:,:,iv)*  alpha(:,:,jv)*k2ef(jmode)/bigr2)
     $              + dt*fdivb*divbd*dalpdrc(:,:,iv)*dalpdrc(:,:,jv),1)

            int(2,1,:,jv,iv)=SUM(
     $                         -ziso* dalpdz(:,:,iv)*dalpdr (:,:,jv)
     $              + dt*fdivb*divbd*dalpdrc(:,:,iv)*dalpdz (:,:,jv),1)

            int(1,2,:,iv,jv)=int(2,1,:,jv,iv)
            int(3,1,:,jv,iv)=SUM(
     $                          ziso*  alpha(:,:,iv)*keff(jmode)/bigr
     $                              *dalpdrc(:,:,jv)
     $              + dt*fdivb*divbd*dalpdrc(:,:,iv)
     $                              * alpha (:,:,jv)*keff(jmode)/bigr,1)

            int(1,3,:,iv,jv)=int(3,1,:,jv,iv)

            int(2,2,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                          ziso*(dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $              +  alpha(:,:,iv)*  alpha(:,:,jv)*k2ef(jmode)/bigr2)
     $              + dt*fdivb*divbd*dalpdz (:,:,iv)*dalpdz(:,:,jv),1)

            int(3,2,:,jv,iv)=SUM(
     $                          ziso*  alpha(:,:,iv)*keff(jmode)/bigr
     $                              * dalpdz(:,:,jv)
     $              + dt*fdivb*divbd*dalpdz (:,:,iv)
     $                              * alpha (:,:,jv)*keff(jmode)/bigr,1)

            int(2,3,:,iv,jv)=int(3,2,:,jv,iv)

            int(3,3,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                          ziso*(dalpdz(:,:,iv)* dalpdz(:,:,jv)
     $                              +dalpdrc(:,:,iv)*dalpdrc(:,:,jv))
     $              + dt*fdivb*divbd*alpha(:,:,iv)
     $                              *alpha(:,:,jv)*k2ef(jmode)/bigr2,1)
          ENDDO
        ENDDO
c-----------------------------------------------------------------------
c     same but no div(b) cleaning.
c-----------------------------------------------------------------------
      ELSE
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                        ziso*(dalpdz(:,:,iv)*dalpdz(:,:,jv)
     $              +  alpha(:,:,iv)*alpha(:,:,jv)*k2ef(jmode)/bigr2),1)
            int(2,1,:,jv,iv)=SUM(
     $                       -ziso* dalpdz(:,:,iv)*dalpdr(:,:,jv),1)
            int(1,2,:,iv,jv)=int(2,1,:,jv,iv)
            int(3,1,:,jv,iv)=SUM(
     $                        ziso*  alpha(:,:,iv)*keff(jmode)/bigr
     $                            *dalpdrc(:,:,jv),1)
            int(1,3,:,iv,jv)=int(3,1,:,jv,iv)

            int(2,2,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                        ziso*(dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $              +  alpha(:,:,iv)*alpha(:,:,jv)*k2ef(jmode)/bigr2),1)
            int(3,2,:,jv,iv)=SUM(
     $                        ziso*  alpha(:,:,iv)*keff(jmode)/bigr
     $                             *dalpdz(:,:,jv),1)
            int(2,3,:,iv,jv)=int(3,2,:,jv,iv)

            int(3,3,:,jv,iv)=SUM( alpha(:,:,iv)*alpha(:,:,jv)+
     $                        ziso*(dalpdz(:,:,iv)* dalpdz(:,:,jv)
     $                            +dalpdrc(:,:,iv)*dalpdrc(:,:,jv)),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE curl_de_iso
c-----------------------------------------------------------------------
c     subprogram 19. curl_de_ciso.
c     compute the integrand for the symmetric part of the resistive
c     diffusion operator for magnetic field.  the matrix is used
c     for preconditioning matrix-free iterations for 3d resistivity.
c-----------------------------------------------------------------------
      SUBROUTINE curl_de_ciso(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: elecd_n0,dp
      REAL(r8), DIMENSION(1,1,1) :: dv
      REAL(r8) :: ziso,bigr2,massfac
      INTEGER(i4) :: iv,jv,nv,ix,iy,ncx,ncy
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and the symmetric
c     part of the electrical diffusivity.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                     inode,elecd_n0,dp,dp,0_i4)
      nv=SIZE(int,5)
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
c-----------------------------------------------------------------------
c     compute the curl(alpha_i)*Z*curl(alpha_j) operator.
c     Poloidal components are either Cartesian or cylindrical.
c     note:  jmode is set in the calling routine.
c
c     multiply electrical diffusivity by the centering parameter and dt.
c
c     if div(b) cleaning is not time split, add the grad(div) operator.
c-----------------------------------------------------------------------
      IF (divbd>0.AND..NOT.split_divb) THEN
       DO iv=1,nv
        DO jv=1,nv
         DO iy=1,ncy
          int(:  ,1,iy,jv,iv)=0._r8
          int(2:3,2,iy,jv,iv)=0._r8
          int(3  ,3,iy,jv,iv)=0._r8
          DO ix=1,ncx
            massfac=alpha(ix,iy,iv)*alpha(ix,iy,jv)
            ziso=dt*feta*elecd_n0(1,ix,iy)
            bigr2=bigr(ix,iy)*bigr(ix,iy)
            int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+massfac+
     $                  ziso*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $        +massfac*k2ef(jmode)/bigr2)
     $        +dt*fdivb*divbd*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
            int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $                  ziso*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)
     $        +dt*fdivb*divbd*dalpdrc(ix,iy,iv)*dalpdz (ix,iy,jv)
            int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)+
     $                  (0,1)*keff(jmode)/bigr(ix,iy)*
     $        (          ziso*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv)
     $        +dt*fdivb*divbd*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv))

            int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+massfac+
     $                  ziso*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $          +massfac*k2ef(jmode)/bigr2)
     $          +dt*fdivb*divbd*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
            int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)+
     $                  (0,1)*keff(jmode)/bigr(ix,iy)*
     $        (          ziso*dalpdz(ix,iy,jv)*alpha(ix,iy,iv)
     $        +dt*fdivb*divbd*dalpdz(ix,iy,iv)*alpha(ix,iy,jv))

            int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+massfac+ziso*
     $        (dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
     $        +dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv))
     $        +dt*fdivb*divbd*massfac*
     $            k2ef(jmode)/bigr2
          ENDDO
          int(1,2,iy,iv,jv)=CONJG(int(2,1,iy,jv,iv))
          int(1,3,iy,iv,jv)=CONJG(int(3,1,iy,jv,iv))
          int(2,3,iy,iv,jv)=CONJG(int(3,2,iy,jv,iv))
         ENDDO
        ENDDO
       ENDDO
c-----------------------------------------------------------------------
c     same without div(b) cleaning.
c-----------------------------------------------------------------------
      ELSE
       DO iv=1,nv
        DO jv=1,nv
         DO iy=1,ncy
          int(:  ,1,iy,jv,iv)=0._r8
          int(2:3,2,iy,jv,iv)=0._r8
          int(3  ,3,iy,jv,iv)=0._r8
          DO ix=1,ncx
            massfac=alpha(ix,iy,iv)*alpha(ix,iy,jv)
            ziso=dt*feta*elecd_n0(1,ix,iy)
            bigr2=bigr(ix,iy)*bigr(ix,iy)
            int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+massfac+
     $                   ziso*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                  +massfac*k2ef(jmode)/bigr2)
            int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $                   ziso*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)
            int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)+
     $                   (0,1)*keff(jmode)/bigr(ix,iy)*
     $                   ziso*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv)

            int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+massfac+
     $                   ziso*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                  +massfac*k2ef(jmode)/bigr2)
            int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)+
     $                   (0,1)*keff(jmode)/bigr(ix,iy)*
     $                   ziso*dalpdz(ix,iy,jv)*alpha(ix,iy,iv)

            int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+massfac+ziso*
     $        (dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
     $        +dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv))
          ENDDO
          int(1,2,iy,iv,jv)=CONJG(int(2,1,iy,jv,iv))
          int(1,3,iy,iv,jv)=CONJG(int(3,1,iy,jv,iv))
          int(2,3,iy,iv,jv)=CONJG(int(3,2,iy,jv,iv))
         ENDDO
        ENDDO
       ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE curl_de_ciso
c----------------------------------------------------------------------
c     subprogram 20. b_hmhd_op.
c
c     find the diagonal-in-Fourier-index part of the implicit 
c     operator used for the combined Hall-MHD advance of B.  this is
c     used as the implicit operator for linear HMHD computations and for
c     the preconditioner matrix with nonlinear HMHD.  In addition, part
c     of it can be used to reduce the number of matrix-free computations
c     in nonlinear advances.
c
c     the constructed operator is curl(A*).de-div(A*)div(db), where A 
c     is a test vector, de is
c
c       de=[dj X (Beq+b0) + (Jeq+j0) X db]/ne - (Veq+v0) X db + eta*dj
c
c     and the div(db) terms are for divergence cleaning.
c
c     this routine is also called for implicit advection for MHD; two-
c     fluid terms are switched-off according to integrand_flag.
c-----------------------------------------------------------------------
      SUBROUTINE b_hmhd_op(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: aldis,dalddr,dalddz

      REAL(r8), DIMENSION(:,:,:), POINTER :: b0,b0r,b0z,v0,nd_eq,n0,beq,
     $          jeq,veq,elecd_n0,ds2,nl_pres,dp
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be,ja
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: temp1,ziso

      COMPLEX(r8), DIMENSION(3,3,SIZE(bigr,1),SIZE(bigr,2)) :: curlxb0

      REAL(r8) :: fdb,hdt,hfac,bigr2,ziss,krow,massfac,disdbfac,fhyp,
     $            fhdb
      INTEGER(i4) :: iv,jv,nv,ix,iy,ncx,ncy,ioff,nvc,nvd,ivd,jvd
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions
c
c     if poly_divb is non-negative, divergence error is controlled with
c     an auxiliary field that is discontinuous at element borders, and
c     these bases are also needed.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      IF (poly_divb>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlmat',aldis,
     $                          dalddr,dalddz,0_i4,poly_divb,
     $                          polydmin=poly_divb_min,
     $                          polydmax=poly_divb_max)
      ENDIF
      nvc=ncontb
      nvd=ndiscb
      nv=nvc+nvd
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
c-----------------------------------------------------------------------
c     dissipation coefficients.
c-----------------------------------------------------------------------
      IF (elecd>0) THEN
        IF (eta_model/='fixed') THEN
          CALL generic_ptr_set(rb%qelecd_n0,tb%qelecd_n0,tb%tgeom,
     $                         inode,elecd_n0,dp,dp,0_i4)
          ziso=dt*feta*elecd_n0(1,:,:)
        ELSE IF (ds_use=='elecd'.OR.ds_use=='both') THEN
          CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
          ziso=dt*feta*elecd*MAX(ds2(1,:,:),0._r8)
        ELSE
          ziso=dt*feta*elecd
        ENDIF
      ELSE
        ziso=0
      ENDIF
      IF (divbd>0.AND..NOT.split_divb) THEN
        fdb=dt*fdivb*divbd
      ELSE
        fdb=0
      ENDIF
      disdbfac=SQRT(disc_dbd*dt*fdivb)  !  diffusive correction
c-----------------------------------------------------------------------
c     coefficients for the Hall term and the partial(J)/partial(t)
c     electron inertia term.
c-----------------------------------------------------------------------
      hdt=0.5*dt
      IF (integrand_flag=='mhd') THEN
        hfac=0
      ELSE
        hfac=hdt*(1-meomi)/(elementary_q*(1+meomi))
      ENDIF
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      IF (nonlinear.AND.
     $    (continuity=='full'.OR.continuity=='n=0 only')) THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                       n0,dp,dp,0_i4)
        temp1=hfac/(nd_eq(1,:,:)+n0(1,:,:))
        IF (ohms=='2fl') 
     $      ziso=ziso+si_fac_hall/(mu0*(nd_eq(1,:,:)+n0(1,:,:))*
     $                   (qs(1)**2/ms(1)+qs(2)**2/(zeff*ms(2))))
      ELSE
        temp1=hfac/nd_eq(1,:,:)
        IF (ohms=='2fl'.AND.nonlinear) THEN
          ziso=ziso+si_fac_hall/(mu0*nd_eq(1,:,:)*
     $                 (qs(1)**2/ms(1)+qs(2)**2/(zeff*ms(2))))
        ELSE IF (ohms=='2fl') THEN
          ziso=ziso+1._r8/(mu0*nd_eq(1,:,:)*
     $                 (qs(1)**2/ms(1)+qs(2)**2/(zeff*ms(2))))
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     evaluate fields needed for the implicit operator.
c     note that be->0.5*dt*B/mu0*ne and ja->0.5*dt*(J/ne - V)
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,beq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                     inode,jeq,dp,dp,0_i4)
      IF (nonlinear) THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,inode,
     $                       b0,b0r,b0z,1_i4)
        CALL generic_ptr_set(rb%qve_n0,tb%qve_n0,tb%tgeom,inode,
     $                       v0,dp,dp,0_i4)
        be(1,:,:)=temp1*(b0(1,:,:)+beq(1,:,:))/mu0
        be(2,:,:)=temp1*(b0(2,:,:)+beq(2,:,:))/mu0
        be(3,:,:)=temp1*(b0(3,:,:)+beq(3,:,:))/mu0
        ja(1,:,:)=temp1*(jeq(1,:,:)+b0z(3,:,:)/mu0)-hdt*v0(1,:,:)
        IF (geom=='tor') THEN
          ja(2,:,:)=temp1*(jeq(2,:,:)-(b0r(3,:,:)+b0(3,:,:)/bigr)/mu0)-
     $                    hdt*v0(2,:,:)
        ELSE
          ja(2,:,:)=temp1*(jeq(2,:,:)-b0r(3,:,:)/mu0)-hdt*v0(2,:,:)
        ENDIF
        ja(3,:,:)=temp1*(jeq(3,:,:)+(b0r(2,:,:)-b0z(1,:,:))/mu0)-
     $                  hdt*v0(3,:,:)
      ELSE
        be(1,:,:)=temp1*beq(1,:,:)/mu0
        be(2,:,:)=temp1*beq(2,:,:)/mu0
        be(3,:,:)=temp1*beq(3,:,:)/mu0
        ja(1,:,:)=temp1*jeq(1,:,:)
        ja(2,:,:)=temp1*jeq(2,:,:)
        ja(3,:,:)=temp1*jeq(3,:,:)
      ENDIF 
      IF (eq_flow/='none') THEN
        CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                       inode,veq,dp,dp,0_i4)
        ja=ja-hdt*veq
      ENDIF
c-----------------------------------------------------------------------
c     add a numerical term to the isotropic impedance that is
c     proportional to dt times a cyclotron frequency that is based
c     on the variation of B in the periodic coordinate.
c-----------------------------------------------------------------------
c     IF (nonlinear) THEN
c       CALL generic_ptr_set(rb%qsi_nl_pres,tb%qsi_nl_pres,tb%tgeom,
c    $                       inode,nl_pres,dp,dp,0_i4)
c       IF (continuity=='full'.OR.continuity=='n=0 only') THEN
c         ziso=ziso+si_fac_hall*dt*0.5_r8*nl_pres(1,:,:)/
c    $              (SQRT(SUM((b0+beq)**2,1))
c    $               *elementary_q*mu0*(nd_eq(1,:,:)+n0(1,:,:)))
c       ELSE
c         ziso=ziso+si_fac_hall*dt*0.5_r8*nl_pres(1,:,:)/
c    $              (SQRT(SUM((b0+beq)**2,1))
c    $               *elementary_q*mu0*nd_eq(1,:,:))
c       ENDIF
c     ENDIF
c-----------------------------------------------------------------------
c     to help construct the hall implicit operator, we create
c     a matrix that holds the curl of the three vector test functions
c     and a matrix that holds curl(Ar,Az,Aphi) X be + ja X (Ar,Az,Aphi)
c     the first two indices of the temporary matrices are defined as
c     ( result vector index, operand vector index ), where the vector
c     indices runs from 1 to 3 as (r,z,phi), and the last index is
c     the basis function index.  curl(Ar,Az,Aphi) goes first.
c
c     in this version, curl is not formed, and curlxb0 is within the
c     outer jv loop.
c-----------------------------------------------------------------------
      krow=keff(jmode)
      DO jv=1,nvc
        curlxb0(1,1,:,:)= be(3,:,:)*(0,1)*krow*alpha(:,:,jv)/bigr+
     $                    be(2,:,:)*dalpdz(:,:,jv)
        curlxb0(2,1,:,:)=-be(1,:,:)*dalpdz(:,:,jv)+
     $                    ja(3,:,:)*alpha(:,:,jv)
        curlxb0(3,1,:,:)=-be(1,:,:)*(0,1)*krow*alpha(:,:,jv)/bigr-
     $                    ja(2,:,:)*alpha(:,:,jv)
        curlxb0(1,2,:,:)=-be(2,:,:)*dalpdr(:,:,jv)-
     $                    ja(3,:,:)*alpha(:,:,jv)
        curlxb0(2,2,:,:)= be(1,:,:)*dalpdr(:,:,jv)+
     $                    be(3,:,:)*(0,1)*krow*alpha(:,:,jv)/bigr
        curlxb0(3,2,:,:)=-be(2,:,:)*(0,1)*krow*alpha(:,:,jv)/bigr+
     $                    ja(1,:,:)*alpha(:,:,jv)
        curlxb0(1,3,:,:)=-be(3,:,:)*dalpdrc(:,:,jv)+
     $                    ja(2,:,:)*alpha(:,:,jv)
        curlxb0(2,3,:,:)=-be(3,:,:)*dalpdz(:,:,jv)-
     $                    ja(1,:,:)*alpha(:,:,jv)
        curlxb0(3,3,:,:)= be(2,:,:)*dalpdz(:,:,jv)+
     $                    be(1,:,:)*dalpdrc(:,:,jv)
c-----------------------------------------------------------------------
c     compute the implicit hmhd operator.
c
c       CC(curl(A)).( (curl(dB))xB0/mu*ne + (J0/ne-V0)xdB + eta*dJ )
c     
c     where A is the vector test function, whose indices correspond
c     to row indices. the diagonal term is added elsewhere.
c-----------------------------------------------------------------------
       DO iv=1,nvc
        DO iy=1,ncy
         int(:,:,iy,jv,iv)=0._r8
         DO ix=1,ncx
          massfac=alpha(ix,iy,iv)*alpha(ix,iy,jv)
          bigr2=bigr(ix,iy)*bigr(ix,iy)
          ziss=ziso(ix,iy)

          int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+massfac-
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,1,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,1,ix,iy)+
     $        fdb*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)+
     $        ziss*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $              +massfac*k2ef(jmode)/bigr2)

          int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,2,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,2,ix,iy)+
     $        fdb*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)-
     $        ziss*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)             

          int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)-
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,3,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,3,ix,iy)+
     $        (0,1)*krow/bigr(ix,iy)*
     $        (fdb *dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)+
     $         ziss*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv))

          int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv)+
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,1,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,1,ix,iy)+
     $        fdb*dalpdrc(ix,iy,jv)*dalpdz (ix,iy,iv)-
     $        ziss*dalpdz(ix,iy,jv)*dalpdr(ix,iy,iv)             

          int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+massfac+
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,2,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,2,ix,iy)+
     $        fdb*  dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $        ziss*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $              +massfac*k2ef(jmode)/bigr2)

          int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)+
     $        (0,1)*krow*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,3,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,3,ix,iy)+
     $        (0,1)*krow/bigr(ix,iy)*
     $        (fdb* dalpdz(ix,iy,iv)*alpha(ix,iy,jv)+
     $         ziss*dalpdz(ix,iy,jv)*alpha(ix,iy,iv))

          int(1,3,iy,jv,iv)=int(1,3,iy,jv,iv)+
     $        dalpdz (ix,iy,iv)*curlxb0(1,1,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,1,ix,iy)-
     $        (0,1)*krow/bigr(ix,iy)*
     $        (fdb *dalpdrc(ix,iy,jv)*alpha(ix,iy,iv)+
     $         ziss*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv))

          int(2,3,iy,jv,iv)=int(2,3,iy,jv,iv)+
     $        dalpdz (ix,iy,iv)*curlxb0(1,2,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,2,ix,iy)-
     $        (0,1)*krow/bigr(ix,iy)*
     $        (fdb* dalpdz(ix,iy,jv)*alpha(ix,iy,iv)+
     $         ziss*dalpdz(ix,iy,iv)*alpha(ix,iy,jv))

          int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+massfac+
     $        dalpdz (ix,iy,iv)*curlxb0(1,3,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,3,ix,iy)+
     $        fdb*massfac*k2ef(jmode)/bigr2+
     $        ziss*(dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
     $             +dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv))
         ENDDO
        ENDDO
       ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     add contributions from the numerical hyper-resistivity if it
c     is used without a split step.  in the rows for dB, the following
c     adds
c
c       -fhyp*curl(A^*).curl(D)
c       -fhdb* div(A^*)* div(D)
c
c     and the rows for the auxiliary vector D are
c
c       (C^*).D + fhyp*curl(C^*).curl(dB)
c               + fhdb* div(C^*)* div(dB)
c
c     where A is the test vector for dB and C is the test vector for D.
c-----------------------------------------------------------------------
      IF ((hyp_eta>0.OR.hyp_dbd>0).AND..NOT.split_hypeta) THEN
        fhyp=SQRT(hyp_eta*fhyp_eta*dt)
        fhdb=SQRT(hyp_dbd*fhyp_dbd*dt)
        DO iv=1,nvc
         DO jv=1,nvc
          DO iy=1,ncy
           DO ix=1,ncx
             massfac=alpha(ix,iy,iv)*alpha(ix,iy,jv)
             bigr2=bigr(ix,iy)*bigr(ix,iy)

             int(4,1,iy,jv,iv)=int(4,1,iy,jv,iv)-
     $           fhyp*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)-
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(5,1,iy,jv,iv)=int(5,1,iy,jv,iv)+
     $           fhyp*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)-
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(6,1,iy,jv,iv)=int(6,1,iy,jv,iv)-
     $           (fhyp*alpha(ix,iy,iv)*dalpdrc(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdrc(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)

             int(4,2,iy,jv,iv)=int(4,2,iy,jv,iv)+
     $           fhyp*dalpdr(ix,iy,iv)*dalpdz(ix,iy,jv)-
     $           fhdb*dalpdz(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(5,2,iy,jv,iv)=int(5,2,iy,jv,iv)-
     $           fhyp*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)-
     $           fhdb*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(6,2,iy,jv,iv)=int(6,2,iy,jv,iv)-
     $           (fhyp*alpha(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdz(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)

             int(4,3,iy,jv,iv)=int(4,3,iy,jv,iv)+
     $           (fhyp*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(5,3,iy,jv,iv)=int(5,3,iy,jv,iv)+
     $           (fhyp*dalpdz(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdz(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(6,3,iy,jv,iv)=int(6,3,iy,jv,iv)-
     $           fhyp*( dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv)
     $                 +dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv))-
     $           fhdb*massfac*k2ef(jmode)/bigr2

             int(1,4,iy,jv,iv)=int(1,4,iy,jv,iv)+
     $           fhyp*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)+
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(2,4,iy,jv,iv)=int(2,4,iy,jv,iv)-
     $           fhyp*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)+
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(3,4,iy,jv,iv)=int(3,4,iy,jv,iv)+
     $           (fhyp*alpha(ix,iy,iv)*dalpdrc(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdrc(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(4,4,iy,jv,iv)=int(4,4,iy,jv,iv)+massfac

             int(1,5,iy,jv,iv)=int(1,5,iy,jv,iv)-
     $           fhyp*dalpdr(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $           fhdb*dalpdz(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(2,5,iy,jv,iv)=int(2,5,iy,jv,iv)+
     $           fhyp*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)+
     $           fhdb*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(3,5,iy,jv,iv)=int(3,5,iy,jv,iv)+
     $           (fhyp*alpha(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdz(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(5,5,iy,jv,iv)=int(5,5,iy,jv,iv)+massfac

             int(1,6,iy,jv,iv)=int(1,6,iy,jv,iv)-
     $           (fhyp*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(2,6,iy,jv,iv)=int(2,6,iy,jv,iv)-
     $           (fhyp*dalpdz(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdz(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(3,6,iy,jv,iv)=int(3,6,iy,jv,iv)+
     $           fhyp*( dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv)
     $                 +dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv))+
     $           fhdb*massfac*k2ef(jmode)/bigr2
             int(6,6,iy,jv,iv)=int(6,6,iy,jv,iv)+massfac
           ENDDO
          ENDDO
         ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     find the contributions associated with the rows and columns of the
c     discontinuous scalar for divb control.
c     
c     the equations are expressed directly in weak form:
c
c       int[A^*.delta(B)] = -dt*int[curl(A^*).E]
c               - int[ sqrt(disc_dbd*dt*f)*auxb*div(A^*) ]
c       int[(phi^*)*auxb]-int[sqrt(disc_dbd*dt*f)*(phi^*)*div(delta(B))]
c         = int[ sqrt(disc_dbd*nu*dt/f)*(phi^*)*div(B_old)]
c
c     where f is the implicit centering and phi is the scalar test
c     field for the auxiliary field auxb.
c-----------------------------------------------------------------------
      IF (poly_divb>=0) THEN
        DO iv=1,nvc
          DO jv=nvc+1,nv
            jvd=jv-nvc
            DO iy=1,ncy
              int(:,:,iy,jv,iv)=0._r8
              DO ix=1,ncx
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+
     $            disdbfac*dalpdrc(ix,iy,iv)*aldis(ix,iy,jvd)
                int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv)+
     $            disdbfac*dalpdz(ix,iy,iv)*aldis(ix,iy,jvd)
                int(1,3,iy,jv,iv)=int(1,3,iy,jv,iv)-(0,1)*krow*
     $            disdbfac*alpha(ix,iy,iv)*aldis(ix,iy,jvd)/bigr(ix,iy)
              ENDDO
            ENDDO
          ENDDO
        ENDDO
        DO iv=nvc+1,nv
          ivd=iv-nvc
          DO jv=1,nvc
            DO iy=1,ncy
              int(:,1,iy,jv,iv)=0._r8
              DO ix=1,ncx
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)-
     $            disdbfac*dalpdrc(ix,iy,jv)*aldis(ix,iy,ivd)
                int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $            disdbfac*dalpdz(ix,iy,jv)*aldis(ix,iy,ivd)
                int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)-(0,1)*krow*
     $            disdbfac*alpha(ix,iy,jv)*aldis(ix,iy,ivd)/bigr(ix,iy)
              ENDDO
            ENDDO
          ENDDO
          DO jv=nvc+1,nv
            jvd=jv-nvc
            DO iy=1,ncy
              int(1,1,iy,jv,iv)=0._r8
              DO ix=1,ncx
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+
     $            aldis(ix,iy,jvd)*aldis(ix,iy,ivd)
              ENDDO
            ENDDO
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE b_hmhd_op
c----------------------------------------------------------------------
c     subprogram 21. b_hmhd_cpop.
c
c     find an off diagonal in Fourier index part of the implicit 
c     operator used for the combined Hall-MHD advance of B.  this is
c     used for SSOR steps while preconditioning the implicit 3D
c     system defined by the b_hmhd_dot routine.
c
c     the constructed operator is curl(A*).de-div(A*)div(db), where A 
c     is a test vector, de is
c
c       de=[dj X (Beq+b0) + (Jeq+j0) X db]/ne - (Veq+v0) X db + eta*dj
c
c     there are no off diagonal divb terms and off diagonal eta*dj only
c     occurs with 3D resistivity.  similarly, the electron inertia term
c     has off diagonal terms only with full continuity.
c-----------------------------------------------------------------------
      SUBROUTINE b_hmhd_cpop(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc

      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,n0,b0,b0r,b0z,beq,
     $          jeq,dp
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: temp1

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: ndp,bep,jap,vep,dcp
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: ziso
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be,ja
      COMPLEX(r8), DIMENSION(3,3,SIZE(bigr,1),SIZE(bigr,2)) :: curlxb0

      REAL(r8) :: hdt,hfac,bigr2,kcol
      COMPLEX(r8) :: ziss
      INTEGER(i4) :: iv,jv,nv,ix,iy,ncx,ncy,lintf,foff,fcol
      LOGICAL :: useconj
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nv=SIZE(int,5)
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
c-----------------------------------------------------------------------
c     determine Fourier index offset relative to the diagonal.  note
c     that the Fourier row index is nindex(jmode).
c-----------------------------------------------------------------------
      lintf=LEN_TRIM(integrand_flag)
      READ(integrand_flag(lintf-2:lintf),'(i3)') foff
      IF (integrand_flag(1:10)=='bhmhd plus') THEN
        useconj=.true.
        fcol=nindex(jmode)+foff
        IF (fcol>nmodes_total-1) THEN
          int=0._r8
          RETURN
        ELSE
          kcol=keff_total(fcol+1)
        ENDIF
      ELSE
        useconj=.false.
        fcol=nindex(jmode)-foff
        IF (fcol<0) THEN
          fcol=-fcol
          kcol=-keff_total(fcol+1)
        ELSE
          kcol=keff_total(fcol+1)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     coefficients for the Hall term and the partial(J)/partial(t)
c     electron inertia term, where a binomial approximation is
c     used for the denominator.
c-----------------------------------------------------------------------
      hdt=0.5*dt
      IF (integrand_flag=='mhd') THEN
        hfac=0
      ELSE
        hfac=hdt*(1-meomi)/(elementary_q*(1+meomi))
      ENDIF
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      IF (continuity=='full'.OR.continuity=='n=0 only') THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,inode,
     $                       n0,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qnd_off,tb%qnd_off,tb%tgeom,inode,
     $                       ndp,dcp,dcp,0_i4)
        temp1=hfac/(nd_eq(1,:,:)+n0(1,:,:))
        IF (ohms=='2fl'.AND.continuity=='full') THEN
            ziso=-ndp(1,:,:,foff)/(mu0*(nd_eq(1,:,:)+n0(1,:,:))**2*
     $                   (qs(1)**2/ms(1)+qs(2)**2/(zeff*ms(2))))
        ELSE
          ziso=0._r8
        ENDIF
      ELSE
        temp1=hfac/nd_eq(1,:,:)
        ziso=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     evaluate fields needed for the implicit operator.
c     note that be->0.5*dt*B/mu0*ne and ja->0.5*dt*(J/ne - V)
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_off,tb%qbe_off,tb%tgeom,inode,
     $                     bep,dcp,dcp,0_i4)
      CALL generic_ptr_set(rb%qja_off,tb%qja_off,tb%tgeom,inode,
     $                     jap,dcp,dcp,0_i4)
      CALL generic_ptr_set(rb%qve_off,tb%qve_off,tb%tgeom,inode,
     $                     vep,dcp,dcp,0_i4)
      be(1,:,:)=temp1*bep(1,:,:,foff)/mu0
      be(2,:,:)=temp1*bep(2,:,:,foff)/mu0
      be(3,:,:)=temp1*bep(3,:,:,foff)/mu0
      ja(1,:,:)=temp1*jap(1,:,:,foff)-hdt*vep(1,:,:,foff)
      ja(2,:,:)=temp1*jap(2,:,:,foff)-hdt*vep(2,:,:,foff)
      ja(3,:,:)=temp1*jap(3,:,:,foff)-hdt*vep(3,:,:,foff)
c-----------------------------------------------------------------------
c     include quadratic perturbed-n terms if continuity is full.
c     note that temp1 changes to hfac**2/nd-sym**2
c-----------------------------------------------------------------------
      IF (ohms/='mhd'.AND.continuity=='full') THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,inode,
     $                       b0,b0r,b0z,1_i4)
        CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                       inode,beq,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                       inode,jeq,dp,dp,0_i4)
        temp1=temp1**2
        be(1,:,:)=be(1,:,:)-
     $    temp1*ndp(1,:,:,foff)*(b0(1,:,:)+beq(1,:,:))/(hfac*mu0)
        be(2,:,:)=be(2,:,:)-
     $    temp1*ndp(1,:,:,foff)*(b0(2,:,:)+beq(2,:,:))/(hfac*mu0)
        be(3,:,:)=be(3,:,:)-
     $    temp1*ndp(1,:,:,foff)*(b0(3,:,:)+beq(3,:,:))/(hfac*mu0)
        ja(1,:,:)=ja(1,:,:)-temp1*ndp(1,:,:,foff)/hfac*
     $                      (jeq(1,:,:)+b0z(3,:,:)/mu0)
        IF (geom=='tor') THEN
          ja(2,:,:)=ja(2,:,:)-temp1*ndp(1,:,:,foff)/hfac*
     $                      (jeq(2,:,:)-(b0r(3,:,:)+b0(3,:,:)/bigr)/mu0)
        ELSE
          ja(2,:,:)=ja(2,:,:)-temp1*ndp(1,:,:,foff)/hfac*
     $                      (jeq(2,:,:)-b0r(3,:,:)/mu0)
        ENDIF
        ja(3,:,:)=ja(3,:,:)-temp1*ndp(1,:,:,foff)/hfac*
     $                      (jeq(3,:,:)+(b0r(2,:,:)-b0z(1,:,:))/mu0)
      ENDIF
c-----------------------------------------------------------------------
c     the conjugate of the Fourier coefficients appearing in coupling
c     terms is needed for contributions from n>nindex(jmode).
c-----------------------------------------------------------------------
      IF (useconj) THEN
        be=CONJG(be)
        ja=CONJG(ja)
        IF (ohms=='2fl') ziso=CONJG(ziso)
      ENDIF
c-----------------------------------------------------------------------
c     to help construct the hall implicit operator, we create
c     a matrix that holds the curl of the three vector test functions
c     and a matrix that holds curl(Ar,Az,Aphi) X be + ja X (Ar,Az,Aphi)
c     the first two indices of the temporary matrices are defined as
c     ( result vector index, operand vector index ), where the vector
c     indices runs from 1 to 3 as (r,z,phi), and the last index is
c     the basis function index.  curl(Ar,Az,Aphi) goes first.
c
c     the wavenumber of the operand is kcol and the wavenumber of the
c     result is keff(jmode).  here, they are not the same, and this
c     affects the curlxb0 term.
c
c     the curl is no longer saved explicitly and curlxb0 is now in the
c     outer jv loop.
c-----------------------------------------------------------------------
      DO jv=1,nv
        curlxb0(1,1,:,:)= be(3,:,:)*(0,1)*kcol*alpha(:,:,jv)/bigr+
     $                    be(2,:,:)*dalpdz(:,:,jv)
        curlxb0(2,1,:,:)=-be(1,:,:)*dalpdz(:,:,jv)+
     $                    ja(3,:,:)* alpha(:,:,jv)
        curlxb0(3,1,:,:)=-be(1,:,:)*(0,1)*kcol*alpha(:,:,jv)/bigr-
     $                    ja(2,:,:)* alpha(:,:,jv)
        curlxb0(1,2,:,:)=-be(2,:,:)*dalpdr(:,:,jv)-
     $                    ja(3,:,:)* alpha(:,:,jv)
        curlxb0(2,2,:,:)= be(1,:,:)*dalpdr(:,:,jv)+
     $                    be(3,:,:)*(0,1)*kcol*alpha(:,:,jv)/bigr
        curlxb0(3,2,:,:)=-be(2,:,:)*(0,1)*kcol*alpha(:,:,jv)/bigr+
     $                    ja(1,:,:)* alpha(:,:,jv)
        curlxb0(1,3,:,:)=-be(3,:,:)*dalpdrc(:,:,jv)+
     $                    ja(2,:,:)* alpha(:,:,jv)
        curlxb0(2,3,:,:)=-be(3,:,:)*dalpdz(:,:,jv)-
     $                    ja(1,:,:)* alpha(:,:,jv)
        curlxb0(3,3,:,:)= be(2,:,:)*dalpdz(:,:,jv)+
     $                    be(1,:,:)*dalpdrc(:,:,jv)
c-----------------------------------------------------------------------
c     compute the implicit hmhd operator.
c
c       CC(curl(A)).( (curl(dB))xB0/mu*ne + (J0/ne-V0)xdB + eta*dJ )
c     
c     where A is the vector test function, whose indices correspond
c     to row indices. the diagonal term is added elsewhere.
c-----------------------------------------------------------------------
       DO iv=1,nv
        DO iy=1,ncy
         int(:,:,iy,jv,iv)=0._r8
         DO ix=1,ncx
          bigr2=bigr(ix,iy)*bigr(ix,iy)
          ziss=ziso(ix,iy)

          int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)-
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,1,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,1,ix,iy)+
     $        ziss*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $              +alpha(ix,iy,iv)*alpha (ix,iy,jv)*
     $               keff(jmode)*kcol/bigr2)

          int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,2,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,2,ix,iy)-
     $        ziss*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)             

          int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)-
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(2,3,ix,iy)-
     $        dalpdz(ix,iy,iv)*curlxb0(3,3,ix,iy)+
     $        (0,1)*keff(jmode)/bigr(ix,iy)*
     $        ziss*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv)

          int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv)+
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,1,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,1,ix,iy)-
     $        ziss*dalpdz(ix,iy,jv)*dalpdr(ix,iy,iv)             

          int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,2,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,2,ix,iy)+
     $        ziss*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $              +alpha(ix,iy,iv)*alpha (ix,iy,jv)*
     $               keff(jmode)*kcol/bigr2)

          int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)+
     $        (0,1)*keff(jmode)*alpha(ix,iy,iv)/bigr(ix,iy)
     $                        *curlxb0(1,3,ix,iy)+
     $        dalpdr(ix,iy,iv)*curlxb0(3,3,ix,iy)+
     $        (0,1)*keff(jmode)/bigr(ix,iy)*
     $        ziss*dalpdz(ix,iy,jv)*alpha(ix,iy,iv)

          int(1,3,iy,jv,iv)=int(1,3,iy,jv,iv)+
     $        dalpdz (ix,iy,iv)*curlxb0(1,1,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,1,ix,iy)-
     $        (0,1)*kcol/bigr(ix,iy)*
     $        ziss*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)

          int(2,3,iy,jv,iv)=int(2,3,iy,jv,iv)+
     $        dalpdz (ix,iy,iv)*curlxb0(1,2,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,2,ix,iy)-
     $        (0,1)*kcol/bigr(ix,iy)*
     $        ziss*dalpdz(ix,iy,iv)*alpha(ix,iy,jv)

          int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+
     $        dalpdz (ix,iy,iv)*curlxb0(1,3,ix,iy)-
     $        dalpdrc(ix,iy,iv)*curlxb0(2,3,ix,iy)+
     $        ziss*(dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
     $             +dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv))
         ENDDO
        ENDDO
       ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE b_hmhd_cpop
c----------------------------------------------------------------------
c     subprogram 21.1 b_hyp_op.
c
c     find the implicit operator used for the time-split hyper-
c     diffusion advance of B, where the rows for delta-B are:
c
c       (A^*).dB - fhyp*curl(A^*).curl(D)
c                - fhdb* div(A^*)* div(D)
c
c     and the rows for the auxiliary vector D are
c
c       (C^*).D + fhyp*curl(C^*).curl(dB)
c               + fhdb* div(C^*)* div(dB)
c
c     where A is the test vector for dB and C is the test vector for D.
c-----------------------------------------------------------------------
      SUBROUTINE b_hyp_op(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8) :: fhyp,fhdb,krow,bigr2,massfac
      INTEGER(i4) :: iv,jv,ix,iy,ncx,ncy,nvc
c-----------------------------------------------------------------------
c     set index limits and real constants.
c-----------------------------------------------------------------------
      nvc=ncontb
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      krow=keff(jmode)
      fhyp=SQRT(hyp_eta*fhyp_eta*dt)
      fhdb=SQRT(hyp_dbd*fhyp_dbd*dt)
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
c-----------------------------------------------------------------------
c     loop over basis functions and add contributions for the
c     mass matrices and for the hyper-resistivity operator.
c-----------------------------------------------------------------------
      DO iv=1,nvc
        DO jv=1,nvc
          DO iy=1,ncy
           int(:,:,iy,jv,iv)=0._r8
           DO ix=1,ncx
             massfac=alpha(ix,iy,iv)*alpha(ix,iy,jv)
             bigr2=bigr(ix,iy)*bigr(ix,iy)

             int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+massfac
             int(4,1,iy,jv,iv)=int(4,1,iy,jv,iv)-
     $           fhyp*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)-
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(5,1,iy,jv,iv)=int(5,1,iy,jv,iv)+
     $           fhyp*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)-
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(6,1,iy,jv,iv)=int(6,1,iy,jv,iv)-
     $           (fhyp*alpha(ix,iy,iv)*dalpdrc(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdrc(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)

             int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+massfac
             int(4,2,iy,jv,iv)=int(4,2,iy,jv,iv)+
     $           fhyp*dalpdr(ix,iy,iv)*dalpdz(ix,iy,jv)-
     $           fhdb*dalpdz(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(5,2,iy,jv,iv)=int(5,2,iy,jv,iv)-
     $           fhyp*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)-
     $           fhdb*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(6,2,iy,jv,iv)=int(6,2,iy,jv,iv)-
     $           (fhyp*alpha(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdz(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)

             int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+massfac
             int(4,3,iy,jv,iv)=int(4,3,iy,jv,iv)+
     $           (fhyp*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(5,3,iy,jv,iv)=int(5,3,iy,jv,iv)+
     $           (fhyp*dalpdz(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdz(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(6,3,iy,jv,iv)=int(6,3,iy,jv,iv)-
     $           fhyp*( dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv)
     $                 +dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv))-
     $           fhdb*massfac*k2ef(jmode)/bigr2

             int(1,4,iy,jv,iv)=int(1,4,iy,jv,iv)+
     $           fhyp*(dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)+
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(2,4,iy,jv,iv)=int(2,4,iy,jv,iv)-
     $           fhyp*dalpdz(ix,iy,iv)*dalpdr(ix,iy,jv)+
     $           fhdb*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(3,4,iy,jv,iv)=int(3,4,iy,jv,iv)+
     $           (fhyp*alpha(ix,iy,iv)*dalpdrc(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdrc(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(4,4,iy,jv,iv)=int(4,4,iy,jv,iv)+massfac

             int(1,5,iy,jv,iv)=int(1,5,iy,jv,iv)-
     $           fhyp*dalpdr(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $           fhdb*dalpdz(ix,iy,iv)*dalpdrc(ix,iy,jv)
             int(2,5,iy,jv,iv)=int(2,5,iy,jv,iv)+
     $           fhyp*(dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                 +massfac*k2ef(jmode)/bigr2)+
     $           fhdb*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
             int(3,5,iy,jv,iv)=int(3,5,iy,jv,iv)+
     $           (fhyp*alpha(ix,iy,iv)*dalpdz(ix,iy,jv)+
     $            fhdb*alpha(ix,iy,jv)*dalpdz(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(5,5,iy,jv,iv)=int(5,5,iy,jv,iv)+massfac

             int(1,6,iy,jv,iv)=int(1,6,iy,jv,iv)-
     $           (fhyp*dalpdrc(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdrc(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(2,6,iy,jv,iv)=int(2,6,iy,jv,iv)-
     $           (fhyp*dalpdz(ix,iy,iv)*alpha(ix,iy,jv)+
     $            fhdb*dalpdz(ix,iy,jv)*alpha(ix,iy,iv))*
     $           (0,1)*krow/bigr(ix,iy)
             int(3,6,iy,jv,iv)=int(3,6,iy,jv,iv)+
     $           fhyp*( dalpdz (ix,iy,iv)*dalpdz (ix,iy,jv)
     $                 +dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv))+
     $           fhdb*massfac*k2ef(jmode)/bigr2
             int(6,6,iy,jv,iv)=int(6,6,iy,jv,iv)+massfac
           ENDDO
          ENDDO
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE b_hyp_op
c-----------------------------------------------------------------------
c     subprogram 22. n_iso_op.
c     compute an isotropic operator for the number density advance.
c-----------------------------------------------------------------------
      SUBROUTINE n_iso_op(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: nd_eq,dart,dp
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: bigr2,kprp
      REAL(r8), DIMENSION(1,1,1) :: dv
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: nd,dcp
      INTEGER(i4) :: nv,iv,jv,imode,iq,jq
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      nv=SIZE(int,5)
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     electrons and ions have different coefficients.  this routine
c     may also be used for a diffusive operator for the continuity eqn.
c-----------------------------------------------------------------------
      IF (integrand_flag(1:2)=='n ') THEN
        IF (nonlinear.AND.nd_floor>0.AND.nd_diff>0) THEN
          CALL generic_ptr_set(rb%qdart,tb%qdart,tb%tgeom,
     $                         inode,dart,dp,dp,0_i4)
          kprp=fthc*dt*nd_diff+dart(1,:,:)
        ELSE
          kprp=fthc*dt*nd_diff
        ENDIF
      ELSE IF (integrand_flag(1:6)=='all ti') THEN
        kprp=fthc*dt*k_perpi
      ELSE
        kprp=fthc*dt*k_perpe
      ENDIF
c-----------------------------------------------------------------------
c     Compute grad(alpha(i))*grad(alpha(j)).  it is assumed that
c     k_perp has a factor of gamma-1, so that k_perp is the diffusivity
c     for temperature.
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
           int(1,1,:,jv,iv)=SUM((dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $                          +dalpdz(:,:,iv)*dalpdz(:,:,jv)
     $                          +k2ef(jmode)*alpha(:,:,iv)*alpha(:,:,jv)
     $                          /bigr2)*kprp
     $                          +alpha(:,:,iv)*alpha(:,:,jv),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE n_iso_op
c-----------------------------------------------------------------------
c     subprogram 23. t_aniso_op.
c     compute the symmetric part of the operator for temperature advance
c
c     (n*I.-dt*f*D)
c
c     including an implicit isotropic diffusion operator and the
c     diagonal part of an anisotropic temperature diffusivity.
c     when implicit advection is used, the operator is
c
c     (n*I. + 0.5*dt*n*( V.grad(I.)+(gamma-1)*div(V)*I. ) - dt*f*D)
c
c-----------------------------------------------------------------------
      SUBROUTINE t_aniso_op(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: bb,nd_eq,nd_n0,kappl,
     $          kaprp,be_eq,t_b2,ja_eq,ve_eq,divvq,v0,v0r,v0z,
     $          b0,b0r,b0z,nqr,nqz,nd0r,nd0z,vv,upave,bcrgt,t_eq,dp
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: bigr2,n0,
     $          prp_fac,pll_fac,ndivvs,fupw,fupwi
      REAL(r8), DIMENSION(6,SIZE(bigr,1),SIZE(bigr,2)) :: updyad
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: nvs,ja
      REAL(r8), DIMENSION(1,1,1) :: dv
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: ctmp
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: csclr
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      INTEGER(i4) :: nv,iv,jv,imode,iq,jq
      REAL(r8) :: zz,kpll,kprp,jfac,g,gm1
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and the bb dyad.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      IF (p_model(1:5)=='aniso')
     $  CALL generic_ptr_set(rb%qbb,tb%qbb,tb%tgeom,inode,bb,dp,dp,0_i4)
      nv=SIZE(int,5)
      IF (integrand_flag(1:6)=='all ti') THEN
        zz=zeff
        IF (separate_pe) THEN
          kpll=k_plli
          jfac= 0.5*dt*meomi/(zz*elementary_q*(meomi+1))
        ELSE
          kpll=k_plle
          jfac=0
        ENDIF
        kprp=k_perpi
      ELSE
        zz=1
        kpll=k_plle
        kprp=k_perpe
        jfac=-0.5*dt/(elementary_q*(meomi+1))
      ENDIF
      IF (geom=='tor') THEN
        g=1
      ELSE
        g=0
      ENDIF
      IF (p_model=='isothermal') THEN
        gm1=0
      ELSE
        gm1=gamm1
      ENDIF
      IF (p_model=='isotropic') kpll=kprp
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     number density, n, is needed for anisotropic cases, so the
c     equation is n*dT/dt= rather than dT/dt=.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,nqr,nqz,1_i4)
      IF (nonlinear.AND.
     $    (continuity=='n=0 only'.OR.continuity=='full') ) THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,nd_n0,nd0r,nd0z,1_i4)
        n0=(nd_eq(1,:,:)+nd_n0(1,:,:))/zz
      ELSE
        n0=nd_eq(1,:,:)/zz
        NULLIFY(nd0r)
      ENDIF
c-----------------------------------------------------------------------
c     Compute the isotropic diffusion coefficient.  k_perp and k_pll
c     already include the factor of gamma-1.  multiply by the
c     appropriate z, too.  [this will not give the exact n=0 part of
c     the operator when continuity=full, but it should be sufficient
c     for the preconditioner in most cases.]
c
c-PRE also use dyad?
c-----------------------------------------------------------------------
      IF (nonlinear.AND.(p_model=="aniso_plltdep".OR.
     $                   p_model=="aniso_tdep")) THEN
        IF (integrand_flag(1:6)=='all ti') THEN
          CALL generic_ptr_set(rb%qkappli_n0,tb%qkappli_n0,tb%tgeom,
     $                         inode,kappl,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkaprpi_n0,tb%qkaprpi_n0,tb%tgeom,
     $                         inode,kaprp,dp,dp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qkapple_n0,tb%qkapple_n0,tb%tgeom,
     $                         inode,kappl,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkaprpe_n0,tb%qkaprpe_n0,tb%tgeom,
     $                         inode,kaprp,dp,dp,0_i4)
        ENDIF
        pll_fac=dt*fthc*MAX(0._r8,kappl(1,:,:)-kaprp(1,:,:))*n0
        prp_fac=dt*fthc*kaprp(1,:,:)*n0
      ELSE IF (.NOT.(p_model=='adiabat'.OR.p_model=='isothermal')) THEN
        pll_fac=dt*fthc*(kpll-kprp)*n0
        prp_fac=dt*fthc*kprp*n0
      ENDIF
c-----------------------------------------------------------------------
c     artificial thermal diffusivity for upwinding.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        IF (integrand_flag(1:6)=='all ti') THEN
          CALL generic_ptr_set(rb%qupti_n0,tb%qupti_n0,tb%tgeom,
     $                         inode,upave,dp,dp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qupte_n0,tb%qupte_n0,tb%tgeom,
     $                         inode,upave,dp,dp,0_i4)
        ENDIF
        CALL generic_ptr_set(rb%qvv,tb%qvv,tb%tgeom,
     $                       inode,vv,dp,dp,0_i4)
        fupw =t_dart_upw*dt**2*n0*upave(1,:,:)*upw_aniso/
     $        ( SUM(vv(1:3,:,:),1) + smallnum )
        fupwi=t_dart_upw*dt**2*n0*upave(1,:,:)*(1._r8-upw_aniso)
        updyad(1,:,:)= fupw*vv(1,:,:)+fupwi
        updyad(2,:,:)= fupw*vv(2,:,:)+fupwi
        updyad(3,:,:)=(fupw*vv(3,:,:)+fupwi)*k2ef(jmode)/bigr2
        updyad(4,:,:)= fupw*vv(4,:,:)
        updyad(5,:,:)= fupw*vv(5,:,:)*keff(jmode)/bigr
        updyad(6,:,:)= fupw*vv(6,:,:)*keff(jmode)/bigr
      ENDIF
c-----------------------------------------------------------------------
c     Compute
c       prp_fac*grad(alpha(i))*grad(alpha(j))+dt*(kpll-kprp)*
c         (grad(alpha(i)) dot B0) (B0 dot grad(alpha(j)) /B^{2}).
c-----------------------------------------------------------------------
      IF (p_model=='adiabat'.OR.p_model=='isothermal') THEN
        DO iv=1,nv
          DO jv=1,iv
            int(1,1,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv)*n0,1)
          ENDDO
        ENDDO
      ELSE IF (p_model=='isotropic') THEN
        DO iv=1,nv
          DO jv=1,iv
            int(1,1,:,jv,iv)=SUM(
     $        (dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $        +dalpdz(:,:,iv)*dalpdz(:,:,jv))*prp_fac
     $        + alpha(:,:,iv)* alpha(:,:,jv)*(n0+
     $                   k2ef(jmode)*prp_fac/bigr2),1)
          ENDDO
        ENDDO
      ELSE
        DO iv=1,nv
          DO jv=1,iv
            int(1,1,:,jv,iv)=SUM(
     $         dalpdr(:,:,iv)*dalpdr(:,:,jv)*(prp_fac+pll_fac*bb(1,:,:))
     $        +dalpdz(:,:,iv)*dalpdz(:,:,jv)*(prp_fac+pll_fac*bb(2,:,:))
     $        + alpha(:,:,iv)* alpha(:,:,jv)*(n0+
     $                   k2ef(jmode)*(prp_fac+pll_fac*bb(3,:,:))/bigr2)
     $        +( (dalpdr(:,:,iv)*dalpdz(:,:,jv)
     $           +dalpdz(:,:,iv)*dalpdr(:,:,jv))*bb(4,:,:)
     $      +(0,1)*(
     $       (dalpdr(:,:,iv)*alpha(:,:,jv)-dalpdr(:,:,jv)*alpha(:,:,iv))
     $                      *keff(jmode)*bb(5,:,:)/bigr+
     $       (dalpdz(:,:,iv)*alpha(:,:,jv)-dalpdz(:,:,jv)*alpha(:,:,iv))
     $                      *keff(jmode)*bb(6,:,:)/bigr ) )*pll_fac,1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     add the anisotropic artificial diffusivity if needed.
c
c-PRE premultiply terms that do not change with jv?
c-----------------------------------------------------------------------
      IF (nonlinear.AND.impladv.AND.t_dart_upw>0) THEN
        DO iv=1,nv
          DO jv=1,iv
            int(1,1,:,jv,iv)=int(1,1,:,jv,iv)+SUM(
     $          dalpdr(:,:,iv)*dalpdr(:,:,jv)*updyad(1,:,:)
     $         +dalpdz(:,:,iv)*dalpdz(:,:,jv)*updyad(2,:,:)
     $         + alpha(:,:,iv)* alpha(:,:,jv)*updyad(3,:,:)
     $        +(dalpdr(:,:,iv)*dalpdz(:,:,jv)
     $         +dalpdz(:,:,iv)*dalpdr(:,:,jv))*updyad(4,:,:)
     $         +(0,1)*(
     $       (dalpdr(:,:,iv)*alpha(:,:,jv)-dalpdr(:,:,jv)*alpha(:,:,iv))
     $                      *updyad(5,:,:)+
     $       (dalpdz(:,:,iv)*alpha(:,:,jv)-dalpdz(:,:,jv)*alpha(:,:,iv))
     $                      *updyad(6,:,:) ),1)
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     use symmetry to complete the rest of the matrix elements.
c-----------------------------------------------------------------------
      DO iv=1,nv-1
        DO jv=iv+1,nv
          int(1,1,:,jv,iv)=CONJG(int(1,1,:,iv,jv))
        ENDDO
      ENDDO

      IF (.NOT.(separate_pe.AND.k_cross>0.OR.impladv)) RETURN
c-----------------------------------------------------------------------
c     add or subtract the grad(alpha(iv)).B_eqXgrad(alpha(jv)) terms
c     for drift effects.  complete the coefficient in the temporary 
c     prp_fac storage location.
c
c     note that n0->0.5*dt*n0
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      n0=0.5*dt*n0

      IF (separate_pe.AND.k_cross>0.AND.p_model(1:5)=='aniso') THEN
        IF (integrand_flag(1:6)=='all ti') THEN
          CALL generic_ptr_set(rb%qti_b2,tb%qti_b2,tb%tgeom,
     $                         inode,t_b2,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qbcrgti,tb%qbcrgti,tb%tgeom,
     $                         inode,bcrgt,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qtion_eq,tb%qtion_eq,tb%tgeom,
     $                         inode,t_eq,dp,dp,0_i4)
          prp_fac=-t_b2(1,:,:)*n0
        ELSE
          CALL generic_ptr_set(rb%qte_b2,tb%qte_b2,tb%tgeom,
     $                         inode,t_b2,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qbcrgte,tb%qbcrgte,tb%tgeom,
     $                         inode,bcrgt,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qtele_eq,tb%qtele_eq,tb%tgeom,
     $                         inode,t_eq,dp,dp,0_i4)
          prp_fac= t_b2(1,:,:)*n0
        ENDIF
        ctmp(1,:,:)=bcrgt(1,:,:)/t_eq(1,:,:)+
     $              be_eq(2,:,:)*(0,1)*keff(jmode)/bigr
        ctmp(2,:,:)=bcrgt(2,:,:)/t_eq(1,:,:)-
     $              be_eq(1,:,:)*(0,1)*keff(jmode)/bigr
        ctmp(3,:,:)=bcrgt(3,:,:)/t_eq(1,:,:)
      ELSE
        prp_fac=0._r8
        ctmp=0._r8
      ENDIF
c-----------------------------------------------------------------------
c     time-centered implicit advection, n_s*V_s and n_s*div(V_s) for
c     each species s.  the Jeq.grad(neq)=0 approximation is used here.
c
c     for ions, we need
c
c       (dt/2)*n_i*V_i = (dt/2)*(n_e*V/z + J*meomi/(z*e*(1+meomi)) )
c
c     where meomi = z*me/mi.  jfac is the full coefficient for J, and
c     n0 holds dt*n_e/2*z.
c
c     we also need
c
c       (dt/2)*n_i*div(V_i) =
c         (dt/2)*(n_e*div(V)/z - J.grad(n_e)*meomi/(n_e*z*e*(1+meomi))
c
c     the electron terms have z->1 and the meomi in the numerator
c     replaced by -1.
c-----------------------------------------------------------------------
      IF (impladv) THEN
        IF (eq_flow/='none') THEN
          CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                         inode,ve_eq,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qdvv_eq,tb%qdvv_eq,tb%tgeom,
     $                         inode,divvq,dp,dp,0_i4)
        ENDIF

        CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                       inode,ja_eq,dp,dp,0_i4)
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                         inode,b0,b0r,b0z,1_i4)
          IF (.NOT.ASSOCIATED(nd0r))
     $      CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                           inode,nd_n0,nd0r,nd0z,1_i4)
          ja(1,:,:)=  b0z(3,:,:)/mu0
          ja(2,:,:)=-(b0r(3,:,:)+g*b0(3,:,:)/bigr)/mu0
          ja(3,:,:)= (b0r(2,:,:)-b0z(1,:,:))/mu0
          nvs=jfac*(ja_eq+ja)
          jfac=0.5*dt*jfac/zz
          ndivvs=-jfac*(ja_eq(1,:,:)*nd0r(1,:,:)+
     $                  ja_eq(2,:,:)*nd0z(1,:,:)+
     $                  ja(1,:,:)*(nd0r(1,:,:)+nqr(1,:,:))+
     $                  ja(2,:,:)*(nd0z(1,:,:)+nqz(1,:,:)))/n0
        ELSE
          nvs=jfac*ja_eq
          ndivvs=0
        ENDIF

        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qve_n0,tb%qve_n0,tb%tgeom,
     $                         inode,v0,v0r,v0z,1_i4)
          IF (eq_flow/='none') THEN
            nvs(1,:,:)=nvs(1,:,:)+n0*(v0(1,:,:)+ve_eq(1,:,:))
            nvs(2,:,:)=nvs(2,:,:)+n0*(v0(2,:,:)+ve_eq(2,:,:))
            nvs(3,:,:)=nvs(3,:,:)+n0*(v0(3,:,:)+ve_eq(3,:,:))
            ndivvs=ndivvs+
     $        n0*(divvq(1,:,:)+v0r(1,:,:)+v0z(2,:,:)+g*v0(1,:,:)/bigr)
          ELSE
            nvs(1,:,:)=nvs(1,:,:)+n0*v0(1,:,:)
            nvs(2,:,:)=nvs(2,:,:)+n0*v0(2,:,:)
            nvs(3,:,:)=nvs(3,:,:)+n0*v0(3,:,:)
            ndivvs=ndivvs+n0*(v0r(1,:,:)+v0z(2,:,:)+g*v0(1,:,:)/bigr)
          ENDIF
        ELSE
          IF (eq_flow/='none') THEN
            nvs(1,:,:)=nvs(1,:,:)+n0*ve_eq(1,:,:)
            nvs(2,:,:)=nvs(2,:,:)+n0*ve_eq(2,:,:)
            nvs(3,:,:)=nvs(3,:,:)+n0*ve_eq(3,:,:)
            ndivvs=ndivvs+n0*divvq(1,:,:)
          ENDIF
        ENDIF
      ELSE
        nvs=0
        ndivvs=0
      ENDIF
c-----------------------------------------------------------------------
c     add these contributions to the symmetric part of the operator.
c     use ctmp as storage for the coefficient of alpha_jv.
c-----------------------------------------------------------------------
      csclr(:,:)=gm1*ndivvs+(0,1)*keff(jmode)*nvs(3,:,:)/bigr
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=int(1,1,:,jv,iv)+SUM(
     $      prp_fac*(
     $      dalpdr(:,:,iv)*
     $        (ctmp(1,:,:)*alpha(:,:,jv)-be_eq(3,:,:)*dalpdz(:,:,jv))+
     $      dalpdz(:,:,iv)*
     $        (ctmp(2,:,:)*alpha(:,:,jv)+be_eq(3,:,:)*dalpdr(:,:,jv))-
     $      (0,1)*keff(jmode)*alpha(:,:,iv)/bigr*
     $        (be_eq(1,:,:)*dalpdz(:,:,jv)-be_eq(2,:,:)*dalpdr(:,:,jv)+
     $         ctmp(3,:,:)*alpha(:,:,jv)))
     $     +alpha(:,:,iv)*
     $         (nvs(1,:,:)*dalpdr(:,:,jv)+
     $          nvs(2,:,:)*dalpdz(:,:,jv)+
     $          csclr(:,:)* alpha(:,:,jv)),1)
        ENDDO
      ENDDO

      RETURN
      END SUBROUTINE t_aniso_op
c-----------------------------------------------------------------------
c     subprogram 24. t_aniso_cpop.
c     find an off diagonal in Fourier index part of the implicit 
c     operator used for temperature advance.
c
c     (n*I.-dt*f*D)
c
c     including an implicit diffusion operator and the
c     diagonal part of an anisotropic temperature diffusivity.
c     when implicit advection is used, the operator is
c
c     (n*I. + 0.5*dt*n*( V.grad(I.)+(gamma-1)*div(V)*I. ) - dt*f*D)
c
c     NOTE: no upwinding or diamagnetic heat flow terms
c-----------------------------------------------------------------------
      SUBROUTINE t_aniso_cpop(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: 
     $          dp,alpha,dalpdr,dalpdz,
     $          nd_eq,nd_n0,be_eq,be_n0,kappl_n0,kaprp_n0,bb_n0
      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: 
     $          dcp,ndp,bep,kapplp,kaprpp
     
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: 
     $          n0,bigr2,pll_fac_n0,prp_fac_n0
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: 
     $          pll_fac_off,prp_fac_off

      COMPLEX(r8), DIMENSION(6,SIZE(bigr,1),SIZE(bigr,2)) :: bb_off

      COMPLEX(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) ::
     $             nd_off,kappl_off,kaprp_off
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be_off

      LOGICAL :: useconj
      INTEGER(i4) :: nv,iv,jv,imode,iq,jq,lintf,foff,fcol,kcol
      REAL(r8) :: zz,kpll,kprp,jfac,g,gm1
      REAL(r8), DIMENSION(1,1,1) :: dv
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc

c-----------------------------------------------------------------------
c     set rblock and tblock weight functions and the bb dyad.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      IF (p_model(1:5)=='aniso')
     $  CALL generic_ptr_set(rb%qbb,tb%qbb,tb%tgeom,inode,
     $                       bb_n0,dp,dp,0_i4)

      nv=SIZE(int,5)
      IF (integrand_flag(1:2)=='ti') THEN
        zz=zeff
        IF (separate_pe) THEN
          kpll=k_plli
          jfac= 0.5*dt*meomi/(zz*elementary_q*(meomi+1))
        ELSE
          kpll=k_plle
          jfac=0
        ENDIF
        kprp=k_perpi
      ELSE
        zz=1
        kpll=k_plle
        kprp=k_perpe
        jfac=-0.5*dt/(elementary_q*(meomi+1))
      ENDIF
      IF (geom=='tor') THEN
        g=1
      ELSE
        g=0
      ENDIF
      IF (p_model=='isothermal') THEN
        gm1=0
      ELSE
        gm1=gamm1
      ENDIF
      IF (p_model=='isotropic') kpll=kprp
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     determine Fourier index offset relative to the diagonal.  note
c     that the Fourier row index is nindex(jmode).
c-----------------------------------------------------------------------
      lintf=LEN_TRIM(integrand_flag)
      READ(integrand_flag(lintf-2:lintf),'(i3)') foff
      IF (integrand_flag(4:7)=='plus') THEN
        useconj=.true.
        fcol=nindex(jmode)+foff
        IF (fcol>nmodes_total-1) THEN
          int=0._r8
          RETURN
        ELSE
          kcol=keff_total(fcol+1)
        ENDIF
      ELSE
        useconj=.false.
        fcol=nindex(jmode)-foff
        IF (fcol<0) THEN
          fcol=-fcol
          kcol=-keff_total(fcol+1)
        ELSE
          kcol=keff_total(fcol+1)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     number density, n, is needed for anisotropic cases, so the
c     equation is n*dT/dt= rather than dT/dt=.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                     inode,nd_n0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qnd_off,tb%qnd_off,tb%tgeom,
     $                     inode,ndp,dcp,dcp,0_i4)
      IF (continuity=='n=0 only'.OR.continuity=='full') THEN  
        n0=(nd_eq(1,:,:)+nd_n0(1,:,:))/zz
      ELSE
        n0=nd_eq(1,:,:)/zz
      ENDIF

      IF (useconj) THEN
        nd_off=CONJG(ndp(:,:,:,foff))
      ELSE
        nd_off=ndp(:,:,:,foff)
      ENDIF
c-----------------------------------------------------------------------
c     Calculates n=0 and off-diagonal bb-dyads
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,be_eq,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                     inode,be_n0,dp,dp,0_i4)
      CALL generic_ptr_set(rb%qbe_off,tb%qbe_off,tb%tgeom,
     $                     inode,bep,dcp,dcp,0_i4)

      IF (useconj) THEN 
        be_off=CONJG(bep(:,:,:,foff))
      ELSE
        be_off=bep(:,:,:,foff)
      ENDIF

      bb_off(1:3,:,:)=
     $    2*be_off(1:3,:,:)*(be_eq(1:3,:,:)+be_n0(1:3,:,:))
      bb_off(4,:,:)=
     $    be_off(1,:,:)*(be_eq(2,:,:)+be_n0(2,:,:))+
     $    be_off(2,:,:)*(be_eq(1,:,:)+be_n0(1,:,:))
      bb_off(5,:,:)=
     $    be_off(1,:,:)*(be_eq(3,:,:)+be_n0(3,:,:))+
     $    be_off(3,:,:)*(be_eq(1,:,:)+be_n0(1,:,:))
      bb_off(6,:,:)=
     $    be_off(2,:,:)*(be_eq(3,:,:)+be_n0(3,:,:))+
     $    be_off(3,:,:)*(be_eq(2,:,:)+be_n0(2,:,:))
      DO iq=1,6
        bb_off(iq,:,:)=bb_off(iq,:,:)/SUM((be_eq+be_n0)**2,1)
      ENDDO
c-----------------------------------------------------------------------
c     Compute the both the n=0 and off-diagnonal diffusion
c       coefficients, where k_perp_* and k_pll_* already include the
c       factor of gamma-1 and is multiplied by the appropriate z, too.
c       [this will not give the exact n=0 part of the operator when
c       continuity=full, but it should be sufficient
c       for the preconditioner in most cases.] 
c
c-PRE also use dyad?
c-----------------------------------------------------------------------

      IF (p_model=='isotropic') THEN
        prp_fac_n0=dt*fthc*kprp*n0
        prp_fac_off=dt*fthc*kprp*nd_off(1,:,:)


      ELSEIF (p_model=='aniso1') THEN
        pll_fac_n0=dt*fthc*(kpll-kprp)*n0
        prp_fac_n0=dt*fthc*kprp*n0
        pll_fac_off=dt*fthc*(kpll-kprp)*nd_off(1,:,:)
        prp_fac_off=dt*fthc*kprp*nd_off(1,:,:)


      ELSEIF (p_model=='aniso_plltdep') THEN
        IF (integrand_flag(1:2)=='ti') THEN
          CALL generic_ptr_set(rb%qkappli_n0,tb%qkappli_n0,tb%tgeom,
     $                         inode,kappl_n0,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkappli_off,tb%qkappli_off,tb%tgeom,
     $                         inode,kapplp,dcp,dcp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qkapple_n0,tb%qkapple_n0,tb%tgeom,
     $                         inode,kappl_n0,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkapple_off,tb%qkapple_off,tb%tgeom,
     $                         inode,kapplp,dcp,dcp,0_i4)
        ENDIF

        IF (useconj) THEN
          kappl_off=CONJG(kapplp(:,:,:,foff))
        ELSE
          kappl_off=kapplp(:,:,:,foff)
        ENDIF

        pll_fac_n0=dt*fthc*MAX(0._r8,kappl_n0(1,:,:)-kaprp_n0(1,:,:))*n0
        pll_fac_off=dt*fthc*(
     $     MAX(0._r8,kappl_n0(1,:,:)-kaprp_n0(1,:,:))*nd_off(1,:,:)
     $             +(kappl_off(1,:,:)-kaprp_off(1,:,:))*n0)


      ELSEIF (p_model=='aniso_tdep') THEN
        IF (integrand_flag(1:2)=='ti') THEN
          CALL generic_ptr_set(rb%qkappli_n0,tb%qkappli_n0,tb%tgeom,
     $                         inode,kappl_n0,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkappli_off,tb%qkappli_off,tb%tgeom,
     $                         inode,kapplp,dcp,dcp,0_i4)
          CALL generic_ptr_set(rb%qkaprpi_n0,tb%qkaprpi_n0,tb%tgeom,
     $                         inode,kaprp_n0,dp,dp,0_i4)
          IF (.NOT.closure_n0_only)
     $      CALL generic_ptr_set(rb%qkaprpi_off,tb%qkaprpi_off,tb%tgeom,
     $                           inode,kaprpp,dcp,dcp,0_i4)
        ELSE
          CALL generic_ptr_set(rb%qkapple_n0,tb%qkapple_n0,tb%tgeom,
     $                         inode,kappl_n0,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qkapple_off,tb%qkapple_off,tb%tgeom,
     $                         inode,kapplp,dcp,dcp,0_i4)
          CALL generic_ptr_set(rb%qkaprpe_n0,tb%qkaprpe_n0,tb%tgeom,
     $                         inode,kaprp_n0,dp,dp,0_i4)
          IF (.NOT.closure_n0_only)
     $      CALL generic_ptr_set(rb%qkaprpe_off,tb%qkaprpe_off,tb%tgeom,
     $                           inode,kaprpp,dcp,dcp,0_i4)
        ENDIF

        IF (useconj) THEN
          kappl_off=CONJG(kapplp(:,:,:,foff))
          IF (.NOT.closure_n0_only)
     $      kaprp_off=CONJG(kaprpp(:,:,:,foff))
        ELSE
          kappl_off=kapplp(:,:,:,foff)
          IF (.NOT.closure_n0_only)
     $      kaprp_off=kaprpp(:,:,:,foff)
        ENDIF

        prp_fac_n0=dt*fthc*kaprp_n0(1,:,:)*n0
        pll_fac_n0=dt*fthc*MAX(0._r8,kappl_n0(1,:,:)-kaprp_n0(1,:,:))*n0
        pll_fac_off=dt*fthc*(
     $     MAX(0._r8,kappl_n0(1,:,:)-kaprp_n0(1,:,:))*nd_off(1,:,:)
     $             +(kappl_off(1,:,:)-kaprp_off(1,:,:))*n0)
        IF (.NOT.closure_n0_only) THEN
          prp_fac_off=dt*fthc*(kaprp_n0(1,:,:)*nd_off(1,:,:)
     $                        +kaprp_off(1,:,:)*n0)
        ELSE
          prp_fac_off=dt*fthc*kaprp_n0(1,:,:)*nd_off(1,:,:)
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     Compute
c       prp_fac*grad(alpha(i))*grad(alpha(j))+dt*(kpll-kprp)*
c         (grad(alpha(i)) dot B) (B dot grad(alpha(j)) /B^{2}).
c-----------------------------------------------------------------------
      IF (p_model=='adiabat'.OR.p_model=='isothermal') THEN
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv)*
     $                   nd_off(1,:,:),1)
          ENDDO
        ENDDO
      ELSE IF (p_model=='isotropic') THEN
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=SUM(
     $        (dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $        +dalpdz(:,:,iv)*dalpdz(:,:,jv))*prp_fac_off
     $        + alpha(:,:,iv)* alpha(:,:,jv)*(nd_off(1,:,:)+
     $                   keff(jmode)*kcol*prp_fac_off/bigr2),1)
          ENDDO
        ENDDO
      ELSE
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=SUM(
     $         dalpdr(:,:,iv)*dalpdr(:,:,jv)*(prp_fac_off+
     $             pll_fac_off*bb_n0(1,:,:)
     $            +pll_fac_n0*bb_off(1,:,:))
     $        +dalpdz(:,:,iv)*dalpdz(:,:,jv)*(prp_fac_off+
     $             pll_fac_off*bb_n0(2,:,:)
     $            +pll_fac_n0*bb_off(2,:,:))
     $        + alpha(:,:,iv)* alpha(:,:,jv)*(nd_off(1,:,:)+
     $             keff(jmode)*kcol*(prp_fac_off+
     $             pll_fac_off*bb_n0(3,:,:)
     $            +pll_fac_n0*bb_off(3,:,:))/bigr2)
     $        +( ( dalpdr(:,:,iv)*dalpdz(:,:,jv)
     $            +dalpdz(:,:,iv)*dalpdr(:,:,jv))*bb_n0(4,:,:)
     $           +(0,1)*( ( dalpdr(:,:,iv)*alpha(:,:,jv)*kcol
     $                     -dalpdr(:,:,jv)*alpha(:,:,iv)*keff(jmode))
     $                    *bb_n0(5,:,:)/bigr
     $                   +( dalpdz(:,:,iv)*alpha(:,:,jv)*kcol
     $                     -dalpdz(:,:,jv)*alpha(:,:,iv)*keff(jmode))
     $                    *bb_n0(6,:,:)/bigr ) )*pll_fac_off
     $        +( ( dalpdr(:,:,iv)*dalpdz(:,:,jv)
     $            +dalpdz(:,:,iv)*dalpdr(:,:,jv))*bb_off(4,:,:)
     $           +(0,1)*( ( dalpdr(:,:,iv)*alpha(:,:,jv)*kcol
     $                     -dalpdr(:,:,jv)*alpha(:,:,iv)*keff(jmode))
     $                    *bb_off(5,:,:)/bigr
     $                   +( dalpdz(:,:,iv)*alpha(:,:,jv)*kcol
     $                     -dalpdz(:,:,jv)*alpha(:,:,iv)*keff(jmode))
     $                    *bb_off(6,:,:)/bigr ) )*pll_fac_n0,1)
          ENDDO
        ENDDO
      ENDIF

      RETURN
      END SUBROUTINE t_aniso_cpop
c-----------------------------------------------------------------------
c     subprogram 25. vec_lap_op.
c     find the Laplacian operator used in the velocity advances.  this
c     routine is used in different ways according the the number of
c     vector components at each vertex.  if the local block is 3x3,
c     then the vector order is (r,z,phi); this is used for si-ops in
c     inviscid calculations, where free-slip boundary conditions
c     couple the poloidal components.  if the local block is 2x2, the
c     vector order is (r,phi), where toroidal geometry couples the
c     two components, and z is decoupled with no-slip bcs.  if the
c     local block is 1x1, it's just the z vector component.
c
c     the equation is now rho*delta_V, so there is a mass density
c     scale to all terms, and the rho*mass_matrix term is added here.
c-----------------------------------------------------------------------
      SUBROUTINE vec_lap_op(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: beq,b0,peq,p0,nd_eq,n0,
     $          nl_pres,dp,ds2
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: bmagsq,bigr2,
     $          func,md_sink,diff
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      INTEGER(i4) :: nv,iv,jv,nq,np
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the
c     diffusivity shape function.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      IF ((ds_use=='kin_visc'.OR.ds_use=='both').AND.kin_visc>0) THEN
        CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                         inode,ds2,dp,dp,0_i4)
        diff=kin_visc*mtot*MAX(ds2(1,:,:),0._r8)
      ELSE
        diff=kin_visc*mtot
      ENDIF
c-----------------------------------------------------------------------
c     the mass density is the md_sink factor.  for nonlinear runs
c     with density evolution, the operator uses only the n=0 part of
c     the number density.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      IF (nonlinear.AND.
     $    (continuity=='n=0 only'.OR.continuity=='full')) THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,n0,dp,dp,0_i4)
        md_sink=mtot*(n0(1,:,:)+nd_eq(1,:,:))
        diff=diff*(n0(1,:,:)+nd_eq(1,:,:))
      ELSE
        md_sink=mtot*nd_eq(1,:,:)
        diff=diff*nd_eq(1,:,:)
      ENDIF
c-----------------------------------------------------------------------
c     evaluate fields for the semi-implicit operator if needed.
c-----------------------------------------------------------------------
      IF (integrand_flag(1:3)=='mhd'.OR.integrand_flag(1:3)=='all') THEN
        CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                       inode,beq,dp,dp,0_i4)
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,
     $                         inode,b0,dp,dp,0_i4)
          bmagsq=SUM((beq+b0)**2,1)
          CALL generic_ptr_set(rb%qsi_nl_pres,tb%qsi_nl_pres,tb%tgeom,
     $                         inode,nl_pres,dp,dp,0_i4)
        ELSE
          bmagsq=SUM(beq**2,1)
        ENDIF
        IF (beta>0) THEN
          CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                         inode,peq,dp,dp,0_i4)
          IF (nonlinear) THEN
            CALL generic_ptr_set(rb%qpres_n0,tb%qpres_n0,tb%tgeom,inode,
     $                           p0,dp,dp,0_i4)
          ENDIF
        ENDIF
      ENDIF
c-----------------------------------------------------------------------
c     find the spatial function that is used with the operator.
c-----------------------------------------------------------------------
      func=0
      IF (kin_visc>0.AND.integrand_flag(1:3)/='mhd')
     $  func=dt*fvsc*diff
      IF (integrand_flag(1:4)/='visc') THEN
        IF (beta>0) THEN
          IF (nonlinear) THEN
            func=func+0.25*dt**2*
     $             (si_fac_mhd*bmagsq/mu0
     $             +si_fac_pres*gamma*(peq(1,:,:)+p0(1,:,:))
     $             +si_fac_nl*(nl_pres(1,:,:)/mu0+gamma*nl_pres(2,:,:)))
          ELSE
            func=func+0.25*dt**2*
     $             (si_fac_mhd*bmagsq/mu0+si_fac_pres*gamma*peq(1,:,:))
          ENDIF
        ELSE
          IF (nonlinear) THEN
            func=func+0.25*dt**2*
     $             (si_fac_mhd*bmagsq/mu0+si_fac_nl*nl_pres(1,:,:)/mu0)
          ELSE
            func=func+0.25*dt**2*si_fac_mhd*bmagsq/mu0
          ENDIF
        ENDIF
      ENDIF
      nv=SIZE(int,5)
      nq=SIZE(int,1)
      IF (nq==3) THEN
        np=3
      ELSE
        np=2
      ENDIF
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     compute func*grad(alpha(i))*grad(alpha(j)) and place this
c     in the 1-1 matrix elements.  [possibly r-r or z-z]
c
c     the mass density term uses the md_sink factor.
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM( func
     $          *(dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $           +dalpdz(:,:,iv)*dalpdz(:,:,jv)
     $           +k2ef(jmode)*alpha(:,:,iv)*alpha(:,:,jv)/bigr2)
     $         +md_sink*alpha(:,:,iv)*alpha(:,:,jv),1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     for an independent z-z matrix, that's it.
c-----------------------------------------------------------------------
      IF (nq==1) RETURN
c-----------------------------------------------------------------------
c     for a full 3x3 matrix, the z-z is the same as r-r prior to
c     toroidal effects.
c-----------------------------------------------------------------------
      IF (nq==3) THEN
        int(1,2,:,:,:)=0
        int(2,2,:,:,:)=int(1,1,:,:,:)
        int(3,2,:,:,:)=0
        int(2,1,:,:,:)=0
        int(2,3,:,:,:)=0
      ENDIF
c-----------------------------------------------------------------------
c     for toroidal geometry, add the coupling between the real_r
c     and -imag_phi (or imag_r and real_phi) components, along with
c     the additional diagonal terms.
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        DO iv=1,nv
          DO jv=1,nv
            int(1,1,:,jv,iv)=int(1,1,:,jv,iv)+SUM(
     $        +func*alpha(:,:,iv)*alpha(:,:,jv)/bigr2,1)
            int(np,1,:,jv,iv)=SUM(
     $         func*2*keff(jmode)*alpha(:,:,iv)*alpha(:,:,jv)/bigr2,1)
            int(1,np,:,jv,iv)=int(np,1,:,jv,iv)
          ENDDO
        ENDDO
      ELSE
        int(np,1,:,:,:)=0
        int(1,np,:,:,:)=0
      ENDIF
      int(np,np,:,:,:)=int(1,1,:,:,:)
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE vec_lap_op
c-----------------------------------------------------------------------
c     subprogram 26. v_aniso_op.
c     find the semi-implicit operator used in the velocity advances.
c     the integrand is now a complex array.
c
c     the equation is now rho*delta_V, so there is a mass density
c     scale to all terms, and the rho*mass_matrix term is added here.
c-----------------------------------------------------------------------
      SUBROUTINE v_aniso_op(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(:,:,:), POINTER :: almod,dalmr,dalmz
      REAL(r8), DIMENSION(:,:,:), POINTER :: b0,b0r,b0z,beq,beqr,beqz,
     $          j0,nd_eq,n0,p0,p0r,p0z,peq,peqr,peqz,nl_pres,bb,dp,ds2,
     $          ti_sym,v0,v0r,v0z,veq,grdv_eq,kappli
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: be,ber,bez,ja,
     $          ve,ver,vez
      REAL(r8), DIMENSION(1,SIZE(bigr,1),SIZE(bigr,2)) :: pres,presr,
     $          presz
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: bmagsq,bigr2,
     $          iso,grd,md_sink,diff,alpr
      REAL(r8), DIMENSION(1,1,1) :: dv

      COMPLEX(r8), DIMENSION(3,3,SIZE(bigr,1),SIZE(bigr,2),ncontb)
     $             :: crl
      COMPLEX(r8), DIMENSION(9,3,SIZE(bigr,1),SIZE(bigr,2),ncontb)
     $             :: piten
      COMPLEX(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2),ncontb) ::
     $             gr_alpha
      COMPLEX(r8), DIMENSION(3,3,SIZE(bigr,1),SIZE(bigr,2)) :: ptmp
      COMPLEX(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: diva
      COMPLEX(r8), DIMENSION(1,1,1,1) :: dc
      COMPLEX(r8), DIMENSION(3,3) :: wdot,bc_wdot
      COMPLEX(r8), DIMENSION(3) :: ctmp

      INTEGER(i4) :: nv,iv,jv,ix,iy,ncx,ncy,iz,i1,i2,i3,nvc,nvm,ivm,jvm
      REAL(r8) :: g,ani,cj0,third,twoth,b2,gcoef,iso_diag,alprs,dtpv
      COMPLEX(r8) :: dvas,bdga
      LOGICAL :: bbset
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions, and evaluate the
c     diffusivity shape function.
c
c     if poly_divv is non-negative, auxiliary fields are used to
c     stabilize the flow-velocity advance.  the modal bases are
c     needed in this case.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)

      IF (poly_divv>=0) THEN
        CALL generic_alpha_eval(rb,tb%tgeom,inode,'modlmat',almod,dalmr,
     $                          dalmz,0_i4,poly_divv,
     $                          polydmin=poly_divv_min,
     $                          polydmax=poly_divv_max)
      ENDIF

      IF ((ds_use=='kin_visc'.OR.ds_use=='both').AND.
     $         (kin_visc>0.OR.iso_visc>0)) THEN
        CALL generic_ptr_set(rb%qdiff_shape,tb%qdiff_shape,tb%tgeom,
     $                       inode,ds2,dp,dp,0_i4)
        md_sink=mtot
        diff=mtot*MAX(ds2(1,:,:),0._r8)
      ELSE
        md_sink=mtot
        diff=mtot
      ENDIF
      third=1._r8/3
      twoth=2._r8/3
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      nvc=ncontb
      nvm=ndiscb
      nv=nvc+nvm
c-----------------------------------------------------------------------
c     the mass density factor is md_sink.  for nonlinear runs
c     with density evolution, the operator uses only the n=0 part of
c     the number density.  for full 3D density evolution, this operator
c     approximates the diagnonal-in-n part of the full operator and is
c     used to make the preconditioner.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qnd_eq,tb%qnd_eq,tb%tgeom,
     $                     inode,nd_eq,dp,dp,0_i4)
      IF (nonlinear.AND.
     $    (continuity=='n=0 only'.OR.continuity=='full')) THEN
        CALL generic_ptr_set(rb%qnd_n0,tb%qnd_n0,tb%tgeom,
     $                       inode,n0,dp,dp,0_i4)
        md_sink=md_sink*(n0(1,:,:)+nd_eq(1,:,:))
        diff=diff*(n0(1,:,:)+nd_eq(1,:,:))
      ELSE
        md_sink=md_sink*nd_eq(1,:,:)
        diff=diff*nd_eq(1,:,:)
      ENDIF
c-----------------------------------------------------------------------
c     evaluate fields for the semi-implicit operator if needed.
c-----------------------------------------------------------------------
      CALL generic_ptr_set(rb%qbe_eq,tb%qbe_eq,tb%tgeom,
     $                     inode,beq,beqr,beqz,1_i4)
      CALL generic_ptr_set(rb%qja_eq,tb%qja_eq,tb%tgeom,
     $                     inode,j0,dp,dp,0_i4)
      IF (nonlinear) THEN
        CALL generic_ptr_set(rb%qbe_n0,tb%qbe_n0,tb%tgeom,inode,
     $                       b0,b0r,b0z,1_i4)
        ja(1,:,:)=j0(1,:,:)+ b0z(3,:,:)/mu0
        ja(2,:,:)=j0(2,:,:)- b0r(3,:,:)/mu0
        ja(3,:,:)=j0(3,:,:)+(b0r(2,:,:)-b0z(1,:,:))/mu0
        IF (geom=='tor') ja(2,:,:)=ja(2,:,:)-b0(3,:,:)/(mu0*bigr)
        be=b0+beq
        ber=b0r+beqr
        bez=b0z+beqz
        CALL generic_ptr_set(rb%qsi_nl_pres,tb%qsi_nl_pres,tb%tgeom,
     $                       inode,nl_pres,dp,dp,0_i4)
      ELSE
        ja=j0
        be=beq
        ber=beqr
        bez=beqz
      ENDIF
      bmagsq=SUM(be*be,1)
      IF (beta>0) THEN
        CALL generic_ptr_set(rb%qpres_eq,tb%qpres_eq,tb%tgeom,
     $                       inode,peq,peqr,peqz,1_i4)
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qpres_n0,tb%qpres_n0,tb%tgeom,
     $                         inode,p0,p0r,p0z,1_i4)
          pres=p0+peq
          presr=p0r+peqr
          presz=p0z+peqz
        ELSE
          pres=peq
          presr=peqr
          presz=peqz
        ENDIF
      ELSE
        pres=0
        presr=0
        presz=0
      ENDIF
c-----------------------------------------------------------------------
c     find the spatial functions used as coefficients for the isotropic
c     and the anisotropic parts of the  operator.
c-----------------------------------------------------------------------
      ani=0.25*si_fac_mhd* (1-mhd_si_iso)*dt**2/mu0
      cj0=0.125*si_fac_j0*dt**2
      IF (beta>0) THEN
        iso=mhd_si_iso*
     $    (si_fac_mhd*bmagsq/mu0+si_fac_pres*gamma*pres(1,:,:))
        grd=0.25*si_fac_pres*(1-mhd_si_iso)*dt**2*gamma*pres(1,:,:)
      ELSE
        iso=mhd_si_iso*si_fac_mhd*bmagsq/mu0
        grd=0
      ENDIF
      IF (nonlinear) THEN
        iso=iso+si_fac_nl*(nl_pres(1,:,:)/mu0+gamma*nl_pres(2,:,:))
      ENDIF
      iso=iso*(0.25*dt**2)
      IF (kin_visc>0.AND.integrand_flag(1:3)/='mhd')
     $  iso=iso+(dt*fvsc*kin_visc)*diff
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     fields for implicit advection.  note that flows and their
c     gradients are multiplied by the mass density and 0.5*dt here.
c-----------------------------------------------------------------------
      IF ((nonlinear.OR.eq_flow/='none').AND.
     $    (advect=='V only'.OR.advect=='all').AND.impladv) THEN
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qve_n0,tb%qve_n0,tb%tgeom,inode,
     $                         v0,v0r,v0z,1_i4)
          ve=v0
          ver=v0r
          vez=v0z
        ELSE
          ve=0
          ver=0
          vez=0
        ENDIF
        IF (eq_flow/='none') THEN
          CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                         inode,veq,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qgrdveq,tb%qgrdveq,tb%tgeom,
     $                         inode,grdv_eq,dp,dp,0_i4)
          ve=ve+veq
          ver=ver+grdv_eq(1:7:3,:,:)
          vez=vez+grdv_eq(2:8:3,:,:)
        ENDIF
        DO i1=1,3
          ve(i1,:,:)=0.5*dt*md_sink*ve(i1,:,:)
          ver(i1,:,:)=0.5*dt*md_sink*ver(i1,:,:)
          vez(i1,:,:)=0.5*dt*md_sink*vez(i1,:,:)
        ENDDO
      ELSE
        ve=0
        ver=0
        vez=0
      ENDIF
c-----------------------------------------------------------------------
c     the anisotropic operator has the form
c     curl(A_iXB0).curl(A_jXB0) after the integration by parts,
c     where A_i is the vector test function alpha*e_hat_i
c     and each curl is broken into B0.grad(A)-A.grad(B0)
c     -B0(div(A)).  B0 is in cylindrical components.
c     the first two indices of the temporary crl matrix are defined as
c     ( curl vector index, operand vector index ), where the vector
c     indices runs from 1 to 3 as (r,z,phi), and the last index is
c     the basis function index.
c-----------------------------------------------------------------------
      IF (geom=='tor') THEN
        g=1
      ELSE
        g=0
      ENDIF
      DO iv=1,nvc
        crl(1,1,:,:,iv)=
     $    dalpdz(:,:,iv)*  be(2,:,:)
     $    -alpha(:,:,iv)*((be(1,:,:)*g-(0,1)*keff(jmode)*be(3,:,:))/bigr
     $                   +ber(1,:,:))
        crl(1,2,:,:,iv)=
     $    -(dalpdz(:,:,iv)*be(1,:,:)+alpha(:,:,iv)*bez(1,:,:))
        crl(1,3,:,:,iv)=
     $    -alpha(:,:,iv)*(0,1)*keff(jmode)*be(1,:,:)/bigr
        crl(2,1,:,:,iv)=
     $    -(dalpdr(:,:,iv)* be(2,:,:)
     $      +alpha(:,:,iv)*(be(2,:,:)*g/bigr+ber(2,:,:)))
        crl(2,2,:,:,iv)=
     $    dalpdr(:,:,iv)*be(1,:,:)
     $    +alpha(:,:,iv)*((0,1)*keff(jmode)*be(3,:,:)/bigr-bez(2,:,:))
        crl(2,3,:,:,iv)=
     $    -alpha(:,:,iv)*(0,1)*keff(jmode)*be(2,:,:)/bigr
        crl(3,1,:,:,iv)=
     $    -(dalpdr(:,:,iv)*be(3,:,:)+alpha(:,:,iv)*ber(3,:,:))
        crl(3,2,:,:,iv)=
     $    -(dalpdz(:,:,iv)*be(3,:,:)+alpha(:,:,iv)*bez(3,:,:))
        crl(3,3,:,:,iv)=
     $    dalpdr(:,:,iv)*be(1,:,:)+dalpdz(:,:,iv)*be(2,:,:)
     $    -alpha(:,:,iv)*be(1,:,:)*g/bigr
      ENDDO
c-----------------------------------------------------------------------
c     create a stress tensor that allows for different forms of
c     viscous stress.  the isotropic contributions to the semi-implicit
c     operator now appear here, too.  we will also save the gradient
c     of the scalar function alpha.  the first index of piten runs
c     over rows first, then columns:
c
c       (1-9)=(r_r,z_r,phi_r,r_z,z_z,phi_z,r_phi,z_phi,phi_phi)
c
c-----------------------------------------------------------------------
      DO iv=1,nvc
        gr_alpha(1,:,:,iv)=dalpdr(:,:,iv)
        gr_alpha(2,:,:,iv)=dalpdz(:,:,iv)
        gr_alpha(3,:,:,iv)=(0,1)*keff(jmode)*alpha(:,:,iv)/bigr
        piten(1,1,:,:,iv)=-iso*gr_alpha(1,:,:,iv)
        piten(2,1,:,:,iv)=-iso*gr_alpha(2,:,:,iv)
        piten(3,1,:,:,iv)=-iso*gr_alpha(3,:,:,iv)
        piten(4:9,1,:,:,iv)=0._r8
        piten(1:3,2,:,:,iv)=0._r8
        piten(4:6,2,:,:,iv)=piten(1:3,1,:,:,iv)
        piten(7:9,2,:,:,iv)=0._r8
        piten(1:6,3,:,:,iv)=0._r8
        piten(7:9,3,:,:,iv)=piten(1:3,1,:,:,iv)
        IF (geom=='tor') THEN
          piten(9,1,:,:,iv)=-iso*alpha(:,:,iv)/bigr
          piten(3,3,:,:,iv)= iso*alpha(:,:,iv)/bigr
        ENDIF
      ENDDO
c-----------------------------------------------------------------------
c     isotropic viscous stress, i.e. proportional to -W, where
c     W is grad(V)+grad(V)^T-2*I*div(V)/3
c-----------------------------------------------------------------------
        IF (iso_visc>0) THEN
          diff=dt*fvsc*iso_visc*diff
          DO iv=1,nvc
           DO iy=1,ncy
           DO ix=1,ncx
            alprs=g*alpha(ix,iy,iv)/bigr(ix,iy)
            dvas=twoth*(gr_alpha(1,ix,iy,iv)+alprs)
            piten(1,1,ix,iy,iv)=piten(1,1,ix,iy,iv)-diff(ix,iy)*
     $        (2*gr_alpha(1,ix,iy,iv)-dvas)
            piten(2,1,ix,iy,iv)=piten(2,1,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(2,ix,iy,iv)
            piten(3,1,ix,iy,iv)=piten(3,1,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(3,ix,iy,iv)
            piten(4,1,ix,iy,iv)=piten(4,1,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(2,ix,iy,iv)
            piten(5,1,ix,iy,iv)=piten(5,1,ix,iy,iv)+diff(ix,iy)*dvas
            piten(7,1,ix,iy,iv)=piten(7,1,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(3,ix,iy,iv)
            piten(9,1,ix,iy,iv)=piten(9,1,ix,iy,iv)-diff(ix,iy)*
     $        (2*alprs-dvas)

            dvas=twoth*gr_alpha(2,ix,iy,iv)
            piten(1,2,ix,iy,iv)=piten(1,2,ix,iy,iv)+diff(ix,iy)*dvas
            piten(2,2,ix,iy,iv)=piten(2,2,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(1,ix,iy,iv)
            piten(4,2,ix,iy,iv)=piten(4,2,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(1,ix,iy,iv)
            piten(5,2,ix,iy,iv)=piten(5,2,ix,iy,iv)-diff(ix,iy)*
     $        (2*gr_alpha(2,ix,iy,iv)-dvas)
            piten(6,2,ix,iy,iv)=piten(6,2,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(3,ix,iy,iv)
            piten(8,2,ix,iy,iv)=piten(8,2,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(3,ix,iy,iv)
            piten(9,2,ix,iy,iv)=piten(9,2,ix,iy,iv)+diff(ix,iy)*dvas

            dvas=twoth*gr_alpha(3,ix,iy,iv)
            piten(1,3,ix,iy,iv)=piten(1,3,ix,iy,iv)+diff(ix,iy)*dvas
            piten(3,3,ix,iy,iv)=piten(3,3,ix,iy,iv)-diff(ix,iy)*
     $        (gr_alpha(1,ix,iy,iv)-alprs)
            piten(5,3,ix,iy,iv)=piten(5,3,ix,iy,iv)+diff(ix,iy)*dvas
            piten(6,3,ix,iy,iv)=piten(6,3,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(2,ix,iy,iv)
            piten(7,3,ix,iy,iv)=piten(7,3,ix,iy,iv)-diff(ix,iy)*
     $        (gr_alpha(1,ix,iy,iv)-alprs)
            piten(8,3,ix,iy,iv)=piten(8,3,ix,iy,iv)-diff(ix,iy)*
     $           gr_alpha(2,ix,iy,iv)
            piten(9,3,ix,iy,iv)=piten(9,3,ix,iy,iv)-diff(ix,iy)*
     $        (2*gr_alpha(3,ix,iy,iv)-dvas)
           ENDDO
           ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       time-centered gyroviscosity operator.  (the div(V) terms
c       cancel analytically).
c-----------------------------------------------------------------------
        bbset=.false.
        IF (gyr_visc>0) THEN
          CALL generic_ptr_set(rb%qti_n0,tb%qti_n0,tb%tgeom,
     $                         inode,ti_sym,dp,dp,0_i4)
          CALL generic_ptr_set(rb%qbb,tb%qbb,tb%tgeom,inode,bb,
     $                         dp,dp,0_i4)
          bbset=.true.
          IF (nonlinear.AND.
     $        (continuity=='n=0 only'.OR.continuity=='full')) THEN
            diff=0.125*gyr_visc*dt*ms(2)*kboltz*(n0(1,:,:)+nd_eq(1,:,:))
     $          *ti_sym(1,:,:)/(zeff**2*elementary_q*bmagsq)
          ELSE
            diff=0.125*gyr_visc*dt*ms(2)*kboltz*nd_eq(1,:,:)*
     $           ti_sym(1,:,:)/(zeff**2*elementary_q*bmagsq)
          ENDIF
          DO iv=1,nvc
            IF (geom=='tor') THEN
              alpr=alpha(:,:,iv)/bigr
            ELSE
              alpr=0._r8
            ENDIF
            DO iz=1,3
              SELECT CASE(iz)
                CASE(1)
                  ptmp(1,1,:,:)=2._r8*gr_alpha(1,:,:,iv)
                  ptmp(2,1,:,:)=gr_alpha(2,:,:,iv)
                  ptmp(3,1,:,:)=gr_alpha(3,:,:,iv)
                  ptmp(1,2,:,:)=gr_alpha(2,:,:,iv)
                  ptmp(2:3,2,:,:)=0._r8
                  ptmp(1,3,:,:)=gr_alpha(3,:,:,iv)
                  ptmp(2,3,:,:)=0._r8
                  ptmp(3,3,:,:)=2._r8*alpr
                CASE(2)
                  ptmp(1,1,:,:)=0._r8
                  ptmp(2,1,:,:)=gr_alpha(1,:,:,iv)
                  ptmp(3,1,:,:)=0._r8
                  ptmp(1,2,:,:)=gr_alpha(1,:,:,iv)
                  ptmp(2,2,:,:)=2._r8*gr_alpha(2,:,:,iv)
                  ptmp(3,2,:,:)=gr_alpha(3,:,:,iv)
                  ptmp(1,3,:,:)=0._r8
                  ptmp(2,3,:,:)=gr_alpha(3,:,:,iv)
                  ptmp(3,3,:,:)=0._r8
                CASE(3)
                  ptmp(1:2,1,:,:)=0._r8
                  ptmp(3,1,:,:)=gr_alpha(1,:,:,iv)-alpr
                  ptmp(1:2,2,:,:)=0._r8
                  ptmp(3,2,:,:)=gr_alpha(2,:,:,iv)
                  ptmp(1,3,:,:)=gr_alpha(1,:,:,iv)-alpr
                  ptmp(2,3,:,:)=gr_alpha(2,:,:,iv)
                  ptmp(3,3,:,:)=2._r8*gr_alpha(3,:,:,iv)
              END SELECT
              DO iy=1,ncy
                DO ix=1,ncx
                  gcoef=diff(ix,iy)
                  wdot(:,1)=ptmp(:,1,ix,iy)+
     $                     3._r8*(ptmp(:,1,ix,iy)*bb(1,ix,iy)+
     $                            ptmp(:,2,ix,iy)*bb(4,ix,iy)+
     $                            ptmp(:,3,ix,iy)*bb(5,ix,iy))
                  wdot(:,2)=ptmp(:,2,ix,iy)+
     $                     3._r8*(ptmp(:,1,ix,iy)*bb(4,ix,iy)+
     $                            ptmp(:,2,ix,iy)*bb(2,ix,iy)+
     $                            ptmp(:,3,ix,iy)*bb(6,ix,iy))
                  wdot(:,3)=ptmp(:,3,ix,iy)+
     $                     3._r8*(ptmp(:,1,ix,iy)*bb(5,ix,iy)+
     $                            ptmp(:,2,ix,iy)*bb(6,ix,iy)+
     $                            ptmp(:,3,ix,iy)*bb(3,ix,iy))
                  bc_wdot(1,:)=be(2,ix,iy)*wdot(3,:)-
     $                         be(3,ix,iy)*wdot(2,:)
                  bc_wdot(2,:)=be(3,ix,iy)*wdot(1,:)-
     $                         be(1,ix,iy)*wdot(3,:)
                  bc_wdot(3,:)=be(1,ix,iy)*wdot(2,:)-
     $                         be(2,ix,iy)*wdot(1,:)
                  piten(1:3,iz,ix,iy,iv)=piten(1:3,iz,ix,iy,iv)+
     $                         gcoef*(bc_wdot(:,1)+bc_wdot(1,:))
                  piten(4:6,iz,ix,iy,iv)=piten(4:6,iz,ix,iy,iv)+
     $                         gcoef*(bc_wdot(:,2)+bc_wdot(2,:))
                  piten(7:9,iz,ix,iy,iv)=piten(7:9,iz,ix,iy,iv)+
     $                         gcoef*(bc_wdot(:,3)+bc_wdot(3,:))
                ENDDO
              ENDDO
            ENDDO
          ENDDO
        ENDIF
c-----------------------------------------------------------------------
c       parallel viscous stress.
c-----------------------------------------------------------------------
        IF (par_visc>0) THEN
          IF (.NOT.bbset) CALL generic_ptr_set(rb%qbb,tb%qbb,tb%tgeom,
     $                                         inode,bb,dp,dp,0_i4)
          IF (nonlinear.AND.
     $        (continuity=='n=0 only'.OR.continuity=='full')) THEN
            diff=-3*dt*fvsc*par_visc*mtot*(n0(1,:,:)+nd_eq(1,:,:))
          ELSE
            diff=-3*dt*fvsc*par_visc*mtot*nd_eq(1,:,:)
          ENDIF
          IF (parvisc_model=='plltdep') THEN
            CALL generic_ptr_set(rb%qkappli_n0,tb%qkappli_n0,tb%tgeom,
     $                           inode,kappli,dp,dp,0_i4)
            diff=diff*kappli(1,:,:)/k_plli
          ENDIF
          DO iv=1,nvc
            IF (geom=='tor') THEN
              alpr=alpha(:,:,iv)/bigr
            ELSE
              alpr=0
            ENDIF
            DO iy=1,ncy
              DO ix=1,ncx
                b2=-third*SUM(bb(1:3,ix,iy))
                ctmp(1)=(gr_alpha(1,ix,iy,iv)*(bb(1,ix,iy)+b2)+
     $                   gr_alpha(2,ix,iy,iv)* bb(4,ix,iy)+
     $                   gr_alpha(3,ix,iy,iv)* bb(5,ix,iy)+
     $                            alpr(ix,iy)*(bb(3,ix,iy)+b2))*
     $                  diff(ix,iy)
                ctmp(2)=(gr_alpha(1,ix,iy,iv)* bb(4,ix,iy)+
     $                   gr_alpha(2,ix,iy,iv)*(bb(2,ix,iy)+b2)+
     $                   gr_alpha(3,ix,iy,iv)* bb(6,ix,iy))*
     $                  diff(ix,iy)
                ctmp(3)=((gr_alpha(1,ix,iy,iv)-alpr(ix,iy))*bb(5,ix,iy)+
     $                    gr_alpha(2,ix,iy,iv)* bb(6,ix,iy)+
     $                    gr_alpha(3,ix,iy,iv)*(bb(3,ix,iy)+b2))*
     $                  diff(ix,iy)
                piten(1,:,ix,iy,iv)=piten(1,:,ix,iy,iv)+
     $                              ctmp(:)*(bb(1,ix,iy)+b2)
                piten(2,:,ix,iy,iv)=piten(2,:,ix,iy,iv)+
     $                              ctmp(:)* bb(4,ix,iy)
                piten(3,:,ix,iy,iv)=piten(3,:,ix,iy,iv)+
     $                              ctmp(:)* bb(5,ix,iy)
                piten(4,:,ix,iy,iv)=piten(4,:,ix,iy,iv)+
     $                              ctmp(:)* bb(4,ix,iy)
                piten(5,:,ix,iy,iv)=piten(5,:,ix,iy,iv)+
     $                              ctmp(:)*(bb(2,ix,iy)+b2)
                piten(6,:,ix,iy,iv)=piten(6,:,ix,iy,iv)+
     $                              ctmp(:)* bb(6,ix,iy)
                piten(7,:,ix,iy,iv)=piten(7,:,ix,iy,iv)+
     $                              ctmp(:)* bb(5,ix,iy)
                piten(8,:,ix,iy,iv)=piten(8,:,ix,iy,iv)+
     $                              ctmp(:)* bb(6,ix,iy)
                piten(9,:,ix,iy,iv)=piten(9,:,ix,iy,iv)+
     $                              ctmp(:)*(bb(3,ix,iy)+b2)
              ENDDO
            ENDDO
          ENDDO
        ENDIF

c-----------------------------------------------------------------------
c     from here, gr_alpha will be used for the test function, so we
c     perform the conjugate operation.
c-----------------------------------------------------------------------
      gr_alpha=CONJG(gr_alpha)
c-----------------------------------------------------------------------
c     compute isotropic, anisotropic, and pressure contributions
c     together for efficiency.
c
c     the "ani" contributions are the dot product of the curls
c     from the different vertices around each cell and the
c
c     symmetrized contributions from the J0Xcurl(vxB0) term are
c     the "cj0" terms without pressure gradient coefficients.
c
c     grad(P0) terms include the div(A)*div() part of the operator
c     and the symmetrized div(A)*A.grad(P0) part.
c
c     the mass density term uses the md_sink factor.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     symmetric contributions
c-----------------------------------------------------------------------
      DO iv=1,nvc
        DO jv=1,nvc
         DO iy=1,ncy
         int(:  ,1,iy,jv,iv)=0._r8
         int(2:3,2,iy,jv,iv)=0._r8
         int(3  ,3,iy,jv,iv)=0._r8
         DO ix=1,ncx
          iso_diag=md_sink(ix,iy)*alpha(ix,iy,iv)*alpha(ix,iy,jv)

          int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+iso_diag+
     $        ani*(CONJG(crl(1,1,ix,iy,iv))*crl(1,1,ix,iy,jv)+
     $             CONJG(crl(2,1,ix,iy,iv))*crl(2,1,ix,iy,jv)+
     $             CONJG(crl(3,1,ix,iy,iv))*crl(3,1,ix,iy,jv))
     $      -cj0*(alpha(ix,iy,iv)*(-ja(3,ix,iy)*crl(2,1,ix,iy,jv)
     $                             +ja(2,ix,iy)*crl(3,1,ix,iy,jv)
     $                           -dalpdrc(ix,iy,jv)*presr(1,ix,iy))
     $           +alpha(ix,iy,jv)*(-ja(3,ix,iy)*CONJG(crl(2,1,ix,iy,iv))
     $                             +ja(2,ix,iy)*CONJG(crl(3,1,ix,iy,iv))
     $                             -dalpdrc(ix,iy,iv)*presr(1,ix,iy)))
     $       +grd(ix,iy)*dalpdrc(ix,iy,iv)*dalpdrc(ix,iy,jv)

          int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)+
     $        ani*(CONJG(crl(1,1,ix,iy,iv))*crl(1,2,ix,iy,jv)+
     $             CONJG(crl(2,1,ix,iy,iv))*crl(2,2,ix,iy,jv)+
     $             CONJG(crl(3,1,ix,iy,iv))*crl(3,2,ix,iy,jv))
     $      -cj0*(alpha(ix,iy,iv)*(-ja(3,ix,iy)*crl(2,2,ix,iy,jv)
     $                             +ja(2,ix,iy)*crl(3,2,ix,iy,jv)
     $                           -dalpdz(ix,iy,jv)*presr(1,ix,iy))
     $           +alpha(ix,iy,jv)*( ja(3,ix,iy)*CONJG(crl(1,1,ix,iy,iv))
     $                             -ja(1,ix,iy)*CONJG(crl(3,1,ix,iy,iv))
     $                            -dalpdrc(ix,iy,iv)*presz(1,ix,iy)))
     $       +grd(ix,iy)*dalpdrc(ix,iy,iv)*dalpdz(ix,iy,jv)


          int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)+
     $        ani*(CONJG(crl(1,1,ix,iy,iv))*crl(1,3,ix,iy,jv)+
     $             CONJG(crl(2,1,ix,iy,iv))*crl(2,3,ix,iy,jv)+
     $             CONJG(crl(3,1,ix,iy,iv))*crl(3,3,ix,iy,jv))
     $      -cj0*(alpha(ix,iy,iv)*(-ja(3,ix,iy)*crl(2,3,ix,iy,jv)
     $                             +ja(2,ix,iy)*crl(3,3,ix,iy,jv)
     $                           +gr_alpha(3,ix,iy,jv)*presr(1,ix,iy))
     $           +alpha(ix,iy,jv)*(-ja(2,ix,iy)*CONJG(crl(1,1,ix,iy,iv))
     $                           +ja(1,ix,iy)*CONJG(crl(2,1,ix,iy,iv))))
     $       -grd(ix,iy)*dalpdrc(ix,iy,iv)*gr_alpha(3,ix,iy,jv)


          int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+iso_diag+
     $        ani*(CONJG(crl(1,2,ix,iy,iv))*crl(1,2,ix,iy,jv)+
     $             CONJG(crl(2,2,ix,iy,iv))*crl(2,2,ix,iy,jv)+
     $             CONJG(crl(3,2,ix,iy,iv))*crl(3,2,ix,iy,jv))
     $       -cj0*(alpha(ix,iy,iv)*( ja(3,ix,iy)*crl(1,2,ix,iy,jv)
     $                              -ja(1,ix,iy)*crl(3,2,ix,iy,jv)
     $                            -dalpdz(ix,iy,jv)*presz(1,ix,iy))
     $           +alpha(ix,iy,jv)*( ja(3,ix,iy)*CONJG(crl(1,2,ix,iy,iv))
     $                             -ja(1,ix,iy)*CONJG(crl(3,2,ix,iy,iv))
     $                            -dalpdz(ix,iy,iv)*presz(1,ix,iy)))
     $       +grd(ix,iy)*dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)

          int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)+
     $        ani*(CONJG(crl(1,2,ix,iy,iv))*crl(1,3,ix,iy,jv)+
     $             CONJG(crl(2,2,ix,iy,iv))*crl(2,3,ix,iy,jv)+
     $             CONJG(crl(3,2,ix,iy,iv))*crl(3,3,ix,iy,jv))
     $       -cj0*(alpha(ix,iy,iv)*( ja(3,ix,iy)*crl(1,3,ix,iy,jv)
     $                              -ja(1,ix,iy)*crl(3,3,ix,iy,jv)
     $                            +gr_alpha(3,ix,iy,jv)*presz(1,ix,iy))
     $           +alpha(ix,iy,jv)*(-ja(2,ix,iy)*CONJG(crl(1,2,ix,iy,iv))
     $                           +ja(1,ix,iy)*CONJG(crl(2,2,ix,iy,iv))))
     $       -grd(ix,iy)*dalpdz(ix,iy,iv)*gr_alpha(3,ix,iy,jv)


          int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)+iso_diag+
     $        ani*(CONJG(crl(1,3,ix,iy,iv))*crl(1,3,ix,iy,jv)+
     $             CONJG(crl(2,3,ix,iy,iv))*crl(2,3,ix,iy,jv)+
     $             CONJG(crl(3,3,ix,iy,iv))*crl(3,3,ix,iy,jv))
     $       -cj0*(alpha(ix,iy,iv)*(-ja(2,ix,iy)*crl(1,3,ix,iy,jv)
     $                              +ja(1,ix,iy)*crl(2,3,ix,iy,jv))
     $           +alpha(ix,iy,jv)*(-ja(2,ix,iy)*CONJG(crl(1,3,ix,iy,iv))
     $                           +ja(1,ix,iy)*CONJG(crl(2,3,ix,iy,iv))))
     $       +grd(ix,iy)*alpha(ix,iy,iv)*alpha(ix,iy,jv)*
     $                   k2ef(jmode)/bigr2(ix,iy)

         ENDDO
         int(1,2,iy,iv,jv)=CONJG(int(2,1,iy,jv,iv))
         int(1,3,iy,iv,jv)=CONJG(int(3,1,iy,jv,iv))
         int(2,3,iy,iv,jv)=CONJG(int(3,2,iy,jv,iv))
         ENDDO
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     possibly asymmetric stresses and implicit advection.  dvas and
c     iso_diag are used as temporary storage.
c
c     implicit advection includes Ai*.V.grad(Aj)+Ai*.Aj.grad(V), where
c     V is the n=0 part of the solution plus equilibrium center-of-mass
c     flow velocity.  [V is also multiplied by the symmetric part of
c     rho above, so it's really momentum density.]  Ai* and Aj are the
c     vector test- and basis-functions, where i and j represent the
c     finite element expansion-function row and column indices.  thus,
c     Aj are the basis functions for the delta-V solution vector.
c
c     with ei being the unit direction vector and alpha_i being
c     the scalar expansion function for the i-th basis, the
c     Ai*.V.grad(Aj) term includes:
c
c       (alpha_i*)V.grad(alpha_j)(ei.ej)  --  the dvas terms below
c
c       -(Vphi)(alpha_i*)(alpha_j)(ei.er)(ej.ephi)/r  --  in int(3,1)
c
c       +(Vphi)(alpha_i*)(alpha_j)(ei.ephi)(ej.er)/r  --  in int(1,3)
c
c     Ai*.Aj.grad(V) term includes:
c
c       (alpha_i*)(alpha_j)(ej.grad(V.ei))  --  ver & vez terms
c
c       -(Vphi)(alpha_i*)(alpha_j)(ei.er)(ej.ephi)/r  --  in int(3,1)
c
c       +(Vr)(alpha_i*)(alpha_j)(ei.ephi)(ej.ephi)/r  --  in int(3,3)
c-----------------------------------------------------------------------
      DO iv=1,nvc
        DO jv=1,nvc
         DO iy=1,ncy
         DO ix=1,ncx
          alprs=g*alpha(ix,iy,iv)/bigr(ix,iy)
          iso_diag=alpha(ix,iy,iv)*alpha(ix,iy,jv)
          dvas=ve(1,ix,iy)*dalpdr(ix,iy,jv)+
     $         ve(2,ix,iy)*dalpdz(ix,iy,jv)+
     $         ve(3,ix,iy)*(0,1)*keff(jmode)*alpha(ix,iy,jv)/bigr(ix,iy)

          int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(1,1,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(2,1,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(3,1,ix,iy,jv)+
     $         alprs*piten(9,1,ix,iy,jv))
     $       +alpha(ix,iy,iv)*(dvas+alpha(ix,iy,jv)*ver(1,ix,iy))

          int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(1,2,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(2,2,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(3,2,ix,iy,jv)+
     $         alprs*piten(9,2,ix,iy,jv))
     $       +iso_diag*vez(1,ix,iy)

          int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(1,3,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(2,3,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(3,3,ix,iy,jv)+
     $         alprs*(piten(9,3,ix,iy,jv)+
     $                2.*ve(3,ix,iy)*alpha(ix,iy,jv)))

          int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(4,1,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(5,1,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(6,1,ix,iy,jv))
     $       +iso_diag*ver(2,ix,iy)

          int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(4,2,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(5,2,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(6,2,ix,iy,jv))
     $       +alpha(ix,iy,iv)*(dvas+alpha(ix,iy,jv)*vez(2,ix,iy))

          int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(4,3,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(5,3,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(6,3,ix,iy,jv))

          int(1,3,iy,jv,iv)=int(1,3,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(7,1,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(8,1,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(9,1,ix,iy,jv))
     $       +alprs*(piten(3,1,ix,iy,jv)+ve(3,ix,iy)*alpha(ix,iy,jv))
     $       +iso_diag*ver(3,ix,iy)

          int(2,3,iy,jv,iv)=int(2,3,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(7,2,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(8,2,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(9,2,ix,iy,jv))
     $       +alprs*piten(3,2,ix,iy,jv)
     $       +iso_diag*vez(3,ix,iy)

          int(3,3,iy,jv,iv)=int(3,3,iy,jv,iv)
     $       -(gr_alpha(1,ix,iy,iv)*piten(7,3,ix,iy,jv)+
     $         gr_alpha(2,ix,iy,iv)*piten(8,3,ix,iy,jv)+
     $         gr_alpha(3,ix,iy,iv)*piten(9,3,ix,iy,jv))
     $       +alprs*(piten(3,3,ix,iy,jv)+ve(1,ix,iy)*alpha(ix,iy,jv))
     $       +alpha(ix,iy,iv)*dvas
         ENDDO
         ENDDO
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     find the contributions associated with the rows and columns of the
c     discontinuous scalar for stabilizing the compression with an
c     auxiliary field.  the equations are expressed directly in weak
c     form:
c
c       int[rho*A^*.delta(V)] 
c         + int[ sqrt(ddivv*nu*dt*f)*auxv*div(A^*) ]=dt*int[A^*.forces]
c       int[(ups^*)*auxv]
c         - int[ sqrt(ddivv*nu*dt*f)*(ups^*)*div(delta(V))]
c         = int[ sqrt(ddivv*nu*dt/f)*(ups^*)*div(V_old)]
c
c     where cma is sqrt(rho) times the magneto-acoustic speed, ups
c     is the scalar test field for the auxiliary field auxv, f is the
c     implicit centering and nu is a numerical viscosity coefficient,
c     nu=dt*cma**2.
c
c     md_sink is redefined as sqrt(rho) times the magneto-
c     acoustic speed, multiplied by numerical factors of dt and the
c     divv coefficient.
c
c     note that gr_alpha here is grad(alpha^*).
c
c     find the contributions associated with the rows and columns of the
c     discontinuous scalar for stabilizing parallel vorticity using an
c     auxiliary field.
c
c       int[rho*A^*.delta(V)] = dt*int[A^*.forces]
c               - int[ dt*sqrt(dpvrt*f/mu0)*auxw*B.curl(A^*) ]
c       int[(lam^*)*auxw]
c        - dt*sqrt(dpvrt*f/mu0)*int[(lam^*)*B.curl(delta(V))]
c        = dt*sqrt(dpvrt/(f*mu0))*int[(lam^*)*B.curl(V_old)]
c
c     where A is the test vector for V, lam is the test function for
c     the discontinuous scalar field auxw, f is the centering
c     coefficient.  here, the effective viscosity is dt*v_Alven**2.
c
c     here we create the time-centered operator for the implicit part
c     of the discretized system.
c
c     the implicit centering parameter is fpvrt.
c-----------------------------------------------------------------------
      IF (poly_divv>=0) THEN
        IF (beta>0) THEN
          md_sink=dt*SQRT(fdivv*ddivv)*
     $            SQRT(gamma*pres(1,:,:)+SUM(be**2,1)/mu0)
        ELSE
          md_sink=dt*SQRT(fdivv*ddivv)*SQRT(SUM(be**2,1)/mu0)
        ENDIF
        dtpv=dt*SQRT(fpvrt*dpvrt/mu0)

        DO iv=1,nvc
          DO jv=nvc+1,nv
            jvm=jv-nvc
            DO iy=1,ncy
              int(:,:,iy,jv,iv)=0._r8
              DO ix=1,ncx
                gcoef=bigr(ix,iy)
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+
     $            md_sink(ix,iy)*dalpdrc(ix,iy,iv)*almod(ix,iy,jvm)
                int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv) - dtpv*
     $            almod(ix,iy,jvm)*(
     $              be(2,ix,iy)*(0,1)*alpha(ix,iy,iv)*keff(jmode)/gcoef+
     $              be(3,ix,iy)*dalpdz(ix,iy,iv) )
                int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv)+
     $            md_sink(ix,iy)*dalpdz(ix,iy,iv)*almod(ix,iy,jvm)
                int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv) + dtpv*
     $            almod(ix,iy,jvm)*(
     $              be(1,ix,iy)*(0,1)*alpha(ix,iy,iv)*keff(jmode)/gcoef+
     $              be(3,ix,iy)*dalpdr(ix,iy,iv) )
                int(1,3,iy,jv,iv)=int(1,3,iy,jv,iv)-(0,1)*keff(jmode)*
     $            md_sink(ix,iy)*alpha(ix,iy,iv)*almod(ix,iy,jvm)/gcoef
                int(2,3,iy,jv,iv)=int(2,3,iy,jv,iv) - dtpv*
     $            almod(ix,iy,jvm)*(
     $              be(2,ix,iy)*dalpdrc(ix,iy,iv)-
     $              be(1,ix,iy)*dalpdz (ix,iy,iv) )
              ENDDO
            ENDDO
          ENDDO
        ENDDO
        DO iv=nvc+1,nv
          ivm=iv-nvc
          DO jv=1,nvc
            DO iy=1,ncy
              int(:,:,iy,jv,iv)=0._r8
              DO ix=1,ncx
                gcoef=bigr(ix,iy)
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)-
     $            md_sink(ix,iy)*dalpdrc(ix,iy,jv)*almod(ix,iy,ivm)
                int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-
     $            md_sink(ix,iy)*dalpdz(ix,iy,jv)*almod(ix,iy,ivm)
                int(3,1,iy,jv,iv)=int(3,1,iy,jv,iv)-(0,1)*keff(jmode)*
     $            md_sink(ix,iy)*alpha(ix,iy,jv)*almod(ix,iy,ivm)/gcoef
                int(1,2,iy,jv,iv)=int(1,2,iy,jv,iv) - dtpv*
     $            almod(ix,iy,ivm)*(
     $              be(2,ix,iy)*(0,1)*alpha(ix,iy,jv)*keff(jmode)/gcoef-
     $              be(3,ix,iy)*dalpdz(ix,iy,jv) )
                int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv) + dtpv*
     $            almod(ix,iy,ivm)*(
     $              be(1,ix,iy)*(0,1)*alpha(ix,iy,jv)*keff(jmode)/gcoef-
     $              be(3,ix,iy)*dalpdr(ix,iy,jv) )
                int(3,2,iy,jv,iv)=int(3,2,iy,jv,iv) + dtpv*
     $            almod(ix,iy,ivm)*(
     $              be(2,ix,iy)*dalpdrc(ix,iy,jv)-
     $              be(1,ix,iy)*dalpdz (ix,iy,jv) )
              ENDDO
            ENDDO
          ENDDO
          DO jv=nvc+1,nv
            jvm=jv-nvc
            DO iy=1,ncy
              int(:,:,iy,jv,iv)=0._r8
              DO ix=1,ncx
                int(1,1,iy,jv,iv)=int(1,1,iy,jv,iv)+
     $            almod(ix,iy,jvm)*almod(ix,iy,ivm)
                int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+
     $            almod(ix,iy,jvm)*almod(ix,iy,ivm)
              ENDDO
            ENDDO
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE v_aniso_op
c-----------------------------------------------------------------------
c     subprogram 27. grad_div.
c     compute the integrand for the lhs of the time split div(b)
c     diffuser for a single mode.
c-----------------------------------------------------------------------
      SUBROUTINE grad_div(int,bigr,rb,dx,dy,tb,inode)

      REAL(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz,
     $          dalpdrc
      REAL(r8), DIMENSION(SIZE(bigr,1),SIZE(bigr,2)) :: bigr2
      INTEGER(i4) :: iv,jv,nv
c-----------------------------------------------------------------------
c     set the rblock or tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree,dalpdrc)
      nv=SIZE(int,5)
c-----------------------------------------------------------------------
c     evaluate div(alpha)*div() for divergence cleaning.
c     note that the three components of B are either 
c     (real_r,real_z,-imag_phi) or (imag_r,imag_z,real_phi).
c-----------------------------------------------------------------------
      bigr2=bigr**2
      DO iv=1,nv
        DO jv=1,nv
          int(1,1,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv)+
     $       dt*fdivb*divbd*dalpdrc(:,:,iv)*dalpdrc(:,:,jv),1)
          int(2,1,:,jv,iv)=SUM(
     $       dt*fdivb*divbd*dalpdrc(:,:,iv)*dalpdz(:,:,jv),1)
          int(1,2,:,iv,jv)=int(2,1,:,jv,iv)
          int(3,1,:,jv,iv)=SUM(
     $       dt*fdivb*divbd*dalpdrc(:,:,iv)
     $                     *  alpha(:,:,jv)*keff(jmode)/bigr,1)
          int(1,3,:,iv,jv)=int(3,1,:,jv,iv)

          int(2,2,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv)+
     $       dt*fdivb*divbd*dalpdz(:,:,iv)*dalpdz(:,:,jv),1)
          int(3,2,:,jv,iv)=SUM(
     $       dt*fdivb*divbd*dalpdz(:,:,iv)
     $                     * alpha(:,:,jv)*keff(jmode)/bigr,1)
          int(2,3,:,iv,jv)=int(3,2,:,jv,iv)

          int(3,3,:,jv,iv)=SUM(alpha(:,:,iv)*alpha(:,:,jv)+
     $       dt*fdivb*divbd*alpha(:,:,iv)*alpha(:,:,jv)*k2ef(jmode)
     $                     /bigr2,1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE grad_div
c-----------------------------------------------------------------------
c     subprogram 28. cont_op.
c     compute a nonsymmetric advective/diffusion operator for number 
c     density.  integration by parts is applied to both terms.
c-----------------------------------------------------------------------
      SUBROUTINE cont_op(int,bigr,rb,dx,dy,tb,inode)

      COMPLEX(r8), DIMENSION(:,:,:,:,:), INTENT(OUT) :: int
      REAL(r8), DIMENSION(:,:), INTENT(IN) :: bigr
      TYPE(rblock_type), INTENT(INOUT) :: rb
      REAL(r8), INTENT(IN) :: dx,dy
      TYPE(tblock_type), INTENT(INOUT) :: tb
      INTEGER(i4), INTENT(IN) :: inode

      REAL(r8), DIMENSION(:,:,:), POINTER :: alpha,dalpdr,dalpdz
      REAL(r8), DIMENSION(:,:,:), POINTER :: ve_eq,nd_eq,v0,dart,
     $                                       upave,vv,dp
      REAL(r8), DIMENSION(3,SIZE(bigr,1),SIZE(bigr,2)) :: ve
      REAL(r8), DIMENSION(  SIZE(bigr,1),SIZE(bigr,2)) :: bigr2,dfac,pll

      COMPLEX(r8), DIMENSION(:,:,:,:), POINTER :: nd,dcp

      INTEGER(i4) :: nv,iv,jv,imode,iq,jq,ix,iy,ncx,ncy
      REAL(r8) :: fhyp,alpij
c-----------------------------------------------------------------------
c     set rblock and tblock weight functions.
c-----------------------------------------------------------------------
      CALL generic_alpha_eval(rb,tb%tgeom,inode,'mat',alpha,dalpdr,
     $                        dalpdz,1_i4,poly_degree)
      IF (impladv) THEN
        IF (nonlinear) THEN
          CALL generic_ptr_set(rb%qve_n0,tb%qve_n0,tb%tgeom,
     $                         inode,v0,dp,dp,0_i4)
          ve=0.5*dt*v0
        ELSE
          ve=0
        ENDIF
        IF (eq_flow/='none') THEN
          CALL generic_ptr_set(rb%qve_eq,tb%qve_eq,tb%tgeom,
     $                         inode,ve_eq,dp,dp,0_i4)
          ve=ve+0.5*dt*ve_eq
        ENDIF
        fhyp=SQRT(nd_hypd*dt)
      ELSE
        ve=0
      ENDIF
      nv=SIZE(int,5)
      ncx=SIZE(bigr,1)
      ncy=SIZE(bigr,2)
      bigr2=bigr**2
c-----------------------------------------------------------------------
c     artificial particle diffusivities.
c-----------------------------------------------------------------------
      IF (nonlinear.AND.nd_floor>0.AND.nd_diff>0) THEN
        CALL generic_ptr_set(rb%qdart,tb%qdart,tb%tgeom,
     $                       inode,dart,dp,dp,0_i4)
        dfac=fthc*dt*nd_diff+dart(1,:,:)
      ELSE
        dfac=fthc*dt*nd_diff
      ENDIF
      IF (nonlinear.AND.nd_dart_upw>0) THEN
        CALL generic_ptr_set(rb%qupw_n0,tb%qupw_n0,tb%tgeom,
     $                       inode,upave,dp,dp,0_i4)
        CALL generic_ptr_set(rb%qvv,tb%qvv,tb%tgeom,
     $                       inode,vv,dp,dp,0_i4)
        pll=nd_dart_upw*dt**2*upave(1,:,:)*upw_aniso/
     $                  ( SUM(vv(1:3,:,:),1) + smallnum )
        dfac=dfac+nd_dart_upw*dt**2*upave(1,:,:)*(1._r8-upw_aniso)
      ENDIF
c-----------------------------------------------------------------------
c     Find -grad(alpha(i)).V*alpha(j)
c          -nd_diff*grad(alpha(i)).grad(alpha(j)).
c     fthc is the centering parameter for diffusion, but implicit
c     advection is time-centered.
c
c     the second part is for anisotropic upwinding-like diffusion.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     the third part applies a fourth-order hyper-diffusivity using
c     an auxiliary field.
c-----------------------------------------------------------------------
      DO iv=1,nv
        DO jv=1,nv
           int(1,1,:,jv,iv)=SUM(-(ve(1,:,:)*dalpdr(:,:,iv)
     $                           +ve(2,:,:)*dalpdz(:,:,iv)
     $                           -ve(3,:,:)*alpha(:,:,iv)*
     $                                    (0,1)*keff(jmode)/bigr)*
     $                         alpha(:,:,jv)
     $                        +(dalpdr(:,:,iv)*dalpdr(:,:,jv)
     $                         +dalpdz(:,:,iv)*dalpdz(:,:,jv)
     $                         +k2ef(jmode)*alpha(:,:,iv)*alpha(:,:,jv)
     $                         /bigr2)*dfac
     $                        + alpha(:,:,iv)*alpha(:,:,jv),1)

          IF (.NOT.(nonlinear.AND.nd_dart_upw>0)) CYCLE

            int(1,1,:,jv,iv)=int(1,1,:,jv,iv)+SUM(
     $       ( dalpdr(:,:,iv)*dalpdr(:,:,jv)*vv(1,:,:)
     $        +dalpdz(:,:,iv)*dalpdz(:,:,jv)*vv(2,:,:)
     $        + alpha(:,:,iv)* alpha(:,:,jv)*k2ef(jmode)*vv(3,:,:)/bigr2
     $        + (dalpdr(:,:,iv)*dalpdz(:,:,jv)
     $          +dalpdz(:,:,iv)*dalpdr(:,:,jv))*vv(4,:,:)
     $      +(0,1)*(
     $       (dalpdr(:,:,iv)*alpha(:,:,jv)-dalpdr(:,:,jv)*alpha(:,:,iv))
     $                              *keff(jmode)*vv(5,:,:)/bigr+
     $       (dalpdz(:,:,iv)*alpha(:,:,jv)-dalpdz(:,:,jv)*alpha(:,:,iv))
     $                              *keff(jmode)*vv(6,:,:)/bigr))*pll,1)
        ENDDO
      ENDDO
c-----------------------------------------------------------------------
c     add contributions for a fourth-order hyper-diffusivity using
c     an auxiliary field.
c-----------------------------------------------------------------------
      IF (nd_hypd>0.AND.impladv) THEN
        DO iv=1,nv
          DO jv=1,nv
            DO iy=1,ncy
              int(2,1,iy,jv,iv)=0._r8
              int(2,2,iy,jv,iv)=0._r8
              DO ix=1,ncx
                alpij=alpha(ix,iy,iv)*alpha(ix,iy,jv)
                int(2,1,iy,jv,iv)=int(2,1,iy,jv,iv)-fhyp*
     $                         (dalpdr(ix,iy,iv)*dalpdr(ix,iy,jv)
     $                         +dalpdz(ix,iy,iv)*dalpdz(ix,iy,jv)
     $                         +k2ef(jmode)*alpij/bigr2(ix,iy))
                int(2,2,iy,jv,iv)=int(2,2,iy,jv,iv)+alpij
              ENDDO
              int(1,2,iy,jv,iv)=-int(2,1,iy,jv,iv)
            ENDDO
          ENDDO
        ENDDO
      ENDIF
c-----------------------------------------------------------------------
c     terminate.
c-----------------------------------------------------------------------
      RETURN
      END SUBROUTINE cont_op
c-----------------------------------------------------------------------
c     close module.
c-----------------------------------------------------------------------
      END MODULE integrands_mat


c-----------------------------------------------------------------------
c     global module referencing mat and rhs modules.
c-----------------------------------------------------------------------
      MODULE integrands

      USE integrands_rhs
      USE integrands_mat
c-----------------------------------------------------------------------
c     close module.
c-----------------------------------------------------------------------
      END MODULE integrands
